// ----------------------------------------------------------------------------
// ASTPrinterHTML.java
// Print syntax tree for embedding in HTML documents.
// $Id: ASTPrinterHTML.java,v 1.12 2011/09/28 16:08:08 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.InOut;

import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Program.*;

import org.w3c.dom.*;

/**
 * 
 * Print syntax tree for embedding in HTML documents.
 *
 */
public final class ASTPrinterHTML extends ASTPrinter
{
  // local information
  private final Element parent;       // the parent node
  private final Presenter presenter;  // the presenter for writing formulas
  private final Document document;    // the document of the parent node
  private int radioNumber;            // the number of radio buttons printed
  private boolean radioButtons;       // print radio buttons?
  private Statement selected;         // a statement to be highlighted
  
  private Element radio;   // the radio button for the current label
  private Element label;   // the label for the current statement
  private Element code;    // the code for the current statement
  
  /***************************************************************************
   * Create printer which appends text to parent node and may create
   * links to newly generated files in virtual directory.
   * @param parent the parent node to which the text is to be appended.
   * @param presenter the presenter for writing formulas
   * @param radioButtons true if to print a radio button for each statement.
   * @param selected a statement to be highlighted (may be null)
   **************************************************************************/
  public ASTPrinterHTML(Element parent, Presenter presenter, boolean radioButtons,
    Statement selected)
  {
    super(null, ASTPrinter.PrintStatementSpec.NONE, true);
    this.parent = parent;
    this.presenter = presenter;
    this.document = parent.getOwnerDocument();
    this.radioNumber = 0;
    this.radioButtons = radioButtons;
    this.selected = selected;
    startLine();
  }
  
  /***************************************************************************
   * Start a new line in the output.
   **************************************************************************/
  private void startLine()
  {
    Element tr = document.createElement("tr");
    parent.appendChild(tr);
    tr.setAttribute("class", "statblock");
    if (radioButtons)
    {
      Element radioCell = document.createElement("td");
      tr.appendChild(radioCell);
      radioCell.setAttribute("class", "statradio");
      radio = document.createElement("input");
      radioCell.appendChild(radio);
      radio.setAttribute("type", "hidden");
    }
    label = document.createElement("td");
    tr.appendChild(label);
    code = document.createElement("td");
    tr.appendChild(code);
    code.setAttribute("class", "statcode");
  }
  
  /****************************************************************************
   * Create indentation up to right level.
   ***************************************************************************/
  protected void tab()
  {
    super.tab();
  }
  
  /****************************************************************************
   * Print string.
   * @param s the string.
   ***************************************************************************/
  protected void print(String s)
  {
    // make spaces non-breakable
    String t = s.replace(' ', '\u00a0');
    Node node = document.createTextNode(t);
    code.appendChild(node);
  }
  
  /****************************************************************************
   * Start a new line and indent appropriately.
   ***************************************************************************/
  protected void newLine()
  {
    startLine();
  }
  
  /****************************************************************************
   * Append tree to output.
   * @param tree the tree
   ***************************************************************************/
  public void append(AST tree)
  {
    tree.accept(this);
  }
  
  /****************************************************************************
   * Visit tree.
   * @param tree the tree
   ***************************************************************************/
  public void visit(ParamDeclarationBase tree) 
  {
    label.setAttribute("class", "statlabel");
    label.setAttribute("onmouseover", 
        "top.frames['" + Presenter.detailId + 
        "'].document.location.href='" + Presenter.detailName2 + "'");
    super.visit(tree);
  }
  
  /****************************************************************************
   * Print body of an enclosing entity.
   * @param body the body to be printed.
   * @param block iff true body is to be printed as block with "{ ... }".
   ***************************************************************************/
  protected void printBody(Statement body, boolean block)
  {
    if (!(body instanceof BlockStatement))
    {
      super.printBody(body, block);
      return;
    }
    newLine();
    tab();
    visitStatement(body);
    BlockStatement b = (BlockStatement)body;
    printBlock(b.stats, false);
  }
  
  /****************************************************************************
   * Visit tree.
   * @param tree the tree
   ***************************************************************************/
  public void visit(StatementBase tree) 
  {
    visitStatement(tree);
    super.visit(tree);
  }
  
  /****************************************************************************
   * Visit tree.
   * @param tree the tree
   ***************************************************************************/
  public void visit(VariableStatement tree) 
  {
    visitStatement(tree);
    super.visit(tree);
  }
  
  /****************************************************************************
   * Visit statement.
   * @param stat the statement.
   ***************************************************************************/
  private void visitStatement(Statement stat)
  {
    // statements of for loop header are not considered
    if (label.getAttribute("class") != "") return;
    
    // position number is increased
    radioNumber = radioNumber+1;
    
    // the radio button
    if (radioButtons)
    {
      radio.setAttribute("type", "radio");
      radio.setAttribute("name", Presenter.nameStatement);
      radio.setAttribute("value", Integer.toString(radioNumber-1));
      if (radioNumber == 1) radio.setAttribute("checked", "");
    }
    
    // the label
    String file = presenter.writeStatement(stat);
    label.setAttribute("class", "statlabel");
    label.setAttribute("onmouseover", 
        "top.frames['" + Presenter.detailId + 
        "'].document.location.href='" + file + "'");
    if (stat == selected)
    {
      label.setAttribute("bgcolor", "#00ff00");
      presenter.setDetailFrame(file);
    }
    
    /*
    Element div = document.createElement("div");
    label.appendChild(div);
    div.setAttribute("class", "statlabel");
    Element tnode = document.createElement("table");
    div.appendChild(tnode);
    tnode.setAttribute("class", "menu");
    Element cnode = document.createElement("caption");
    tnode.appendChild(cnode);
    cnode.appendChild(document.createTextNode("Statement Formula"));
    appendReference(tnode, "Statement Details", file);
    */
  }
  
  /***************************************************************************
   * Append to parent node a menu entry for referencing a particular url.
   * @param parent the parent to append the node to.
   * @param label the label appearing in the command menu.
   * @param url the url to be referenced.
   **************************************************************************/
  protected void appendReference(Element parent,
    String label, String url)
  {
    Element trnode0 = document.createElement("tr");
    parent.appendChild(trnode0);
    Element tdnode0 = document.createElement("td");
    trnode0.appendChild(tdnode0);
    tdnode0.setAttribute("class", "entry");
    Element anode0 = document.createElement("a");
    tdnode0.appendChild(anode0);
    try
    {
      String url0 = URLEncoder.encode(url, "UTF-8");
      anode0.setAttribute("href", url0);
      anode0.setAttribute("target", Presenter.detailId);
    }
    catch(UnsupportedEncodingException e)
    {
      Main.internalError("Error in encoding command: " + e);
    }
    anode0.appendChild(document.createTextNode(label));
  }
}
