// ----------------------------------------------------------------------------
// Breaks.java
// Compute line break information for pretty-printing abstract syntax trees.
// $Id: Breaks.java,v 1.9 2017/07/04 08:04:36 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.InOut;

import java.io.*;

import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;

/**
 * 
 * Compute line break information for pretty-printing abstract syntax trees.
 *
 */
public final class Breaks extends ASTVisitorBase
{
  /**
   * 
   * The information compute for a syntax tree.
   * 
   */
  public static class Info
  {
    // the tree for which the info is computed
    public final AST tree;        
    
    // the length of its linear representation (children included)
    private int length = 0;  
    
    // the info for the children of the tree (null, if none)
    private Info[] children = null; 
    
    /*************************************************************************
     * Create info for the denoted tree.
     * @param tree the tree for which the info is to be computed.
     ************************************************************************/
    public Info(AST tree)
    {
      this.tree = tree;
    }
    
    /*************************************************************************
     * Get the length of the linear representation (children included).
     * @return the length of the linear representation.
     ************************************************************************/
    public int getLength()
    {
      return length;
    }
    
    /*************************************************************************
     * Get the children information.
     * @return the children information.
     ************************************************************************/
    public Info[] getChildren()
    {
      return children;
    }
    
    /*************************************************************************
     * Print break information in tree form.
     * @param out the stream to print to.
     ************************************************************************/
    public void print(PrintWriter out)
    {
      print(0, out);
    }
    
    /*************************************************************************
     * Print break information in tree form.
     * @param indent the current indentation level.
     * @param out the stream to print to.
     ************************************************************************/
    private void print(int indent, PrintWriter out)
    {
      for (int i=0; i<indent; i++) out.print("  ");
      String string = tree.toString();
      out.println(indent+":"+length+":"+string.length()+":"+string);
      if (children != null)
      {
        for (Info child : children)
          child.print(indent+1, out);
      }
    }
  }
  
  // the information for the currently visited node
  private Info current;
  
  // auxiliary variable used during construction
  // true iff mathml representation is considered in computing the length
  private boolean mathml;
  
  /****************************************************************************
   * Construct the line break information for the denoted formula.
   * @param formula a formula.
   * @param mathml true iff the mathml representation is considered.
   * @return the information.
   ***************************************************************************/
  public static Info compute(Formula formula, boolean mathml)
  {
    Breaks breaks = new Breaks(mathml);
    return breaks.compute(formula);
  }
  
  /****************************************************************************
   * Construct the line break information for the denoted term.
   * @param term a term.
   * @param mathml true iff the mathml representation is considered.
   * @return the information.
   ***************************************************************************/
  public static Info compute(Term term, boolean mathml)
  {
    Breaks breaks = new Breaks(mathml);
    return breaks.compute(term);
  }
  
  /****************************************************************************
   * Get the sum of the lengths of the info objects.
   * @param infos the info objects.
   * @return the sum of the lengths.
   ***************************************************************************/
  public static int length(Info[] infos)
  {
    int len = 0;
    for (Info info : infos)
      len = len+info.length;
    return len;
  }
  
  /****************************************************************************
   * Construct the line break information for a formula.
   * @param mathml true iff the mathml representation is considered.
   ***************************************************************************/
  private Breaks(boolean mathml)
  {
    this.mathml = mathml;
  }
  
  /***************************************************************************
   * Set break information for a given node.
   * @param tree the tree for which the break information is to be computed.
   * @return the break information for that node.
   **************************************************************************/
  private Info compute(AST tree)
  {
    Info info = new Info(tree);
    current = info;
    tree.accept(this);
    return info;
  }
  
  /****************************************************************************
  * Compute for a sequence of nodes trees[0,...] the break infos infos[pos,...]
  * @param trees the sequence of nodes
  * @param the uninitialized array of info objects 
  * @param pos the position where the first information is to be computed.
  * @param sep the length of a separator.
  * @return the sum of the lengths of all computed info objects 
  *         plus the length of the required separators;
  ****************************************************************************/
  private int compute(AST trees[], Info[] infos, int pos, int sep)
  {
    int n = trees.length;
    int length = 0;
    for (int i=0; i<n; i++)
    {
      infos[i+pos] = compute(trees[i]);
      length = length + infos[i+pos].length;
    }
    if (n > 0) length = length+sep*(n-1);
    return length;
  }
  
  /***************************************************************************
   * Return length required to print parentheses if exp is 
   * a left subexpression of outer.
   * @param exp the subexpression.
   * @param outer the outer expression.
   * @return the length of parentheses to print
   *************************************************************************/
  private int leftPar(Expression exp, Expression outer)
  {
    if (exp.getPrecedence() < outer.getPrecedence() &&
      !(exp.getClass() == outer.getClass() && exp.isLeftAssociative()))
      return 2; // "()"
    else
      return 0;
  }
  
  /***************************************************************************
   * Return length required to print parentheses if exp is 
   * a right subexpression of outer.
   * @param exp the subexpression.
   * @param outer the outer expression.
   * @return the length of parentheses to print
   *************************************************************************/
  private int rightPar(Expression exp, Expression outer)
  {
    if (exp.getPrecedence() < outer.getPrecedence() &&
     !(exp.getClass() == outer.getClass() && exp.isRightAssociative()))
      return 2; // "()"
    else
      return 0;
  }
  
  // -------------------------------------------------------------------------
  //
  // the visit() methods for this visitor
  //
  // -------------------------------------------------------------------------
  
  public void visit(Identifier exp)
  {
    // information for this node
    Info info = current;
    
    // set the information for this node
    info.length = exp.getString().length();
    info.children = null;
    // return true;
  }
  
  public void visit(Name exp)
  {
    // information for this node
    Info info = current;
    
    // set the information for this node
    info.length = exp.toString().length();
    info.children = null;
    // return true;
  }
  
  public void visit(ApplicationExpression exp)
  {
    // information for this node
    Info info = current;
    
    // process children and compute their length
    Expression fun = exp.getFunctionExpression();
    Expression[] args = exp.getArgumentExpressions();
    int n = args.length;
    Info[] infos = new Info[1+n];
    infos[0] = compute(fun);
    int alength = compute(args, infos, 1, 2);
    
    // set the information for this node
    info.length = infos[0].length+alength+2; // "fun()"
    info.children = infos;
    // return true;
  }
  
  public void visit(BinaryExpression exp)
  { 
    // information for this node
    Info info = current;
    
    // process children
    Expression first = exp.getFirst();
    Expression second = exp.getSecond();
    Info firstInfo = compute(first);
    Info secondInfo = compute(second);
    
    // compute length of this node
    int length = exp.getOperator().length();       // "op"
    if (mathml && (exp instanceof AndFormula || exp instanceof OrFormula))
      length = 1;                                  // "/\", "\/"
    if (exp instanceof Formula) length = length+2; // " op "

    // set the information for this node
    info.length = 
      length+leftPar(first, exp)+rightPar(second, exp)+
      firstInfo.length+secondInfo.length;
    info.children = new Info[] { firstInfo, secondInfo };
    // return true;
  }
  
  public void visit(UnaryExpression exp)
  { 
    // information for this node
    Info info = current;
    
    // process children
    Expression first = exp.getBase();
    Info firstInfo = compute(first);
    
    // compute length of this node
    int length = exp.getOperator().length();       // "op"
    if (mathml)
    {
      if (exp instanceof NotFormula) length = 1;   // "~"
    }
    else 
    {
      String op = exp.getOperator();
      if (Character.isLetter(op.charAt(op.length()-1)))
        length = length+1;                        // " "
    }

    // set the information for this node
    info.length = length+firstInfo.length+rightPar(first, exp);
    info.children = new Info[] { firstInfo };
    // return true;
  }
  
  public void visit(QuantifiedExpression exp)
  {
    // information for this node
    Info info = current;
    
    // process children and compute their length
    String quantifier = exp.getQuantifier();
    TypedIdentifier[] vars = exp.getVariables();
    Expression body = exp.getBody();
    int length = quantifier.length();
    if (mathml && (exp instanceof ForallFormula || // A (forall)
                   exp instanceof ExistsFormula || // E (exists)
                   exp instanceof LambdaFormula))  // L (lambda)
      length = 1;
    int n = vars.length;
    Info[] infos = new Info[n+1];
    int vlength = compute(vars, infos, 0, 2);
    infos[n] = compute(body);
    
    // set the information for this node "quant(vars): body
    info.length = length+vlength+infos[n].length+rightPar(body, exp)+4;
    info.children = infos;
    // return true;
  }
  
  public void visit(ArrayType exp)
  {
    // information for this node
    Info info = current;
    
    // process children
    Expression first = exp.getIndex();
    Expression second = exp.getBase();
    Info firstInfo = compute(first);
    Info secondInfo = compute(second);

    // set the information for this node "ARRAY OF "
    info.length = 9 + firstInfo.length + secondInfo.length;
    info.children = new Info[] { firstInfo, secondInfo };
    // return true;
  }
  
  public void visit(AtomicType exp)
  { 
    // information for this node
    Info info = current;
    info.length = exp.getName().length();
    info.children = null;
    // return true;
  }
  
  public void visit(Constant exp)
  { 
    // information for this node
    Info info = current;
    info.length = exp.getName().length();
    info.children = null;
    // return true;
  }
  
  public void visit(FunctionType exp)
  { 
    // information for this node
    Info info = current;

    Expression[] first = exp.getDomain();
    Expression second = exp.getRange();
    int n = first.length;
    
    // special case of a predicate type
    if (second instanceof BooleanType)
    {
      Info[] infos = new Info[n];
      compute(first, infos, 0, 2);
      info.length = 11; // "PREDICATE()"
      info.children = infos;
      return;
    }
    
    // general case of a function type
    Info[] infos = new Info[n+1];
    compute(first, infos, 0, 2);
    infos[n] = compute(second);
    
    // compute length of this node
    int length = 4; // " -> "
    if (n == 1)
      length = length+leftPar(first[0], exp);
    else
      length = length+2; // "()"

    // set the information for this node
    info.length = length + rightPar(second, exp);
    info.children = infos;
    // return true;
  }
  
  public void visit(IfThenElseExpression exp)
  { 
    // information for this node
    Info info = current;
    
    // process children
    Info firstInfo = compute(exp.getCondition());
    Info secondInfo = compute(exp.getThenBranch());
    Info thirdInfo = compute(exp.getElseBranch());
    
    // compute length of this node // "IF THEN [ELSIF | ELSE ] ENDIF"
    int length = 
      IfThenElseExpression.ifKeyword.length()+2+
      IfThenElseExpression.thenKeyword.length()+2+
      IfThenElseExpression.endifKeyword.length();
    if (exp.getElseBranch() instanceof IfThenElseExpression)
      length = length
                 -(IfThenElseExpression.endifKeyword.length()+2)
                 -(IfThenElseExpression.elseKeyword.length()+
                   IfThenElseExpression.ifKeyword.length()-
                   IfThenElseExpression.elsifKeyword.length());
    else
      length = length+IfThenElseExpression.elseKeyword.length()+2;

    // set the information for this node
    info.length = length+firstInfo.length+secondInfo.length+thirdInfo.length;
    info.children = new Info[] { firstInfo, secondInfo, thirdInfo };
    // return true;
  }
  
  public void visit(LetExpression exp)
  { 
    // information for this node
    Info info = current;
    
    // process children
    Definition[] defs = exp.getDefinitions();
    Expression body = exp.getBody();
    int n = defs.length;
    Info[] infos = new Info[n+1];
    int length = compute(defs, infos, 0, 2);
    infos[n] = compute(body);
    
    // set the information for this node
    info.length = length+infos[n].length+7+rightPar(body, exp); // "LET IN "
    info.children = infos;
    // return true;
  }
  
  public void visit(PostfixApplication exp)
  { 
    // if (exp instanceof StateThrowsException)
    // {
    //  visit((StateThrowsException)exp);
    //  return;
    // }
    
    // information for this node
    Info info = current;
    
    // process children
    Expression arg = exp.getArgument();
    Info ainfo = compute(arg);
    
    // set the information for this node
    info.length = 
      exp.getIdentifier().getString().length()+
      PostfixApplication.operator.length()+
      ainfo.length+
      rightPar(arg, exp);
    info.children = new Info[]{ ainfo };
    // return true;
  }
  
  public void visit(StateThrowsException exp)
  { 
    // information for this node
    Info info = current;
    
    // process children
    int length = exp.getIdentifier().getString().length()
               +PostfixApplication.operator.length();
    Expression arg = exp.getArgument();
    Name ex = exp.getException();
    Info ainfo = compute(arg);
    Info einfo = compute(ex);
    
    // set the information for this node
    info.length = length+ainfo.length+3+ex.toString().length();
    info.children = new Info[]{ ainfo, einfo };
    // return true;
  }
  
  public void visit(ProgramVariableBase exp)
  { 
    // information for this node
    Info info = current;
    
    // set the information for this node
    info.length = exp.getKeyword().length()+1+exp.getName().toString().length();
    info.children = null;
    // return true;
  }
  
  public void visit(WritesOnlyFormula exp)
  { 
    // information for this node
    Info info = current;
    
    // process children
    int length = exp.getKeyword().length();
    Name[] vars = exp.getVariables();
    Info[] infos = null;
    if (vars != null)
    {
      int n = vars.length;
      infos = new Info[n];
      length = length+1+compute(vars, infos, 0, 2);
    }
    
    // set the information for this node
    info.length = length;
    info.children = infos;
    // return true;
  }
  
  public void visit(RecordTerm exp)
  { 
    // information for this node
    Info info = current;
   
    // process children
    ValuedIdentifier[] ids = exp.getBase();
    int n = ids.length;
    Info[] infos = new Info[n];
    int length = compute(ids, infos, 0, 2);
    
    // set the information for this node
    info.length = length+
      RecordTerm.beginToken.length()+RecordTerm.endToken.length();
    info.children = infos;
    // return true;
  }
  
  public void visit(RecordType exp)
  { 
    // information for this node
    Info info = current;
   
    // process children
    TypedIdentifier[] ids = exp.getBase();
    int n = ids.length;
    Info[] infos = new Info[n];
    int length = compute(ids, infos, 0, 2);
    
    // set the information for this node
    info.length = length+
      RecordType.beginToken.length()+RecordType.endToken.length();
    info.children = infos;
    // return true;
  }
  
  public void visit(Reference exp)
  { 
    // information for this node
    Info info = current;
    
    // set the information for this node
    info.length = exp.getName().toString().length();
    info.children = null;
    // return true;
  }
  
  public void visit(SelectionTerm exp)
  { 
    // information for this node
    Info info = current;
    
    // process children
    Expression base = exp.getBase();
    Selector sel = exp.getSelector();
    Info baseInfo = compute(base);
    Info selInfo = compute(sel);
    
    // set the information for this node
    info.length = baseInfo.length+selInfo.length;
    info.children = new Info[] { baseInfo, selInfo };
    // return true;
  }
  
  public void visit(SelectorIdentifier exp)
  { 
    // information for this node
    Info info = current;
    
    // set the information for this node
    info.length = 1+exp.getIdentifier().getString().length(); // ".ident"
    info.children = null;
    // return true;
  }
 
  public void visit(SelectorIndex exp)
  { 
    // information for this node
    Info info = current;
    
    // process children
    Expression index = exp.getIndex();
    Info indexInfo = compute(index);

    // set the information for this node
    info.length = 2+indexInfo.length; // "[index]"
    info.children = new Info[] { indexInfo };
    // return true;
  }
  
  public void visit(SelectorNumber exp)
  { 
    // information for this node
    Info info = current;
    
    // set the information for this node
    info.length = 1+exp.getNumber().getName().length(); // ".number"
    info.children = null;
    // return true;
  }
  
  public void visit(StateType exp)
  { 
    // information for this node
    Info info = current;
    
    // state may have no result type
    Type rtype = exp.getReturn();
    if (rtype == null)
    {
      info.length = StateType.stateKeyword.length(); // "STATE"
      info.children = null;
      return;
    }
    
    // process children
    Info rinfo = compute(rtype);
    
    // set the information for this node
    info.length = StateType.stateKeyword.length()+2+rinfo.length; // "STATE(type)"
    info.children = new Info[] { rinfo };
    // return true;
  }
  
  public void visit(SubrangeType exp)
  { 
    // information for this node
    Info info = current;
    
    // process children
    Info info1 = compute(exp.getFrom());
    Info info2 = compute(exp.getTo());
    
    // set the information for this node
    info.length = 
      SubrangeType.beginToken.length()+info1.length+
      SubrangeType.separatorToken.length()+info2.length+
      SubrangeType.endToken.length();
    info.children = new Info[] { info1, info2 };
    // return true;
  }
  
  public void visit(SubType exp)
  { 
    // information for this node
    Info info = current;
    
    // process children
    Info info1 = compute(exp.getPredicate());
    
    // set the information for this node "SUBTYPE()"
    info.length = SubType.keyword.length()+info1.length+2;
    info.children = new Info[] { info1 };
    // return true;
  }
  
  public void visit(TupleTerm exp)
  {
    // information for this node
    Info info = current;
    
    // process children and compute their length
    Expression[] base = exp.getBase();
    int n = base.length;
    Info[] infos = new Info[n];
    int blength = compute(base, infos, 0, 2);
    
    // compute the length of the children
    // set the information for this node
    info.length = 
      TupleTerm.beginToken.length()+
      blength+
      TupleTerm.endToken.length();
    info.children = infos;
    // return true;
  }
  
  public void visit(TupleType exp)
  {
    // information for this node
    Info info = current;
    
    // process children and compute their length
    Expression[] base = exp.getBase();
    int n = base.length;
    Info[] infos = new Info[n];
    int blength = compute(base, infos, 0, 2);
    
    // set the information for this node
    info.length = 
      TupleType.beginToken.length()+
      blength+
      TupleType.endToken.length();
    info.children = infos;
    // return true;
  }
  
  public void visit(TypedIdentifier exp)
  {
    // information for this node
    Info info = current;
    
    // process children and compute their length
    Info info1 = compute(exp.getName());
    Info info2 = compute(exp.getType());

    // set the information for this node "Id: Type"
    info.length = info1.length+2+info2.length;
    info.children = new Info[] { info1, info2 };
    // return true;
  }
  
  public void visit(UpdateTerm exp)
  {
    // information for this node
    Info info = current;
    
    // process children and compute their length
    Selector[] sels = exp.getSelectors();
    int n = sels.length;
    Info[] infos = new Info[n+2];
    infos[0] = compute(exp.getBase());
    int slength = compute(sels, infos, 1, 0);
    infos[n+1] = compute(exp.getValue());

    // set the information for this node "base WITH sels:=value"
    info.length = 
      infos[0].length+UpdateTerm.withKeyword.length()+2+
      slength+UpdateTerm.assignToken.length()+infos[n+1].length;
    info.children = infos;
    // return true;
  }
  
  public void visit(ValueDeclarationInitialized exp)
  {
    // information for this node
    Info info = current;
    
    // process children and compute their length
    Info info1 = compute(exp.getName());
    Info info2 = compute(exp.getType());
    Info info3 = compute(exp.getValue());

    // set the information for this node "name: type = value"
    info.length = 5+info1.length+info2.length+info3.length;
    info.children = new Info[] { info1, info2, info3 };
    // return true;
  }
  
  public void visit(ValueDefinitionClass exp)
  {
    // information for this node
    Info info = current;
    
    // process children and compute their length
    Info info1 = compute(exp.getName());
    Info info2 = compute(exp.getValue());

    // set the information for this node "name = value"
    info.length = 3+info1.length+info2.length;
    info.children = new Info[] { info1, info2 };
    // return true;
  }
  
  public void visit(ValuedIdentifier exp)
  {
    // information for this node
    Info info = current;
    
    // process children and compute their length
    Info info1 = compute(exp.getIdentifier());
    Info info2 = compute(exp.getValue());

    // set the information for this node "name:=value"
    info.length = 2+info1.length+info2.length;
    info.children = new Info[] { info1, info2 };
    // return true;
  }
}
