// ----------------------------------------------------------------------------
// Package.java
// Handling of packages.
// $Id: Package.java,v 1.11 2010/01/21 17:15:46 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.InOut;

import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Syntax.*;
import java.io.*;

/**
 * 
 * Handling of packages.
 *
 */
public class Package
{
  /****************************************************************************
   * Create a package handler.
   * @param dirs a list of directories whose contents represent the package.
   * @param error the stream on which to write error messages.
   * @param wdir the directory to write information to
   * @return the package (or null if an error occurred).
   ***************************************************************************/
  public static PackageSymbol construct(File[] dirs, 
    ErrorStream error, File wdir)
  {
    PackageSymbol psymbol = 
      new PackageSymbol(Identifier.construct(""), error, wdir);
    for (File dir : dirs)
    {
      boolean okay = process(psymbol, dir, error);
      if (!okay) return null;
    }
    return psymbol;
  }
  
  /****************************************************************************
   * Filter for class files.
   ***************************************************************************/
  private static class ClassFilter implements FileFilter
  {
    public boolean accept(File file)
    {
      return Source.isClassSource(file);
    }
  }
  
  /****************************************************************************
   * Filter for theory files.
   ***************************************************************************/
  private static class TheoryFilter implements FileFilter
  {
    public boolean accept(File file)
    {
      return Source.isTheorySource(file);
    }
  }
  
  /****************************************************************************
   * Filter for packages.
   ***************************************************************************/
  private static class PackageFilter implements FileFilter
  {
    public boolean accept(File file)
    {
      return (!file.getName().startsWith(".")) && file.isDirectory();
    }
  }
  
  // static instances of above classes
  private static ClassFilter classFilter = new ClassFilter();
  private static TheoryFilter theoryFilter = new TheoryFilter();
  private static PackageFilter packageFilter = new PackageFilter();
  
  /****************************************************************************
   * Add content of directory to package.
   * @param parent the package to add to.
   * @param dir the directory.
   * @param error the error stream to write to
   * @return true iff everything went fine.
   ***************************************************************************/
  private static boolean process(PackageSymbol parent, File dir,
    ErrorStream error)
  {
    SymbolTable<ClassSymbol> classes = parent.getClasses();
    SymbolTable<TheorySymbol> theories = parent.getTheories();
    
    // add classes and class theories
    File[] files = dir.listFiles(classFilter);
    for (File f : files)
    {
      String base = getBase(f.getName());
      Identifier name = Identifier.construct(base);
      Source source = Source.create(f, error);
      ClassReader reader = new ClassReader(source, error);
      ClassSymbol c = new ClassSymbol(parent, name, reader);
      boolean okay = classes.put(c);
      if (!okay)
      {
        error.newError("Class " + name + 
            " is already contained in package " + parent.getLongName());
        return false;
      }
      TheorySymbol t = new TheorySymbol(c);
      okay = theories.put(t);
      if (!okay)
      {
        error.newError("Theory " + name + 
            " is already contained in package " + parent.getLongName());
        return false;
      }
    }
    
    // add manually created theories
    files = dir.listFiles(theoryFilter);
    for (File f : files)
    {
      String base = getBase(f.getName());
      Identifier name = Identifier.construct(base);
      Source source = Source.create(f, error);
      TheoryReader reader = new TheoryReader(source, error);
      TheorySymbol c = new TheorySymbol(parent, name, reader);
      boolean okay = theories.put(c);
      if (!okay)
      {
        error.newError(name, "Theory " + name + 
            " is already contained in package " + parent.getLongName());
        return false;
      }
    }
    
    // add packages and process them recursively
    File[] subdirs = dir.listFiles(packageFilter); 
    if (subdirs == null)
    {
      error.newError(
          "Directory \"" + dir.getAbsolutePath() + "\" cannot be read.");
      return false;
    }
    SymbolTable<PackageSymbol> packages = parent.getPackages();
    for (File d : subdirs)
    {
      Identifier name = Identifier.construct(d.getName());
      PackageSymbol p = new PackageSymbol(parent, name, error);
      p.setDirectory(d);
      boolean okay = packages.put(p);
      if (!okay)
      {
        error.newError("Subpackage " + name + 
            " is already contained in package " + parent.getLongName());
        return false;
      }
      okay = process(p, d, error);
      if (!okay) return false;
    }
    return true;
  }
  
  /***************************************************************************
   * Get base of file name.
   * @param name a file name.
   * @return the name without the trailing extension of form ".ext".
   **************************************************************************/
  private static String getBase(String name)
  {
    int i = name.lastIndexOf(".");
    if (i == -1) return name;
    return name.substring(0, i);
  }   
}
