// ----------------------------------------------------------------------------
// PropagatorCommand.java
// The command of propagating a formula through the body of a method.
// $Id: PropagatorCommand.java,v 1.17 2013/06/26 16:24:26 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.InOut;

import java.util.*;
import java.io.*;
import org.antlr.runtime.*;
import org.w3c.dom.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Parser.*;
import fmrisc.ProgramExplorer.SWT.*;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Program.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Syntax.Spec.*;

/**
 * 
 * The command of propagating a formula through the body of a method.
 *
 */
public final class PropagatorCommand
{
  private final TopWindow topWindow;
  private final PackageSymbol topPackage;
  private final PrintWriter writer;
  
  /***************************************************************************
   * Execute command.
   * @param topWindow the top window giving the context of the command.
   * @param arguments a mapping of names to values.
   * @return the content of the output document.
   *************************************************************************/
  public static String execute(TopWindow topWindow,
    Map<String, String> arguments)
  {
    // get arguments
    String method = getValue(arguments, Presenter.nameMethod);
    String statement = getValue(arguments, Presenter.nameStatement);
    String prepost = getValue(arguments, Presenter.namePrePost);
    String formula = getValue(arguments, Presenter.nameFormula);
    
    // execute command
    StringWriter writer = new StringWriter();
    PrintWriter error = new PrintWriter(writer);
    PropagatorCommand command = new PropagatorCommand(topWindow, error);
    command.execute(method, statement, prepost, formula);
    
    // construct result document
    int fontSize = topWindow.getFontSize()-1;
    Document document = Presenter.createDocument("Output", fontSize);
    Element top = Presenter.createTopNode(document, fontSize);
    Element pre = document.createElement("pre");
    top.appendChild(pre);
    Node text = document.createTextNode(writer.toString());
    pre.appendChild(text);
    StringWriter result = new StringWriter();
    Presenter.writeDocument(document, new PrintWriter(result));
    return result.toString();
  }
  
  /***************************************************************************
   * Get value from argument mapping.
   * @param arguments a mapping of names to values.
   * @param name a name.
   * @return the corresponding value (aborts program, if no such value).
   **************************************************************************/
  private static String getValue(Map<String,String> arguments, String name)
  {
    String value = arguments.get(name);
    if (value == null) Main.internalError("no value for key " + name);
    return value;
  }
  
  /***************************************************************************
   * Create a command executor.
   * @param topWindow the context of the application.
   * @param writer the stream for printing messages
   **************************************************************************/
  private PropagatorCommand(TopWindow topWindow, PrintWriter writer)
  {
    this.topWindow = topWindow;
    this.topPackage = topWindow.getTopPackage();
    this.writer = writer;
  }
  
  /***************************************************************************
   * Propagate condition in body of method.
   * @param method the method.
   * @param statement the statement string
   * @param prepost is "pre", if formula applies to prestate
   * @param fstring the string representation of the formula
   **************************************************************************/
  private void execute(String method, String statement, String prepost, String fstring)
  {
    ParamSymbol msymbol = getMethod(topPackage, method);
    int number = Integer.parseInt(statement);
    ErrorStream error = new ErrorStream("ERROR", "   ", "", writer, null);
    Environment env = new Environment(topPackage, error);
    Statement stat = getStatement(msymbol, number, env);
    Formula formula = getFormula(fstring, msymbol, env, error);
    if (formula == null) return;
    boolean pre = prepost.equals("pre");
    topWindow.propagateCondition(msymbol, stat, formula, pre);
  }
  
  /****************************************************************************
   * Find method with denoted name.
   * @param topPackage the top package.
   * @param method the name of the method.
   * @return the method denoted by the name in the package (null, if not found)
   ***************************************************************************/
  private static ParamSymbol getMethod(PackageSymbol topPackage, String method)
  {
    Symbol symbol = topPackage;
    String[] names = method.split("\\.");
    int n = names.length;
    for (int i=0; i<n; i++)
    {
      String name = names[i];
      if (symbol instanceof PackageSymbol)
      {
        PackageSymbol symbol0 = (PackageSymbol)symbol;
        SymbolTable<PackageSymbol> packages = symbol0.getPackages();
        PackageSymbol psymbol = packages.get(Identifier.construct(name));
        if (psymbol != null)
        {
          symbol = psymbol;
          continue;
        }
        SymbolTable<ClassSymbol> classes = symbol0.getClasses();
        ClassSymbol csymbol = classes.get(Identifier.construct(name));
        if (csymbol == null) Main.internalError("no class " + name);
        symbol = csymbol;
        continue;
      }
      if (symbol instanceof ClassSymbol)
      {
        ClassSymbol symbol0 = (ClassSymbol)symbol;
        if (i != n-2) Main.internalError("wrong class position: " + i);
        String name0 = names[n-1];
        int offset = Integer.parseInt(name0);
        Collection<ConstructorSymbol> constructors = symbol0.getConstructorSet();
        ParamSymbol result = getConstructor(constructors, offset);
        if (result != null) return result;
        Collection<ObjectMethodSymbol> omethods = symbol0.getObjectMethodSet();
        result = getObjectMethod(omethods, offset);
        if (result != null) return result;
        Collection<ClassMethodSymbol> cmethods = symbol0.getClassMethodSet();
        result = getClassMethod(cmethods, offset);
        if (result != null) return result;
        Main.internalError("method not found: " + method);
        continue;
      }
      Main.internalError("unknown kind of symbol: " + symbol.getName());
    }
    Main.internalError("error in finding method: " + method);
    return null;
  }
  
  /****************************************************************************
   * Get symbol whose declaration has denoted offset in source.
   * @param symbols a collection of parameterized symbols.
   * @param offset a desired offset.
   * @return that symbol in the collection that has the denoted offset
   *         (null, if none)
   ***************************************************************************/
  private static ParamSymbol getConstructor(Collection<ConstructorSymbol> symbols, int offset)
  {
    for (ConstructorSymbol symbol : symbols)
    {
      int offset0 = symbol.getDeclaration().getPosition().getOffset();
      if (offset == offset0) return symbol;
    }
    return null;
  }
  
  /****************************************************************************
   * Get symbol whose declaration has denoted offset in source.
   * @param symbols a collection of parameterized symbols.
   * @param offset a desired offset.
   * @return that symbol in the collection that has the denoted offset
   *         (null, if none)
   ***************************************************************************/
  private static ClassMethodSymbol getClassMethod(Collection<ClassMethodSymbol> symbols, int offset)
  {
    for (ClassMethodSymbol symbol : symbols)
    {
      int offset0 = symbol.getDeclaration().getPosition().getOffset();
      if (offset == offset0) return symbol;
    }
    return null;
  }
  
  /****************************************************************************
   * Get symbol whose declaration has denoted offset in source.
   * @param symbols a collection of parameterized symbols.
   * @param offset a desired offset.
   * @return that symbol in the collection that has the denoted offset
   *         (null, if none)
   ***************************************************************************/
  private static ObjectMethodSymbol getObjectMethod(Collection<ObjectMethodSymbol> symbols, int offset)
  {
    for (ObjectMethodSymbol symbol : symbols)
    {
      int offset0 = symbol.getDeclaration().getPosition().getOffset();
      if (offset == offset0) return symbol;
    }
    return null;
  }
  
  /***************************************************************************
   * Get statement of method body.
   * @param method the method.
   * @param number the number of the statement.
   * @param env a top-level environment which is set to the 
   *        environment of the statement.
   * @return the statement.
   ***************************************************************************/
  private Statement getStatement(ParamSymbol method, int number, Environment env)
  {
    ClassSymbol csymbol = method.getClassSymbol();
    env.enterPackage(csymbol.getPackage());
    env.enterClass(csymbol);
    if (method instanceof ConstructorSymbol || method instanceof ObjectMethodSymbol)
      env.enterObject();
    env.enterMethod(method);
    env.enterBody();
    StatementFinder visitor = new StatementFinder(number, env);
    Statement body = method.getDeclaration().getBody();
    body.accept(visitor);
    if (visitor.result == null) Main.internalError("statement not found");
    return visitor.result;
  }
  
  /**************************************************************************
   * A visitor that determines a statement identified by number.
   *************************************************************************/
  private static class StatementFinder extends ASTVisitorBase
  {
    // the number of the statement to be determined
    private final int number;
    
    // the number of statements visited so far
    private int counter;
    
    // if not-null, the statement with the denoted number
    private Statement result;
    
    // the environment built up in collecting the statement
    private Environment env;
    
    /*************************************************************************
     * Construct a finder for determining a particular statement.
     * @param number the number of the statement.
     * @param env the environment which is updated to reflect the
     *        environment of the statement.
     ***********************************************************************/
    public StatementFinder(int number, Environment env)
    {
      this.number = number;
      this.counter = 0;
      this.result = null;
      this.env = env;
    }
    
    // concrete statements
    public void visit(AssignmentStatement stat) { visitCore(stat); }
    public void visit(BlockStatement stat) { visitCore(stat); }
    public void visit(BreakStatement stat) { visitCore(stat); }
    public void visit(ConditionalStatement stat) { visitCore(stat); }
    public void visit(ContinueStatement stat) { visitCore(stat); }
    public void visit(EmptyStatement stat) { visitCore(stat); }
    public void visit(ReturnStatement stat) { visitCore(stat); }
    public void visit(TryCatchStatement stat) { visitCore(stat); }
    public void visit(VariableStatement stat) { visitCore(stat); }
    
    // groups of statements
    public void visit(LoopStatementBase stat) { visitCore(stat); }
    public void visit(CallStatement stat) { visitCore(stat); }
    
    private void visitCore(Statement stat)
    {
      if (stat instanceof VariableDeclaration)
      {
        VariableDeclaration stat0 = (VariableDeclaration)stat;
        BodySymbol symbol = (BodySymbol)stat0.getName().getSymbol();
        boolean okay = env.put(symbol);
        if (!okay) Main.internalError("error in adding symbol to environment");
      }
      counter = counter+1;
      if (counter == number+1)
      {
        result = stat;
        return;
      }
      if (stat instanceof BlockStatement)
      {
        env.enterScope();
        BlockStatement stat0 = (BlockStatement)stat;
        for (Statement s : stat0.stats)
        {
          s.accept(this);
          if (result != null) return;
        }
        env.exitScope();
        return;
      }
      if (stat instanceof ForLoopStatement)
      {
        ForLoopStatement stat0 = (ForLoopStatement)stat;
        counter = counter-4;
        visitCore(stat0.whileReplacement);
        // stat0.body.accept(this);
        return;
      }
      stat.acceptChildren(this);
    }
  }
  
  /***************************************************************************
   * Get formula processed in denoted method and program context.
   * @param text the string representation of the formula.
   * @param method the method symbol.
   * @param env the program environment.
   * @param folder the folder for generating type checking tasks.
   * @param error the error stream to which error messages are written.
   * @return the formula (null, if parsing or type-checking failed)
   **************************************************************************/
  private Formula getFormula(String text, ParamSymbol method, Environment env,
    ErrorStream error)
  {
    // parse formula
    Formula formula;
    try
    {
      int n = error.getErrorNumber();
      CharStream input = new ANTLRStringStream(text);
      SpecLangLexer lexer = new SpecLangLexer(input, error, (SourcePosition)null);
      CommonTokenStream tokens = new CommonTokenStream(lexer);
      SpecLangParser parser = new SpecLangParser(tokens, error, null, text);
      formula = parser.formulaEOF();
      int m = error.getErrorNumber();
      if (m > n)
      {
        error.explainError("could not parse formula");
        return null;
      }     
    }
    catch (RecognitionException e)
    {
      error.newError("could not parse formula");
      return null;
    }
    
    // type-check formula in denoted context
    ClassSymbol csymbol = method.getClassSymbol();
    LogicChecking checker = LogicChecking.createChecker(csymbol, env, error);
    error.setPosition(new StringSourcePosition(text, 0, 0)); 
    Formula preCondition = null;
    Name[] assignVars = null;
    Name[] throwExcepts = null;
    MethodSpec spec = method.getSpecification();
    if (spec != null)
    {
      preCondition = spec.getPreCondition();
      assignVars = spec.getVariables();
      throwExcepts = spec.getExceptions();
    }
    Type returnType = null;
    if (method instanceof MethodSymbol)
    {
      MethodSymbol method0 = (MethodSymbol)method;
      returnType = method0.getLogicReturnType();
    }
    StateType stateType = StateType.construct(returnType);
    checker.setStateContext(preCondition, stateType, assignVars, throwExcepts,
        method.getTCCs());
    boolean okay = checker.processFormula(formula);
    if (!okay) return null;
    
    // new type checking tasks may have been created
    topWindow.redrawExplorers();
    
    // formula was successfully processed
    return formula;
  }
}
