// ----------------------------------------------------------------------------
// SourcePositionClass.java
// A position in a source file.
// $Id: SourcePositionClass.java,v 1.3 2010/11/17 17:27:03 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for identifieric Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.InOut;

import fmrisc.ProgramExplorer.Syntax.*;

/**
 * 
 * A position in a source file.
 *
 */
public final class SourcePositionClass implements SourcePosition
{
  private Source source;  // the source file
  
  private int line;       // the line number
  private int column;     // the column number
  
  private Identifier ident;  // the registered identifier (may be null)
  
  /***************************************************************************
   * Construct a source position.
   * @param source the source file.
   * @param line the line number.
   * @param column the column number.
   **************************************************************************/
  public SourcePositionClass(Source source, int line, int column)
  {
    this.source = source;
    this.line = line;
    this.column = column;
    source.register(this);
  }
  
  /***************************************************************************
   * Get the source file.
   * @return the source file.
   **************************************************************************/
  public Source getSource()
  {
    return source;
  }
  
  /***************************************************************************
   * Get the line number of the position.
   * @return the line number.
   **************************************************************************/
  public int getLine()
  {
    return line;
  }
  
  /***************************************************************************
   * Get the column number of the position.
   * @return the column number.
   **************************************************************************/
  public int getColumn()
  {
    return column;
  }
  
  /****************************************************************************
   * Get offset of denoted position
   * @return the number of characters up to (excluding) the position denoted
   * by the line and column index.
   ***************************************************************************/
  public int getOffset()
  {
    return source.getOffset(line, column);
  }
  
  /**************************************************************************
   * Get number of characters from current position to end of line.
   * @return the number of characters.
   *************************************************************************/
  public int getLength()
  {
    return source.getLength(line, column);
  }
  
  /**************************************************************************
   * Get printable text representation of position.
   * @return the text representation.
   *************************************************************************/
  public String toString()
  {

    return source.getFile().getName() + ":" + (line+1) + ":" + (column+1);
  }
  
  /**************************************************************************
   * Get source line corresponding to position.
   * @return the source line of the position
   *************************************************************************/
  public String getSourceLine()
  {
    return source.getLine(line);
  }
  
  /***************************************************************************
   * Register identifier with denoted position.
   * @param ident the identifier to be registered.
   **************************************************************************/
  public void setIdentifier(Identifier ident)
  {
    this.ident = ident;
  }
  
  /***************************************************************************
   * Get identifier registered with denoted position.
   * @return the registered identifier (may be null).
   **************************************************************************/
  public Identifier getIdentifier()
  {
    return ident;
  }
}


