// ----------------------------------------------------------------------------
// Pre.java
// The judgement "se, Is, VS |- pre(C, Q) = P"
// $Id: Pre.java,v 1.15 2013/09/09 09:23:40 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Judgements;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Syntax.Program.Statement;
import fmrisc.ProgramExplorer.Semantics.*;

import java.util.*;

/**
 * 
 * The judgement "se, Is, VS |- pre(C, Q) = P"
 * 
 */
public final class Pre
{
  // the (base) name of a mathematical variable for the old/new state
  private final static String oldStateString = 
    OldState.oldStateKeyword.toLowerCase() + "_";
  
  // the context of the judgement processor
  private final StateType stype; // the state type
  
  /***************************************************************************
   * Create a judgement processor.
   * @param stype the current state type.
   **************************************************************************/
  public Pre(StateType stype)
  {
    this.stype = stype;
  }
  
  /***************************************************************************
   * Apply the judgement "se, Is, VS |- pre(C, Q) = P"
   * to compute the precondition P of command C with respect to postcondition Q
   * and to generate proving tasks as a side effect
   * (see Section 4.11 of "A Program Calculus", se, Is, Vs are not needed
   * because identifiers have already been resolved to symbols).
   * @param C the command C with judgement attached.
   * @param Q the formula Q
   * @return the precondition P of command C with respect to postcondition Q.
   **************************************************************************/
  public Formula derive(Statement C, Formula Q)
  {
    // determine specification of command
    StatJudgement spec = C.getJudgement();
    Formula F = spec.formula;
    
    // the program variables modified in the command
    int n = spec.variables.size();
    VariableSymbol[] symbols = new VariableSymbol[n];
    spec.variables.toArray(symbols);
    
    // unused names for mathematical variables substituting program variables
    Collection<String> used = UsedVariables.compute(F);
    UsedVariables.add(Q, used);
    String[] vars = NewVariables.newVariables(symbols, "", used);
    UsedVariables.add(vars, used);
    String stateVar = NewVariables.newVariable(oldStateString, used);
        
    // identifiers/symbols for the new variables
    // later ensure that generated references hold corresponding symbols
    TypedIdentifier[] tids = new TypedIdentifier[n+1];
    ValueSymbol[] symbols0 = new ValueSymbol[n+1];
    for (int i=0; i<n; i++)
    {
      Identifier ident = Identifier.construct(vars[i]);
      Type type = symbols[i].getLogicType();
      TypedIdentifier tid = TypedIdentifier.construct(ident, type);
      tids[i] = tid;
      symbols0[i] = ValueSymbol.construct(ident, type, tid);
    }
    Identifier ident = Identifier.construct(stateVar);
    TypedIdentifier tid = TypedIdentifier.construct(ident, stype);
    tids[n] = tid;
    symbols0[n] = ValueSymbol.construct(ident, stype, tid);
    
    // the substitution of Q
    Map<VariableSymbol, Substitution.Value> substQ = 
      new MapClass<VariableSymbol, Substitution.Value>();
    for (int i=0; i<n; i++)
    {
      Term term = Reference.construct(tids[i].getName());
      Substitution.Value value = new Substitution.Value(
          Substitution.Action.term, term, 
          Substitution.Action.term, term);
      substQ.put(symbols[i], value);
    }
    Term I_s = Reference.construct(tids[n].getName());
    Formula Q0 = (Formula)Substitution.perform(Q, I_s, I_s, substQ);
    
    // the substitution of F
    Map<VariableSymbol, Substitution.Value> substF = 
      new MapClass<VariableSymbol, Substitution.Value>();
    for (int i=0; i<n; i++)
    {
      Term term = Reference.construct(tids[i].getName());
      Substitution.Value value = new Substitution.Value(
          Substitution.Action.none, null, 
          Substitution.Action.term, term);
      substF.put(symbols[i], value);
    }
    Formula F0 = (Formula)Substitution.perform(F, null, I_s, substF);
    
    // construct result formula
    Term literal = OldState.construct();
    literal.setType(stype);
    return ForallFormula.construct(tids,
        ImpliesFormula.construct(
            AndFormula.construct(
                StateExecutes.construct(literal), F0),
            Q0));
  }
}
