// ----------------------------------------------------------------------------
// Simplification.java
// Simplify a specification formula to make it easier understandable.
// $Id: Simplification.java,v 1.41 2020/12/07 09:23:18 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General private License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General private License for more details.
//
// You should have received a copy of the GNU General private License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Judgements;

import java.util.*;
import java.math.BigInteger;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Semantics.*;

/**
 * 
 * Simplify a specification formula to make it easier understandable.
 *
 */
public final class Simplification extends ASTCloner
{
  // the set holding the names/occurrence numbers of free variables 
  // private Map<String,Integer> free = null;
  
  /***************************************************************************
   * Create simplifier.
   **************************************************************************/
  private Simplification()
  {
    // free = new MapClass<String,Integer>();
  }
  
  /**************************************************************************
   * Simplify the given formula. 
   * @param formula the formula to be simplified..
   * @return a simplified version of the formula.
   *************************************************************************/
  public static Formula perform(Formula formula)
  {
    int n = 2;
    Formula formula0 = formula;
    while (true)
    {
      // iterate until fixed point or iteration limit is reached
      n = n-1;
      Simplification simplifier = new Simplification();
      // System.out.println("*** " + n + "0: " + formula0);
      Formula formula1 = (Formula)simplifier.clone0(formula0);
      // System.out.println("*** " + n + "1: " + formula1);
      Formula formula2 = Simplification2.perform(formula1);
      // System.out.println("*** " + n + "2: " + formula2);
      if (n == 0) return formula2;
      String text0 = formula0.toString();
      String text2 = formula2.toString();
      if (text0.equals(text2)) return formula2;
      formula0 = formula2;
    }
  }
  
  /***************************************************************************
   * Simplify a formula with this simplifier.
   * @param formula a formula.
   * @return the simplified version of the formula.
   ***************************************************************************/
  private Formula simplify(Formula formula)
  {
    return (Formula)clone0(formula);
  }
  
  /***************************************************************************
   * Simplify a term with this simplifier.
   * @param formula a formula.
   * @return the simplified version of the formula.
   ***************************************************************************/
  private Term simplify(Term term)
  {
    if (term instanceof IfThenElseTerm)
    {
      IfThenElseTerm term0 = (IfThenElseTerm)term;
      return ifThenElseTerm(simplify(term0.getCondition()),
          simplify(term0.getThenBranch()), 
          simplify(term0.getElseBranch()),
          term.getType());
    }
    Term result = simplifyStatic((Term)clone0(term));
    return result;
  }

  // ------------------------------------------------------------------------
  //
  // the simplification rules for the individual kinds of formulas.
  //
  // -------------------------------------------------------------------------
  
  /***************************************************************************
   * The following simplifications are implemented:
   * 
   * o Some "quasi-normalization":
   * 
   * (A /\ B) /\ C   to A /\ (B /\ C) 
   * (A \/ B) \/ C   to A \/ (B \/ C) 
   * A imp (B imp C) to (A /\ B) imp C
   * 
   * This simplifies pretty-printing and further processing.
   * 
   * o Pulling negation inside
   * 
   * ~~A to A 
   * ~(A /\ B) to ~A \/ ~B 
   * ~(A \/ B) to ~A /\ ~B 
   * ~(A imp B) to A /\ ~B 
   * ~(A equiv B) to A xor B 
   * ~(A xor B) to A equiv B 
   * ~(FORALL x: A) to EXISTS x: ~A 
   * ~(EXISTS x: A) to FORALL x: ~A 
   * ~(LET ... in A) to LET ... in ~A) 
   * ~(x=a) to x /= a, ~(x/=a) to x=a 
   * ~(x leq a) to x gt a, ~(x lt a) to x geq a
   * ~(x leq a) to x lt a, ~(x gt a) to x leq a
   * 
   * As a consequence, negation can appear only at the level of 
   * atomic predicates and IF-THEN-ELSE (for which there is no 
   * convenient simplification).
   * 
   * o Removing superfluous quantifiers
   * 
   * (Q x: A) to A, if x not free in A
   * (let x=... in A) to A, if x not free in A
   * (FORALL x: x=T imp A) to A[T/x], if x occurs at most once in A
   * (EXISTS x: x=T /\ A) to A[T/x], if x occurs at most once in A
   * (let x=T in A) to A[T/x], if x occurs at most once in A
   * 
   * o Pulling quantifiers outside
   * 
   * Conjunction and disjunction:
   * 
   * (FORALL x: A) /\ B to (FORALL y: A[y/x] /\ B)
   *    y not free in A or B
   *    (if possible, choose y=x to avoid substitution)
   *    
   * The same for \/ instead of /\.
   * The same for EXISTS instead of FORALL.
   * The same for both sides switched.
   * 
   * Above rule is overridden by:
   *    
   * (FORALL x: A) /\ (FORALL y: B) to (FORALL z: A[z/x] /\ B[z/x])
   * (EXISTS x: A) \/ (EXISTS y: B) to (EXISTS z: A[z/x] \/ B[z/x])
   *    z not free in A or B
   *    (if possible, choose z in {x,y} to avoid one substitution)
   *    
   * Implication:
   * 
   * (FORALL x: A) imp B to (EXISTS y: A[y/x] imp B)
   * (EXISTS x: A) imp B to (FORALL y: A[y/x] imp B)
   * A imp (FORALL x: B) to (FORALL y: A imp B[y/x])
   * A imp (EXISTS x: B) to (EXISTS y: A imp B[y/x])
   * 
   * No convenient rules for equiv, XOR, IF.
   * 
   * Future ideas (not yet implemented):
   * 
   * o Sort defining (in)equalities in topological dependence order
   * 
   * y = ..x.. /\ x = ... to x = ... /\ y = ..x..
   *  
   **************************************************************************/
  
  // tags for expression types
  public enum Tag { AND, OR, IMPLIES, FORALL, EXISTS,
    ARRAYTERM, LAMBDATERM, LAMBDAFORMULA,
    EQUALS, NOTEQUALS, LESS, LESSEQUAL, GREATER, GREATEREQUAL, SIMILAR};
  
  /***************************************************************************
   * Construct a negation.
   * @param base base of the negation.
   * @return the negation.
   **************************************************************************/
  private Formula notFormula(Formula base)
  {
    if (base instanceof NotFormula)
    {
      // ~~A -> A
      NotFormula base0 = (NotFormula)base;
      return base0.getBase();
    }
    if (base instanceof AndFormula)
    {
      // ~(A /\ B) -> ~A \/ ~B
      AndFormula base0 = (AndFormula)base;
      Formula formula1 = notFormula(base0.getFirst());
      Formula formula2 = notFormula(base0.getSecond());
      return orFormula(formula1, formula2);
    }
    if (base instanceof OrFormula)
    {
      // ~(A \/ B) -> ~A /\ ~B
      OrFormula base0 = (OrFormula)base;
      Formula formula1 = notFormula(base0.getFirst());
      Formula formula2 = notFormula(base0.getSecond());
      return andFormula(formula1, formula2);
    }
    if (base instanceof ImpliesFormula)
    {
      // ~(A => B) -> A /\ ~B
      ImpliesFormula base0 = (ImpliesFormula)base;
      Formula formula1 = base0.getFirst();
      Formula formula2 = notFormula(base0.getSecond());
      return andFormula(formula1, formula2);
    }
    if (base instanceof EquivalentFormula)
    {
      // ~(A <=> B) -> A xor B
      EquivalentFormula base0 = (EquivalentFormula)base;
      Formula formula1 = base0.getFirst();
      Formula formula2 = base0.getSecond();
      return notEquivalentFormula(formula1, formula2);
    }
    if (base instanceof NotEquivalentFormula)
    {
      // ~(A xor B) -> A <=> B
      NotEquivalentFormula base0 = (NotEquivalentFormula)base;
      Formula formula1 = base0.getFirst();
      Formula formula2 = base0.getSecond();
      return equivalentFormula(formula1, formula2);
    }
    if (base instanceof ForallFormula)
    {
      // ~(FORALL x: A) -> EXISTS x: ~A
      ForallFormula base0 = (ForallFormula)base;
      Formula body = base0.getBody();
      return existsFormula(base0.getVariables(), notFormula(body));
    }
    if (base instanceof ExistsFormula)
    {
      // ~(EXISTS x: A) -> FORALL x: ~A
      ExistsFormula base0 = (ExistsFormula)base;
      Formula body = base0.getBody();
      return forallFormula(base0.getVariables(), notFormula(body));
    }
    if (base instanceof LetFormula)
    {
      // ~(LET ... in A) -> LET ... in ~A)
      LetFormula base0 = (LetFormula)base;
      Formula body = base0.getBody();
      return letFormula(base0.getDefinitions(), notFormula(body));
    }
    if (base instanceof BinaryAtomicFormula)
    {
      // ~(x=a) -> x /= a, ~(x/=a) -> x=a
      // ~(x<=a) -> x > a, ~(x<a) -> x >= a
      // ~(x>=a) -> x < a, ~(x>a) -> x <= a
      BinaryAtomicFormula base0 = (BinaryAtomicFormula)base;
      Term term1 = base0.getFirst();
      Term term2 = base0.getSecond();
      if (base0 instanceof EqualsFormula)
        return NotEqualsFormula.construct(term1, term2);
      if (base0 instanceof NotEqualsFormula)
        return EqualsFormula.construct(term1, term2);
      if (base0 instanceof GreaterEqualFormula)
        return LessFormula.construct(term1, term2);
      if (base0 instanceof GreaterFormula)
        return LessEqualFormula.construct(term1, term2);
      if (base0 instanceof LessEqualFormula)
        return GreaterFormula.construct(term1, term2);
      if (base0 instanceof LessFormula)
        return GreaterEqualFormula.construct(term1, term2);
      Main.internalError("unknown binary atomic formula");
    }
    // the default result
    return NotFormula.construct(base);
  }
  
  /***************************************************************************
   * Construct a conjunction.
   * @param formula1 the first subformula (may be null).
   * @param formula2 the second subformula (may be null)
   * @return the conjunction.
   **************************************************************************/
  private Formula andFormula(Formula formula1, Formula formula2)
  {
    // formula may be reduced to single formula
    if ((formula1 == null || formula1 instanceof True) && 
        (formula2 == null || formula2 instanceof True)) 
      return ExpressionLogic.booleanFormula(true);
    if (formula1 == null || formula1 instanceof True) return formula2;
    if (formula2 == null || formula2 instanceof True) return formula1;
    
    // pull quantifiers outside
    if (formula1 instanceof QuantifiedFormula ||
        formula2 instanceof QuantifiedFormula)
      return pullQuantifiers(Tag.AND, formula1, formula2);
    
    // (A /\ B) /\ C -> A /\ (B /\ C)
    Vector<Formula> formulas = new Vector<Formula>();
    collect(Tag.AND, formula1, formulas);
    collect(Tag.AND, formula2, formulas);
    
    // A /\ (if B then C else D) -> if B then A /\ C else A /\ D
    IfThenElseFormula cformula = null;
    Vector<Formula> formulas0 = new Vector<Formula>();
    for (Formula f : formulas)
    {
      if (cformula == null && f instanceof IfThenElseFormula)
        cformula = (IfThenElseFormula)f;
      else
        formulas0.add(f);
    }
    if (cformula != null)
    {
      int n = formulas0.size();
      formulas0.add(cformula.getThenBranch());
      Formula f1 = combine(Tag.AND, formulas0);
      formulas0.removeElementAt(n);
      formulas0.add(cformula.getElseBranch());
      Formula f2 = combine(Tag.AND, formulas0);
      return ifThenElseFormula(cformula.getCondition(), f1, f2);
    }
    
    // i > n /\ i <= n+1 --> i = n+1
    formulas = combineInequalities(formulas);
    return combine(Tag.AND, formulas);
  }
  
  /****************************************************************************
   * Construct if-then-else formula.
   * @param cond the condition.
   * @param thenBranch the then branch.
   * @param elseBranch the else branch.
   * @return the if-then-else formula.
   **************************************************************************/
  private Formula ifThenElseFormula(Formula cond, 
    Formula thenBranch, Formula elseBranch)
  {
    // poor man's logical simplification
    cond = simplify(cond);
    Formula f1 = thenBranch;
    Formula f2 = elseBranch;
    if (cond instanceof EqualsFormula || cond instanceof NotEqualsFormula)
    {
      boolean equals = cond instanceof EqualsFormula;
      BinaryAtomicFormula cond0 = (BinaryAtomicFormula)cond;
      Term c1 = cond0.getFirst();
      Term c2 = cond0.getSecond();
      if (!(c2 instanceof Constant))
      {
        Term c0 = c1; c1 = c2; c2 = c0;
      }
      if (c2 instanceof Constant)
      {
        if (equals)
          f1 = (Formula)ExpSubstitution.perform(f1, c1, c2);
        else
          f2 = (Formula)ExpSubstitution.perform(f2, c1, c2);
        if (c2 instanceof BitLiteral)
        {
          Type t = c2.getType();
          Constant c3 = null;
          if (c2 instanceof BitTrue)
            c3 = BitFalse.construct();
          else
            c3 = BitTrue.construct();
          c3.setType(t);
          if (equals)
            f2 = (Formula)ExpSubstitution.perform(f2, c1, c3);
          else
            f1 = (Formula)ExpSubstitution.perform(f1, c1, c3);
        }
        else if (c2 instanceof BooleanLiteral)
        {
          Type t = c2.getType();
          Constant c3 = null;
          if (c2 instanceof True)
            c3 = False.construct();
          else
            c3 = True.construct();
          c3.setType(t);
          if (equals)
            f2 = (Formula)ExpSubstitution.perform(f2, c1, c3);
          else
            f1 = (Formula)ExpSubstitution.perform(f1, c1, c3);
        }
      }
    }
    f1 = (Formula)ExpSubstitution.perform(f1, 
           cond, ExpressionLogic.booleanFormula(true));
    f2 = (Formula)ExpSubstitution.perform(f2, 
           cond, ExpressionLogic.booleanFormula(false));
    f1 = simplify(f1);
    f2 = simplify(f2);
    if (cond instanceof True) return f1;
    if (cond instanceof False) return f2;
    if (f1 instanceof True) return orFormula(cond, f2);
    if (f1 instanceof False) return andFormula(notFormula(cond), f2);
    if (f2 instanceof True) return impliesFormula(cond, f1);
    if (f2 instanceof False) return andFormula(cond, f1);
    return IfThenElseFormula.construct(cond, f1, f2);
  }
  
  /****************************************************************************
   * Construct if-then-else term.
   * @param cond the condition.
   * @param thenBranch the then branch.
   * @param elseBranch the else branch.
   * @param type the type of the term.
   * @return the if-then-else term.
   **************************************************************************/
  private Term ifThenElseTerm(Formula cond, 
    Term thenBranch, Term elseBranch, Type type)
  {
    // poor man's logical simplification
    cond = simplify(cond);
    Term t1 = (Term)ExpSubstitution.perform(thenBranch, 
        cond, ExpressionLogic.booleanFormula(true));
    t1 = simplify(t1);
    Term t2 = (Term)ExpSubstitution.perform(elseBranch, 
        cond, ExpressionLogic.booleanFormula(false));
    t2 = simplify(t2);
    if (cond instanceof True) return t1;
    if (cond instanceof False) return t2;
    Term result = IfThenElseTerm.construct(cond, t1, t2);
    result.setType(type);
    return result;
  }
  
  /***************************************************************************
   * Construct a disjunction.
   * @param formula1 the first subformula.
   * @param formula2 the second subformula.
   * @return the disjunction.
   **************************************************************************/
  private Formula orFormula(Formula formula1, Formula formula2)
  {
    // formula may be reduced to single formula
    if ((formula1 == null || formula1 instanceof False) && 
        (formula2 == null || formula2 instanceof False)) 
      return ExpressionLogic.booleanFormula(false);
    if (formula1 == null || formula1 instanceof False) return formula2;
    if (formula2 == null || formula2 instanceof False) return formula1;
    if (formula1 instanceof True || formula2 instanceof True)
      return ExpressionLogic.booleanFormula(true);
    
    // pull quantifiers outside
    if (formula1 instanceof QuantifiedFormula ||
        formula2 instanceof QuantifiedFormula)
      return pullQuantifiers(Tag.OR, formula1, formula2);
    
    // (A \/ B) \/ C -> A \/ (B \/ C) 
    return associativeFormula(Tag.OR, formula1, formula2);
  }
  
  /***************************************************************************
   * Construct an implication.
   * @param formula1 the first subformula.
   * @param formula2 the second subformula.
   * @return the implication.
   **************************************************************************/
  private Formula impliesFormula(Formula formula1, Formula formula2)
  {
    if ((formula1 == null || formula1 instanceof True) && 
        (formula2 == null || formula2 instanceof False)) 
      return ExpressionLogic.booleanFormula(false);
    if (formula1 == null || formula1 instanceof True || formula2 instanceof True) 
      return formula2;
    if (formula2 == null || formula2 instanceof False) return notFormula(formula1);
    if (formula1 instanceof False)
      return ExpressionLogic.booleanFormula(true);
    
    // simplify ... /\ A /\ ... => ... \/ A \/ ...
    Vector<Formula> formulas1 = new Vector<Formula>();
    Vector<Formula> formulas2 = new Vector<Formula>();
    collect(Tag.AND, formula1, formulas1);
    collect(Tag.OR,  formula2, formulas2);
    Set<String> texts1 = new SetClass<String>();
    for (Formula f1 : formulas1)
      texts1.add(f1.toString());
    for (Formula f2 : formulas2)
    {
      if (texts1.contains(f2.toString()))
        return ExpressionLogic.booleanFormula(true);
    }
    
    // simplify ... /\ A /\ ... => ... /\ A /\ ...
    Vector<Formula> formulas3 = new Vector<Formula>();
    Vector<Formula> formulas4 = new Vector<Formula>();
    collect(Tag.AND, formula2, formulas3);
    for (Formula f3 : formulas3)
    {
       if (!texts1.contains(f3.toString())) 
         formulas4.add(f3);
    }
    if (formulas4.isEmpty())
      return ExpressionLogic.booleanFormula(true);
    formula2 = combine(Tag.AND, formulas4);
    
    // pull quantifiers outside
    if (formula1 instanceof QuantifiedFormula ||
        formula2 instanceof QuantifiedFormula)
      return pullQuantifiers(Tag.IMPLIES, formula1, formula2);
    
    // A => (B => C) -> (A /\ B) => C
    if (!(formula2 instanceof ImpliesFormula)) 
      return ImpliesFormula.construct(formula1, formula2);
    ImpliesFormula implies = (ImpliesFormula)formula2;
    return impliesFormula(
        andFormula(formula1, implies.getFirst()),
        implies.getSecond());
  }
  
  /***************************************************************************
   * Construct an equivalence.
   * @param formula1 the first subformula.
   * @param formula2 the second subformula.
   * @return the equivalence.
   **************************************************************************/
  private Formula equivalentFormula(Formula formula1, Formula formula2)
  {
    return EquivalentFormula.construct(formula1, formula2);
  }
  
  /***************************************************************************
   * Construct a negated equivalence.
   * @param formula1 the first subformula.
   * @param formula2 the second subformula.
   * @return the negated equivalence.
   **************************************************************************/
  private Formula notEquivalentFormula(Formula formula1, Formula formula2)
  {
    return NotEquivalentFormula.construct(formula1, formula2);
  }
  
  /***************************************************************************
   * Construct a universally quantified formula
   * @param vars the variables.
   * @param body the formula body.
   * @return the quantified formula.
   **************************************************************************/
  private Formula forallFormula(TypedIdentifier[] variables, Formula body)
  {
    // too costly
    // return (Formula)quantifiedExpression(Tag.FORALL, variables, body);
    return (Formula)mergeQuantifiedVariables(Tag.FORALL, variables, simplify(body));
  }
  
  /***************************************************************************
   * Construct an existentially quantified formula
   * @param vars the variables.
   * @param body the formula body.
   * @return the quantified formula.
   **************************************************************************/
  private Formula existsFormula(TypedIdentifier[] variables, Formula body)
  {
    // too costly
    // return (Formula)quantifiedExpression(Tag.EXISTS, variables, body);
    return (Formula)mergeQuantifiedVariables(Tag.EXISTS, variables, simplify(body));
  }

  /***************************************************************************
   * Construct a let formula
   * @param defs the definitions.
   * @param body the formula body.
   * @return the let formula.
   **************************************************************************/
  private Formula letFormula(ValueDefinition[] defs, Formula body)
  {
    return LetFormula.construct(defs, body);
  }
    
  /***************************************************************************
   * Construct a let term
   * @param defs the definitions.
   * @param body the term body.
   * @return the let term
   **************************************************************************/
  private Term letTerm(ValueDefinition[] defs, Term body)
  {
    return LetTerm.construct(defs, body);
  }
  
  /***************************************************************************
   * Simplify binary atomic formulas.
   * @param tag the tag identifying the formula.
   * @param term0 the left side of the formula.
   * @param term1 the right side of the formula.
   * @return the simplified formula.
   **************************************************************************/
  private static Formula binaryAtomicFormula(Tag tag, Term term0, Term term1)
  {
    if (LogicChecking.isIntegerType(term0.getType()) &&
        LogicChecking.isIntegerType(term1.getType()))
    {
      Vector<Term> plus0 = new Vector<Term>();
      Vector<Term> minus0 = new Vector<Term>();
      Vector<Term> plus1 = new Vector<Term>();
      Vector<Term> minus1 = new Vector<Term>();
      int constant0 = summands(term0, plus0, minus0);
      int constant1 = summands(term1, plus1, minus1);
      if (plus0.isEmpty() && minus0.isEmpty())
      {
        term0 = sum(plus0, minus0, constant0-constant1);
        term1 = sum(plus1, minus1, 0);
      }
      else if (plus1.isEmpty() && minus1.isEmpty())
      {
        term0 = sum(plus0, minus0, 0);
        term1 = sum(plus1, minus1, constant1-constant0);
      }
      else
      {
        int const0 = constant0 >= constant1 ? constant0-constant1 : 0;
        int const1 = constant0 >= constant1 ? 0 : constant1-constant0;
        term0 = sum(plus0, minus0, const0);
        term1 = sum(plus1, minus1, const1);
      }
    }
    if (term0 instanceof IntLiteral && term1 instanceof IntLiteral)
    {
      int value0 = Integer.parseInt(((IntLiteral)term0).getName());
      int value1 = Integer.parseInt(((IntLiteral)term1).getName());
      switch (tag)
      {
      case EQUALS: return ExpressionLogic.booleanFormula(value0 == value1);
      case NOTEQUALS: return ExpressionLogic.booleanFormula(value0 != value1);
      case LESS: return ExpressionLogic.booleanFormula(value0 < value1);
      case GREATER: return ExpressionLogic.booleanFormula(value0 > value1);
      case LESSEQUAL: return ExpressionLogic.booleanFormula(value0 <= value1);
      case GREATEREQUAL: return ExpressionLogic.booleanFormula(value0 >= value1);
      default: 
        Main.internalError("unknown binary atomic formula tag: " + tag);
        return null;
      }
    }
    boolean equals = term0.toString().equals(term1.toString());
    if (equals)
    {
      switch (tag)
      {
      case EQUALS: return ExpressionLogic.booleanFormula(true);
      case NOTEQUALS: return ExpressionLogic.booleanFormula(false);
      case LESS: return ExpressionLogic.booleanFormula(false);
      case GREATER: return ExpressionLogic.booleanFormula(false);
      case LESSEQUAL: return ExpressionLogic.booleanFormula(true);
      case GREATEREQUAL: return ExpressionLogic.booleanFormula(true);
      default: 
        Main.internalError("unknown binary atomic formula tag: " + tag);
        return null;
      }
    }
    switch (tag)
    {
    case EQUALS: return EqualsFormula.construct(term0, term1);
    case NOTEQUALS: return NotEqualsFormula.construct(term0, term1);
    case LESS: return LessFormula.construct(term0, term1);
    case GREATER: return GreaterFormula.construct(term0, term1);
    case LESSEQUAL: return LessEqualFormula.construct(term0, term1);
    case GREATEREQUAL: return GreaterEqualFormula.construct(term0, term1);
    case SIMILAR: return SimilarFormula.construct(term0, term1);
    default: 
      Main.internalError("unknown binary atomic formula tag: " + tag);
      return null;
    }
  }
  
  // ------------------------------------------------------------------------
  //
  // the visitors
  //
  // -------------------------------------------------------------------------
  
  /***************************************************************************
   * Process binary atomic formula.
   * @param formula the formula to be simplified.
   **************************************************************************/
  public void visit(EqualsFormula formula)
  {
    Term term0 = simplify(formula.getFirst());
    Term term1 = simplify(formula.getSecond());
    if (term0 instanceof Constant && term1 instanceof Constant)
    {
      if (term0.toString().equals(term1.toString()))
        result = True.construct();
      else
        result = False.construct();
      return;
    }
    result = binaryAtomicFormula(Tag.EQUALS, term0, term1);
  }
  
  /***************************************************************************
   * Process binary atomic formula.
   * @param formula the formula to be simplified.
   **************************************************************************/
  public void visit(GreaterEqualFormula formula)
  {
    Term term0 = simplify(formula.getFirst());
    Term term1 = simplify(formula.getSecond());
    result = binaryAtomicFormula(Tag.GREATEREQUAL, term0, term1);
  }
  
  /***************************************************************************
   * Process binary atomic formula.
   * @param formula the formula to be simplified.
   **************************************************************************/
  public void visit(GreaterFormula formula)
  {
    Term term0 = simplify(formula.getFirst());
    Term term1 = simplify(formula.getSecond());
    result = binaryAtomicFormula(Tag.GREATER, term0, term1);
  }
  
  /***************************************************************************
   * Process binary atomic formula.
   * @param formula the formula to be simplified.
   **************************************************************************/
  public void visit(LessEqualFormula formula)
  {
    Term term0 = simplify(formula.getFirst());
    Term term1 = simplify(formula.getSecond());
    result = binaryAtomicFormula(Tag.LESSEQUAL, term0, term1);
  }
  
  /***************************************************************************
   * Process binary atomic formula.
   * @param formula the formula to be simplified.
   **************************************************************************/
  public void visit(LessFormula formula)
  {
    Term term0 = simplify(formula.getFirst());
    Term term1 = simplify(formula.getSecond());
    result = binaryAtomicFormula(Tag.LESS, term0, term1);
  }
  
  /***************************************************************************
   * Process binary atomic formula.
   * @param formula the formula to be simplified.
   **************************************************************************/
  public void visit(NotEqualsFormula formula)
  {
    Term term0 = simplify(formula.getFirst());
    Term term1 = simplify(formula.getSecond());
    if (term0 instanceof Constant && term1 instanceof Constant)
    {
      if (term0.toString().equals(term1.toString()))
        result = False.construct();
      else
        result = True.construct();
      return;
    }
    result = binaryAtomicFormula(Tag.NOTEQUALS, term0, term1);
  }
  
  /***************************************************************************
   * Process binary atomic formula.
   * @param formula the formula to be simplified.
   **************************************************************************/
  public void visit(SimilarFormula formula)
  {
    Term term0 = simplify(formula.getFirst());
    Term term1 = simplify(formula.getSecond());
    result = binaryAtomicFormula(Tag.SIMILAR, term0, term1);
  }
  
  /***************************************************************************
   * Process negation.
   * @param formula the formula to be simplified.
   **************************************************************************/
  public void visit(NotFormula formula)
  {
    Formula base = simplify(formula.getBase());
    result = notFormula(base);
  }
  
  /***************************************************************************
   * Process conjunction.
   * @param formula the formula to be simplified.
   **************************************************************************/
  public void visit(AndFormula formula)
  {
    Formula formula1 = simplify(formula.getFirst());
    Formula formula2 = simplify(formula.getSecond());
    result = andFormula(formula1, formula2);
  }
  
  /***************************************************************************
   * Process disjunction.
   * @param formula the formula to be simplified.
   **************************************************************************/
  public void visit(OrFormula formula)
  {
    Formula formula1 = simplify(formula.getFirst());
    Formula formula2 = simplify(formula.getSecond());
    result = orFormula(formula1, formula2);
  }
  
  /***************************************************************************
   * Process implication.
   * @param formula the formula to be simplified.
   **************************************************************************/
  public void visit(ImpliesFormula formula)
  {
    Formula formula1 = simplify(formula.getFirst());
    Formula formula2 = simplify(formula.getSecond());
    result = impliesFormula(formula1, formula2);
  }
  
  /***************************************************************************
   * Process disjunction.
   * @param formula the formula to be simplified.
   **************************************************************************/
  public void visit(IfThenElseFormula formula)
  {
    // simplification is performed inside
    Formula cond = formula.getCondition();
    Formula formula1 = formula.getThenBranch();
    Formula formula2 = formula.getElseBranch();
    result = ifThenElseFormula(cond, formula1, formula2);
  }
  
  /****************************************************************************
   * Process quantified expression
   * @param exp a quantified expression.
   ***************************************************************************/
  public void visit(ForallFormula exp)
  {    
    result = quantifiedExpression(Tag.FORALL, exp.getVariables(), exp.getBody());
  }
  
  /****************************************************************************
   * Process quantified expression
   * @param exp a quantified expression.
   ***************************************************************************/
  public void visit(ExistsFormula exp)
  {    
    result = quantifiedExpression(Tag.EXISTS, exp.getVariables(), exp.getBody());
  }
  
  /****************************************************************************
   * Process quantified expression
   * @param exp a quantified expression.
   ***************************************************************************/
  public void visit(ArrayTerm exp)
  {    
    result = quantifiedExpression(Tag.ARRAYTERM, exp.getVariables(), exp.getBody());
  }
  
  /****************************************************************************
   * Process quantified expression
   * @param exp a quantified expression.
   ***************************************************************************/
  public void visit(LambdaTerm exp)
  {    
    result = quantifiedExpression(Tag.LAMBDATERM, exp.getVariables(), exp.getBody());
  }
  
  /****************************************************************************
   * Process quantified expression
   * @param exp a quantified expression.
   ***************************************************************************/
  public void visit(LambdaFormula exp)
  {    
    result = quantifiedExpression(Tag.LAMBDAFORMULA, exp.getVariables(), exp.getBody());
  }
  
  /****************************************************************************
   * Process quantified expression
   * @param exp a quantified expression.
   ***************************************************************************/
  public void visit(LetExpression exp)
  { 
    Definition[] defs = exp.getDefinitions();
    Expression body = exp.getBody();
    
    // recursion base
    int n = defs.length;
    if (n == 0)
    {
      // simplify body
      if (body instanceof Formula)
        result = simplify((Formula)body);
      else if (body instanceof Term)
        result = simplify((Term)body);
      else
        Main.internalError("unknown expression");
      return;
    }
    
    // simplify let expression for all but first definition
    Expression body0;
    if (n == 1)
    {
      if (body instanceof Formula)
        body0 = simplify((Formula)body);
      else if (body instanceof Term)
        body0 = simplify((Term)body);
      else
      {
        Main.internalError("unknown expression");
        body0 = null;
      }
    }
    else if (exp instanceof LetFormula)
    {
      LetFormula exp0 = (LetFormula)exp;
      ValueDefinition[] defs0 = exp0.getDefinitions();
      ValueDefinition[] defs1 = new ValueDefinition[n-1];
      for (int i=0; i<n-1; i++) defs1[i] = defs0[i+1];
      Formula exp1 = letFormula(defs1, exp0.getBody());
      body0 = simplify(exp1);
    }
    else if (exp instanceof LetTerm)
    {
      LetTerm exp0 = (LetTerm)exp;
      ValueDefinition[] defs0 = exp0.getDefinitions();
      ValueDefinition[] defs1 = new ValueDefinition[n-1];
      for (int i=0; i<n-1; i++) defs1[i] = defs0[i+1];
      Term exp1 = letTerm(defs1, exp0.getBody());
      body0 = simplify(exp1);
    }
    else
    {
      Main.internalError("unknown let expression");
      body0 = null;
    }
    
    // collect free variables of expression body
    Map<String, Integer> f = new Counter(body0).getFree(); 
    
    // determine how often definition was referenced
    ValueDefinition def = (ValueDefinition)defs[0];
    Identifier id = def.getName();
    String name = id.getString();
    int counter = getCounter(f, name);
    
    // replace let expression by body
    if (counter == 0)
    {
      result = body0;
      return;
    }
    
    // simplify defined value
    Term value0 = simplify(def.getValue());
    
    // inline definition and simplify again
    if (counter == 1)
    {
      Map<String, Term> map = new MapClass<String, Term>();
      map.put(name, value0);
      body0 = (Expression)RefSubstitution.perform(map, body0);
      result = clone0(body0);
      return;
    }
    
    // construct resulting let expression
    ValueDefinition def0 = ValueDefinitionClass.construct(id, value0);
    if (body0 instanceof LetFormula)
    {
      LetFormula body1 = (LetFormula)body0;
      ValueDefinition[] defs1 = body1.getDefinitions();
      int n1 = defs1.length;
      ValueDefinition[] defs0 = new ValueDefinition[n1+1];
      defs0[0] = def0;
      for (int i=0; i<n1; i++) defs0[i+1] = defs1[i];
      result = letFormula(defs0, body1.getBody());
      return;
    }
    if (body0 instanceof LetTerm)
    {
      LetTerm body1 = (LetTerm)body0;
      ValueDefinition[] defs1 = body1.getDefinitions();
      int n1 = defs1.length;
      ValueDefinition[] defs0 = new ValueDefinition[n1+1];
      defs0[0] = def0;
      for (int i=0; i<n1; i++) defs0[i+1] = defs1[i];
      result = letTerm(defs0, body1.getBody());
      return;
    }
    ValueDefinition[] defs0 = new ValueDefinition[]{ def0 };
    if (body0 instanceof Formula)
      result = letFormula(defs0, (Formula)body0);
    else if (body0 instanceof Term)
      result = letTerm(defs0, (Term)body0);      
    else
      Main.internalError("unknown expression");
  }
  
  /****************************************************************************
   * Process quantified expression
   * @param exp a quantified expression.
   ***************************************************************************/
  public void visit(LetTerm exp)
  { 
    visit((LetExpression)exp);
  }
  
  /****************************************************************************
   * Process quantified expression
   * @param exp a quantified expression.
   ***************************************************************************/
  public void visit(LetFormula exp)
  { 
    visit((LetExpression)exp);
  }
  
  // ------------------------------------------------------------------------
  //
  // Auxiliaries
  //
  // -------------------------------------------------------------------------
    
  /***************************************************************************
   * Construct an associative formula.
   * @param tag the tag indicating the type of the formula
   * @param formula1 the first subformula.
   * @param formula2 the second subformula.
   * @return the simplification result.
   **************************************************************************/
  private static Formula associativeFormula(Tag tag,
    Formula formula1, Formula formula2)
  {
    Vector<Formula> formulas = new Vector<Formula>();
    collect(tag, formula1, formulas);
    collect(tag, formula2, formulas);
    return combine(tag, formulas);
  }
  
  /****************************************************************************
   * Decompose formula into subformulas if it is of denoted formula type.
   * @param tag a tag indicating the type of the formula.
   * @param formula a formula to be decomposed.
   * @param result the result of the decomposition.
   ***************************************************************************/
  public static void collect(Tag tag, Formula formula, 
    Collection<Formula> result)
  {
    switch (tag)
    {
    case AND:
    {
      if (!(formula instanceof AndFormula))
      {
        result.add(formula);
        return;
      }
      break;
    }
    case OR:
    {
      if (!(formula instanceof OrFormula))
      {
        result.add(formula);
        return;
      }
      break;
    }
    case IMPLIES:
    {
      if (!(formula instanceof ImpliesFormula))
      {
        result.add(formula);
        return;
      }
      break;
    }
    default:
    {
      Main.internalError("unknown tag");
    }
    }
    BinaryFormula formula0 = (BinaryFormula)formula;
    collect(tag, formula0.getFirst(), result);
    collect(tag, formula0.getSecond(), result);
  }
  
  /***************************************************************************
   * Combine formulas according to formula tag.
   * @param tag the tag indicating the type of the formula
   * @param formulas the formulas
   **************************************************************************/
  private static Formula combine(Tag tag, Vector<Formula> formulas)
  {
    // avoid duplicate/redundant formulas
    Vector<Formula> formulas0 = new Vector<Formula>();
    Set<String> formulas1 = new SetClass<String>();
    for (Formula f : formulas)
    {
      if (f instanceof True)
      {
        if (tag == Tag.AND) continue;
        if (tag == Tag.OR) return f;
      }
      else if (f instanceof False)
      {
        if (tag == Tag.AND) return f;
        if (tag == Tag.OR) continue;
      }
      String f1 = f.toString();
      if (formulas1.contains(f1)) continue;
      formulas1.add(f1);
      formulas0.add(f);
    }
          
    // construct result right-to-left for pretty-printing
    Formula result = null;
    int n = formulas0.size();
    for (int i=n-1;  i>=0; i--)
    {
      Formula f = formulas0.get(i);
      if (result == null)
        result = f;
      else if (tag == Tag.AND)
        result = AndFormula.construct(f, result);
      else if (tag == Tag.OR)
        result = OrFormula.construct(f, result);
      else
        Main.internalError("unknown tag");
    }
    return result;
  }
  
  /***************************************************************************
   * Remove superfluous variables from quantified expression.
   * @param tag the tag indicating the type of the expression.
   * @param vars the quantified variables.
   * @param body the body of the expression.
   * @return the slim-lined variant of the expression.
   **************************************************************************/
  private Expression quantifiedExpression(Tag tag,
    TypedIdentifier[] vars, Expression body)
  { 
    // simplify expression body
    Simplification simplifier = new Simplification();
    body.accept(simplifier);
    Expression body0 = (Expression)simplifier.result;
    
    // merge quantified variables and push body quantifiers inside
    QuantifiedExpression exp = mergeQuantifiedVariables(tag, vars, body0);
    vars = exp.getVariables();
    body0 = Simplification2.perform(exp.getBody());
        
    // collect free variables of expression body
    Map<String,Integer> f = new Counter(body0).getFree(); // simplifier.free;
    Set<String> evars = new SetClass<String>();
    
    // determine referenced bound variables and remove them
    Vector<TypedIdentifier> vars0 = new Vector<TypedIdentifier>();
    int n = vars.length;
    for (int i=0; i<n; i++) vars0.add(vars[i]);
    while (true)
    {       
      // use information to remove uniquely defined variable
      boolean extracted = false;
      Map<String, Term> map = new MapClass<String, Term>();
      for (TypedIdentifier var : vars0)
      {
        String name = var.getName().getString();
        if (evars.contains(name)) continue;
        int counter = getCounter(f, name);
        if (counter == 0) continue;
        if (tag == Tag.FORALL || tag == Tag.EXISTS)
        {
          TermFormula tf = extractValue((Formula)body0, name, tag == Tag.FORALL);
          if (tf != null)
          {
            Term value = tf.term;
            if (counter <= 2 ||
                value instanceof Reference ||
                value instanceof ProgramVariable ||
                value instanceof Constant ||
                value instanceof StateValue ||
                value.getType() instanceof StateType)
            {
              extracted = true;
              evars.add(name);
              map.put(name, value);
              body0 = tf.formula;
              // expression is tautologically true
              if (body0 == null) return ExpressionLogic.booleanFormula(true);
              break;
            }
          }
        }
      }

      // if no more variable could be extracted, we are done
      if (!extracted) break;
      
      // perform substitution and repeat simplification
      body0 = (Expression)RefSubstitution.perform(map, body0);
      //simplifier = new Simplification();
      //body0.accept(simplifier);
      //body0 = (Expression)simplifier.result;
      //body0 = Simplification2.perform(body0);
      f = new Counter(body0).getFree(); //simplifier.free;
    }
    
    // pull body quantifiers outside
    //simplifier = new Simplification();
    //body0.accept(simplifier);
    //body0 = (Expression)simplifier.result;
    f = new Counter(body0).getFree(); 
    
    // remove unreferenced variables
    Vector<TypedIdentifier> vars1 = new Vector<TypedIdentifier>();
    for (TypedIdentifier var : vars0)
    {
      String name = var.getName().getString();
      if (f.containsKey(name))
      {
        vars1.add(var);
        f.remove(name);
      }
    }
    vars0 = vars1;
    
    // add remaining free variables to collection
    // add(free, f);
    
    // (Q x: A) -> A, if x not free in A
    int n0 = vars0.size();
    if (n0 == 0) return body0;
    
    // determine bound variables and construct result
    TypedIdentifier[] vars2 = new TypedIdentifier[n0];
    vars0.toArray(vars2);
    return constructQuantifiedExpression(tag, vars2, body0);
  }
  
  /***************************************************************************
   * Get number of occurrences of key in map.
   * @param map the map.
   * @param key the key.
   * @return the number of occurrences (0, if none)
   **************************************************************************/
  private static int getCounter(Map<String, Integer> map, String key)
  {
    Integer result = map.get(key);
    if (result == null) return 0;
    return result;
  }
  
  /***************************************************************************
   * Add to map one instance of key.
   * @param map the map.
   * @param key the key.
   **************************************************************************/
  private static void add(Map<String, Integer> map, String key)
  {
    add(map, key, 1);
  }
  
  /***************************************************************************
   * Add to map n instances of key.
   * @param map the map.
   * @param key the key.
   * @param n the number of instances
   **************************************************************************/
  private static void add(Map<String, Integer> map, String key, int n)
  {
    Integer counter = map.get(key);
    if (counter == null)
      map.put(key, n);
    else
      map.put(key, counter+n);
  }
  
  /***************************************************************************
   * Add to map all instances from another map.
   * @param to the map to which to add.
   * @param from the map from which to add.
   **************************************************************************/
  private static void add(Map<String, Integer> to, Map<String, Integer> from)
  {
    for (Map.Entry<String, Integer> entry : from.entrySet())
    {
      String key = entry.getKey();
      int counter = entry.getValue();
      add(to, key, counter);
    }
  }
  
  // a pair of a term and a formula
  private static class TermFormula
  {
    public final Term term;
    public final Formula formula;
    public TermFormula(Term term, Formula formula)
    {
      this.term = term;
      this.formula = formula;
    }
  }
    
  /****************************************************************************
   * Extract disjunct x /= T or conjunct x = T from formula.
   * @param formula the formula.
   * @param name the name of the variable (x).
   * @param disjunct true iff to look for disjunct.
   * @return either null or the pair of T and the formula 
   *         without the disjunction/conjunct.
   ***************************************************************************/
  private TermFormula extractValue(Formula formula, String name,
    boolean disjunct)
  {
    // atomic case
    if (disjunct && formula instanceof NotEqualsFormula ||
        (!disjunct && formula instanceof EqualsFormula))
    {
      BinaryAtomicFormula formula0 = (BinaryAtomicFormula)formula;
      Term first = formula0.getFirst();
      Term second = formula0.getSecond();
      String name1 = getName(first);
      if (name.equals(name1)) return new TermFormula(second, null);
      String name2 = getName(second);
      if (name.equals(name2)) return new TermFormula(first, null);
      Vector<Term> plus0 = new Vector<Term>();
      Vector<Term> plus1 = new Vector<Term>();
      Vector<Term> minus0 = new Vector<Term>();
      Vector<Term> minus1 = new Vector<Term>();
      int const0 = summands(first, plus0, minus0);
      int const1 = summands(second, plus1, minus1);
      int pos = getPos(plus0, name);
      if (pos != -1) 
      { 
        add(minus1, plus0, pos);
        plus1.addAll(minus0);
        return new TermFormula(sum(plus1, minus1, const1-const0), null);
      }
      pos = getPos(minus0, name);
      if (pos != -1)
      {
        add(plus1, minus0, pos);
        minus1.addAll(plus0);
        return new TermFormula(sum(minus1, plus1, const0-const1), null);
      }
      pos = getPos(plus1, name);
      if (pos != -1) 
      { 
        add(minus0, plus1, pos);
        plus0.addAll(minus1);
        return new TermFormula(sum(plus0, minus0, const0-const1), null);
      }
      pos = getPos(minus1, name);
      if (pos != -1)
      {
        add(plus0, minus1, pos);
        minus0.addAll(plus1);
        return new TermFormula(sum(minus0, plus0, const1-const0), null);
      }
      return null;
    }
    
    // we are not investigating negations
    // since we assume that negations have been pushed inside already
    
    // are only interested in binary formulas
    if (!(formula instanceof BinaryFormula)) return null;
    BinaryFormula formula0 = (BinaryFormula)formula;
    Formula first = formula0.getFirst();
    Formula second = formula0.getSecond();
    
    if (disjunct && formula instanceof OrFormula)
    {
      TermFormula tf = extractValue(first, name, true);
      if (tf != null)
        return new TermFormula(tf.term, orFormula(tf.formula, second));
      tf = extractValue(second, name, true);
      if (tf != null)
        return new TermFormula(tf.term, orFormula(first, tf.formula));
      return null;
    }
    if (disjunct && formula instanceof ImpliesFormula)
    {
      TermFormula tf = extractValue(first, name, false);
      if (tf != null)
        return new TermFormula(tf.term, impliesFormula(tf.formula, second));
      tf = extractValue(second, name, true);
      if (tf != null)
        return new TermFormula(tf.term, impliesFormula(first, tf.formula));
      return null;
    }
    if (!disjunct && formula instanceof AndFormula)
    {
      TermFormula tf = extractValue(first, name, false);
      if (tf != null)
        return new TermFormula(tf.term, andFormula(tf.formula, second));
      tf = extractValue(second, name, false);
      if (tf != null)
        return new TermFormula(tf.term, andFormula(first, tf.formula));
      return null;
    }
    
    // base case
    return null;
  }
  
  /****************************************************************************
   * Get name of plain reference.
   * @param term a term.
   * @return the name of the plain reference denoted by term
   *         (null if term does not denote such a reference).
   ***************************************************************************/
  private static String getName(Term term)
  {
    if (!(term instanceof Reference)) return null;
    Reference term0 = (Reference)term;
    Name name = term0.getName();
    if (name.idents.length != 1) return null;
    return name.idents[0].getString();
  }
  
  /****************************************************************************
   * Get position of plain reference in vector of terms.
   * @param terms a vector of terms.
   * @param name a name.
   * @return the position of the reference denoted by the name in the vector
   *         (-1 if vector does not contain such a reference).
   ***************************************************************************/
  private static int getPos(Vector<Term> terms, String name)
  {
    int n = terms.size();
    for (int i=0; i<n; i++)
    {
      String name0 = getName(terms.get(i));
      if (name0 == null) continue;
      if (name0.equals(name)) return i;
    }
    return -1;
  }
  
  /***************************************************************************
   * Add to vector output all elements of vector input except one.
   * @param output the output vector.
   * @param input the input vector.
   * @param except the position of the exception.
   **************************************************************************/
  private static void add(Vector<Term> output, Vector<Term> input, int except)
  {
    int n = input.size();
    for (int i=0; i<n; i++)
    {
      if (i == except) continue;
      output.add(input.get(i));
    }
  }
    
  /***************************************************************************
   * Pull quantifiers out of formulas combined by connective indicated by tag
   * @param tag the tag indicating the connective OP (AND, OR, IMPLIES)
   * @param formula1 the first subformula
   * @param formula2 the second subformula
   * @return a formula that is equivalent to formula1 OP formula2
   *         but with quantifiers pulled outside wherever possible
   **************************************************************************/
  private Formula pullQuantifiers(Tag tag, Formula formula1, Formula formula2)
  {
    // we assume that formulas have been processed such that
    // all quantified variables occur among the free variables
    
    // (FORALL x: A) /\ (FORALL y: B) -> (FORALL z: A[z/x] /\ B[z/x])
    // (EXISTS x: A) \/ (EXISTS y: B) -> (EXISTS z: A[z/x] \/ B[z/x])
    if ((tag == Tag.AND && formula1 instanceof ForallFormula
                       && formula2 instanceof ForallFormula) ||
        (tag == Tag.OR && formula1 instanceof ExistsFormula
                        && formula2 instanceof ExistsFormula))
    {
      // get components
      QuantifiedFormula f1 = (QuantifiedFormula)formula1;
      QuantifiedFormula f2 = (QuantifiedFormula)formula2;
      TypedIdentifier[] vars1 = f1.getVariables();
      TypedIdentifier[] vars2 = f2.getVariables();
      Formula body1 = f1.getBody();
      Formula body2 = f2.getBody();

      // process formulas
      Collection<String> free1 = FreeVariables.compute(body1);
      Collection<String> free2 = FreeVariables.compute(body2);
      Collection<String> free = new SetClass<String>();
      free.addAll(free1);
      free.addAll(free2);
      Vector<TypedIdentifier> vars = new Vector<TypedIdentifier>();
      Collection<String> forbidden1 = new  SetClass<String>(free2);
      Collection<String> forbidden2 = new  SetClass<String>(free1);
      forbidden1.addAll(BoundVariables.compute(body1));
      forbidden2.addAll(BoundVariables.compute(body2));
      Formula g1 = processBody(vars1, body1, forbidden1, free, vars);
      Formula g2 = processBody(vars2, body2, forbidden2, free, vars);
      
      // construct result
      int n = vars.size();
      TypedIdentifier[] vars0 = new TypedIdentifier[n];
      vars.toArray(vars0);
      if (tag == Tag.AND)
        return forallFormula(vars0, andFormula(g1, g2));
      else
        return existsFormula(vars0, orFormula(g1, g2));
    }
    
    // (FORALL x: A) /\ B -> (FORALL y: A[y/x] /\ B)
    // all combinations of FORALL and EXISTS, /\ and \/
    if ((formula1 instanceof ForallFormula || 
        formula1 instanceof ExistsFormula) &&
        (tag == Tag.AND || tag == Tag.OR))
    {
      // get components
      QuantifiedFormula f1 = (QuantifiedFormula)formula1;
      TypedIdentifier[] vars1 = f1.getVariables();
      Formula body1 = f1.getBody();

      // process formula
      Collection<String> free1 = FreeVariables.compute(body1);
      Collection<String> free2 = FreeVariables.compute(formula2);
      Collection<String> free = new SetClass<String>();
      free.addAll(free1);
      free.addAll(free2);
      Collection<String> forbidden1 = new  SetClass<String>(free2);
      forbidden1.addAll(BoundVariables.compute(body1));
      Vector<TypedIdentifier> vars = new Vector<TypedIdentifier>();
      Formula g1 = processBody(vars1, body1, forbidden1, free, vars);
      
      // construct result
      int n = vars.size();
      TypedIdentifier[] vars0 = new TypedIdentifier[n];
      vars.toArray(vars0);
      Formula body;
      if (tag == Tag.AND)
        body = andFormula(g1, formula2);
      else
        body = orFormula(g1, formula2);
      if (formula1 instanceof ForallFormula)
        return forallFormula(vars0, body);
      else 
        return existsFormula(vars0, body);
    }
    
    // A /\ (FORALL x: B) -> (FORALL y: A /\ B[y/x])
    // all combinations of FORALL and EXISTS, /\ and \/
    if ((formula2 instanceof ForallFormula || 
        formula2 instanceof ExistsFormula) &&
        (tag == Tag.AND || tag == Tag.OR))
    {
      // get components
      QuantifiedFormula f2 = (QuantifiedFormula)formula2;
      TypedIdentifier[] vars2 = f2.getVariables();
      Formula body2 = f2.getBody();

      // process formula
      Collection<String> free1 = FreeVariables.compute(formula1);
      Collection<String> free2 = FreeVariables.compute(body2);
      Collection<String> free = new SetClass<String>();
      free.addAll(free1);
      free.addAll(free2);
      Collection<String> forbidden2 = new  SetClass<String>(free1);
      forbidden2.addAll(BoundVariables.compute(body2));
      Vector<TypedIdentifier> vars = new Vector<TypedIdentifier>();
      Formula g2 = processBody(vars2, body2, forbidden2, free, vars);
      
      // construct result
      int n = vars.size();
      TypedIdentifier[] vars0 = new TypedIdentifier[n];
      vars.toArray(vars0);
      Formula body;
      if (tag == Tag.AND)
        body = andFormula(formula1, g2);
      else
        body = orFormula(formula1, g2);
      if (formula2 instanceof ForallFormula)
        return forallFormula(vars0, body);
      else 
        return existsFormula(vars0, body);
    }
    
    // (FORALL x: A) => B -> (EXISTS y: A[y/x] => B)
    // (EXISTS x: A) => B -> (FORALL y: A[y/x] => B)
    if ((formula1 instanceof ForallFormula || 
        formula1 instanceof ExistsFormula) &&
        tag == Tag.IMPLIES)
    {
      // get components
      QuantifiedFormula f1 = (QuantifiedFormula)formula1;
      TypedIdentifier[] vars1 = f1.getVariables();
      Formula body1 = f1.getBody();

      // process formula
      Collection<String> free1 = FreeVariables.compute(body1);
      Collection<String> free2 = FreeVariables.compute(formula2);
      Collection<String> free = new SetClass<String>();
      free.addAll(free1);
      free.addAll(free2);
      Collection<String> forbidden1 = new  SetClass<String>(free2);
      forbidden1.addAll(BoundVariables.compute(body1));
      Vector<TypedIdentifier> vars = new Vector<TypedIdentifier>();
      Formula g1 = processBody(vars1, body1, forbidden1, free, vars);
      
      // construct result
      int n = vars.size();
      TypedIdentifier[] vars0 = new TypedIdentifier[n];
      vars.toArray(vars0);
      Formula body = impliesFormula(g1, formula2);
      if (formula1 instanceof ForallFormula)
        return existsFormula(vars0, body);
      else 
        return forallFormula(vars0, body);
    }
    
    // A => (FORALL x: B) -> (FORALL y: A => B[y/x])
    // A => (EXISTS x: B) -> (EXISTS y: A => B[y/x])
    if ((formula2 instanceof ForallFormula || 
        formula2 instanceof ExistsFormula) &&
        tag == Tag.IMPLIES)
    {
      // get components
      QuantifiedFormula f2 = (QuantifiedFormula)formula2;
      TypedIdentifier[] vars2 = f2.getVariables();
      Formula body2 = f2.getBody();

      // process formula
      Collection<String> free1 = FreeVariables.compute(formula1);
      Collection<String> free2 = FreeVariables.compute(body2);
      Collection<String> free = new SetClass<String>();
      free.addAll(free1);
      free.addAll(free2);
      Collection<String> forbidden2 = new  SetClass<String>(free1);
      forbidden2.addAll(BoundVariables.compute(body2));
      Vector<TypedIdentifier> vars = new Vector<TypedIdentifier>();
      Formula g2 = processBody(vars2, body2, forbidden2, free, vars);
      
      // construct result
      int n = vars.size();
      TypedIdentifier[] vars0 = new TypedIdentifier[n];
      vars.toArray(vars0);
      Formula body = impliesFormula(formula1, g2);
      if (formula2 instanceof ForallFormula)
        return forallFormula(vars0, body);
      else 
        return existsFormula(vars0, body);
    }
    
    Main.internalError("unknown case for pulling out quantifiers");
    return null;
  }
  
  /***************************************************************************
   * Process body of quantified formula for pulling out quantifiers.
   * @param vars the quantifiers of the formula.
   * @param body the body of the formula.
   * @param forbidden the variables that are forbidden as quantified variables.
   * @param free the set of free variables (modified by the processing).
   * @param newVars the vector to which to add the modified quantifiers.
   * @return the new body of the formula for the modified quantifiers.
   ****************************************************************************/
  private Formula processBody(TypedIdentifier[] vars, 
    Formula body, Collection<String> forbidden,
    Collection<String> free, Vector<TypedIdentifier> newVars)
  {
    MapClass<String,Term> map = new MapClass<String,Term>();
    for (TypedIdentifier var : vars)
    {
      String name = var.getName().getString();
      Type type = var.getType();
      if (forbidden.contains(name))
      {
        Collection<String> forbidden0 = new SetClass<String>(forbidden);
        forbidden0.addAll(free);
        String name0 = NewVariables.newVariable(name, forbidden0);
        Identifier id0 = Identifier.construct(name0);
        TypedIdentifier var0 = TypedIdentifier.construct(id0, type);
        ValueSymbol.construct(id0, type, var0); // link symbol to identifier
        map.put(name, Reference.construct(id0));
        free.remove(name);
        free.add(name0);
        var = var0;
      }
      newVars.add(var);
    }
    return (Formula)RefSubstitution.perform(map, body);
  }
  
  /***************************************************************************
   * Merge quantified variables.
   * @param tag the tag for the type of the outside quantifier.
   * @param vars the variables to be merged.
   * @param exp the inner expression.
   * @return a version of the outer expression with the variables potentially
   *         merged into the quantified variables
   **************************************************************************/
  private QuantifiedExpression mergeQuantifiedVariables(Tag tag, 
    TypedIdentifier[] vars, Expression exp)
  {
    // check applicability
    switch (tag)
    {
    case FORALL:
      if (!(exp instanceof ForallFormula))
        return constructQuantifiedExpression(tag, vars, exp);
      break;
    case EXISTS:
      if (!(exp instanceof ExistsFormula))
        return constructQuantifiedExpression(tag, vars, exp);
      break;
    case ARRAYTERM:
      if (!(exp instanceof ArrayTerm))
        return constructQuantifiedExpression(tag, vars, exp);
      break;
    case LAMBDAFORMULA:
      if (!(exp instanceof LambdaFormula))
        return constructQuantifiedExpression(tag, vars, exp);
      break;
    case LAMBDATERM:
      if (!(exp instanceof LambdaTerm))
        return constructQuantifiedExpression(tag, vars, exp);
      break;
    default:
      Main.internalError("unknown quantified expression");
      return null;
    }
    
    QuantifiedExpression exp0 = (QuantifiedExpression)exp;
    Vector<TypedIdentifier> vars0 = new Vector<TypedIdentifier>();
    Vector<TypedIdentifier> vars1 = new Vector<TypedIdentifier>();
    Set<String> names = new SetClass<String>();
    for (TypedIdentifier var : vars)
    {
      vars0.add(var);
      names.add(var.getName().getString());
    }
    for (TypedIdentifier var: exp0.getVariables())
    {
      String name = var.getName().getString();
      if (names.contains(name))
        vars1.add(var);
      else
        vars0.add(var);
    }
    Expression body = exp0.getBody();
    int n0 = vars0.size();
    TypedIdentifier[] qvars0 = new TypedIdentifier[n0];
    vars0.toArray(qvars0);
    if (!vars1.isEmpty())
    {
      int n1 = vars1.size();
      TypedIdentifier[] qvars1 = new TypedIdentifier[n1];
      vars1.toArray(qvars1);
      body = constructQuantifiedExpression(tag, qvars1, body);
    }
    return constructQuantifiedExpression(tag, qvars0, body);
  }
  
  /****************************************************************************
   * Construct quantified expression.
   * @param tag the kind of quantifier.
   * @param vars the bound variables.
   * @param body the expression body.
   * @return the quantified expression.
   ***************************************************************************/
  private static QuantifiedExpression constructQuantifiedExpression(Tag tag, 
    TypedIdentifier[] vars, Expression body)
  {
    switch (tag)
    {
    case FORALL:
      return ForallFormula.construct(vars, (Formula)body);
    case EXISTS:
      return ExistsFormula.construct(vars, (Formula)body);
    case ARRAYTERM:
      return ArrayTerm.construct(vars, (Term)body);
    case LAMBDAFORMULA:
      return LambdaFormula.construct(vars, (Formula)body);
    case LAMBDATERM:
      return LambdaTerm.construct(vars, (Term)body);
    default:
      Main.internalError("unknown quantified expression");
      return null;
    }
  }
  
  // ------------------------------------------------------------------------
  //
  // Auxiliaries for solving certain inequalities over the integers as
  // frequently arise from loop invariants: a <= b+1 /\ a > b /\  --> a=b 
  //
  // -------------------------------------------------------------------------
  
  /**
   * A value bound
   */
  private final static class ValueBound
  {
    public final int formula;   // the index of a formula indicating the bound
    public final Term bound;    // the (lower/upper) bound indicated by the formula.
    public ValueBound(int formula, Term bound)
    {
      this.formula = formula;
      this.bound = bound;
    }
  }
  
  /**
   * Sets of bounds for a term
   */
  private final static class TermBounds
  {
    // the the term
    public final Term term;
    
    // sets of lower/upper bounds for the term
    public final Set<ValueBound> lower = new SetClass<ValueBound>(); 
    public final Set<ValueBound> upper = new SetClass<ValueBound>(); 
    
    public TermBounds(Term term)
    {
      this.term = term;
    }
  }
  
  /**
   * A term which is given a value.
   */
  private final static class TermValue
  {
    public final Term term;
    public final Term value;
    public final int lower;
    public final int upper;
    public TermValue(Term term, Term value, int lower, int upper)
    {
      this.term = term;
      this.value = value;
      this.lower = lower;
      this.upper = upper;
    }
  }
  
  /***************************************************************************
   * Combine inequalities of formulas that identify values uniquely.
   * @param formulas the formulas.
   * @return the combined version of the formulas.
   **************************************************************************/
  private Vector<Formula> combineInequalities(Vector<Formula> formulas)
  {
    Map<String, TermBounds> bounds = collectBounds(formulas);
    Set<TermValue> values = getValues(bounds.values());
    if (values.isEmpty()) return formulas;
    Vector<Formula> result = new Vector<Formula>();
    int n = formulas.size();
    boolean[] remove = new boolean[n];
    for (int i=0; i<n; i++) remove[i] = false;
    for (TermValue value : values)
    {
      result.add(EqualsFormula.construct(value.term, value.value));
      remove[value.lower] = true;
      remove[value.upper] = true;
    }
    for (int i=0; i<n; i++)
    {
      if (!remove[i]) result.add(formulas.get(i));
    }
    return result;
  }
  
  /***************************************************************************
   * Get term values from a set of bounds
   * @param boundSet the set of term bounds.
   * @return the set of derived term values.
   **************************************************************************/
  private static Set<TermValue> getValues(Collection<TermBounds> boundSet)
  {
    Set<TermValue> values = new SetClass<TermValue>();
    boundLoop: for (TermBounds bounds : boundSet)
    {
      for (ValueBound lower : bounds.lower)
      {
        String bound = lower.bound.toString();
        for (ValueBound upper : bounds.upper)
        {
          if (bound.equals(upper.bound.toString()))
          {
            values.add(new TermValue(bounds.term, lower.bound, 
                lower.formula, upper.formula));
            continue boundLoop;
          }
        }
      }
    }
    return values;
  }
  
  /***************************************************************************
   * Collect inequality bounds from a sequence of formulas.
   * @param formulas the sequence of formulas.
   * @return bounds the derived bounds.
   **************************************************************************/
  private Map<String, TermBounds> collectBounds(Vector<Formula> formulas)
  {
    Map<String, TermBounds> bounds = new MapClass<String, TermBounds>();
    int n = formulas.size();
    for (int i=0; i<n; i++)
    {
      Formula formula = formulas.get(i);
      if (!(formula instanceof BinaryAtomicFormula)) continue;
      BinaryAtomicFormula formula0 = (BinaryAtomicFormula)formula;
      Type type1 = formula0.getFirst().getType();
      if (!LogicChecking.isIntegerType(type1)) continue;
      Type type2 = formula0.getSecond().getType();
      if (!LogicChecking.isIntegerType(type2)) continue;
      if (formula instanceof LessEqualFormula)
      {
        Term first = simplify(formula0.getFirst());
        Term second = simplify(formula0.getSecond());
        TermBounds fbounds = getBounds(first, bounds);
        TermBounds sbounds = getBounds(second, bounds);
        fbounds.upper.add(new ValueBound(i, second));
        sbounds.lower.add(new ValueBound(i, first));
        continue;
      }
      if (formula instanceof GreaterEqualFormula)
      {
        Term first = simplify(formula0.getFirst());
        Term second = simplify(formula0.getSecond());
        TermBounds fbounds = getBounds(first, bounds);
        TermBounds sbounds = getBounds(second, bounds);
        fbounds.lower.add(new ValueBound(i, second));
        sbounds.upper.add(new ValueBound(i, first));
        continue;
      }
      if (formula instanceof LessFormula)
      {
        Term first = simplify(formula0.getFirst());
        Term second = simplify(formula0.getSecond());
        TermBounds fbounds = getBounds(first, bounds);
        TermBounds sbounds = getBounds(second, bounds);
        fbounds.upper.add(new ValueBound(i, add(second, -1)));
        sbounds.lower.add(new ValueBound(i, add(first, +1)));
        continue;
      }
      if (formula instanceof GreaterFormula)
      {
        Term first = simplify(formula0.getFirst());
        Term second = simplify(formula0.getSecond());
        TermBounds fbounds = getBounds(first, bounds);
        TermBounds sbounds = getBounds(second, bounds);
        fbounds.lower.add(new ValueBound(i, add(second, +1)));
        sbounds.upper.add(new ValueBound(i, add(first, -1)));
        continue;
      }
    }
    return bounds;
  }
  
  /***************************************************************************
   * Get set of bounds for denoted term (creating it, if necessary)
   * @param term the term.
   * @param bounds the current mapping of terms to bounds (may be updated).
   * @return the set of bounds for the term (may be new and empty).
   **************************************************************************/
  private static TermBounds getBounds(Term term, Map<String, TermBounds> bounds)
  {
    String key = term.toString();
    TermBounds result = bounds.get(key);
    if (result != null) return result;
    result = new TermBounds(term);
    bounds.put(key, result);
    return result;
  }
  
  /****************************************************************************
   * Construct term denoting a sum.
   * @param term the term.
   * @param number the summand.
   * @return the resulting term.
   ***************************************************************************/
  private Term add(Term term, int number)
  {
    IntLiteral literal = ExpressionLogic.intLiteral(Integer.toString(number));
    Term result = PlusTerm.construct(term, literal);
    result.setType(LogicTypeTable.intType);
    return simplify(result);
  }
  
  /****************************************************************************
   * Simplify a term.
   * @param term the term.
   * @return its simplified form.
   ***************************************************************************/
  private static Term simplifyStatic(Term term)
  {
    Type type = term.getType();
    if (!LogicChecking.isIntegerType(type)) return term;
    Vector<Term> plus = new Vector<Term>();
    Vector<Term> minus = new Vector<Term>();
    int constant = summands(term, plus, minus);
    return sum(plus, minus, constant);
  }
  
  /***************************************************************************
   * Decompose term into summands.
   * @param term the term to be decomposed.
   * @param plus the sequence to which to add the positive summands.
   * @param minus the sequence to which to add the negative summands.
   * @return a constant summand.
   **************************************************************************/
  private static int summands(Term term, Vector<Term> plus, Vector<Term> minus)
  {
    // avoid overflows in simplification
    if (term instanceof PlusTerm)
    {
      PlusTerm term0 = (PlusTerm)term;
      int constant1 = summands(term0.getFirst(), plus, minus);
      int constant2 = summands(term0.getSecond(), plus, minus);
      BigInteger result = BigInteger.valueOf(constant1).add(BigInteger.valueOf(constant2));
      int result0 = result.intValue();
      if (result.equals(BigInteger.valueOf(result0)))
        return result0;
      plus.add(term);
      return 0;
    }
    if (term instanceof MinusTerm)
    { 
      MinusTerm term0 = (MinusTerm)term;
      int constant1 = summands(term0.getFirst(), plus, minus);
      int constant2 = summands(term0.getSecond(), minus, plus);
      BigInteger result = BigInteger.valueOf(constant1).subtract(BigInteger.valueOf(constant2));
      int result0 = result.intValue();
      if (result.equals(BigInteger.valueOf(result0)))
        return result0;
      plus.add(term);
      return 0;
    }
    if (term instanceof NegationTerm)
    { 
      NegationTerm term0 = (NegationTerm)term;
      int constant = summands(term0.getBase(), minus, plus);
      BigInteger result = BigInteger.valueOf(constant).negate();
      int result0 = result.intValue();
      if (result.equals(BigInteger.valueOf(result0)))
        return result0;
      plus.add(term);
      return 0;
    }
    if (term instanceof IntLiteral)
    {
      IntLiteral term0 = (IntLiteral)term;
      return Integer.parseInt(term0.getName());
    }
    plus.add(term);
    return 0;
  }
  
  /***************************************************************************
   * Create sum of terms.
   * @param plus the terms that account positively.
   * @param minus the terms that account negatively.
   * @param constant a constant term.
   * @return the sum of the terms.
   **************************************************************************/
  private static Term sum(Vector<Term> plus, Vector<Term> minus, int constant)
  {
    Term result = null;
    for (Term term : plus)
    {
      if (result == null)
        result = term;
      else
        result = PlusTerm.construct(result, term);
      result.setType(LogicTypeTable.intType);
    }
    for (Term term : minus)
    {
      if (result == null)
        result = NegationTerm.construct(term);
      else
        result = MinusTerm.construct(result, term);
      result.setType(LogicTypeTable.intType);
    }
    if (constant != 0)
    {
      Term constant0 = 
        ExpressionLogic.intLiteral(Integer.toString(Math.abs(constant)));
      if (constant > 0)
      {
        if (result == null)
          result = constant0;
        else
          result = PlusTerm.construct(result, constant0);
      }
      else
      {
        if (result == null)
          result = NegationTerm.construct(constant0);
        else
          result = MinusTerm.construct(result, constant0);
      } 
      result.setType(LogicTypeTable.intType);
    }
    if (result == null) result = ExpressionLogic.intLiteral("0");
    return result;
  }
  
  /**
   * 
   * A counter for the free variables in an expression.
   *
   */
  private static class Counter extends ASTVisitorBase
  {
    // the set holding the names/occurrence numbers of free variables 
    private Map<String,Integer> free;
    
    /***************************************************************************
     * Create free variable counter for expression.
     * @param exp the expression.
     **************************************************************************/
    public Counter(Expression exp)
    {
      free = new MapClass<String,Integer>();
      exp.accept(this);
    }
    
    /*************************************************************************
     * Get free variables
     * @return the set of free variables.
     ************************************************************************/
    public Map<String,Integer> getFree()
    {
      return free;
    }
    
    /**************************************************************************
     * Process quantified expression
     * @param exp a quantified expression
     *************************************************************************/
    public void visit(QuantifiedExpression exp)
    {
      Map<String,Integer> free0 = new Counter(exp.getBody()).getFree();
      for (TypedIdentifier tid : exp.getVariables())
      {
        String name = tid.getName().getString();
        free0.remove(name);
      }
      add(free, free0);
    }
    
    /**************************************************************************
     * Process quantified expression
     * @param exp a quantified expression
     *************************************************************************/
    public void visit(LetExpression exp)
    {
      Map<String,Integer> free0 = new Counter(exp.getBody()).getFree();
      Definition[] defs = exp.getDefinitions();
      int n = defs.length;
      for (int i=n-1; i>=0; i--)
      {
        Definition def = defs[i];
        String name = def.getName().getString();
        free0.remove(name);
        def.accept(this);
      }
      add(free, free0);
    }
    
    /**************************************************************************
     * Process reference.
     * @param ref a reference.
     *************************************************************************/
    public void visit(Reference ref) 
    { 
      // find first identifier in name that denotes value
      Name name = ref.getName();
      for (Identifier ident : name.idents)
      {
        Symbol symbol = ident.getSymbol();
        if (symbol instanceof ValueSymbol)
        {
          // strings from identifiers are unique by construction
          String var = ident.getString();
          add(free, var);
          break;
        }
      }
      super.visit(ref);
    }
  }
}
