// ----------------------------------------------------------------------------
// Terminates.java
// The judgement "se, Is, VS |- C |v J F"
// $Id: Terminates.java,v 1.23 2015/03/10 14:30:39 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C, J, F) 2011-, Research Institute for Symbolic Computation (RISC, J, F)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General private License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General private License for more details.
//
// You should have received a copy of the GNU General private License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Judgements;

import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Judgements.Substitution.*;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Syntax.Spec.*;
import fmrisc.ProgramExplorer.Syntax.Program.LeftValue;
import fmrisc.ProgramExplorer.Tasks.*;

/**
 * 
 * The judgement "se, Is, VS |- C |v J F"
 *
 */
public final class Terminates
{
  // the context of the judgement processor
  private final ParamSymbol method;        // the method of the judgements
  private final StateType stype;           // the state type
  private final TaskFolder tasks;          // the tasks folder
  private final Vector<TheorySymbol> theories; // the referenced theories
  private final Declaration[][] decls;         // the current declarations
  private final ErrorStream out;               // the error output stream
  
  // the auxiliary judgement processors            
  private final ExpressionLogic logic; // translate expression to term/formula
  private final Pre pre;               // compute preconditions
  private final Post post;             // compute postconditions
  
  /***************************************************************************
   * Create a termination judgement processor.
   * @param method the method in which the judged commands arise.
   * @param stype the type of the state for the context of a command.
   * @param tasks the folder to which to add termination-related tasks.
   * @param theories the referenced theories (first is base theory)
   * @param decls a sequence of declaration lists.
   * @param out the stream to which to write error messages.
   **************************************************************************/
  public Terminates(ParamSymbol method,
    StateType stype, TaskFolder tasks, 
    Vector<TheorySymbol> theories, Declaration[][] decls, ErrorStream out)
  {
    this.method = method;
    this.stype = stype;
    this.tasks = tasks;
    this.theories = theories;
    this.decls = decls;
    this.out = out;
    this.logic = new ExpressionLogic(theories.firstElement());
    this.pre = new Pre(stype);
    this.post = new Post(stype);
  }
  
  /***************************************************************************
   * Apply the judgement 
   * 
   *   se, Is, Vs |- C |v J F
   *   
   * to prove that command C terminates when executed in a state in which
   * F[J] holds where J is a vector of free variables in F that denotes the
   * measure of the recursive methods in Is that may be called in C
   * (J may be null which indicates that the current method is not recursive).
   * 
   * Must be called only after C has been annotated with its judgement.
   * 
   * As a side effect, generate the verification conditions to be proved
   * and extend the judgement of C by a condition T on a state that is 
   * sufficient to ensure that the execution of C in that state terminates
   * (excluding the termination of recursive function calls). T may be
   * null which indicates TRUE (C terminates on every state).
   * 
   * Returns the condition T.
   * 
   * See Section 4.10 of "A Program Calculus", Is is set in the method,
   * while se and Vs are not needed because identifiers have already 
   * been resolved to symbols.
   * 
   * @param C the command C
   * @param J the variable vector representing the termination measure
   * @param F the condition on the state in which C is called
   * 
   * @return the condition T (possibly null which indicates TRUE).
   **************************************************************************/
  public Formula derive(
    fmrisc.ProgramExplorer.Syntax.Program.Statement C,
    ValueSymbol[] J, Formula F)
  {
    Formula T = null;
    if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.AssertionStatement) 
      T = derive((fmrisc.ProgramExplorer.Syntax.Program.AssertionStatement)C, J, F);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.AssignmentStatement) 
      T = derive((fmrisc.ProgramExplorer.Syntax.Program.AssignmentStatement)C, J, F);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.BlockStatement) 
      T = derive((fmrisc.ProgramExplorer.Syntax.Program.BlockStatement)C, J, F);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.BreakStatement) 
      T = derive((fmrisc.ProgramExplorer.Syntax.Program.BreakStatement)C, J, F);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.CallStatement) 
      T = derive((fmrisc.ProgramExplorer.Syntax.Program.CallStatement)C, J, F);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.ConditionalStatement) 
      T = derive((fmrisc.ProgramExplorer.Syntax.Program.ConditionalStatement)C, J, F);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.ContinueStatement) 
      T = derive((fmrisc.ProgramExplorer.Syntax.Program.ContinueStatement)C, J, F);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.EmptyStatement) 
      T = derive((fmrisc.ProgramExplorer.Syntax.Program.EmptyStatement)C, J, F);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.ForLoopStatement) 
      T = derive((fmrisc.ProgramExplorer.Syntax.Program.ForLoopStatement)C, J, F);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.ReturnStatement) 
      T = derive((fmrisc.ProgramExplorer.Syntax.Program.ReturnStatement)C, J, F);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.TryCatchStatement) 
      T = derive((fmrisc.ProgramExplorer.Syntax.Program.TryCatchStatement)C, J, F);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.VariableStatement) 
      T = derive((fmrisc.ProgramExplorer.Syntax.Program.VariableStatement)C, J, F);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.WhileLoopStatement) 
      T = derive((fmrisc.ProgramExplorer.Syntax.Program.WhileLoopStatement)C, J, F);
    else
      Main.internalError("unknown kind of statement");
    if (T != null)
    {
      // if command does not execute, it terminates anyway
      StateLiteral literal = OldState.construct();
      literal.setType(stype);
      T = ImpliesFormula.construct(StateExecutes.construct(literal), T);
    }
    C.getJudgement().setTermination(T);
    return T;
  }
  
  /****************************************************************************
   * The assertion statement "assert E"
   ***************************************************************************/
  private Formula derive(
    fmrisc.ProgramExplorer.Syntax.Program.AssertionStatement C,
    ValueSymbol[] J, Formula F)
  {
    return null;
  }
  
  /****************************************************************************
   * The assignment statement "lval = exp"
   ***************************************************************************/
  private Formula derive(
    fmrisc.ProgramExplorer.Syntax.Program.AssignmentStatement C,
    ValueSymbol[] J, Formula F)
  {
    return null;
  }
  
  /****************************************************************************
   * The block statement "{ ... }"
   * The rule is derived by combining the rules of local variable blocks
   * with that of statement sequences. Since we refer to variables by symbols
   * rather than by identifiers, only one block rule has to be considered.
   ***************************************************************************/
  private Formula derive(
    fmrisc.ProgramExplorer.Syntax.Program.BlockStatement C,
    ValueSymbol[] J, Formula F)
  { 
    // first the trivial case
    int n = C.stats.length;
    if (n == 0) return null;
    
    // determine individual termination conditions
    VariableSymbol[] Isymbols = new VariableSymbol[n];
    TypedIdentifier[] tids = new TypedIdentifier[n];
    Substitution.Value[] subst = new Substitution.Value[n];
    fmrisc.ProgramExplorer.Syntax.Program.Statement[] Cs =
      new fmrisc.ProgramExplorer.Syntax.Program.Statement[n];
    Formula[] Ts = new Formula[n];
    Formula F0 = F;
    for (int i=0; i<n; i++)
    {
      // determine statement to be processed (empty statement replaced by null)
      fmrisc.ProgramExplorer.Syntax.Program.Statement C0 = C.stats[i];
      if (C0 instanceof fmrisc.ProgramExplorer.Syntax.Program.EmptyStatement)
        C0 = null;
      Cs[i] = C0;
      
      // statement may introduce a variable, apply variable declaration rule
      // and replace statement by one without declaration
      if (C0 instanceof fmrisc.ProgramExplorer.Syntax.Program.VariableDeclaration)
      { 
        // determine variable I introduced by statement
        fmrisc.ProgramExplorer.Syntax.Program.VariableDeclaration C1 =
          (fmrisc.ProgramExplorer.Syntax.Program.VariableDeclaration)C0;
        Identifier I = C1.getName();
        VariableSymbol Isymbol = (VariableSymbol)I.getSymbol();
        Isymbols[i] = Isymbol;
        
        // introduce mathematical variable K for substituting I
        Collection<String> used = UsedVariables.compute(F0);
        Identifier K = Identifier.construct(NewVariables.newVariable(I.getString(), used));
        Type type = Isymbol.getLogicType();
        TypedIdentifier tid = TypedIdentifier.construct(K, type);
        ValueSymbol.construct(K, type, tid);
        tids[i] = tid;
        
        // compute substitution formula EXISTS(K): F[K/I]
        subst[i] = new Substitution.Value(
            Substitution.Action.term, Satisfies.reference(K),
            Substitution.Action.term, Satisfies.reference(K));
        Map<VariableSymbol,Value> map = new MapClass<VariableSymbol,Value>();
        map.put(Isymbol, subst[i]);
        F0 = (Formula)Substitution.perform(F, null, null, map);
        F0 = ExistsFormula.construct(new TypedIdentifier[] { tid }, F0);
        
        // determine core statement (null on pure variable declaration)
        Cs[i] = coreStatement(C1);
      }
      
      // apply judgement to (possibly replaced) statement and
      // apply sequence rule to determine condition for next statement
      if (Cs[i] != null)
      {
        Ts[i] = derive(Cs[i], J, F0); 
        F0 = post.derive(Cs[i], F0);
      }
    }
    
    // combine termination conditions to condition FORALL(K): T[K/I]
    Formula T = null;
    for (int i=n-1; i>=0; i--)
    {
      fmrisc.ProgramExplorer.Syntax.Program.Statement C0 = Cs[i];
      Formula T0 = T;
      if (C0 != null && T0 != null) T0 = pre.derive(C0, T0);
      if (T0 != null && tids[i] != null)
      {
        Map<VariableSymbol,Value> map = new MapClass<VariableSymbol,Value>();
        map.put(Isymbols[i], subst[i]);
        T0 = (Formula)Substitution.perform(T0, null, null, map);
        T0 = ForallFormula.construct(new TypedIdentifier[] { tids[i] }, T0);
      }
      T = Satisfies.andFormula(Ts[i], T0);
    }
    return T;
  }
  
  /***************************************************************************
   * Determine core statement of a variable declaration.
   * @param C a variable declaration.
   * @return the statement describing the effect of the declaration
   *         (null, if the empty statement)
   **************************************************************************/
  private fmrisc.ProgramExplorer.Syntax.Program.Statement
    coreStatement(fmrisc.ProgramExplorer.Syntax.Program.VariableDeclaration C)
  {
    LeftValue lval = 
      LeftValue.construct(C.getName(), 
        new fmrisc.ProgramExplorer.Syntax.Program.Selector[0]);
    if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.VariableStatement)
    {
      fmrisc.ProgramExplorer.Syntax.Program.VariableStatement C0 =
        (fmrisc.ProgramExplorer.Syntax.Program.VariableStatement)C;
      // pure variable declaration has no effect
      if (C0.exp == null) 
        return null;
      // initialized variable declaration becomes assignment
      else
        return attachInfo(C0,
            fmrisc.ProgramExplorer.Syntax.Program.AssignmentStatement.construct(
                lval, C0.exp));
    }
    if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.DeclNewStatement)
    {
      // declaration new becomes assignment new
      fmrisc.ProgramExplorer.Syntax.Program.DeclNewStatement C0 =
        (fmrisc.ProgramExplorer.Syntax.Program.DeclNewStatement)C;
      return attachInfo(C0,
          fmrisc.ProgramExplorer.Syntax.Program.AssignNewStatement.construct(
              lval, C0.method, C0.args));
    }
    if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.DeclCallStatement)
    {
      // declaration call becomes assignment call
      fmrisc.ProgramExplorer.Syntax.Program.DeclCallStatement C0 =
        (fmrisc.ProgramExplorer.Syntax.Program.DeclCallStatement)C;
      return attachInfo(C0,
          fmrisc.ProgramExplorer.Syntax.Program.AssignCallStatement.construct(
              lval, C0.method, C0.args));
    }
    Main.internalError("unknown kind of variable declaration");
    return null;
  }
  
  /***************************************************************************
   * Attach to statement information from other statement.
   * @param from the source of the information.
   * @param to the destination of the information.
   * @return the destination.
   **************************************************************************/
  private fmrisc.ProgramExplorer.Syntax.Program.Statement attachInfo(
    fmrisc.ProgramExplorer.Syntax.Program.Statement from,
    fmrisc.ProgramExplorer.Syntax.Program.Statement to)
  {
    to.setPosition(from.getPosition());
    to.setJudgement(from.getJudgement());
    if (to instanceof fmrisc.ProgramExplorer.Syntax.Program.CallStatement)
    {
      ((fmrisc.ProgramExplorer.Syntax.Program.CallStatement)to).setMethodSymbol(
          ((fmrisc.ProgramExplorer.Syntax.Program.CallStatement)from).getMethodSymbol());
    }
    return to;
  }
  
  /****************************************************************************
   * The statement "break"
   ***************************************************************************/
  private Formula derive(
    fmrisc.ProgramExplorer.Syntax.Program.BreakStatement C,
    ValueSymbol[] J, Formula F)
  {
    return null;
  }
  
  /****************************************************************************
   * The conditional statement "if (E) C1 [ else C2 ]"
   ***************************************************************************/
  private Formula derive(
    fmrisc.ProgramExplorer.Syntax.Program.ConditionalStatement C,
    ValueSymbol[] J, Formula F)
  {
    Formula G = logic.deriveFormula(C.cond, null);
    Formula F1 = Satisfies.andFormula(F, G);
    Formula T1 = derive(C.thenstat, J, F1);
    if (C.elsestat == null) return Satisfies.impliesFormula(G, T1);
    Formula F2 = Satisfies.andFormula(F, Satisfies.notFormula(G));
    Formula T2 = derive(C.elsestat, J, F2);
    return Satisfies.ifThenElseFormula(G, T1, T2);
  }
  
  /****************************************************************************
   * The statement "continue"
   ***************************************************************************/
  private Formula derive(
    fmrisc.ProgramExplorer.Syntax.Program.ContinueStatement C,
    ValueSymbol[] J, Formula F)
  {
    return null;
  }
  
  /****************************************************************************
   * The empty statement ";"
   ***************************************************************************/
  private Formula derive(
    fmrisc.ProgramExplorer.Syntax.Program.EmptyStatement C,
    ValueSymbol[] J, Formula F)
  {
    return null;
  }
  
  /****************************************************************************
   * The statement "return [E]"
   ***************************************************************************/
  private Formula derive(
    fmrisc.ProgramExplorer.Syntax.Program.ReturnStatement C,
    ValueSymbol[] J, Formula F)
  {
    return null;
  }
  
  /****************************************************************************
   * The statement "throw new I(E)"
   ***************************************************************************/
  private Formula derive(
    fmrisc.ProgramExplorer.Syntax.Program.ThrowStatement C,
    ValueSymbol[] J, Formula F)
  {
    return null;
  }
  
  /****************************************************************************
   * The protected block statement "try { C1 } ( catch(T I) C2 )+"
   ***************************************************************************/
  private Formula derive(
    fmrisc.ProgramExplorer.Syntax.Program.TryCatchStatement C,
    ValueSymbol[] J, Formula F)
  {
    // process protected code block
    Formula T = derive(C.body, J, F);
    Formula G = post.derive(C.body, F);
    Collection<String> used = UsedVariables.compute(G);

    // process handlers and determine termination conditions
    int n = C.handlers.length;
    for (int i=0; i<n; i++)
    {
      // apply judgement to handler and derive its termination condition
      Identifier param = C.handlers[i].param.getName();
      Collection<String> used0 = new SetClass<String>(used);
      Identifier param0 = 
        Identifier.construct(NewVariables.newVariable(param.getString(), used0));
      used0.add(param0.getString());
      Identifier state =
        Identifier.construct(NewVariables.newVariable("state", used0));
      VariableSymbol symbol = (VariableSymbol)param.getSymbol();
      Type type = symbol.getLogicType();
      TypedIdentifier tid0 = TypedIdentifier.construct(param0, type);
      ValueSymbol.construct(param0, type, tid0);
      TypedIdentifier tid1 = TypedIdentifier.construct(state, stype);
      ValueSymbol.construct(state, stype, tid1);
      Map<VariableSymbol,Value> map = new MapClass<VariableSymbol,Value>();
      map.put(symbol, 
            new Substitution.Value(
                Substitution.Action.term, Satisfies.reference(param0),
                Substitution.Action.term, Satisfies.reference(param0))); 
      Formula G0 = (Formula)Substitution.perform(G,
          Satisfies.reference(state), Satisfies.reference(state), map);
      Name exception = 
        ((fmrisc.ProgramExplorer.Syntax.Program.NamedType)
            C.handlers[i].param.type).name;
      Formula G1 = ExistsFormula.construct(
          new TypedIdentifier[] { tid0, tid1 },
          AndFormula.construct(
              StateThrowsException.construct(
                  Satisfies.reference(state), exception),
                  AndFormula.construct(
                      EqualsFormula.construct(
                          StateValue.construct(Satisfies.reference(state)),
                          Satisfies.reference(param0)),G0)));
      Formula T0 = derive(C.handlers[i].body, J, G1);  
      if (T0 == null) continue;
      
      // combine handler's termination condition with overall condition
      Formula T1 = (Formula)Substitution.perform(T0,
          Satisfies.reference(state), Satisfies.reference(state), map);
      Formula T2 =
        ImpliesFormula.construct(
            StateThrowsException.construct(stateLiteral(true), exception),
            ForallFormula.construct(
                new TypedIdentifier[] { tid0, tid1 }, T1));
      Formula T3 = pre.derive(C.body, T2);
      T = Satisfies.andFormula(T, T3);
    }
    
    // the final result
    return T;
  }
  
  /****************************************************************************
   * Construct a state literal giving it the current state type.
   * @param old true if it is the old state (the new state, otherwise)
   * @return the literal with the state type attached.
   ***************************************************************************/
  private StateLiteral stateLiteral(boolean old)
  {
    StateLiteral literal = old ? OldState.construct() : NewState.construct();
    if (stype == null) Main.internalError("no state type");
    literal.setType(stype);
    return literal;
  }
  
  /****************************************************************************
   * The variable declaration statement "Type var;"
   ***************************************************************************/
  private Formula derive(
    fmrisc.ProgramExplorer.Syntax.Program.VariableStatement C,
    ValueSymbol[] J, Formula F)
  {
    return null; 
  }
  
  /****************************************************************************
   * A while loop "while (E) C"
   ***************************************************************************/
  private Formula derive(
    fmrisc.ProgramExplorer.Syntax.Program.WhileLoopStatement C,
    ValueSymbol[] J, Formula F)
  {
    // determine body judgement
    StatJudgement cSpec = C.body.getJudgement();
    Formula bodyStep = cSpec.formula;
    Set<VariableSymbol> variables = cSpec.variables;
    Set<ClassSymbol> exceptions = cSpec.exceptions;
    
    // determine invariant and termination term
    Formula invariant = null;
    Term[] termination = null;
    LoopSpec spec = (LoopSpec)C.getLoopAnnotation();
    if (spec != null)
    {
      invariant = spec.getLoopInvariant();
      if (invariant != null) 
        invariant = Normalization.perform(invariant, cSpec.variables);
      termination = spec.getTerminationTerm();
      if (termination != null)
      {
        int n = termination.length;
        Term[] termination0 = new Term[n];
        for (int i=0; i<n; i++)
          termination0[i] = Normalization.perform(termination[i]);
        termination = termination0;
      }
    }
    
    // if either one is missing, loop must not be entered
    Formula cond = logic.deriveFormula(C.cond, null);
    if (invariant == null || termination == null)
    {
      /*
      Formula negCond = NotFormula.construct(cond);
      tasks.add(new TerminationTask("loop must not be entered (no invariant/measure)",
          method, ImpliesFormula.construct(F, negCond), 
          C, null, tasks.getDirectory(), stype,
          variables, exceptions,
          theories, decls, out));
      */
      tasks.add(new FailedTask(method.getLongName()+
          ":"+C.getPosition().getLine(),
          "loop may not terminate", "loop has no invariant/measure",
          C.getPosition(), tasks.getDirectory()));
      return null;
    }
    
    // determine substitutions
    Collection<String> used = UsedVariables.compute(F);
    UsedVariables.add(invariant, used);
    UsedVariables.add(cond, used);
    int n = variables.size();
    VariableSymbol[] symbols = new VariableSymbol[n];
    variables.toArray(symbols);
    TypedIdentifier[] tids = new TypedIdentifier[n];
    Map<VariableSymbol,Value> mapFormula = new MapClass<VariableSymbol,Value>();
    Map<VariableSymbol,Value> mapInvariant = new MapClass<VariableSymbol,Value>();
    Map<VariableSymbol,Value> mapTerm = new MapClass<VariableSymbol,Value>();
    for (int i=0; i<n; i++)
    {
      VariableSymbol symbol = symbols[i];
      String name = NewVariables.newVariable(symbol.getName().getString(), used);
      used.add(name);
      Identifier ident = Identifier.construct(name);
      Type type = symbol.getLogicType();
      TypedIdentifier tid = TypedIdentifier.construct(ident, type);
      ValueSymbol.construct(ident, type, tid);
      tids[i] = tid;
      mapFormula.put(symbol, 
          new Substitution.Value(
              Substitution.Action.term, Satisfies.reference(ident),
              Substitution.Action.term, Satisfies.reference(ident)));
      mapInvariant.put(symbol, 
          new Substitution.Value(
              Substitution.Action.term, Satisfies.reference(ident),
              Substitution.Action.flip, null));
      mapTerm.put(symbol, 
          new Substitution.Value(
              Substitution.Action.flip, null,
              Substitution.Action.none, null));
    }
    String name0 = NewVariables.newVariable("os", used);
    used.add(name0);
    Identifier ident0 = Identifier.construct(name0);
    TypedIdentifier tid0 = TypedIdentifier.construct(ident0, stype);
    ValueSymbol.construct(ident0, stype, tid0);
    String name1 = NewVariables.newVariable("ns", used);
    used.add(name1);
    Identifier ident1 = Identifier.construct(name1);
    TypedIdentifier tid1 = TypedIdentifier.construct(ident1, stype);
    ValueSymbol.construct(ident1, stype, tid1);
    
    // verify termination of loop body
    Formula F0 =
      (Formula)Substitution.perform(F, 
          Satisfies.reference(ident0), Satisfies.reference(ident0), mapFormula);
    Formula invariant0 =
      (Formula)Substitution.perform(invariant,
          Satisfies.reference(ident0), Satisfies.reference(ident1), mapInvariant);
    int tn = termination.length;
    Formula coreNotNegative0 = null;
    for (int i=tn-1; i>=0; i--)
    {
      Term t0 = termination[i];
      Formula notNegative =
        GreaterEqualFormula.construct(t0, ExpressionLogic.intLiteral("0"));
      if (coreNotNegative0 == null)
        coreNotNegative0 = notNegative;
      else
        coreNotNegative0 = AndFormula.construct(notNegative, coreNotNegative0);
    }
    Formula coreFormula =
        AndFormula.construct(StateExecutes.construct(Satisfies.reference(ident0)),
            AndFormula.construct(F0, invariant0));
    Formula bodyExecutesCore =
        AndFormula.construct(coreFormula,
            AndFormula.construct(cond,
                AndFormula.construct(
                    OrFormula.construct(
                        StateExecutes.construct(Satisfies.reference(ident1)),
                        StateContinues.construct(Satisfies.reference(ident1))),
                        coreNotNegative0))
            );
    Formula bodyExecutes =
      Satisfies.existsFormula(tids,
          ExistsFormula.construct(new TypedIdentifier[] { tid0, tid1 },
              bodyExecutesCore));
    Formula bodyTerminatesCore =  derive(C.body, J, bodyExecutes); 
    if (bodyTerminatesCore != null)
    {
      Formula bodyTerminates =
          ForallFormula.construct(tids,
              ForallFormula.construct(new TypedIdentifier[] { tid0, tid1 },
                  ImpliesFormula.construct(
                      coreFormula, 
                      bodyTerminatesCore)));
        tasks.add(new TerminationTask("body terminates",
            method, bodyTerminates, 
            C, null, tasks.getDirectory(), stype,
            variables, exceptions,
            theories, decls, out));
    }
    
    // verify that termination term does not become negative by iteration
    Formula coreNotNegative = null;
    for (int i=tn-1; i>=0; i--)
    {
      Term t0 = termination[i];
      Term t1 = (Term)Substitution.perform(t0, stateLiteral(false), null, mapTerm);
      Formula notNegative =
        GreaterEqualFormula.construct(t1, ExpressionLogic.intLiteral("0"));
      if (coreNotNegative == null)
        coreNotNegative = notNegative;
      else
        coreNotNegative = AndFormula.construct(notNegative, coreNotNegative);
    }
    Formula notNegative =
      ForallFormula.construct(tids,
          ForallFormula.construct(new TypedIdentifier[] { tid0, tid1 },
              ImpliesFormula.construct(
                  AndFormula.construct(
                      bodyExecutesCore,
                      AndFormula.construct(
                          bodyStep,
                          OrFormula.construct(
                              StateExecutes.construct(stateLiteral(false)),
                              StateContinues.construct(stateLiteral(false))))), 
                  coreNotNegative)));
    tasks.add(new TerminationTask("measure is well-formed",
        method, notNegative, 
        C, null, tasks.getDirectory(), stype,
        variables, exceptions,
        theories, decls, out));
    
    // verify that termination term is decreased by iteration
    Formula decreasesCore = null;
    Formula equalities = null;
    for (int i=0; i<tn; i++)
    {
      Term t0 = termination[i];
      Term t1 = (Term)Substitution.perform(t0, stateLiteral(false), null, mapTerm);
      Formula less = LessFormula.construct(t1, t0);
      Formula equals = EqualsFormula.construct(t1, t0);
      if (decreasesCore == null)
      {
        decreasesCore = less;
        equalities = equals;
      }
      else
      {
        decreasesCore = OrFormula.construct(decreasesCore, 
            AndFormula.construct(equalities, less));
        equalities = AndFormula.construct(equalities, equals);
      }
    }
    Formula decreases =
      ForallFormula.construct(tids,
          ForallFormula.construct(new TypedIdentifier[] { tid0, tid1 },
              ImpliesFormula.construct(
                  AndFormula.construct(bodyExecutesCore, 
                      AndFormula.construct(
                          bodyStep,
                          OrFormula.construct(
                              StateExecutes.construct(stateLiteral(false)),
                              StateContinues.construct(stateLiteral(false))))),
                  decreasesCore)));
    tasks.add(new TerminationTask("measure is decreased",
        method, decreases, 
        C, null, tasks.getDirectory(), stype,
        variables, exceptions,
        theories, decls, out));
    
    // termination term must not be negative when loop is entered
    return coreNotNegative;
  }
  
  /****************************************************************************
   * A for loop "for(init; cond; update) C"
   ***************************************************************************/
  private Formula derive(
    fmrisc.ProgramExplorer.Syntax.Program.ForLoopStatement C,
    ValueSymbol[] J, Formula F)
  {
    return derive(C.whileReplacement, J, F);
  }
  
  /****************************************************************************
   * A call statement "[ ... = ] p(...) "
   ***************************************************************************/
  private Formula derive(
    fmrisc.ProgramExplorer.Syntax.Program.CallStatement C,
    ValueSymbol[] J, Formula F)
  {
    if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.AssignCallStatement) 
      return derive((fmrisc.ProgramExplorer.Syntax.Program.AssignCallStatement)C, J, F);
    if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.DeclCallStatement) 
      return derive((fmrisc.ProgramExplorer.Syntax.Program.DeclCallStatement)C, J, F);
    if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.ThrowStatement) 
      return derive((fmrisc.ProgramExplorer.Syntax.Program.ThrowStatement)C, J, F);
    if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.VoidCallStatement) 
      return derive((fmrisc.ProgramExplorer.Syntax.Program.VoidCallStatement)C, J, F);
    Main.internalError("unknown kind of method call");
    return null;
  }
  
  private Formula derive(
    fmrisc.ProgramExplorer.Syntax.Program.AssignCallStatement C,
    ValueSymbol[] J, Formula F)
  {
    return deriveCall(C, J, F);
  }
  
  private Formula derive(
    fmrisc.ProgramExplorer.Syntax.Program.DeclCallStatement C,
    ValueSymbol[] J, Formula F)
  {
    Identifier ident = C.name;
    VariableSymbol symbol = (VariableSymbol)ident.getSymbol();
    Collection<String> used = UsedVariables.compute(F);
    String name0 = NewVariables.newVariable(ident.getString(), used);
    Identifier ident0 = Identifier.construct(name0);
    Type type = symbol.getLogicType();
    TypedIdentifier tid = TypedIdentifier.construct(ident0, type);
    ValueSymbol.construct(ident0, type, tid);
    Map<VariableSymbol, Value> map = new MapClass<VariableSymbol, Value>();
    map.put(symbol, 
        new Substitution.Value(
            Substitution.Action.term, Satisfies.reference(ident0),
            Substitution.Action.term, Satisfies.reference(ident0)));
    Formula F0 = (Formula)Substitution.perform(F, null, null, map);
    Formula F1 = ExistsFormula.construct(new TypedIdentifier[] { tid }, F0);
    Formula T = deriveCall(C, J, F1);
    Formula T0 = (Formula)Substitution.perform(T, null, null, map);
    return ForallFormula.construct(new TypedIdentifier[] { tid }, T0);
  }
  
  private Formula derive(
    fmrisc.ProgramExplorer.Syntax.Program.VoidCallStatement C,
    ValueSymbol[] J, Formula F)
  {
    return deriveCall(C, J, F);
  }
  
  private Formula deriveCall(
    fmrisc.ProgramExplorer.Syntax.Program.CallStatement C,
    ValueSymbol[] J, Formula F)
  {    
    // determine method specification 
    ParamSymbol msymbol = (ParamSymbol)C.getMethodSymbol();
    MethodSpec spec = msymbol.getSpecification();
    Formula diverges = null;
    Term[] termination = null;
    if (spec != null) 
    { 
      diverges = spec.getDivergesCondition();
      if (diverges != null) diverges = Normalization.perform(diverges);
      termination = spec.getTerminationTerm();
      if (termination != null)
      {
        int n = termination.length;
        Term[] termination0 = new Term[n];
        for (int i=0; i<n; i++)
          termination0[i] = Normalization.perform(termination[i]);
        termination = termination0;
      }
    }
    
    // non-recursive method that has no diverges condition terminates
    boolean recursive = this.method.getRecursion().contains(msymbol);
    if (!recursive && diverges == null)
    {
      return null;
    }
        
    // unspecified method or recursive method without measure
    // must be assumed not to terminate
    if (spec == null || (recursive && termination == null)) 
    {
      String message =
        spec == null ? "method is unspecified" : 
          "call is recursive but method has no recursion measure";
      tasks.add(new FailedTask(method.getLongName()+
          ":"+method.getDeclaration().getPosition(), 
          "call of method " + method.getLongName() + " may not terminate", message, 
          C.getPosition(), tasks.getDirectory()));
      return null;
    }
    
    // the remaining case deals with a specified method that is recursive
    // with a termination measure and/or has a diverges condition
    
    // determine used variables
    Collection<String> used = UsedVariables.compute(F);
    if (diverges != null) UsedVariables.add(diverges, used);
    if (recursive)
    {
      for (Term t : termination)
        UsedVariables.add(t, used);
    }
      
    // determine quantified variables and substitution map
    int n = C.args.length;
    int tn = 0;
    if (J != null) tn = J.length;
    TypedIdentifier[] tids = new TypedIdentifier[n+tn];
    Map<VariableSymbol,Value> map = new MapClass<VariableSymbol,Value>();
    fmrisc.ProgramExplorer.Syntax.Program.Parameter[] params =
      msymbol.getDeclaration().getParams();
    for (int i=0; i<n; i++)
    {
      String pname = params[i].name.getString();
      String pname0 = NewVariables.newVariable(pname, used);
      used.add(pname0);
      Identifier ident = Identifier.construct(pname0);
      ParameterSymbol psymbol = (ParameterSymbol)params[i].name.getSymbol();
      Type type = psymbol.getLogicType();
      TypedIdentifier tid = TypedIdentifier.construct(ident, type);
      ValueSymbol.construct(ident, type, tid);
      tids[i] = tid;
      map.put(psymbol, 
          new Substitution.Value(
              Substitution.Action.term, Satisfies.reference(ident),
              Substitution.Action.term, Satisfies.reference(ident)));
    }
    
    // formula that defines argument values
    Formula argsFormula = null;
    for (int i=n-1; i>=0; i--)
    {
      Term arg = logic.deriveTerm(C.args[i], null);
      Formula equals = EqualsFormula.construct(
          Satisfies.reference(tids[i].getName()), arg);
      if (argsFormula == null)
        argsFormula = equals;
      else
        argsFormula = AndFormula.construct(equals, argsFormula);
    }
    
    // formula that termination term is decreased
    Formula decreases = null;
    if (recursive)
    {
      Formula equalities = null;
      for (int i=0; i<tn; i++)
      {
        Term t0 = Satisfies.reference(J[i].getName());
        Term t1 = (Term)Substitution.perform(termination[i], 
            null, stateLiteral(true), map);
        Formula less = LessFormula.construct(t1, t0);
        Formula equals = EqualsFormula.construct(t1, t0);
        if (decreases == null)
        {
          decreases = less;
          equalities = equals;
        }
        else
        {
          decreases = OrFormula.construct(decreases, 
              AndFormula.construct(equalities, less));
          equalities = AndFormula.construct(equalities, equals);
        }
      }
    }
    
    // formula that method call does not diverge
    Formula terminatesCore = decreases;
    Formula notDiverges = null;
    if (diverges != null)
    {
      notDiverges =
        (Formula)Substitution.perform(NotFormula.construct(diverges),
            null, stateLiteral(true), map);
      terminatesCore = Satisfies.andFormula(notDiverges, terminatesCore);
    }
    
    // return state condition that implies termination
    // quantification over J is not needed
    TypedIdentifier[] tids0 = new TypedIdentifier[n];
    for (int i=0; i<n; i++) tids0[i] = tids[i];
    Formula terminates =
      Satisfies.forallFormula(tids0,
          ImpliesFormula.construct(argsFormula, terminatesCore));
    return terminates;
    
    /*
    // clone F and substitute reference to J by reference to quantified variable
    Map<String, Term> map0 = new MapClass<String, Term>();
    for (int i=0; i<tn; i++)
    {
      ValueSymbol J0 = J[i];
      String name = J0.getName().getString();
      String name0 = NewVariables.newVariable(name, used);
      used.add(name0);
      Identifier ident = Identifier.construct(name0);
      Type type = J0.getType();
      TypedIdentifier tid = TypedIdentifier.construct(ident, type);
      ValueSymbol.construct(ident, type, tid);
      tids[n+i] = tid;
      map0.put(name, Satisfies.reference(ident));
    }
    Formula F0 = (Formula)RefSubstitution.perform(map0, F);
    Formula terminates =
      Satisfies.forallFormula(tids,
          ImpliesFormula.construct(
              AndFormula.construct(
                  StateExecutes.construct(stateLiteral(true)),
                  AndFormula.construct(F0, argsFormula)),
                  terminatesCore));
    Set<VariableSymbol> variables = NameUtils.getVariableSymbols(spec.getVariables());
    Set<ClassSymbol> exceptions = NameUtils.getClassSymbols(spec.getExceptions());
    tasks.add(new TerminationTask("method call terminates",
        msymbol, terminates, 
        C, null, tasks.getDirectory(), stype,
        variables, exceptions,
        theories, decls, out));        

    // for recursive methods, no prestate can guarantee termination
    if (recursive) return ExpressionLogic.booleanLiteral(false);

    // for a non-recursive method, we state non-divergence (J is not needed)
    TypedIdentifier[] tids0 = new TypedIdentifier[n];
    for (int i=0; i<n; i++) tids0[i] = tids[i];
    return Satisfies.forallFormula(tids0,
        ImpliesFormula.construct(argsFormula, notDiverges));   
    */ 
  }
}
