// ----------------------------------------------------------------------------
// Main.java
// Command line interface to the RISC ProgramExplorer
// $Id: Main.java,v 1.92 2020/12/07 09:28:38 schreine Exp schreine $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer;

import java.io.*;
import java.util.*;

import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Program.NamedType;
import fmrisc.ProgramExplorer.Syntax.Program.TypeExpression;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.InOut.Package;
import fmrisc.ProgramExplorer.SWT.*;
import fmrisc.ProgramExplorer.Tasks.MethodTasks;
import fmrisc.ProgramExplorer.Tasks.Task;
import fmrisc.ProgramExplorer.Proving.*;

/**
 * Command line interface to the RISC ProgramExplorer
 */
public class Main
{
  // software version
  public static final String version =
    "2.27 (December 7, 2020)";
  
  // message printed on startup
  public static final String copyright =
    "RISC ProgramExplorer Version " + version + "\n" + 
    "http://www.risc.jku.at/research/formal/software/ProgramExplorer\n" +
    "(C) 2008-, Research Institute for Symbolic Computation (RISC)\n" +
    "This is free software distributed under the terms of the GNU GPL.\n" +
    "Execute \"ProgramExplorer -h\" to see the options available.\n" +
    "------------------------------------------------------------------";
 
  // command line options
  private static String unitName = null;   // name of unit to be processed
  private static boolean nogui = false;    // no GUI is to be started
  private static boolean silent = false;   // true iff no copyright information
  private static File[] path = null;       // directories for top-level package
  private static boolean printing = false; // print class
  private static boolean theory = false;   // unit is theory
  private static boolean firstOrder = true; // check first-order constraint
  private static boolean toLogical = true; // translate BOOLEAN to LOGICAL
  
  // System output and error stream
  private static PrintWriter output = new PrintWriter(System.out);
  private static ErrorStream error = null; // initialized by init()
  
  // compilation/execution properties
  private static String compileCommand = null;
  private static String executeCommand = null;
  private static String workingDirectory = null;
  private static String mainClass = null;
  private static String cvclCommand = null;
  private static String cvc3Command = null;
  private static String cvc4Command = null;
  
  // the currently selected prover command
  private static String proverCommand = null;
  
  // the currently selected prover type
  private static fmrisc.ProofNavigator.State.ProverVersion proverVersion = null;
  private static String proverVersionString = null;
  
  // time of the current invocation
  private static long timeStamp = 0;
  
  // the task directory
  private static File taskDirectory = null;
  
  /***************************************************************************
   * Command line interface to the RISC ProgramExplorer.
   * The program reads the following environment variables:
   * 
   * PE_CLASSPATH 
   *   the directories (separated by ":") representing the top package
   *   (default the current working directory)
   * PE_CVCL
   *   the command for executing the cvcl checker (default "cvcl")
   * PE_JAVAC 
   *   the command for compiling java programs (default "javac")
   * PE_JAVA  
   *   the command for executing java programs (default "java")
   * PE_CWD   
   *   the directory used for compiling/executing 
   *   (default the current working directory)
   * PE_MAIN  
   *   the name of the main class of the program (default "Main")
   *   
   * @param args command line arguments, see help()
   **************************************************************************/
  public static void main(String[] args)
  {     
    // process arguments
    processArgs(args);

    // initialize system properties
    initProperties();
 
    // if GUI is requested, start it
    if (nogui)
      runNoGUI();
    else
      MainSWT.main();
  }
  
  /****************************************************************************
   * Print help message.
   * @param exit iff true, then terminate program after message
   ***************************************************************************/
  private static void help(boolean exit)
  {
    System.out.println(copyright);
    System.out.println();
    System.out.println("Usage: ProgramExplorer [OPTION]...");
    // System.out.println("UNIT: (qualified) name of unit (class/theory) to be processed.");
    System.out.println("OPTION: one of the following options:");
    System.out.println("\t-h, --help: print this message.");
    // System.out.println("\t-n, --nogui: do not start GUI.");
    // System.out.println("\t-s, --silent: omit startup message.");
    System.out.println("\t-cp, --classpath [PATH]: directories representing top package.");
    // System.out.println("\t-p, --print: print unit declaration.");
    // System.out.println("\t-t, --theory: unit is theory (default: unit is class) ");
    System.out.println();
    System.out.println("Environment Variables:");
    System.out.println("\tPE_CLASSPATH:");
    System.out.println("\t\tthe directories (separated by \":\") " +
        "representing the top package");
    System.out.println("\t\t(default the current working directory)");
    System.out.println("\tPE_JAVAC"); 
    System.out.println("\t\tthe command for compiling java programs (default \"javac\")");
    System.out.println("\tPE_JAVA");  
    System.out.println("\t\tthe command for executing java programs (default \"java\")");
    System.out.println("\tPE_CWD");   
    System.out.println("\t\tthe directory used for compiling/executing ");
    System.out.println("\t\t(default the current working directory)");
    System.out.println("\tPE_MAIN");  
    System.out.println("\t\tthe name of the main class of the program (default \"Main\")");
    System.out.println("\tPE_CVCL"); 
    System.out.println("\t\tthe command for executing the cvcl checker (default \"cvcl\")");
    System.out.println("\tPE_CVC3"); 
    System.out.println("\t\tthe command for executing the cvc3 checker (default \"cvc3\")");
    System.out.println("\tPE_CVC4"); 
    System.out.println("\t\tthe command for executing the cvc4 checker (default \"cvc4\")");
    System.out.println("\tPE_CHECKER"); 
    System.out.println("\t\tthe initially selected checker (CVCL, CVC3, CVC4)");
    System.out.println("\t\t(default CVC3)");
    System.out.println("\tPE_HIGHERORDER"); 
    System.out.println("\t\tset for higher-order logic (only for CVCL)");
    System.out.println("\tPE_BOOLEANFORMULA"); 
    System.out.println("\t\tset if formula type is BOOLEAN (else LOGICAL)");
    if (exit) exit(0);
  }
  
  /****************************************************************************
   * Initialize the system properties.
   ***************************************************************************/
  private static void initProperties()
  { 
    // set workspace path
    String classPath = System.getenv("PE_CLASSPATH");
    if (classPath != null) path = filePaths(classPath);
    if (path == null) path = 
      new File[] { new File(System.getProperty("user.dir")) } ;
    
    // validate path
    for (File dir : path)
    {
      if (!dir.isDirectory() || !dir.canRead())
      {
        System.out.print("The RISC ProgramExplorer cannot start because ");
        if (!dir.isDirectory())
        {
          System.out.print("the specified path\n  " + dir.getAbsolutePath() + 
          "\ndoes not denote a directory. ");
        }
        else
        {
          System.out.print("it cannot open the specified directory\n  " + 
              dir.getAbsolutePath()); 
        }
        System.out.println("Check the command line option\n  -cp(--classpath)\n" +
          "respectively the environment variable\n  PE_CLASSPATH.");
        System.out.println();
        help(true);
      }
    }
    
    // set information required for compiling/executing a program
    if (compileCommand == null) compileCommand = System.getenv("PE_JAVAC");
    if (compileCommand == null) compileCommand = "javac";
    if (executeCommand == null) executeCommand = System.getenv("PE_JAVA");
    if (executeCommand == null) executeCommand = "java";
    if (workingDirectory == null) workingDirectory = System.getenv("PE_CWD");
    if (workingDirectory == null) 
      workingDirectory = System.getProperty("user.dir");
    if (mainClass == null) mainClass = System.getenv("PE_MAIN");
    if (mainClass == null) mainClass = "Main";
    if (cvclCommand == null) cvclCommand = System.getenv("PE_CVCL");
    if (cvclCommand == null) cvclCommand = "cvcl";
    if (cvc3Command == null) cvc3Command = System.getenv("PE_CVC3");
    if (cvc3Command == null) cvc3Command = "cvc3";
    if (cvc4Command == null) cvc4Command = System.getenv("PE_CVC4");
    if (cvc4Command == null) cvc4Command = "cvc4";
    if (proverVersion == null)
    {
      String version = System.getenv("PE_CHECKER");
      if (version != null)
      {
        switch (version)
        {
          case "CVCL": 
            proverVersion = fmrisc.ProofNavigator.State.ProverVersion.CVCL;
            proverVersionString = "CVCL";
            break;
          case "CVC3": 
            proverVersion = fmrisc.ProofNavigator.State.ProverVersion.CVC3;
            proverVersionString = "CVC3";
            break;
          case "CVC4": 
            proverVersion = fmrisc.ProofNavigator.State.ProverVersion.CVC4;
            proverVersionString = "CVC4";
            break;
          default:
            System.out.println("unsupported value of environment variable PE_CHECKER");
            System.out.println();
            help(true);
        }
      }
    }
    if (proverVersion == null) 
    {
      proverVersion = fmrisc.ProofNavigator.State.ProverVersion.CVC3;
      proverVersionString = "CVC3";
    }
    if (proverVersion == fmrisc.ProofNavigator.State.ProverVersion.CVCL &&
        System.getenv("PE_HIGHERORDER") != null)
      firstOrder = false;
    if (System.getenv("PE_BOOLEANFORMULA") != null)
      toLogical = false;
        
    // set prover command
    switch (proverVersion)
    {
      case CVCL: 
        proverCommand = cvclCommand;
        break;
      case CVC3: 
        proverCommand = cvc3Command;
        break;
      case CVC4: 
        proverCommand = cvc4Command;
        break;
      default:
        internalError("unknown type of prover command");
    } 
    
    // set also ProofNavigatorState accordingly
    fmrisc.ProofNavigator.Main.setProperties(
        workingDirectory+"/.ProofNavigator",
        proverCommand, proverVersion,
        firstOrder, toLogical);
  }
  
  /***************************************************************************
   * Perform initialization.
   * @param top the top-level window (may be null)
   **************************************************************************/
  public static void init(TopWindow top)
  {    
    // print copyright message
    if (!silent) output.println(copyright);
    
    // initialize error stream
    error = new ErrorStream("ERROR", "   ", "", output, top);
  }
  
  /***************************************************************************
   * Program execution without GUI.
   **************************************************************************/
  private static void runNoGUI()
  {
    init(null);
    PackageSymbol top = processPackages();
    if (top == null || unitName == null) return;
    Symbol symbol = getUnit(top, unitName, theory);
    if (symbol == null) return;
    if (theory)
      processTheory((TheorySymbol)symbol);
    else
      processClass((ClassSymbol)symbol);
  }
  
  /***************************************************************************
   * Reset system state.
   **************************************************************************/
  public static void resetState()
  {
    ProofNavigatorProblem.closeProof();
    fmrisc.ProofNavigator.Main.init();
  }
  
  /***************************************************************************
   * Process packages as denoted by user path.
   * @return the top-level package (null, if an error occurred).
   **************************************************************************/
  public static PackageSymbol processPackages()
  {
    // reset time stamp
    timeStamp = new Date().getTime();
    
    // reset number of error messages
    error.reset();
    
    // set up directories
    File dir = new File(workingDirectory);
    if (!dir.isDirectory() || !dir.canWrite())
    {
      error.explainError("cannot write to directory " + dir);
      return null;
    }
    taskDirectory = Directory.setup(dir, error);
    if (taskDirectory == null) return null;
     
    // construct top-level package
    PackageSymbol top = Package.construct(path, error, taskDirectory);
    if (top == null) 
    {
      error.explainError("could not construct package environment");
      return null;
    }
    
    // initialize type translator (build base theory)
    top.getTypeTranslator().init(top, error);
    
    // process String class
    TypeExpression stringType = top.getTypeTable().getStringType();
    if (stringType instanceof NamedType)
    {
      NamedType stringType0 = (NamedType)stringType;
      boolean okay = stringType0.getSymbol().read(error);
      if (!okay)
      {
        error.newMessage("error in processing class " + stringType0);
        return top;
      }
    }
    
    // we are done
    error.newMessage("initialization completed.");
    
    // generate tasks
    MethodTasks.generateTasks(output);
    output.println("done.");
    
    // return top-level package
    return top;
  }
  
  /***************************************************************************
   * Process class.
   * @param csymbol the symbol denoting the class.
   **************************************************************************/
  public static void processClass(ClassSymbol csymbol)
  { 
    // reset number of error messages
    error.reset();
    
    // process declaration
    csymbol.reset(); // force re-processing
    csymbol.read(error);
    String text1 = "class " + csymbol.getLongName() + " was processed with ";
    String text2;
    int number = error.getErrorNumber();
    if (number == 0)
      text2 = "no errors.";
    else if (number == 1)
      text2 = "1 error.";
    else
      text2 = number + " errors.";
    error.newMessage(text1 + text2);
    
    // optionally print class
    if (printing)
    {
      output.println("Print class declaration");
      output.println("-----------------------");
      PrintWriter out = new PrintWriter(output);
      csymbol.getDeclaration().print(out);
    }
  }
  
  /***************************************************************************
   * Process theory.
   * @param tsymbol the symbol denoting the theory.
   **************************************************************************/
  public static void processTheory(TheorySymbol tsymbol)
  { 
    // generated theory is processed when class is processed
    if (tsymbol.isFromClass()) return;
    
    // reset number of error messages
    error.reset();
    
    // force re.processing of declaration
    tsymbol.reset(); 
    tsymbol.read(error);
    int number = error.getErrorNumber();
    
    // generate output
    String text1 = "theory " + tsymbol.getLongName() + " was processed with ";
    String text2;
    if (number == 0)
      text2 = "no errors";
    else if (number == 1)
      text2 = "1 error";
    else
      text2 = number + " errors";
    error.newMessage(text1 + text2);
    
    // optionally print theory
    if (printing)
    {
      output.println("Print theory declaration");
      output.println("-----------------------");
      PrintWriter out = new PrintWriter(output);
      tsymbol.getDeclaration().print(out);
    }
  }

  /****************************************************************************
   * Process arguments and set corresponding flags.
   * @param args the command line arguments.
   ***************************************************************************/
  private static void processArgs(String[] args)
  {
    int n = args.length;
    int i=0;
    while (i<n)
    {
      String arg = args[i];
      if (arg.equals("-h") || arg.equals("--help"))
        help(true);
      // else if (arg.equals("-s") || arg.equals("--silent"))
      //  silent = true;
      // else if (arg.equals("-n") || arg.equals("--nogui"))
      //   nogui = true;
      else if (arg.equals("-cp") || arg.equals("--classpath"))
      {
        i = i+1;
        if (i == n) help(true);
        path = filePaths(args[i]);
      }
      // else if (arg.equals("-p") || arg.equals("--print"))
      // {
      //   printing = true;
      // }
      // else if (arg.equals("-t") || arg.equals("--theory"))
      //   theory = true;
      // else if (arg.charAt(0) != '-' && unitName == null)
      // {
      //  if (i == n-1)
      //    unitName = args[i];
      //  else
      //    help(true);
      // }
      else
        help(true);
      i = i+1;
    }
  }
 
  /***************************************************************************
   * Convert string to file paths
   * @param string a sequence of file paths separated by ":"
   * @return the corresponding array of file paths.
   **************************************************************************/
  private static File[] filePaths(String string)
  {
    String[] strings = string.split(":");
    int len = strings.length;
    File[] path = new File[len];
    for (int j=0; j<len; j++)
      path[j] = new File(strings[j]);
    return path;
  }
  
  /***************************************************************************
   * Get symbol of unit denoted by qualified name in top-level package.
   * @param top the top-level package
   * @param name the qualified name of the package
   * @param theory true if unit is theory (class, otherwise)
   * @return the package symbol (or null, if an error occurred)
   **************************************************************************/
  public static Symbol getUnit(PackageSymbol top, String name, boolean theory)
  {
    // parse string
    String[] names = name.split("\\.");
    int n = names.length;
    if (n == 0)
    {
      error.newError("empty class name encountered");
      return null;
    }
    
    // find parent package of class
    PackageSymbol psymbol = top;
    for (int i=0; i<n-1; i++)
    {
      Identifier ident = Identifier.construct(names[i]);
      PackageSymbol psymbol0 = psymbol.getPackages().get(ident);
      if (psymbol0 == null)
      {
        String pname;
        if (i == 0)
          pname = "default package";
        else
          pname = "package " + psymbol.getLongName().toString();
        error.newError("no package named \"" + ident + 
            "\" found in " + pname);
        return null;
      }
      psymbol = psymbol0;        
    }
    
    // find unit in parent package
    Identifier uname = Identifier.construct(names[n-1]);
    Symbol symbol;
    if (theory)
      symbol = psymbol.getTheories().get(uname);
    else
      symbol = psymbol.getClasses().get(uname);
    if (symbol == null)
    {
      String pname;
      if (n == 1)
        pname = "default package";
      else
        pname = "package " + psymbol.getLongName().toString();
      if (theory)
        error.newError("no theory named \"" + uname + "\" found in " + pname);
      else
        error.newError("no class named \"" + uname + "\" found in " + pname);
      return null;
    }
    
    // return result
    return symbol;
  }
  
  /***************************************************************************
   * Terminate program because of an internal error.
   * @param message the message indicating the error.
   **************************************************************************/
  public static void internalError(String message)
  {
    System.out.println("INTERNAL ERROR: " + message + "\n");
    Thread.dumpStack();
    exit(-1);
  }
  
  /***************************************************************************
   * Terminate program with denoted exit code.
   * @param code the exit code.
   **************************************************************************/
  public static void exit(int code)
  {
    cleanDirectory(taskDirectory);
    System.exit(code);
  }
  
  /***************************************************************************
   * Clean directory from auxiliary files (.log and .xhtml).
   * @param dir the directory
   **************************************************************************/
  private static void cleanDirectory(File dir)
  {
    if (dir == null) return;   // directory may not have been initialized yet
    File[] files = dir.listFiles(new JunkFilter());
    if (files == null) return; // directory may not exist any more
    for (File file : files)
    {
      if (file.isDirectory())
      {
        cleanDirectory(file);
        continue;
      }
      String name = file.getName();
      if (name.endsWith(".xhtml") || name.endsWith(".log"))
        file.delete();
    }
  }
  
  /****************************************************************************
   * Filter for junk files.
   ***************************************************************************/
  private static class JunkFilter implements FileFilter
  {
    public boolean accept(File file)
    {
      if (file.isDirectory()) return true;
      String name = file.getName();
      if (name.endsWith(".xhtml")) return true;
      if (name.endsWith(".log")) return true;
      return false;
    }
  }
  
  /***************************************************************************
   * Return version string.
   * @return the version string.
   **************************************************************************/
  public static String getVersion()
  {
    return version;
  }
  
  /***************************************************************************
   * Set output stream for printing messages.
   * @param out the output stream.
   **************************************************************************/
  public static void setOutput(PrintWriter out)
  {
    output = out;
  }
  
  /***************************************************************************
   * Get error stream for printing error messages.
   * @return the error stream.
   **************************************************************************/
  public static ErrorStream getError()
  {
    return error;
  }
  
  /***************************************************************************
   * Reset number of errors to zero.
   **************************************************************************/
  public static void resetErrors()
  {
    error.reset();
  }
  
  /***************************************************************************
   * Get path for looking up source files.
   * @return the path.
   **************************************************************************/
  public static File[] getPath()
  {
    return path;
  } 
  
  /***************************************************************************
   * Set path for looking up source files.
   * @param path the path.
   **************************************************************************/
  public static void setPath(File[] path)
  {
    Main.path = path;
  } 
  
  /***************************************************************************
   * Get the system properties.
   * @return an array of 7 elements holding 
   *         - the compilation command
   *         - the execution command
   *         - the path of the working directory
   *         - the name of the main class
   *         - the CVCL command
   *         - the CVC3 command
   *         - the CVC4 command
   *         - the version of the prover (as a string)
   *         - TRUE if firstOrder is set
   *         - TRUE if toLogical is set
   **************************************************************************/
  public static String[] getProperties()
  {
    return new String[] 
      { compileCommand, executeCommand, workingDirectory, mainClass,
        cvclCommand, cvc3Command, cvc4Command,
        proverVersionString,
        firstOrder ? "TRUE" : "FALSE",
        toLogical ? "TRUE" : "FALSE" };
  }
  
  /***************************************************************************
   * Set the system properties
   * @param compileCommand the compilation command
   * @param executeCommand the execution command
   * @param workingDirectory the path of the working directory
   * @param mainClass the name of the main class
   * @param cvclCommand the CVCL command
   * @param cvc3Command the CVC3 command
   * @param cvc4Command the CVC4 command
   * @param proverVersion the version of the prover
   * @param firstOrder TRUE if prover is first order
   * @param toLogical TRUE if type LOGICAL is to be used
   **************************************************************************/
  public static void setProperties(String compileCommand,
    String executeCommand, String workingDirectory, String mainClass,
    String cvclCommand, String cvc3Command, String cvc4Command,
    String proverVersion, String firstOrder, String toLogical)
  {
    // set properties except for proverCommand and workingDirectory
    Main.compileCommand = compileCommand;
    Main.executeCommand = executeCommand;
    Main.mainClass = mainClass;
    Main.cvclCommand = cvclCommand;
    Main.cvc3Command = cvc3Command;
    Main.cvc4Command = cvc4Command;
    Main.proverVersionString = proverVersion;
    Main.firstOrder = firstOrder.equals("TRUE");
    Main.toLogical = toLogical.equals("TRUE");
    
    // set prover version and determine new prover command
    String proverCommand = null;
    switch (proverVersion)
    { 
      case "CVCL":
        Main.proverVersion = fmrisc.ProofNavigator.State.ProverVersion.CVCL;
        proverCommand = cvclCommand;
        break;
      case "CVC3":
        Main.proverVersion = fmrisc.ProofNavigator.State.ProverVersion.CVC3;
        proverCommand = cvc3Command;
        break;
      case "CVC4":
        Main.proverVersion = fmrisc.ProofNavigator.State.ProverVersion.CVC4;
        proverCommand = cvc4Command;
        break;
      default:
        internalError("unknown prover version");
    }
    
    // ProofNavigator may have to be informed about new properties
    if (proverCommand != Main.proverCommand || 
        workingDirectory != Main.workingDirectory)
    {
      String contextDir = workingDirectory + "/.ProofNavigator";
      boolean okay = 
          fmrisc.ProofNavigator.State.init(contextDir, proverCommand, Main.proverVersion,
              Main.firstOrder, Main.toLogical);
      if (okay)
        fmrisc.ProofNavigator.Main.setProperties(contextDir, proverCommand, Main.proverVersion,
            Main.firstOrder, Main.toLogical);
      else
      {
        error.newError("cannot change prover command and/or working directory");
        cvclCommand = Main.proverCommand;
        workingDirectory = Main.workingDirectory;
        okay = 
            fmrisc.ProofNavigator.State.init(contextDir, proverCommand, Main.proverVersion,
                Main.firstOrder, Main.toLogical);
        if (!okay) Main.internalError("cannot return to valid state");
      }
      // fmrisc.ProofNavigator.Main.displayDeclarations();
    }
    
    // set remaining properties
    Main.proverCommand = proverCommand;
    Main.workingDirectory = workingDirectory;
  }
  
  /***************************************************************************
   * Called if a proof is completed by the ProofNavigator.
   * @param proof the completed proof.
   **************************************************************************/
  public static void proofDone(fmrisc.ProofNavigator.Proving.Proof proof)
  {
    MainSWT.displayAnalyzeView();
    ProofNavigatorProblem.proofDone(proof);
  }
  
  /***************************************************************************
   * Get time stamp of current program execution.
   * @return the time stamp.
   **************************************************************************/
  public static long getTimeStamp()
  {
    return timeStamp;
  }
  
  /***************************************************************************
   * Display info (proof) associated to task.
   * @param task the denoted task
   **************************************************************************/
  public static void displayInfo(Task task)
  {
    Object info = task.getInfo();
    if (info instanceof fmrisc.ProofNavigator.Proving.Proof)
    {
      MainSWT.displayVerifyView();
      File dir = new File(task.getDirectory(), Directory.CONTEXT);
      fmrisc.ProofNavigator.State.setContextDir(dir);
      fmrisc.ProofNavigator.Proving.Proof proof =
        (fmrisc.ProofNavigator.Proving.Proof)info;
      // fmrisc.ProofNavigator.State.setProof(proof);
      fmrisc.ProofNavigator.SWT.MainSWT.setProof(proof, false);
      return;
    }
  }
  
  /***************************************************************************
   * Return true if first order check is to be applied.
   * @return true if first order check is to be applied.
   **************************************************************************/
  public static boolean firstOrder()
  {
    return firstOrder;
  }
  
  /***************************************************************************
   * Return true if BOOLEAN is translated to LOGICAL.
   * @return true if BOOLEAN is translated to LOGICAL.
   **************************************************************************/
  public static boolean toLogical()
  {
    return firstOrder && toLogical;
  }
  
  /***************************************************************************
   * Return version of default prover.
   * @return the prover version.
   **************************************************************************/
  public static fmrisc.ProofNavigator.State.ProverVersion proverVersion()
  {
    return proverVersion;
  }
}
//----------------------------------------------------------------------------
//end of file
//----------------------------------------------------------------------------