// ----------------------------------------------------------------------------
// SymbolTree.java
// A tree displaying symbol structures.
// $Id: SymbolTree.java,v 1.32 2017/10/24 08:28:35 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.SWT;

import java.io.*;
import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Program.ClassDeclaration;
import fmrisc.ProgramExplorer.Syntax.Spec.TheoryDeclaration;

import org.eclipse.swt.*;
import org.eclipse.swt.widgets.*;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.layout.*;

/**
 * 
 * A tree displaying symbol structures.
 *
 */
public final class SymbolTree
{
  // the tree (subclassing not allowed)
  private Tree tree;
  
  // the images used for tree nodes
  private Image packageImage;
  private Image classImage;
  private Image theoryImage;
  private Image classVarImage;
  private Image objectVarImage;
  private Image constructorImage;
  private Image classMethodImage;
  private Image objectMethodImage;
  private Image parameterImage;
  private Image typeImage;
  private Image formulaImage;
  private Image valueImage;
  
  // other images
  private Image runImage;
  private Image analyzeImage;
  private Image errorImage;
  private Image detailImage;
  
  // all items in the tree and the associated symbols
  private Map<TreeItem,Symbol> itemSet;
  private Map<Symbol,SymbolItem> symbolSet;
  
  // the top-level package
  private PackageSymbol topPackage = null;
  
  // the top-level window
  private TopWindow topWindow = null;
  
  /***************************************************************************
   * A tree entry.
   **************************************************************************/
  private final class SymbolItem
  {
    TreeItem node; // the tree item
    
    /*************************************************************************
     * Create a node linked to parent node.
     * @param parent the parent (object of type Tree or SymbolItem)
     * @param file the symbol represented by the node.
     ************************************************************************/
    public SymbolItem(Object parent, Symbol symbol)
    {
      // initialize node
      if (parent instanceof Tree)
        node = new TreeItem((Tree)parent, SWT.NULL);
      else if (parent instanceof SymbolItem)
        node = new TreeItem(((SymbolItem)parent).node, SWT.NULL);
      
      // register symbol item
      itemSet.put(node, symbol);
      symbolSet.put(symbol, this);
      
      // set text 
      if (symbol instanceof ParamSymbol)
      {
        ParamSymbol psymbol = (ParamSymbol)symbol;
        String name = Syntax.procHeader(symbol.getName().getString(),
            psymbol.getArgumentTypes());
        node.setText(name);
      }
      else
        node.setText(symbol.getName().getString());
      
      // set image
      if (symbol instanceof PackageSymbol)
        node.setImage(packageImage);
      else if (symbol instanceof ClassSymbol)
        node.setImage(classImage);
      else if (symbol instanceof TheorySymbol)
        node.setImage(theoryImage);
      else if (symbol instanceof ClassVariableSymbol)
        node.setImage(classVarImage);
      else if (symbol instanceof ObjectVariableSymbol)
        node.setImage(objectVarImage);
      else if (symbol instanceof ConstructorSymbol)
        node.setImage(constructorImage);
      else if (symbol instanceof ClassMethodSymbol)
        node.setImage(classMethodImage);
      else if (symbol instanceof ObjectMethodSymbol)
        node.setImage(objectMethodImage);
      else if (symbol instanceof ParameterSymbol)
        node.setImage(parameterImage);
      else if (symbol instanceof TypeSymbol)
        node.setImage(typeImage);
      else if (symbol instanceof FormulaSymbol)
        node.setImage(formulaImage);
      else if (symbol instanceof ValueSymbol)
        node.setImage(valueImage);
    }
  }
  
  /****************************************************************************
   * Popup information on a symbol
   ***************************************************************************/
  private static class SymbolTip 
  {
    private Shell shell;
    private Label label;
    
    // only one instance of symbol tip at a time
    private static SymbolTip tip = null;
    
    /**************************************************************************
     * Create popup window displaying information on given symbol.
     * @param parent the parent window.
     * @param pos the position of the window.
     * @param symbol the symbol.
     *************************************************************************/
    public SymbolTip(Shell parent, Point pos, Symbol symbol)
    {     
      close();
      tip = this;
      
      shell = new Shell(parent, SWT.ON_TOP | SWT.TOOL); 
      GridLayout gridLayout = new GridLayout();
      gridLayout.numColumns = 1;
      gridLayout.marginWidth = 2;
      gridLayout.marginHeight = 2;
      shell.setLayout(gridLayout);
      Color color = shell.getDisplay().getSystemColor(SWT.COLOR_INFO_BACKGROUND);
      Color tcolor = shell.getDisplay().getSystemColor(SWT.COLOR_INFO_FOREGROUND);
      shell.setBackground(color);
      
      // determine text
      StringWriter out = new StringWriter();
      symbol.print(new PrintWriter(out));
      String text = out.toString();
      
      // set label
      label = new Label(shell, SWT.NONE);
      label.setLayoutData(new GridData(GridData.FILL_HORIZONTAL
          | GridData.VERTICAL_ALIGN_CENTER));
      label.setText(text);
      label.setBackground(color);
      label.setForeground(tcolor);
      
      shell.addMouseTrackListener(new MouseTrackAdapter() { 
        public void mouseEnter(MouseEvent e)
        {
          close();
        }
      });
      
      // open window
      shell.pack();
      shell.setLocation(pos.x+10, pos.y+10);
      shell.open();
    }
    
    /*************************************************************************
     * Close tip instance.
     ************************************************************************/
    public static void close()
    {
      if (tip != null) tip.shell.close();
      tip = null;
    }
  }
  
  /***************************************************************************
   * Create a tree displaying symbol structures.
   * @param parent the parent widget.
   * @param top the top window.
   **************************************************************************/
  public SymbolTree(Composite parent, final TopWindow top)
  {    
    // store top window
    topWindow = top;
    
    // set images
    packageImage = top.getPackageImage();
    classImage = top.getClassImage();
    theoryImage = top.getTheoryImage();
    classVarImage = top.getClassVarImage();
    objectVarImage = top.getObjectVarImage();
    constructorImage = top.getConstructorImage();
    classMethodImage = top.getClassMethodImage();
    objectMethodImage = top.getObjectMethodImage();
    parameterImage = top.getParameterImage();
    typeImage = top.getTypeImage();
    formulaImage = top.getFormulaImage();
    valueImage = top.getValueImage();
    runImage = top.getRunImage();
    analyzeImage = top.getAnalyzeImage();
    errorImage = top.getErrorImage();
    detailImage = top.getDetailImage();
    
    // initialize tree
    tree = new Tree(parent, SWT.SINGLE);
  
    // initialize node/symbol sets
    initSets();
    
    // set menu
    final Menu menu = new Menu(tree.getShell(), SWT.POP_UP);
    tree.setMenu(menu);
    menu.addListener (SWT.Show, new Listener () {
      public void handleEvent (Event event) {
        MenuItem [] menuItems = menu.getItems ();
        for (int i=0; i<menuItems.length; i++) {
          menuItems [i].dispose ();
        }
        TreeItem [] treeItems = tree.getSelection ();
        if (treeItems.length != 1) return;
        final TreeItem item = treeItems[0];
        final Symbol symbol = itemSet.get(item);
        if (symbol instanceof ClassSymbol || symbol instanceof TheorySymbol)
        {
          MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
          menuItem.setText ("Process");
          menuItem.setImage(runImage);
          menuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
              processSymbol(item);
            }
          });
        }
        MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
        menuItem.setText ("Print Symbol");
        menuItem.setImage(analyzeImage);
        menuItem.addSelectionListener(new SelectionAdapter() {
          public void widgetSelected(SelectionEvent e) {
            topWindow.print(symbol);
          }
        });
        MenuItem menuItem2 = new MenuItem (menu, SWT.PUSH);
        menuItem2.setText ("Print Declaration");
        menuItem2.setImage(analyzeImage);
        menuItem2.addSelectionListener(new SelectionAdapter() {
          public void widgetSelected(SelectionEvent e) {
            topWindow.printDecl(symbol);
          }
        });
        if (symbol instanceof ParamSymbol)
        {
          MenuItem menuItem0 = new MenuItem (menu, SWT.PUSH);
          menuItem0.setText ("Show Semantics");
          menuItem0.setImage(detailImage);
          menuItem0.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
              topWindow.showMethodDetail((ParamSymbol)symbol);
            }
          });
        }
      }
    });

    // set click handlers
    tree.addSelectionListener(new SelectionAdapter() {
      
      // item is double clicked
      public void widgetDefaultSelected(SelectionEvent e) 
      { 
        TreeItem item = (TreeItem)e.item;
        processSymbol(item);
      }
      
      // item is selected by single click
      public void widgetSelected(SelectionEvent e) {
      }
    });
    
    // set hover handler
    final Composite parent0 = parent;
    tree.addMouseTrackListener(new MouseTrackAdapter() {
      public void mouseHover(MouseEvent e) {
        Tree tree = (Tree)e.widget;
        Symbol symbol = find(tree.getItems(), e.x, e.y);
        if (symbol == null) return;
        new SymbolTip(parent0.getShell(), tree.toDisplay(e.x, e.y), symbol);
      }
      public void mouseExit(MouseEvent e) {
        SymbolTip.close();
      }
    });
  }
  
  /****************************************************************************
   * Find symbol associated to item in set of items denoted by mouse position.
   * @param items the set of tree items to be recursively searched.
   * @param x horizontal coordinate of mouse position.
   * @param y vertical coordinate of mouse position.
   * @return the denoted symbol (or null, if none was found)
   ***************************************************************************/
  private Symbol find(TreeItem[] items, int x, int y)
  {
    for (TreeItem item : items)
    {
      TreeItem[] items0 = item.getItems();
      Symbol symbol = find(items0, x, y);
      if (symbol != null) return symbol;
      symbol = itemSet.get(item);
      if (symbol == null) continue;
      Rectangle b = item.getBounds();
      if (x < b.x || x > b.x + b.width) continue;
      if (y < b.y || y > b.y + b.height) continue;
      return symbol;
    }
    return null;
  }
  
  /***************************************************************************
   * Process symbol denoted by tree item.
   * @param item the tree item.
   **************************************************************************/
  private void processSymbol(TreeItem item)
  {
    Symbol symbol = itemSet.get(item);
    if (symbol instanceof ClassSymbol)
    {
      ClassSymbol csymbol = (ClassSymbol)symbol;
      // make sure class is displayed 
      topWindow.displaySource(csymbol.getSource());
      if (!csymbol.isRead())
      {
        topWindow.processClass(csymbol);
      }
    }
    else if (symbol instanceof TheorySymbol)
    {
      TheorySymbol tsymbol = (TheorySymbol)symbol;
      // make sure class is displayed 
      Source source = tsymbol.getSource();
      if (source != null) topWindow.displaySource(source);
      if (!tsymbol.isRead())
      {
        topWindow.processTheory(tsymbol);
      }
    }
    item.setExpanded(true);
    SourcePosition pos = getPosition(symbol);
    if (pos == null) 
    {
      // symbol without source, print definition instead
      if (symbol.getDeclaration() != null)
        topWindow.printDecl(symbol);
    }
    else if (symbol instanceof ClassSymbol || symbol instanceof TheorySymbol)
      topWindow.displaySource(pos, false);
    else
      topWindow.displaySource(pos, true);
    redraw(topPackage);
    topWindow.displayTasks(symbol);
  }
  
  /****************************************************************************
   * Initizialize node/symbol sets.
   ***************************************************************************/
  private void initSets()
  {
    itemSet = new MapClass<TreeItem,Symbol>();
    symbolSet = new MapClass<Symbol,SymbolItem>();
  }
  
  /****************************************************************************
   * Get visual representation of tree.
   * @return the visual representation.
   ***************************************************************************/
  public Tree getTree()
  {
    return tree;
  }
  
  
  /****************************************************************************
   * Get symbols selected in tree.
   * @return the symbols.
   ***************************************************************************/
  public Symbol[] getSelection()
  {
    TreeItem[] items = tree.getSelection();
    int n = items.length;
    Symbol[] symbols = new Symbol[n];
    for (int i=0; i<n; i++)
    {
      TreeItem item = items[i];
      Symbol symbol = itemSet.get(item);
      if (symbol == null) Main.internalError("no symbol for selection");
      symbols[i] = symbol;
    }
    return symbols;
  }
  
  /****************************************************************************
   * Redraw symbol structures.
   * @param root the symbol of the top-level package
   *        (may be null, the tree is then erased)
   ***************************************************************************/
  public void redraw(PackageSymbol root)
  {
    topPackage = root;
    if (root == null)
    {
      tree.removeAll();
      initSets();
      return;
    }
    redraw(tree, root, false);
  }
  
  // a class for ordering symbols by names
  private static class SymbolComparator implements Comparator<Object>
  {
    /**************************************************************************
     * The comparison operation for a comparator
     * @param o1 first object (must be of type Symbol)
     * @param o2 second object (must be of type Symbol)
     * @return negative, zero, positive number if name of o1 is lexically
     *         less than, equal, greater than the name of o2
     ************************************************************************/
    public int compare(Object o1, Object o2)
    {
      Symbol s1 = (Symbol)o1;
      Symbol s2 = (Symbol)o2;
      String name1 = s1.getName().getString();
      String name2 = s2.getName().getString();
      return name1.compareTo(name2);
    }
    
    private static SymbolComparator SC = new SymbolComparator();
    
    /*************************************************************************
     * Sort array of symbols by name in increasing order.
     * @param array array of symbols to be sorted.
     ************************************************************************/
    public static void sort(Symbol[] array)
    {
      Arrays.sort(array, SC);
    }
  }
  
  /***************************************************************************
   * Get children of parent node.
   * @param parent an object of type Tree or SymbolItem.
   * @return the children of the parent.
   **************************************************************************/
  private TreeItem[] getChildren(Object parent)
  {
    if (parent instanceof Tree)
      return ((Tree)parent).getItems();
    else if (parent instanceof SymbolItem)
      return ((SymbolItem)parent).node.getItems();
    else
    {
      Main.internalError("unknown kind of node");
      return null;
    }
  }
  
  /***************************************************************************
   * Indicate whether node is expanded.
   * @param node (object of Type SymbolItem or Tree)
   * @return true if node is expanded.
   **************************************************************************/
  private boolean isExpanded(Object node)
  {
    if (node instanceof SymbolItem)
      return ((SymbolItem)node).node.getExpanded();
    return false;
  }
  
  /***************************************************************************
   * Set expansion status of node.
   * @param node (object of Type SymbolItem or Tree)
   * @param expanded true iff node is to be expanded.
   **************************************************************************/
  private void setExpanded(Object node, boolean expanded)
  {
    if (node instanceof SymbolItem)
      ((SymbolItem)node).node.setExpanded(expanded);
  }
  
  /***************************************************************************
   * Dispose subtree denoted by node.
   * @param node the root of the subtree to be disposed.
   **************************************************************************/
  private void dispose(TreeItem node)
  {    
    disposeChildren(node);
    Symbol symbol = itemSet.get(node);
    symbolSet.remove(symbol);
    itemSet.remove(node);
    node.dispose();
  }
  
  /***************************************************************************
   * Dispose children of subtree denoted by node.
   * @param node the root of the subtree to be disposed.
   **************************************************************************/
  private void disposeChildren(TreeItem node)
  {
    TreeItem[] children = node.getItems();
    for (TreeItem c : children) dispose(c);
  }
  
  /***************************************************************************
   * Redraw symbol linked into parent component.
   * @param parent the parent component (type Tree or SymbolItem).
   * @param symbol the symbol to be displayed.
   **************************************************************************/
  private void redraw(Object parent, Symbol symbol)
  {
    redraw(parent, symbol, true);
  }
  
  /***************************************************************************
   * Redraw symbol linked into parent component.
   * @param parent the parent component (type Tree or SymbolItem).
   * @param symbol the symbol to be displayed.
   * @param create if a new node is to be created for symbol
   *        (otherwise its content will be linked into parent node).
   **************************************************************************/
  private void redraw(Object parent, Symbol symbol, boolean create)
  {
    // display symbol
    Object node;
    if (create)
    {
      // reuse existing tree node, if possible
      node = symbolSet.get(symbol);
      if (node == null)
        node = new SymbolItem(parent, symbol);
    }
    else
      node = parent;
    
    // display contents of packages
    if (symbol instanceof PackageSymbol)
    {
      PackageSymbol psymbol = (PackageSymbol)symbol;
      redraw(psymbol, node, false);
      return;
    }
    
    // display contents of classes
    if (symbol instanceof ClassSymbol)
    {
      ClassSymbol csymbol = (ClassSymbol)symbol;
      redraw(csymbol, false);
      return;
    }
    
    // display contents of theories
    if (symbol instanceof TheorySymbol)
    {
      TheorySymbol tsymbol = (TheorySymbol)symbol;
      redraw(tsymbol, false);
      return;
    }
    
    // display parameterized entity
    if (symbol instanceof ParamSymbol)
    {
      ParamSymbol psymbol = (ParamSymbol)symbol;
      redraw(psymbol);
      return;
    }
  }
  
  /****************************************************************************
   * Redraw node denoting a package symbol.
   * @param csymbol the package symbol.
   * @param node the parent node (object of type Tree or SymbolItem)
   * @param expanded true if the node is to be expanded
   *                 (otherwise it preserves its expansion status)
   ***************************************************************************/
  private void redraw(PackageSymbol psymbol, Object node, boolean expanded)
  {    
    // (re)set package image (may have been set to error image)
    SymbolItem mynode = symbolSet.get(psymbol);
    if (mynode != null) mynode.node.setImage(packageImage);
      
    // remember expansion status and children
    expanded = expanded || isExpanded(node);
    TreeItem[] children = getChildren(node);
    
    // packages
    Collection<PackageSymbol> psymbols = psymbol.getPackageSet();
    PackageSymbol[] parray = new PackageSymbol[psymbols.size()];
    psymbols.toArray(parray);
    SymbolComparator.sort(parray);
    for (PackageSymbol psym : parray)
      redraw(node, psym);
    
    // classes
    Collection<ClassSymbol> csymbols = psymbol.getClassSet();
    ClassSymbol[] carray = new ClassSymbol[csymbols.size()];
    csymbols.toArray(carray);
    SymbolComparator.sort(carray);
    for (ClassSymbol csym : carray)
      redraw(node, csym);
    
    // theories
    Collection<TheorySymbol> tsymbols = psymbol.getTheorySet();
    TheorySymbol[] tarray = new TheorySymbol[tsymbols.size()];
    tsymbols.toArray(tarray);
    SymbolComparator.sort(tarray);
    for (TheorySymbol tsym : tarray)
      redraw(node, tsym);
    
    // remove children that are not in above lists
    Collection<Symbol> allSymbols = new SetClass<Symbol>();
    allSymbols.addAll(psymbols);
    allSymbols.addAll(csymbols);
    allSymbols.addAll(tsymbols);
    for (TreeItem child : children)
    {
      Symbol childSymbol = itemSet.get(child);
      if (!allSymbols.contains(childSymbol))
        dispose(child);
    }
    
    // reset expansion status
    setExpanded(node, expanded);
  }
  
  /****************************************************************************
   * Redraw node denoting a class symbol.
   * @param csymbol the class symbol.
   * @param expanded true if the node is to be expanded
   *                 (otherwise it preserves its expansion status)
   ***************************************************************************/
  private void redraw(ClassSymbol csymbol, boolean expanded)
  {        
    // class may have errors
    if (csymbol.hasErrors())
    {
      signalError(csymbol);
      return;
    }

    // get node information
    SymbolItem node = symbolSet.get(csymbol);
    if (node == null) return;
    
    // set class image
    node.node.setImage(classImage);
    
    // if class has not been read, we are done
    if (!csymbol.isRead()) return;
    
    // remember expansion status and children
    expanded = expanded || isExpanded(node);
    TreeItem[] children = getChildren(node);
    
    // local theory of class
    TheorySymbol tlsymbol = csymbol.getLocalTheory();
    if (tlsymbol != null) redraw(node, tlsymbol);
    
    // class variables
    Collection<ClassVariableSymbol> cvsymbols = csymbol.getClassVariableSet();
    ClassVariableSymbol[] cvarray = new ClassVariableSymbol[cvsymbols.size()];
    cvsymbols.toArray(cvarray);
    SymbolComparator.sort(cvarray);
    for (ClassVariableSymbol cvsym : cvarray)
      redraw(node, cvsym);
    
    // object variables
    Collection<ObjectVariableSymbol> ovsymbols = csymbol.getObjectVariableSet();
    ObjectVariableSymbol[] ovarray = new ObjectVariableSymbol[ovsymbols.size()];
    ovsymbols.toArray(ovarray);
    SymbolComparator.sort(ovarray);
    for (ObjectVariableSymbol ovsym : ovarray)
      redraw(node, ovsym);
    
    // constructors 
    Collection<ConstructorSymbol> cosymbols = csymbol.getConstructorSet();
    ConstructorSymbol[] coarray = new ConstructorSymbol[cosymbols.size()];
    cosymbols.toArray(coarray);
    SymbolComparator.sort(coarray);
    for (ConstructorSymbol cosym : coarray)
      redraw(node, cosym);
    
    // class methods
    Collection<ClassMethodSymbol> cmsymbols = csymbol.getClassMethodSet();
    ClassMethodSymbol[] cmarray = new ClassMethodSymbol[cmsymbols.size()];
    cmsymbols.toArray(cmarray);
    SymbolComparator.sort(cmarray);
    for (ClassMethodSymbol cmsym : cmarray)
      redraw(node, cmsym);
    
    // object methods
    Collection<ObjectMethodSymbol> omsymbols = csymbol.getObjectMethodSet();
    ObjectMethodSymbol[] omarray = new ObjectMethodSymbol[omsymbols.size()];
    omsymbols.toArray(omarray);
    SymbolComparator.sort(omarray);
    for (ObjectMethodSymbol omsym : omarray)
      redraw(node, omsym);
    
    // remove children that are not in above lists
    Collection<Symbol> allSymbols = new SetClass<Symbol>();
    if (tlsymbol != null) allSymbols.add(tlsymbol);
    allSymbols.addAll(cvsymbols);
    allSymbols.addAll(ovsymbols);
    allSymbols.addAll(cosymbols);
    allSymbols.addAll(cmsymbols);
    allSymbols.addAll(omsymbols);
    for (TreeItem child : children)
    {
      Symbol childSymbol = itemSet.get(child);
      if (!allSymbols.contains(childSymbol))
        dispose(child);
    }
    
    // reset expansion status
    setExpanded(node, expanded);
  }
  
  /****************************************************************************
   * Redraw node denoting a theory symbol.
   * @param tsymbol the theory symbol.
   * @param expanded true if the node is to be expanded
   *                 (otherwise it preserves its expansion status)
   ***************************************************************************/
  private void redraw(TheorySymbol tsymbol, boolean expanded)
  {        
    // theory may have errors
    if (tsymbol.hasErrors())
    {
      signalError(tsymbol);
      return;
    }
    
    // get node information
    SymbolItem node = symbolSet.get(tsymbol);
    if (node == null) return;
    
    // set theory image
    node.node.setImage(theoryImage);
    
    // if class has not been read, we are done
    if (!tsymbol.isRead()) return;
    
    // remember expansion status and children
    expanded = expanded || isExpanded(node);
    TreeItem[] children = getChildren(node);
    
    // types
    Collection<TypeSymbol> tsymbols = tsymbol.getTypes().getSymbols();
    TypeSymbol[] tsarray = new TypeSymbol[tsymbols.size()];
    tsymbols.toArray(tsarray);
    SymbolComparator.sort(tsarray);
    for (TypeSymbol tssym : tsarray)
      redraw(node, tssym);
    
    // values
    Collection<ValueSymbol> vsymbols = tsymbol.getValues().getSymbols();
    ValueSymbol[] vsarray = new ValueSymbol[vsymbols.size()];
    vsymbols.toArray(vsarray);
    SymbolComparator.sort(vsarray);
    for (ValueSymbol vssym : vsarray)
      redraw(node, vssym);
    
    // formulas
    Collection<FormulaSymbol> fsymbols = tsymbol.getFormulas().getSymbols();
    FormulaSymbol[] fsarray = new FormulaSymbol[fsymbols.size()];
    fsymbols.toArray(fsarray);
    SymbolComparator.sort(fsarray);
    for (FormulaSymbol fssym : fsarray)
      redraw(node, fssym);
    
    // remove children that are not in above lists
    Collection<Symbol> allSymbols = new SetClass<Symbol>();
    allSymbols.addAll(tsymbols);
    allSymbols.addAll(vsymbols);
    allSymbols.addAll(fsymbols);
    for (TreeItem child : children)
    {
      Symbol childSymbol = itemSet.get(child);
      if (!allSymbols.contains(childSymbol))
        dispose(child);
    }
    
    // reset expansion status
    setExpanded(node, expanded);
  }
  
  /****************************************************************************
   * Indicate that symbol has errors.
   * @param symbol the symbol that has errors.
   ***************************************************************************/
  public void signalError(Symbol symbol)
  {
    // get node information
    SymbolItem node = symbolSet.get(symbol);
    if (node == null) return;
   
    // set error image and dispose children
    node.node.setImage(errorImage);
    disposeChildren(node.node);
    
    // set error image on all ancestor packages
    TreeItem tnode = node.node.getParentItem();
    while (tnode != null)
    {
      tnode.setImage(errorImage);
      tnode = tnode.getParentItem();
    }
  }
  
  /****************************************************************************
   * Redraw node denoting a parameterized (constructor/method) symbol.
   * @param psymbol the parameterized symbol.
   ***************************************************************************/
  private void redraw(ParamSymbol psymbol)
  {
    // get node information
    SymbolItem node = symbolSet.get(psymbol);
    if (node == null) return;
    
    // redraw symbols in original order
    ParameterSymbol[] parray = psymbol.getParameterArray();
    for (ParameterSymbol psym : parray)
      redraw(node, psym);
  }
  
  /***************************************************************************
   * Get position in source denoting declaration of symbol.
   * @param symbol a symbol.
   * @return the corresponding source position (may be null).
   **************************************************************************/
  private SourcePosition getPosition(Symbol symbol)
  {
    if (symbol instanceof ClassSymbol)
    {
      ClassSymbol csymbol = (ClassSymbol)symbol;
      if (!csymbol.isRead()) return null;
      ClassDeclaration decl = csymbol.getDeclaration();
      return decl.getPosition();
    }
    if (symbol instanceof TheorySymbol)
    {
      TheorySymbol tsymbol = (TheorySymbol)symbol;
      if (!tsymbol.isRead()) return null;
      TheoryDeclaration decl = tsymbol.getDeclaration();
      return decl.getPosition();
    }
    if (symbol instanceof ParameterSymbol)
    {
      ParameterSymbol symbol0 = (ParameterSymbol)symbol;
      ParamSymbol symbol1 = symbol0.getParamSymbol();
      return getPosition(symbol1);
    }
    Declaration decl = symbol.getDeclaration();
    if (decl == null) return null;
    return decl.getPosition();
  }
  
  /****************************************************************************
   * Expand tree node associated to source (if any).
   * @param source the source whose visual representation is to be expanded.
   ***************************************************************************/
  public void expand(Source source)
  {
    Set<Map.Entry<TreeItem,Symbol>> entries = itemSet.entrySet();
    for (Map.Entry<TreeItem,Symbol> e : entries)
    {
      Symbol symbol = e.getValue();
      if (symbol instanceof ClassSymbol) 
      {
        ClassSymbol csymbol = (ClassSymbol)symbol;
        if (csymbol.getSource() == source)
        {
          e.getKey().setExpanded(true);
          return;
        }
      }
      else if (symbol instanceof TheorySymbol) 
      {
        TheorySymbol tsymbol = (TheorySymbol)symbol;
        if (tsymbol.getSource() == source)
        {
          e.getKey().setExpanded(true);
          return;
        }
      }
    }
  }
  
  /***************************************************************************
   * Get all symbols in symbol tree denoting units (classes/theories).
   * @return the set of all symbols denoting units.
   **************************************************************************/
  public Collection<Symbol> getUnitSymbols()
  {
    Collection<Symbol> result = new SetClass<Symbol>();
    Collection<Symbol> candidates = symbolSet.keySet();
    for (Symbol s : candidates)
    {
      if (s instanceof ClassSymbol || s instanceof TheorySymbol)
        result.add(s);
    }
    return result;
  }
  
  /***************************************************************************
   * Remove subtree denoted by symbol.
   * @param symbol the symbol whose subtree (if any) is to be disposed.
   **************************************************************************/
  public void remove(Symbol symbol)
  {
    SymbolItem item = symbolSet.get(symbol);
    if (item != null) dispose(item.node);
  }
  
  /***************************************************************************
   * Show tree item denoted by symbol.
   * @param symbol the symbol whose subtree (if any) is to be expanded.
   **************************************************************************/
  public void show(Symbol symbol)
  {
    SymbolItem item = symbolSet.get(symbol);
    if (item != null) 
    {
      tree.showItem(item.node);
      tree.select(item.node);
    }
  }
}
