// ----------------------------------------------------------------------------
// BoundVariables.java
// Compute the bound variables of an expression.
// $Id: BoundVariables.java,v 1.1 2020/12/07 09:23:17 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2020, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Semantics;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import java.util.*;

/**
 * 
 * Compute the bound variables of an expression.
 *
 */
public final class BoundVariables extends ASTVisitorBase
{
  /***************************************************************************
   * Compute bound variables of expression.
   * @param exp an expression.
   * @return a collection of the names of the bound variables of the expression
   **************************************************************************/
  public static Collection<String> compute(Expression exp)
  {
    BoundVariables fv = new BoundVariables();
    exp.accept(fv);
    return fv.getVariables();
  }
  
  /***************************************************************************
   * Compute bound  variables of expression.
   * @param exp an expression.
   * @param result a collection to which the names of the bound variables 
   *        of the expression are added.
   **************************************************************************/
  public static void add(Expression exp, Collection<String> result)
  {
    BoundVariables fv = new BoundVariables(result);
    exp.accept(fv);
  }
  
  // the collection holding the bound variable symbols computed up to now
  private Collection<String> bound = null;
  
  /***************************************************************************
   * Create visitor with empty result collection.
   **************************************************************************/
  private BoundVariables()
  {
    this(new SetClass<String>());
  }
  
  /****************************************************************************
   * Create visitor with denoted result collection.
   * @param result the collection to which the bound variables are added.
   ***************************************************************************/
  private BoundVariables(Collection<String> result)
  {
    bound = result;
  }
  
  /****************************************************************************
   * Get free variables computed by the expression visitor.
   * @return a collection of the names of the bound variables.
   ***************************************************************************/
  private Collection<String> getVariables()
  {
    return bound;
  }
  
  /****************************************************************************
   * Process quantified expression
   * @param exp a quantified expression.
   ***************************************************************************/
  public void visit(QuantifiedExpression exp)
  {    
    // collect bound variables of expression body
    Collection<String> b = compute(exp.getBody());
    
    // add bound variables
    TypedIdentifier[] variables = exp.getVariables();
    for (TypedIdentifier var : variables)
      b.add(var.getName().getString());
    
    // add remaining variables to collection
    bound.addAll(b);
  }
  
  /****************************************************************************
   * Process quantified expression
   * @param exp a quantified expression.
   ***************************************************************************/
  public void visit(LetExpression exp)
  { 
    // the definitions in the expression
    Definition[] definitions = exp.getDefinitions();
    
    // collect bound variables of expression body and definition values
    Collection<String> b = compute(exp.getBody());
    int n = definitions.length;
    for (int i=n-1; i>=0; i--)
    {
      Definition d = definitions[i];
      b.add(d.getName().getString());
      add(d.getValue(), b);
    }
    
    // add remaining variables to collection
    bound.addAll(b);
  }
}
