// ----------------------------------------------------------------------------
// Checking.java
// Type-check a class declaration and annotate it semantically.
// $Id: Checking.java,v 1.81 2017/12/11 15:55:03 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Semantics;

import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Judgements.Normalization;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Program.*;
import fmrisc.ProgramExplorer.Syntax.Spec.*;
import fmrisc.ProgramExplorer.Syntax.Logic.Formula;
import fmrisc.ProgramExplorer.Syntax.Logic.Term;
import fmrisc.ProgramExplorer.Syntax.Logic.StateType;
import fmrisc.ProgramExplorer.Syntax.Logic.Type;
import fmrisc.ProgramExplorer.Tasks.*;

/**
 * 
 * Type-check a class declaration and annotate it semantically.
 * 
 */
public class Checking
{
  /*
   * MiniJava is a simple subset of Java that is restricted such that no value 
   * - that is represented by a reference and
   * - that can be modified by some statement
   * can be referenced via two different variables. 
   * 
   * Consequently, we can formally replace references by their values  
   * (by translating method calls with references as arguments into
   *  function calls that receive the referenced values as arguments
   *  and return new values as results).
   *  
   * Such values are in Java arrays and class objects (except objects of 
   * class "String", because string values cannot be modified).
   *
   * Above restriction is ensured by three constraints:
   * 
   * 1. For method calls "f(..., arg, ...)" and for constructor calls 
   *    "new C(..., arg, ...)":
   *    
   *    If arg is a reference value, then 
   *    * either arg is a constructor call "new ..." 
   *    * or arg is a local variable or a parameter of the current method
   *      that does not occur as another argument in the same call.
   *      
   * 2. For initializations "type arg = ..." or assignments "arg = ..." 
   *    where arg is a reference value:
   *    
   *    Here "..." must be a constructor call "new ...".
   *    
   * 3. For return statements "return ..." or throw statements "throw ..."
   *    where "..." is a reference value:
   *    
   *    Here "..."  must be either a constructor call "new ..." 
   *    or a selector expression whose base variable is a local variable.
   *    
   *    A "selector expression" arg has form a( '.'... | '['...']'... )*
   *    where a is the "base variable" of the expression.
   * 
   */
  
  /****************************************************************************
   * Process the class symbol i.e. type-check the associated declaration
   * and populate the content of the symbol with it.
   * @param csymbol the class symbol with a non-empty declaration.
   * @param error the stream on which to write the error messages.
   * @return true iff no problem occurred in type-checking.
   **************************************************************************/
  public static boolean process(ClassSymbol csymbol, ErrorStream error)
  {
    return new Checking(csymbol, error).process();
  }
  
  // the class symbol whose declaration is checked.
  private ClassSymbol csymbol;
  
  // the symbol for the top-level package of this class
  private PackageSymbol top;
  
  // the environment in which the class declaration is checked.
  private Environment env;
  
  // the type table to be used for checking
  private TypeTable ttable;
  
  // the stream for generating error messages
  private ErrorStream error;
  
  // the checker for handling logical entities
  private LogicChecking checker;
  
  // the translator for translating program types to logical types
  private TypeTranslator translator;
  
  // information about the current method (only valid in context of a method)
  private ParamSymbol methodSymbol = null;  // the method symbol itself
  private Formula preCondition = null;      // the method precondition 
  private TypeExpression returnType = null; // the method's return type
  private StateType stateType = null;       // the method's state type
  private Name[] assignVars = null;         // the assignable variables
  private Name[] throwExcepts = null;       // the throwable exceptions
  
  /****************************************************************************
   * Construct a checker for checking a class.
   * @param csymbol the symbol of the class to be checked.
   * @param error the stream on which to write the error messages.
   **************************************************************************/
  private Checking(ClassSymbol csymbol, ErrorStream error)
  {
    this.top = csymbol.getPackage().getTopPackage();
    this.csymbol = csymbol;
    this.env = new Environment(top, error);
    this.ttable = top.getTypeTable();
    this.error = error;
    this.translator = top.getTypeTranslator();
  }
  
  // --------------------------------------------------------------------------
  //
  // Classes
  //
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Type-check declaration in class symbol and annotate it semantically and
   * populate the class symbol with the top-level declarations in the class.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process()
  {
    error.newMessage("processing class " + csymbol.getLongName() + "...");
    
    // the class declaration
    ClassDeclaration cdecl = csymbol.getDeclaration();
    
    // initialize error stream appropriately
    SourcePosition pos = cdecl.getPosition();
    error.clearErrors(pos.getSource());
    SourcePosition pos0 = error.getPosition();
    error.setPosition(pos);
    
    // check declaration
    Identifier cident = csymbol.getName();
    String cname = cident.getString();
    Identifier dident = cdecl.getName();
    String dname = dident.getString();
    if (cname != dname)
    {
      error.newError(dident, "found declaration of class " + dname + 
        " rather than declaration of class " + cname);
      return false;
    }
    dident.setSymbol(csymbol);
    
    // check package declaration
    Name pname = cdecl.getPackage();
    if (pname == null) pname = top.getLongName();
    PackageSymbol psymbol = csymbol.getPackage();
    Name pname0 = psymbol.getLongName();
    if (!pname.equals(pname0))
    {
      error.newError(pname.idents[0], "declaration of wrong package " + pname + 
        " in class " + csymbol.getLongName());
      return false;
    }
       
    // process imports
    Import[] imports = cdecl.getImports();
    int n = imports.length;
    for (int i=0; i<n; i++)
    {
      Import imp = imports[i];
      Name iname = imp.getName();
      Symbol symbol = null;
      if (imp instanceof ClassImport)
        symbol = env.getClass(iname);
      else
        symbol = top.getPackage(iname);
      if (symbol == null)
      {
        error.explainError(imp.toString());
        return false;
      }
    }
    
    // for processing mutually dependent class declarations
    // we first populate the class symbol before we process the declarations
    
    // process class (recursive dependencies are allowed in core checking)
    // and process local theory
    env.enterPackage(psymbol);
    boolean okay = env.enterClass(csymbol);
    if (okay) okay = fillEnvironment(csymbol);
    if (okay)
    {
      // set logical type of "this" symbol for class
      ThisSymbol thisSymbol = csymbol.getThisSymbol();
      Type type0 = translator.logicType(thisSymbol.getType());
      okay = type0 != null;
      if (okay) thisSymbol.setLogicType(type0);
    }
    if (okay) okay = processClassAnnotation(csymbol);
    if (okay) okay = processClassSpecification(csymbol);
    if (okay) okay = processClassDeclaration(csymbol);
    env.exitClass();
    env.exitPackage();
    
    // indicate if class had problems
    if (!okay)
    {
      SourcePosition p = csymbol.getDeclaration().getPosition();
      error.explainError(p + ": declaration of class " + csymbol.getLongName());
      error.explainError("Source file " + p.getSource().getFile().getAbsolutePath());
      return false;
    }
    
    // restore position
    error.setPosition(pos0);
    return true;
  }
  
  /****************************************************************************
   * Report that error occured in declaration.
   * @param stat the declaration where the error occurred.
   ***************************************************************************/
  private void declError(Declaration decl)
  {
    SourcePosition p = decl.getPosition();
    Identifier name = decl.getName();
    String text;
    Symbol symbol = name.getSymbol();
    if (symbol == null)
    {
      text = name.getString();
    }
    else
    {
      String sname;
      if (symbol instanceof GlobalSymbol)
      {
        GlobalSymbol symbol0 = (GlobalSymbol)symbol;
        sname = symbol0.getLongName().toString();
      }
      else
        sname = symbol.getName().getString();
      if (symbol instanceof ParamSymbol)
      {
        ParamSymbol symbol0 = (ParamSymbol)symbol;
        text = Syntax.procHeader(sname, symbol0.getArgumentTypes());
      }
      else
        text = sname;
    }
    error.explainError(p.toString() + ": declaration of " + text);
  }
  
  /****************************************************************************
   * Populate the environment with the top-level declarations in the class.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean fillEnvironment(ClassSymbol csymbol)
  { 
    boolean allOkay = true;
    ClassDeclaration cdecl = (ClassDeclaration)csymbol.getDeclaration();
    TopDeclaration[] decls = cdecl.decls;
    
    // first process object variables such that the object type can be determined
    for (TopDeclaration d : decls)
    {
      error.setPosition(d.getPosition());
      if (d instanceof ObjectVariableDeclaration)
      {
        ObjectVariableDeclaration d0 = (ObjectVariableDeclaration)d;
        boolean okay = !d.getName().getString().equals("this");
        if (!okay) error.newError("a variable must not be named 'this'");
        if (okay) okay = process(d0.getType());
        if (!okay)
          declError(d0);
        else
          okay = env.put(d0);
        if (!okay) allOkay = false;
        continue;
      }
    }
    
    // now process the remaining elements
    for (TopDeclaration d : decls)
    {
      error.setPosition(d.getPosition());
      if (d instanceof ObjectVariableDeclaration)
      {
        continue;
      }
      if (d instanceof ClassVariableDeclaration)
      {
        ClassVariableDeclaration d0 = (ClassVariableDeclaration)d;
        boolean okay = !d.getName().getString().equals("this");
        if (!okay) error.newError("a variable must not be named 'this'");
        if (okay) okay = process(d0.getType());
        if (!okay)
          declError(d0);
        else
          okay = env.put(d0);
        if (!okay) allOkay = false;
        continue;
      }
      if (d instanceof ConstructorDeclaration)
      {
        ConstructorDeclaration d0 = (ConstructorDeclaration)d;
        boolean okay = process(d0.getParams());
        if (!okay)
          declError(d0);
        else
          okay = env.put(d0);
        if (!okay) allOkay = false;
        continue;
      }
      if (d instanceof ClassMethodDeclaration)
      {
        ClassMethodDeclaration d0 = (ClassMethodDeclaration)d;
        boolean okay = process(d0.getParams()) && process(d0.type);
        if (!okay)
          declError(d0);
        else
          okay = env.put(d0);
        if (!okay) allOkay = false;
        continue;
      }
      if (d instanceof ObjectMethodDeclaration)
      {
        ObjectMethodDeclaration d0 = (ObjectMethodDeclaration)d;
        boolean okay = process(d0.getParams()) && process(d0.type);
        if (!okay)
          declError(d0);
        else
          okay =  env.put(d0);
        if (!okay) allOkay = false;
        continue;
      }
      Main.internalError("unknown top-level declaration " + d);
    }
    return allOkay;
  }
  
  /****************************************************************************
   * Type-check declaration in class symbol and annotate it semantically.
   * The environment has already been appropriately set up.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean processClassDeclaration(ClassSymbol csymbol)
  {
    boolean allOkay = true;
    ClassDeclaration cdecl = (ClassDeclaration)csymbol.getDeclaration();
    TopDeclaration[] decls = cdecl.decls; 
    
    // process object variables, make sure initialization is in dependence order
    env.enterObjectVariables();
    for (TopDeclaration d : decls)
    {
      error.setPosition(d.getPosition());
      
      // get symbol constructed in previous step 
      Symbol symbol = d.getName().getSymbol();
      if (symbol == null) continue;
      
      // process symbol and enter it into object variable environment
      if (symbol instanceof ObjectVariableSymbol)
      {
        ObjectVariableSymbol symbol0 = (ObjectVariableSymbol)symbol;
        boolean okay =  process(symbol0);
        if (!okay) 
        {
          declError(d);
          allOkay = false;
        }
        env.enterObjectVariable(symbol0);
      }
    }
    env.exitObjectVariables();
    
    // process class variables, make sure initialization is in dependence order
    env.enterClassVariables();
    for (TopDeclaration d : decls)
    {
      error.setPosition(d.getPosition());
      
      // get symbol constructed in previous step 
      Symbol symbol = d.getName().getSymbol();
      if (symbol == null) continue;
      
      // process symbol and enter it into class variable environment
      if (symbol instanceof ClassVariableSymbol)
      {
        ClassVariableSymbol symbol0 = (ClassVariableSymbol)symbol;
        boolean okay =  process(symbol0);
        if (!okay) 
        {
          declError(d);
          allOkay = false;
        }
        env.enterClassVariable(symbol0);
      }
    }
    env.exitClassVariables();
    
    // then process remaining entities
    for (TopDeclaration d : decls)
    {
      error.setPosition(d.getPosition());
      
      // get symbol constructed in previous step 
      Symbol symbol = d.getName().getSymbol();
      if (symbol == null) continue;

      // process symbol
      if (symbol instanceof ObjectVariableSymbol)
      {
        continue;
      }
      if (symbol instanceof ClassVariableSymbol)
      {
        ClassVariableSymbol symbol0 = (ClassVariableSymbol)symbol;
        boolean okay =  process(symbol0);
        if (!okay) 
        {
          declError(d);
          allOkay = false;
        }
        continue;
      }
      if (symbol instanceof ConstructorSymbol)
      {
        ConstructorSymbol symbol0 = (ConstructorSymbol)symbol;
        boolean okay =  process(symbol0);
        if (!okay) 
        {
          declError(d);
          allOkay = false;
        }
        continue;
      }
      if (symbol instanceof ClassMethodSymbol)
      {
        ClassMethodSymbol symbol0 = (ClassMethodSymbol)symbol;
        boolean okay =  process(symbol0);
        if (!okay) 
        {
          declError(d);
          allOkay = false;
        }
        continue;
      }
      if (symbol instanceof ObjectMethodSymbol)
      {
        ObjectMethodSymbol symbol0 = (ObjectMethodSymbol)symbol;
        boolean okay =  process(symbol0);
        if (!okay) 
        {
          declError(d);
          allOkay = false;
        }
        continue;
      }
      Main.internalError("unknown symbol " + symbol.getName());
    }
    if (!allOkay) return false;

    // determine auxiliary information
    LogicEnvironment logicEnv = checker.getLogicEnvironment();
    Declaration[][] decls0 = logicEnv.getDeclarations();

    // determine the relevant theories
    Vector<TheorySymbol> theories = new Vector<TheorySymbol>();
    TheorySymbol.add(theories, translator.getBaseTheory());
    TheorySymbol.add(theories, csymbol.getClassTheory());
    Vector<TheorySymbol> specTheories = logicEnv.getTheories();
    TheorySymbol.add(theories, specTheories);
    Vector<Symbol> units = new Vector<Symbol>();
    ReferencedUnits.computeClosure(ReferencedUnits.compute(cdecl), units);
    for (Symbol unit : units)
    {
      if (unit instanceof ClassSymbol)
      {
        ClassSymbol unit0 = (ClassSymbol)unit;
        TheorySymbol.add(theories, unit0.getClassTheory());
      }
    }

    // register the delayed generation of class tasks
    // (necessary to handle mutually recursive tasks)
    ClassTasks.register(csymbol, theories, decls0, error);
    return true;
  }
  
  /****************************************************************************
   * Process class annotation to construct corresponding theory.
   * @param csymbol the class whose annotation is to be processed.
   * @return the symbol of the theory (null, if an error occurred)
   ***************************************************************************/
  private boolean processClassAnnotation(ClassSymbol csymbol)
  {
    PackageSymbol psymbol = csymbol.getPackage();
    Name pname = null;
    if (psymbol != top) pname = psymbol.getLongName();
    Identifier tname = Identifier.construct("(local)");
    
    ClassDeclaration cdecl = csymbol.getDeclaration();
    ProgramAnnotation annotation = cdecl.getAnnotation();
    TheoryDeclaration tdecl = null;
    if (annotation == null) 
    { 
      // if there is no annotation, create empty pseudo theory
      tdecl = 
        TheoryDeclaration.construct(pname, 
          new Import[] { }, tname, new Name[] { }, new Declaration[] {});
      tdecl.setPosition(cdecl.getPosition());
    }
    else
    {
      // construct a pseudo theory from the annotation
      if (!(annotation instanceof UnitSpec))
        Main.internalError("annotation be UnitSpec");
      UnitSpec spec = (UnitSpec)annotation;
      Import[] imports = spec.getImports();
      Name[] theories = spec.getTheories();
      Declaration[] decls = spec.getDeclarations();
      tdecl = 
        TheoryDeclaration.construct(pname, imports, tname, theories, decls);
      tdecl.setPosition(annotation.getPosition());
    }
    
    // process and set local theory of class
    TheorySymbol tsymbol = new TheorySymbol(csymbol, tname, tdecl, error);
    if (!tsymbol.isRead()) return false;
    csymbol.setLocalTheory(tsymbol);
    
    // create checker
    checker = LogicChecking.createChecker(csymbol, env, error);
    return true;
  }
  
  /****************************************************************************
   * Process class specification.
   * @param csymbol the symbol of the class.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean processClassSpecification(ClassSymbol csymbol)
  {
    ClassDeclaration cdecl = csymbol.getDeclaration();
    ClassSpec spec = cdecl.getClassSpec();
    if (spec == null) return true;
    Formula invariant = spec.getClassInvariant();
    if (invariant == null) return true;

    // process logical entities
    TaskFolder folder = csymbol.getTCCs();
    checker.setStateContext(null, null, null, null, folder);
    env.enterObject();
    boolean okay = processFormula(invariant);
    env.exitObject();
    checker.resetStateContext();
    
    // signal result
    return okay; 
  }
  
  /****************************************************************************
   * Process symbol.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(ClassVariableSymbol symbol)
  {
    ClassVariableDeclaration decl = symbol.getDeclaration();
    ValueExpression value = decl.value;
    if (value == null) return true;
    TypeExpression ltype = symbol.getType();
    TypeExpression etype = process(decl.value);
    if (etype == null) return false;
    boolean okay = match(ltype, etype);
    if (okay) annotateNullLiteral(decl.value, ltype);
    return okay;
  }
  
  /****************************************************************************
   * Process symbol.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(ObjectVariableSymbol symbol)
  {    
    ObjectVariableDeclaration decl = symbol.getDeclaration();
    ValueExpression value = decl.value;
    if (value == null) return true;
    TypeExpression ltype = symbol.getType();
    env.enterObject();
    TypeExpression etype = process(decl.value);
    env.exitObject();
    if (etype == null) return false;
    boolean okay = match(ltype, etype);
    if (okay) annotateNullLiteral(decl.value, ltype);
    return okay;
  }
  
  /****************************************************************************
   * Process symbol.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(ConstructorSymbol symbol)
  {
    env.enterObject();
    env.enterMethod(symbol);
    env.enterBody();
    methodSymbol = symbol;
    methodSymbol.reset();
    preCondition = null;
    returnType = null;
    stateType = StateType.construct(null);
    assignVars = null;
    throwExcepts = null;
    boolean okay = processMethod(symbol);
    methodSymbol = null;
    preCondition = null;
    returnType = null;
    stateType = null;
    assignVars = null;
    throwExcepts = null;
    env.exitBody();
    env.exitMethod();
    env.exitObject();
    return okay;
  }
  
  /****************************************************************************
   * Process symbol.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(ClassMethodSymbol symbol)
  {
    env.enterMethod(symbol);
    env.enterBody();
    methodSymbol = symbol;
    methodSymbol.reset();
    preCondition = null;
    returnType = symbol.getReturnType();
    stateType = StateType.construct(symbol.getLogicReturnType());
    assignVars = null;
    throwExcepts = null;
    boolean okay = processMethod(symbol);
    methodSymbol = null;
    preCondition = null;
    returnType = null;
    stateType = null;
    assignVars = null;
    throwExcepts = null;
    env.exitBody();
    env.exitMethod();
    return okay;
  }
  
  /****************************************************************************
   * Process symbol.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(ObjectMethodSymbol symbol)
  {
    env.enterObject();
    env.enterMethod(symbol);
    env.enterBody();
    methodSymbol = symbol;
    methodSymbol.reset();
    preCondition = null;
    returnType = symbol.getReturnType();
    stateType = StateType.construct(symbol.getLogicReturnType());
    assignVars = null;
    throwExcepts = null;
    boolean okay = processMethod(symbol);
    methodSymbol = null;
    preCondition = null;
    returnType = null;
    stateType = null;
    assignVars = null;
    throwExcepts = null;
    env.exitBody();
    env.exitMethod();
    env.exitObject();
    return okay;
  }
  
  /***************************************************************************
   * Process method/constructor.
   * @param method the method/constructor.
   **************************************************************************/
  private boolean processMethod(ParamSymbol method)
  {
    // reset call set
    method.setCall(new SetClass<ParamSymbol>());
    
    // process specification
    boolean okay = processMethodAnnotation(method);
    if (!okay) return false;
    
    // process implementation
    ParamDeclaration decl = method.getDeclaration();
    Statement body = decl.getBody();
    okay = process(body);
    if (!okay) return false; 
    
    // annotate all declared variables with their logical types
    if (okay) VariableTypes.process(decl, translator);
    
    // set logical return type and state type of method
    returnType = null;
    Type logicType = null;
    if (method instanceof MethodSymbol)
    {
      MethodSymbol method0 = (MethodSymbol)method;
      returnType = method0.getReturnType();
      logicType = method0.getLogicReturnType();
    }
    StateType stateType = StateType.construct(logicType);
    
    // determine auxiliary information
    LogicEnvironment logicEnv = checker.getLogicEnvironment();
    Declaration[][] decls = logicEnv.getDeclarations();
    
    // determine the relevant theories
    Vector<TheorySymbol> theories = new Vector<TheorySymbol>();
    TheorySymbol.add(theories, translator.getBaseTheory());
    TheorySymbol.add(theories, method.getClassSymbol().getClassTheory());
    Vector<TheorySymbol> specTheories = logicEnv.getTheories();
    TheorySymbol.add(theories, specTheories);
    Vector<Symbol> units = new Vector<Symbol>();
    ReferencedUnits.computeClosure(ReferencedUnits.compute(decl), units);
    for (Symbol unit : units)
    {
      if (unit instanceof ClassSymbol)
      {
        ClassSymbol unit0 = (ClassSymbol)unit;
        TheorySymbol.add(theories, unit0.getClassTheory());
      }
    }
    
    // register the delayed generation of method tasks
    // (necessary to handle mutually recursive tasks)
    MethodTasks.register(method,
        stateType, assignVars, throwExcepts, theories, decls, error);
    return true;
  }
  
  /****************************************************************************
   * Process annotation of method and link method symbol to annotation.
   * @param method the method whose annotation is to be processed
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean processMethodAnnotation(ParamSymbol method)
  {
    Declaration decl = method.getDeclaration();
    if (!(decl instanceof ParamDeclaration))
      Main.internalError("unexpected kind of declaration");
    ParamDeclaration decl0 = (ParamDeclaration)method.getDeclaration();
    ProgramAnnotation annotation = decl0.getAnnotation();
    if (annotation == null) return true;
    if (!(annotation instanceof MethodSpec))
      Main.internalError("unexpected kind of annotation");
    MethodSpec spec = (MethodSpec)annotation;
    
    // check the assignable variables
    assignVars = spec.getVariables();
    boolean okay = processVariables(assignVars);
    if (!okay) return false;
    
    // check the possible exceptions
    throwExcepts = spec.getExceptions();
    okay = processExceptions(throwExcepts);
    if (!okay) return false;
    
    // check the logical entities
    Formula pre = spec.getPreCondition();
    Formula div = spec.getDivergesCondition();
    Formula post = spec.getPostCondition();
    Term[] term = spec.getTerminationTerm();
    TaskFolder folder = methodSymbol.getTCCs();
    
    // process the precondition outside of formula context
    checker.setStateContext(null, stateType, assignVars, throwExcepts, folder);
    okay = processFormula(pre);
    checker.resetStateContext();
    
    // set the precondition for processing the rest
    if (okay && pre != null)
    {
      // shift all post-state references to the pre-state
      preCondition = pre; // (Formula)PreStateCloner.clone(pre);
      // checker.setStateContext(null, stateType, assignVars, throwExcepts, folder);
      // okay = processFormula(preCondition);
      // checker.resetStateContext();
      // if (!okay) Main.internalError("error in processing cloned formula");
    }
    
    // process the rest in the context of precondition
    checker.setStateContext(preCondition, stateType, assignVars, throwExcepts, folder);
    if (okay) okay = processFormula(div);
    if (okay) okay = processFormula(post);
    if (term != null)
    {
      int n = term.length;
      for (int i=0; i<n; i++)
      {
        if (okay) okay = processTerm(term[i]);
        if (okay) term[i] = Normalization.perform(term[i]);
      }
    }
    checker.resetStateContext();
    
    // signal result
    method.setSpecification(spec);
    return okay; 
  }
    
  /****************************************************************************
   * Process loop annotation.
   * @param annotation the annotation to be processed.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean processLoopAnnotation(ProgramAnnotation annotation)
  {
    if (annotation == null) return true;
    if (!(annotation instanceof LoopSpec))
      Main.internalError("annotation be LoopSpec");
    LoopSpec spec = (LoopSpec)annotation;
    
    // check logical entities
    Formula invariant = spec.getLoopInvariant();
    Term[] term = spec.getTerminationTerm();
    TaskFolder folder = methodSymbol.getTCCs();
    checker.setStateContext(preCondition, stateType, assignVars, throwExcepts, folder);
    boolean okay = processFormula(invariant);
    if (term != null)
    {
      int n = term.length;
      for (int i=0; i<n; i++)
      {
        if (okay) okay = processTerm(term[i]);
      }
    }
    checker.resetStateContext();
    
    // signal result
    return okay; 
  }
  
  /****************************************************************************
   * Process statement annotation.
   * @param annotation the annotation to be processed.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean processStatementAnnotation(ProgramAnnotation annotation)
  {
    if (annotation == null) return true;
    if (!(annotation instanceof StatementSpec))
      Main.internalError("annotation be StatementSpec");
    StatementSpec spec = (StatementSpec)annotation;
    
    // check logical entities
    Formula assertion = spec.getAssertion();
    TaskFolder folder = methodSymbol.getTCCs();
    checker.setStateContext(preCondition, stateType, assignVars, throwExcepts, folder);
    boolean okay = processFormula(assertion);
    checker.resetStateContext();
    
    // signal result
    return okay; 
  }
  
  /****************************************************************************
   * Process assignable variables.
   * @param vars the names of the assignable variables (may be null)
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean processVariables(Name[] vars)
  {
    if (vars == null) return true;
    for (Name name: vars)
    {
      error.setPosition(name.idents[0].getPosition());
      VariableSymbol symbol = env.getVariable(name);
      if (symbol == null) 
      {
        error.newError("there is no variable " + name + " visible");
        return false;
      }
      // object/array parameters may become assignable
      if (symbol instanceof ParameterSymbol)
      {
        TypeExpression type = symbol.getType();
        if (type == ttable.getStringType() ||
            !(type instanceof ArrayType) && !(type instanceof NamedType))
        {
          error.newError("non-object/array parameter " + name + 
              " must not be declared as assignable");
          return false;
        }
      }
    }
    return true;
  }
  
  /****************************************************************************
   * Process throwable exceptions.
   * @param exceptions the names of the exceptions that may be thrown.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean processExceptions(Name[] exceptions)
  {
    if (exceptions == null) return true;
    for (Name name: exceptions)
    {
      error.setPosition(name.idents[0].getPosition());
      ClassSymbol symbol = env.getClass(name);
      if (symbol == null) return false;
    }
    return true;
  }
  
  /****************************************************************************
   * Process formula.
   * @param formula the formula to be checked (may be null).
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean processFormula(Formula formula)
  {
    if (formula == null) return true;
    error.setPosition(formula.getPosition());
    boolean okay = checker.processFormula(formula);
    return okay;
  }
  
  /****************************************************************************
   * Process term.
   * @param term the term to be checked (may be null).
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean processTerm(Term term)
  {
    if (term == null) return true;
    error.setPosition(term.getPosition());
    boolean okay = checker.processIntTerm(term);
    if (!okay) error.newError("term " + term + " is not of type INT");
    return okay;
  }
  
  // --------------------------------------------------------------------------
  //
  // Statements
  //
  // --------------------------------------------------------------------------

  /****************************************************************************
   * Process statement in method body with given return type.
   * @param stat the statement to be processed.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(Statement stat)
  {
    error.setPosition(stat.getPosition());
    boolean okay = processStatementAnnotation(stat.getAnnotation());
    if (!okay)
    {
      statError(stat);
      return false;
    }
    if (stat instanceof EmptyStatement)
      okay = process((EmptyStatement)stat);
    else if (stat instanceof BlockStatement)
      okay = process((BlockStatement)stat);
    else if (stat instanceof AssignmentStatement)
      okay = process((AssignmentStatement)stat);
    else if (stat instanceof AssignCallStatement)
      okay = process((AssignCallStatement)stat);
    else if (stat instanceof VoidCallStatement)
      okay = process((VoidCallStatement)stat);
    else if (stat instanceof VariableStatement)
      okay = process((VariableStatement)stat); 
    else if (stat instanceof DeclCallStatement)
      okay = process((DeclCallStatement)stat);  
    else if (stat instanceof ConditionalStatement)
      okay = process((ConditionalStatement)stat);
    else if (stat instanceof WhileLoopStatement)
      okay = process((WhileLoopStatement)stat);
    else if (stat instanceof ForLoopStatement)
      okay = process((ForLoopStatement)stat);
    else if (stat instanceof ContinueStatement)
      okay = process((ContinueStatement)stat);
    else if (stat instanceof BreakStatement)
      okay = process((BreakStatement)stat);
    else if (stat instanceof ReturnStatement)
      okay = process((ReturnStatement)stat);
    else if (stat instanceof ThrowStatement)
      okay = process((ThrowStatement)stat);
    else if (stat instanceof TryCatchStatement)
      okay = process((TryCatchStatement)stat);
    else if (stat instanceof AssertionStatement)
      okay = process((AssertionStatement)stat);
    else
      Main.internalError("unknown kind of statement\n" + stat);
    return okay;
  }

  /****************************************************************************
   * Report that error occured in statement.
   * @param stat the statement where the error occurred.
   ***************************************************************************/
  private void statError(Statement stat)
  {
    SourcePosition p = stat.getPosition();
    error.explainError(p.toString() + ":" + p.getSourceLine());
  }
  
  /****************************************************************************
   * Process statement in method body with given return type.
   * @param stat the statement to be processed.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(EmptyStatement stat)
  {
    return true;
  }
  
  /****************************************************************************
   * Process statement in method body with given return type.
   * @param stat the statement to be processed.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(BlockStatement stat)
  {
    env.enterScope();
    Statement[] stats = stat.stats;
    int n=stats.length;
    boolean allOkay = true;
    for (int i=0; i<n; i++)
    {
      boolean okay = process(stats[i]);
      if (!okay) allOkay = false;
    }
    env.exitScope();
    return allOkay;
  }
  
  /****************************************************************************
   * Process statement in method body with given return type.
   * @param stat the statement to be processed.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(AssignmentStatement stat)
  {
    TypeExpression ltype = process(stat.lval);
    if (ltype == null)
      return false;
    TypeExpression rtype = process(stat.exp);
    if (rtype == null)
      return false;
    boolean okay = match(ltype, rtype);
    if (!okay)
    {
      statError(stat);
      return false;
    }
    annotateNullLiteral(stat.exp, ltype);
    return true;
  }

  /****************************************************************************
   * Process statement in method body with given return type.
   * @param stat the statement to be processed.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(VoidCallStatement stat)
  {
    TypeExpression[] atypes = process(stat.args);
    if (atypes == null)
    {
      statError(stat);
      return false;
    }
    MethodSymbol symbol = env.getMethod(stat, atypes);
    if (symbol == null || !argumentsOkay(symbol, stat.args))
    {
      statError(stat);
      return false;     
    }
    methodSymbol.getCall().add(symbol);
    if (symbol.getReturnType() != ttable.voidType)
    {
      error.newError(stat, "method " + stat.method + " has result type " +
          symbol.getReturnType() + " (void expected)");
      statError(stat);
      return false;     
    } 
    annotateNullLiterals(stat.args, symbol.getArgumentTypes());
    return true;
  }
  
  /***************************************************************************
   * Check whether arguments of method call meet syntactic restrictions.
   * @param symbol the method called.
   * @param args the arguments of the call.
   * @return true if the arguments meet the the restrictions of the method.
   **************************************************************************/
  private boolean argumentsOkay(ParamSymbol symbol, ValueExpression[] args)
  {
    ParamDeclaration decl = symbol.getDeclaration();
    MethodSpec spec = (MethodSpec)decl.getAnnotation();
    if (spec == null) return true;
    Name[] vars = spec.getVariables();
    if (vars == null) return true;
    Parameter[] params = decl.getParams();
    int n = params.length;
    for (int i=0; i<n; i++)
    {
      Parameter param = params[i];
      Identifier ident = param.getName();
      ParameterSymbol psymbol = (ParameterSymbol)ident.getSymbol();
      TypeExpression type = psymbol.getType();
      
      // only object/array parameter is of concern
      if (!(type instanceof NamedType) && !(type instanceof ArrayType)) continue;
      if (type == ttable.getStringType()) continue;
      
      // only modifiable parameter is of concern
      for (Name var : vars)
      {
        if (var.idents.length != 1) continue;
        if (var.idents[0].getString() != ident.getString()) continue;
        
        // check constraint on argument for modifiable reference parameter
        ValueExpression arg = args[i];
        while (arg instanceof SelectorExpression)
        {
          SelectorExpression arg0 = (SelectorExpression)arg;
          arg = arg0.base;
        }
        if (arg instanceof VariableValue) break;
        error.newError("assignable object/array parameter " + param + 
            " needs variable argument (argument " + args[i] + " encountered)");
        return false;
      }
    }
    return true;
  }
  
  /****************************************************************************
   * Process statement in method body with given return type.
   * @param stat the statement to be processed.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(AssignCallStatement stat)
  {
    TypeExpression ltype = process(stat.lval);
    if (ltype == null)
    {
      statError(stat);
      return false;
    }
    TypeExpression[] atypes = process(stat.args);
    TypeExpression[] ptypes = null;
    if (atypes == null)
    {
      statError(stat);
      return false;
    }
    TypeExpression etype = null;
    if (stat instanceof AssignNewStatement)
    {
      ConstructorSymbol msymbol = env.getConstructor(stat, atypes);
      if (msymbol == null || !argumentsOkay(msymbol, stat.args))
      {
        statError(stat);
        return false;
      }
      methodSymbol.getCall().add(msymbol);
      etype = NamedType.construct(stat.method);
      etype = NewType.construct(etype);
      etype = ttable.process(etype);
      ptypes = msymbol.getArgumentTypes();
    }
    else
    {
      MethodSymbol msymbol = env.getMethod(stat, atypes);
      if (msymbol == null || !argumentsOkay(msymbol, stat.args))
      {
        statError(stat);
        return false;
      }
      methodSymbol.getCall().add(msymbol);
      etype = msymbol.getReturnType();
      if (etype instanceof NamedType || etype instanceof ArrayType)
      {
        etype = NewType.construct(etype);
        etype = ttable.process(etype);
      }
      ptypes = msymbol.getArgumentTypes();
    }
    boolean okay = match(ltype, etype);
    if (!okay)
    {
      statError(stat);
      return false;
    }
    annotateNullLiterals(stat.args, ptypes);
    return true;
  }
  
  /****************************************************************************
   * Process statement in method body with given return type.
   * @param stat the statement to be processed.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(VariableStatement stat)
  {
    boolean okay = !stat.getName().getString().equals("this");
    if (!okay)
    {
      error.newError("a variable must not be named 'this'");
      statError(stat);
      return false;
    }
    okay = process(stat.type);
    if (!okay)
    {
      statError(stat);
      return false;
    }   
    okay = env.put(stat);
    if (!okay)
    {
      statError(stat);
      return false;
    }
    if (stat.exp == null) return true;
    TypeExpression etype = process(stat.exp);
    if (etype == null)
    {
      statError(stat);
      return false;
    }
    BodySymbol symbol = (BodySymbol)stat.name.getSymbol();
    TypeExpression ltype = symbol.getType();
    okay = match(ltype, etype);
    if (!okay)
    {
      statError(stat);
      return false;
    }
    annotateNullLiteral(stat.exp, ltype);
    return true;
  }
  
  /****************************************************************************
   * Process statement in method body with given return type.
   * @param stat the statement to be processed.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(DeclCallStatement stat)
  {
    boolean okay = !stat.getName().getString().equals("this");
    if (!okay)
    {
      error.newError("a variable must not be named 'this'");
      statError(stat);
      return false;
    }
    okay = process(stat.type);
    if (!okay)
    {
      statError(stat);
      return false;
    }
    okay = env.put(stat);
    if (!okay)
    {
      statError(stat);
      return false;
    }
    TypeExpression[] atypes = process(stat.args);
    TypeExpression[] ptypes = null;
    if (atypes == null)
    {
      statError(stat);
      return false;
    }
    BodySymbol bsymbol = (BodySymbol)(stat.name.getSymbol());
    TypeExpression ltype = bsymbol.getType();
    TypeExpression etype = null;
    if (stat instanceof DeclNewStatement)
    {
      ConstructorSymbol msymbol = env.getConstructor(stat, atypes);
      if (msymbol == null || !argumentsOkay(msymbol, stat.args))
      {
        statError(stat);
        return false;
      }
      methodSymbol.getCall().add(msymbol);
      etype = NamedType.construct(stat.method);
      etype = NewType.construct(etype);
      etype = ttable.process(etype);
      ptypes = msymbol.getArgumentTypes();
    }
    else
    {
      MethodSymbol msymbol = env.getMethod(stat, atypes);
      if (msymbol == null || !argumentsOkay(msymbol, stat.args))
      {
        statError(stat);
        return false;
      }
      methodSymbol.getCall().add(msymbol);
      etype = msymbol.getReturnType();
      if (etype instanceof NamedType || etype instanceof ArrayType)
      {
        etype = NewType.construct(etype);
        etype = ttable.process(etype);
      }
      ptypes = msymbol.getArgumentTypes();
    }
    okay = match(ltype, etype);
    if (!okay)
    {
      statError(stat);
      return false;
    }
    annotateNullLiterals(stat.args, ptypes);
    return true;
  }
  
  /****************************************************************************
   * Process statement in method body with given return type.
   * @param stat the statement to be processed.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(ConditionalStatement stat)
  {
    TypeExpression type = process(stat.cond);
    if (type == null)
    {
      statError(stat);
      return false;
    }
    if (!match(type, ttable.booleanType))
    {
      statError(stat);
      return false;
    }
    boolean okay0 = process(stat.thenstat);  
    if (!okay0) return false;
    if (stat.elsestat == null) return true;
    boolean okay1 = process(stat.elsestat);  
    if (!okay1) return false;  
    return true;
  }
  
  /****************************************************************************
   * Process statement in method body with given return type.
   * @param stat the statement to be processed.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(WhileLoopStatement stat)
  {
    TypeExpression type = process(stat.cond);
    if (type == null)
    {
      statError(stat);
      return false;
    }
    if (!match(type, ttable.booleanType))
    {
      statError(stat);
      return false;
    }
    boolean okay = processLoopAnnotation(stat.getLoopAnnotation());
    if (!okay) 
    {
      statError(stat);
      return false;
    }
    return process(stat.body);  
  }
  
  /****************************************************************************
   * Process statement in method body with given return type.
   * @param stat the statement to be processed.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(ForLoopStatement stat)
  {
    // initialization part may introduce local variable
    env.enterScope();
    boolean okay = processCore(stat);
    env.exitScope();
    return okay;
  }
  
  /****************************************************************************
   * Process statement in method body with given return type.
   * @param stat the statement to be processed.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean processCore(ForLoopStatement stat)
  {
    boolean okay = process(stat.init);
    if (!okay)
    {
      statError(stat);
      return false;
    }
    TypeExpression type = process(stat.cond);
    if (type == null)
    {
      statError(stat);
      return false;
    }
    if (!match(type, ttable.booleanType))
    {
      statError(stat);
      return false;
    }
    okay = process(stat.update);
    if (!okay)
    {
      statError(stat);
      return false;
    }
    okay = processLoopAnnotation(stat.getLoopAnnotation());
    if (!okay)     
    {
      statError(stat);
      return false;
    }
    okay = process(stat.body);
    if (!okay)
    {
      statError(stat);
      return false;
    }
    return true;  
  }
  
  /****************************************************************************
   * Process statement in method body with given return type.
   * @param stat the statement to be processed.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(ContinueStatement stat)
  {
    return true;
  }
  
  /****************************************************************************
   * Process statement in method body with given return type.
   * @param stat the statement to be processed.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(BreakStatement stat)
  {
    return true;
  }
  
  /****************************************************************************
   * Process statement in method body with given return type.
   * @param stat the statement to be processed.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(ReturnStatement stat)
  {
    if (returnType instanceof VoidType)
    {
      if (stat.value != null)
      {
        error.newError(stat, "non-void return statement in void method");
        statError(stat);
        return false;
      }
      return true;
    }
    if (stat.value == null)
    {
      error.newError(stat, "void return statement in non-void method");
      statError(stat);
      return false;
    }
    TypeExpression type = process(stat.value);
    if (type == null)
    {
      statError(stat);
      return false;
    }
    boolean okay = match(returnType, type, !isLocal(stat.value, false));
    if (!okay)
    {
      statError(stat);
      return false;
    } 
    annotateNullLiteral(stat.value, returnType);
    return true;
  }
  
  /***************************************************************************
   * Determine whether value is local to the current body.
   * @param value the value
   * @param param true if value may also depend on parameter of current method.
   * @return true if value is local to the current body.
   **************************************************************************/
  private boolean isLocal(ValueExpression value, boolean param)
  {
    VariableValue var;
    if (value instanceof VariableValue)
      var = (VariableValue)value;
    else if (value instanceof SelectorExpression)
    {
      SelectorExpression sel = (SelectorExpression)value;
      ValueExpression base = sel.base;
      if (!(base instanceof VariableValue)) { return false; }
      var = (VariableValue)base;
    }
    else
      return false;
    Symbol symbol = var.ident.getSymbol();
    if (param && symbol instanceof ParameterSymbol) return true;
    return symbol instanceof BodySymbol;
  }
  
  /****************************************************************************
   * Process statement in method body with given return type.
   * @param stat the statement to be processed.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(ThrowStatement stat)
  {
    // looking up the symbol in environment links it to identifiers
    ClassSymbol symbol = env.getClass(stat.method);
    if (symbol == null)
    {
      statError(stat);
      return false;
    }
    TypeExpression type = process(stat.getValue());
    if (type == null)
    {
      statError(stat);
      return false;
    }
    if (!equals(type,ttable.getStringType()))
    {
      error.newError(stat, stat.getValue() + " is of type " + type + 
          " but must be of type " + ttable.getStringType());
      statError(stat);
      return false;
    }
    TypeExpression[] atypes = ttable.process(new TypeExpression[]{ type });
    if (atypes == null) Main.internalError("error in processing type");
    ConstructorSymbol msymbol = env.getConstructor(stat, atypes);
    if (msymbol == null)
    {
      statError(stat);
      return false;
    }
    methodSymbol.getCall().add(msymbol);
    return true;
  }
  
  /****************************************************************************
   * Process statement in method body with given return type.
   * @param stat the statement to be processed.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(TryCatchStatement stat)
  {
    boolean okay = process(stat.body);
    if (!okay) return false;
    Handler[] handlers = stat.handlers;
    int n = handlers.length;
    for (int i=0; i<n; i++)
    {
      Handler handler = handlers[i];
      Parameter param = handler.param;
      okay = !param.name.getString().equals("this");
      if (!okay)
      {
        error.newError("a parameter must not be named 'this'");
        return false;
      }
      okay = process(param.type);
      if (!okay)
      {
        error.explainError("exception handler parameter " + param);
        return false;
      }
      env.enterScope();
      // enter exception object to denote Exception type
      // thus the only applicable operation on this object is .getMessage()
      // (it can not stored in a variable or passed to a function)
      Parameter param0 = Parameter.construct(
          ExceptionType.construct(), param.name);
      okay = env.put(param0);
      if (!okay)
      {
        error.explainError("exception handler parameter " + param);
        env.exitScope();
        return false;
      }
      Statement body = handler.body;
      okay = process(body);
      env.exitScope();
      if (!okay) return false;
    }
    return true;
  }
  
  /****************************************************************************
   * Process statement in method body with given return type.
   * @param stat the statement to be processed.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(AssertionStatement stat)
  {
    TypeExpression type = process(stat.cond);
    if (type == null)
    {
      statError(stat);
      return false;
    }
    boolean okay = match(type, ttable.booleanType);
    if (!okay)
    {
      statError(stat);
      return false;
    } 
    return true;
  }
  
  // --------------------------------------------------------------------------
  //
  // Value expressions
  //
  // --------------------------------------------------------------------------
 
  /****************************************************************************
   * Process value expressions.
   * @param value expressions to be processed.
   * @return sequence of expression types (null if an error occurred)
   ***************************************************************************/
  private TypeExpression[] process(ValueExpression[] args)
  {
    // determine expression types
    int n = args.length;
    TypeExpression[] atypes = new TypeExpression[n];
    for (int i=0; i<n; i++)
    {
      atypes[i] = process(args[i]);
      if (atypes[i] == null) return null;
    }
    
    // verify "new object" constraint
    for (int i=0; i<n; i++)
    {
      TypeExpression type = atypes[i];
      ValueExpression arg = args[i];
      if (type instanceof NamedType && type != ttable.getStringType())
      {
        if (arg instanceof VariableValue && 
              isLocal(args[i], true) && isUnique(i, args)) continue;
        error.newError("argument " + args[i] + 
            " does not denote a new object");
        return null;
      }
      if (type instanceof ArrayType)
      {
        if (arg instanceof VariableValue && 
            isLocal(args[i], true) && isUnique(i, args)) continue;
        error.newError("argument " + args[i] + 
            " does not denote a new array");
        return null;
      }
      if (type instanceof NewType)
      {
        NewType type0 = (NewType)type;
        atypes[i] = type0.base;
      }
    }
    
    // make types unique
    atypes = ttable.process(atypes);
    return atypes;
  }
  
  /****************************************************************************
   * Check whether variable argument i does not occur multiple times in
   * argument list.
   * @param i the index of a variable in an argument list.
   * @param args the argument list.
   * @return true iff argument i does not occur in args multiple times.
   ***************************************************************************/
  private boolean isUnique(int i, ValueExpression[] args)
  {
    String name = ((VariableValue)args[i]).ident.getString();
    int n = args.length;
    for (int j=0; j<n; j++)
    {
      if (j == i) continue;
      ValueExpression arg = args[j];
      if (!(arg instanceof VariableValue)) continue;
      VariableValue var = (VariableValue)arg;
      if (name == var.ident.getString()) return false;
    }
    return true;
  }
  
  /****************************************************************************
   * Process value expression.
   * @param value expression to be processed.
   * @return type of the expression (null if an error occurred)
   ***************************************************************************/
  private TypeExpression process(ValueExpression value)
  {
    if (value instanceof VariableValue)
      return process((VariableValue)value);
    if (value instanceof IntLiteral)
    {
      IntLiteral value0 = (IntLiteral)value;
      try
      {
        Integer.parseInt(value0.digits);
      }
      catch(NumberFormatException e)
      {
        error.newError("integer literal " + value0.digits + 
            " is too big for type int");
        return null;
      }
      return ttable.intType;
    }
    if (value instanceof BooleanLiteral)
      return ttable.booleanType;
    if (value instanceof CharLiteral)
      return ttable.charType;
    if (value instanceof StringLiteral)
    {
      // annotate literal with string class (may be null)
      StringLiteral value0 = (StringLiteral)value;
      value0.setStringClass(ttable.getStringClass());
      return ttable.process(ttable.getStringType());
    }
    if (value instanceof NullLiteral)
      return ttable.nullType;
    if (value instanceof NewArrayExpression)
      return process((NewArrayExpression)value);
    if (value instanceof UnaryValueExpression)
      return process((UnaryValueExpression)value);  
    if (value instanceof BinaryValueExpression)
      return process((BinaryValueExpression)value);
    if (value instanceof SelectorExpression)
      return process((SelectorExpression)value);
    return null;
  }
  
  /****************************************************************************
   * Process value expression.
   * @param value expression to be processed.
   * @return type of the expression (null if an error occurred)
   ***************************************************************************/
  private TypeExpression process(VariableValue value)
  {
    VariableSymbol symbol = env.getVariable(value.ident);
    if (symbol == null) 
    {
      error.newError(value.ident, "there is no variable " + value + " visible");
      return null;
    }
    return symbol.getType();
  }
  
  /****************************************************************************
   * Process value expression.
   * @param value expression to be processed.
   * @return type of the expression (null if an error occurred)
   ***************************************************************************/
  private TypeExpression process(NewArrayExpression value)
  {
    TypeExpression ltype = process(value.length);
    if (ltype == null)
    {
      error.explainError("new[] expression " + value);
      return null;
    }
    if (!equals(ltype, ttable.intType))
    {
      error.newError("expression " + value + " should have type " + 
          ttable.intType + " but has type " + ltype);
      error.explainError("new[] expression " + value);
      return null;
    } 
    boolean okay = process(value.type);
    if (!okay)
    {
      error.explainError("new[] expression " + value);
      return null;
    } 
    TypeExpression type = NewType.construct(ArrayType.construct(value.type));
    TypeExpression type0 = ttable.process(type);
    if (type0 == null)
    {
      error.explainError("new[] expression " + value);
      return null;
    } 
    return type0;
  }
  
  /****************************************************************************
   * Process value expression.
   * @param value expression to be processed.
   * @return type of the expression (null if an error occurred)
   ***************************************************************************/
  private TypeExpression process(UnaryValueExpression value)
  {
    ValueExpression base = value.base;
    TypeExpression type = process(base); 
    if (type == null) return null;
    if (value instanceof NegationExpression)
    {
      if (!equals(type, ttable.intType))
      {
        error.newError("expression " + base + " has type " + type +
            " but should have type " + ttable.intType);
        return null;
      }
      return ttable.intType;
    }
    if (value instanceof NotExpression)
    {
      if (!equals(type, ttable.booleanType))
      {
        error.newError("expression " + base + " has type " + type +
            " but should have type " + ttable.booleanType);
        return null;
      }
      return ttable.booleanType;
    }
    Main.internalError("unknown unary expression " + value);
    return null;
  }
  
  /****************************************************************************
   * Process value expression.
   * @param value expression to be processed.
   * @return type of the expression (null if an error occurred)
   ***************************************************************************/
  private TypeExpression process(BinaryValueExpression value)
  {
    ValueExpression exp0 = value.first;
    TypeExpression type0 = process(exp0); 
    if (type0 == null) return null;
    ValueExpression exp1 = value.second;
    TypeExpression type1 = process(exp1); 
    if (type1 == null) return null;
    if (value instanceof EqualsExpression || value instanceof NotEqualsExpression)
    {
      if (!equals(type0, type1))
      {
        error.newError("expression " + exp0 + " has type " + type0 +
            " and expression " + exp1 + " has type " + type1 +
            " but they should have same type");
        return null;
      }
      annotateNullLiteral(exp1, type0);
      return ttable.booleanType;
    }
    if (value instanceof AndExpression || value instanceof OrExpression)
    {
      if (!equals(type0, ttable.booleanType))
      {
        error.newError("expression " + exp0 + " has type " + type0 +
            " but should have type " + ttable.booleanType);
        return null;
      }
      if (!equals(type1, ttable.booleanType))
      {
        error.newError("expression " + exp1 + " has type " + type1+
            " but should have type " + ttable.booleanType);
        return null;
      }
      return ttable.booleanType;
    }
    if (value instanceof PlusExpression ||
        value instanceof MinusExpression || value instanceof TimesExpression ||
        value instanceof DividesExpression || value instanceof PercentExpression)
    {
      /*
      if (value instanceof PlusExpression)
      {
        // string concatenation returning a new string
        TypeExpression stype = ttable.getStringType();
        if (equals(type0, stype) || equals(type1, stype))
          return stype;
      }
      */
      if (!equals(type0, ttable.intType))
      {
        error.newError("expression " + exp0 + " has type " + type0 +
            " but should have type " + ttable.intType);
        return null;
      }
      if (!equals(type1, ttable.intType))
      {
        error.newError("expression " + exp1 + " has type " + type1 +
            " but should have type " + ttable.intType);
        return null;
      }    
      return ttable.intType;
    }
    if (value instanceof LessExpression || value instanceof LessEqualExpression ||
        value instanceof GreaterExpression || value instanceof GreaterEqualExpression)
    {
      if (!equals(type0, ttable.intType))
      {
        error.newError("expression " + exp0 + " has type " + type0 +
            " but should have type " + ttable.intType);
        return null;
      }
      if (!equals(type0, ttable.intType))
      {
        error.newError("expression " + exp1 + " has type " + type1 +
            " but should have type " + ttable.intType);
        return null;
      }    
      return ttable.booleanType;
    }
    Main.internalError("unknown binary expression " + value);
    return null;
  }
  
  /****************************************************************************
   * Process value expression.
   * @param value expression to be processed.
   * @return type of the expression (null if an error occurred)
   ***************************************************************************/
  private TypeExpression process(SelectorExpression value)
  {
    ValueExpression base = value.base;
    Selector[] sels = value.sels;
    
    // determine type of base value
    TypeExpression type;
    
    // expression might actually be (start with) a long name
    // therefore we must first lookup the environment by the long name
    
    int i; // number of already processed selectors
    if (base instanceof VariableValue)
    {
      VariableValue base0 = (VariableValue)base;
      Identifier ident = base0.ident;
      type = process(ident, sels);
      if (type == null) return null;
      i = selnum; // output of process()
    }
    else
    {
      type = process(base);
      if (type == null) return null;
      i = 0;
    }
    
    // set type of base value
    value.setBaseType(type);
    
    // process remaining selectors
    return process(type, sels, i);
  }
  
  // output variable for process(ident, sels)
  private int selnum;
  
  /***************************************************************************
   * Determine from identifier and subsequent identifier selectors long name
   * denoting a vale and return its type.
   * @param ident the first identifier
   * @param sels the selectors.
   * @return the type of the value denoted by the long identifier;
   *         sets 'selnum' to index of first selector that is not identifier.
   **************************************************************************/
  private TypeExpression process(Identifier ident, Selector[] sels)
  {
    Vector<Identifier> idents = new Vector<Identifier>();
    idents.add(ident);
    int i = 0;
    int n = sels.length;
    while (i < n)
    {
      Selector sel = sels[i];
      if (!(sel instanceof IdentifierSelector)) break;
      IdentifierSelector sel0 = (IdentifierSelector)sel;
      idents.add(sel0.name);
      i = i+1;
    }
    selnum = i;
    Name name = Name.construct(idents);
    VariableSymbol s = env.getVariable(name);
    if (s == null) return null;
    
    // set base types of true selectors
    TypeExpression type = null;
    Symbol symbol = ident.getSymbol();
    if (symbol instanceof VariableSymbol)
      type = ((VariableSymbol)symbol).getType();
    for (int j=0; j<selnum; j++)
    {
      IdentifierSelector sel = (IdentifierSelector)sels[j];
      if (type != null)
      {
        Type ltype = translator.logicType(type);
        sel.setLogicBaseType(ltype);
      }
      Symbol symbol0 = sel.name.getSymbol();
      if (!(symbol0 instanceof VariableSymbol)) continue;
      type = ((VariableSymbol)symbol0).getType();
    }
    
    // finally the result
    return s.getType();
  }
  
  /****************************************************************************
   * Process selectors
   * @param type the base type of the value to which the selectors are applied.
   * @param sels the selectors
   * @param s the index of the first selector to be processed.
   * @return the result type (or null if an error occurred).
   ***************************************************************************/
  private TypeExpression process(TypeExpression type, Selector[] sels, int s)
  {
    int i = s;
    int n = sels.length;
    while (i < n)
    {
      Selector sel = sels[i];
      Type ltype = translator.logicType(type);
      sel.setLogicBaseType(ltype);
      i = i+1;
      if (sel instanceof IdentifierSelector)
      {
        IdentifierSelector sel0 = (IdentifierSelector)sel;
        Identifier ident = sel0.name;
        if (!(type instanceof NamedType))
        {
          error.newError(ident, "selector ." + ident + 
              " is applied to a value of type " + type + 
              " but should be applied to an object");
          return null;
        }
        NamedType type0 = (NamedType)type;
        ClassSymbol csymbol = type0.getSymbol();
        ObjectVariableSymbol symbol = csymbol.getObjectVariables().get(ident);
        if (symbol == null)
        {
          error.newError(ident, "selector ." + ident + 
              " is applied to an object of type " + type + 
              " which does not have an object variable of this name");
          return null;
        }
        type = symbol.getType();
        sel.setType(type);
        continue;
      }
      if (sel instanceof IndexSelector)
      {
        IndexSelector sel0 = (IndexSelector)sel;
        ValueExpression index = sel0.index;
        TypeExpression itype = process(index);
        if (itype == null) return null;
        if (!equals(itype, ttable.intType))
        {
          error.newError("expression " + index + " should have type " +
              ttable.intType + " but has type " + itype);
          return null;
        }
        if (!(type instanceof ArrayType))
        {
          error.newError("[] is applied to a value of type " +
            type + " but should be applied to an array");
          return null;
        }
        ArrayType type0 = (ArrayType)type;
        type = type0.base;
        sel.setType(type);
        continue;
      }
      if (sel instanceof MessageSelector)
      {
        if (!(type instanceof ExceptionType))
        {
          error.newError(".getMessage() is applied to a value of type " +
              type + " but should be applied to an exception type");
            return null;
        }
        type = ttable.getStringType();
        sel.setType(type);
        continue;
      }
      Main.internalError("unknown selector " + sel);
      return null;
    }
    return type;
  }
  
  /****************************************************************************
   * Process left value
   * @param lval left value to be processed.
   * @return type of the left value (null if an error occurred)
   ***************************************************************************/
  private TypeExpression process(LeftValue lval)
  {
    Identifier name = lval.name;
    Selector[] sels = lval.selectors;
    TypeExpression type = process(name, sels);
    if (type == null) return null;
    if (sels.length == 0)
    {
      VariableSymbol symbol = (VariableSymbol)name.getSymbol();
      if (symbol instanceof ParameterSymbol)
      {
        if (type instanceof ArrayType || type instanceof NamedType)
        {
          error.newError("must not assign to object/array parameter " +
            lval + " a value");
          return null;
        }
      }
    }
    int i = selnum;
    type = process(type, sels, i);
    if (type == null) return null;
    return type;
  }

  // --------------------------------------------------------------------------
  //
  // Types
  //
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Process type expressions in parameter list
   * (setting the symbols for identifiers in type).
   * @return true iff everything worked fine (message is printed otherwise).
   ***************************************************************************/
  private boolean process(Parameter[] params)
  {
    int n = params.length;
    for (int i=0; i<n; i++)
    {
      Parameter param = params[i];
      boolean okay = !param.name.getString().equals("this");
      if (!okay)
      {
        error.newError("a parameter must not be named 'this'");
        return false;
      }
      okay = process(params[i].getType());
      if (!okay) return false;
    }
    return true;
  }
  
  /****************************************************************************
   * Process type expression (setting the symbols for identifiers in type).
   * @return true iff everything worked fine (message is printed otherwise).
   ***************************************************************************/
  private boolean process(TypeExpression type)
  {
    if (type instanceof PrimitiveType)
      return true;
    if (type instanceof NamedType)
    {
      NamedType type0 = (NamedType)type;
      Name name = type0.name;
      // looking up the symbol in environment links it to identifiers
      ClassSymbol symbol = env.getClass(name);
      if (symbol == null)
      {
        error.explainError("type expression " + type);
        return false;
      }
      return true;
    }
    if (type instanceof NewType)
    {
      NewType type0 = (NewType)type;
      return process(type0.base);
    }
    if (type instanceof ArrayType)
    {
      ArrayType type0 = (ArrayType)type;
      return process(type0.base);
    }
    Main.internalError("unknown type " + type);
    return false;
  }
  
  /****************************************************************************
   * Compare two types for equality.
   * @param type0 one type
   * @param type1 the other type
   * @return true iff the types are equal
   **************************************************************************/
  private boolean equals(TypeExpression type0, TypeExpression type1)
  {
    if (type0 instanceof NewType) type0 = ((NewType)type0).base;
    if (type0 instanceof NewType) type1 = ((NewType)type1).base;
    if (type0 == type1) return true;
    if (!(type1 instanceof NullType)) return false;
    if (type0 instanceof StringType) return true;
    if (type0 instanceof NamedType) return true;
    if (type0 instanceof ArrayType) return true;
    return false;
  }
  
  /****************************************************************************
   * Match type of left value to type of expression considering the
   * constraint that if ltype is an object type that etype must be
   * the type of a new object of type ltype.
   * 
   * @param ltype the type of a left value (variable to be assigned to).
   * @param rtype the type of an expression.
   * @return true iff the match suceeds (message is printed otherwise).
   **************************************************************************/
  private boolean match(TypeExpression ltype, TypeExpression etype)
  {
    return match(ltype, etype, true);
  }
  
  /****************************************************************************
   * Match type of left value to type of expression.
   * @param ltype the type of a left value (variable to be assigned to).
   * @param rtype the type of an expression.
   * @param isnew consider the "new object" constraint.
   * @return true iff the match suceeds (message is printed otherwise).
   **************************************************************************/
  private boolean match(TypeExpression ltype, TypeExpression etype,
    boolean isnew)
  {
    if (ltype instanceof NamedType)
    {
      if (etype instanceof NullType)
        return true;
      if (etype instanceof NewType)
      {
        NewType etype0 = (NewType)etype;
        TypeExpression base = etype0.base;
        if (ltype != base)
        {
          error.newError("type " + etype
              + " of expression does not match type " + ltype
              + " of variable");
          return false;
        }
        return true;
      }
      if (etype instanceof NamedType)
      {
        NamedType ltype0 = (NamedType)ltype;
        ClassSymbol lsymbol = ltype0.getSymbol();
        NamedType etype0 = (NamedType)etype;
        ClassSymbol esymbol = etype0.getSymbol();
        if (lsymbol != esymbol)
        {
          error.newError("type " + etype + " of expression does not match type "
              + ltype + " of variable");
          return false;
        }
        if (isnew && etype != ttable.getStringType())
        {
          error.newError("expression does not denote a new object");
          return false;
        }
        return true;
      }
      error.newError("type " + etype + " of expression does not match type "
          + ltype + " of variable");
      return false;
    }
    if (ltype instanceof ArrayType)
    {
      if (etype instanceof NullType)
        return true;
      if (etype instanceof NewType)
      {
        NewType etype0 = (NewType)etype;
        TypeExpression base = etype0.base;
        if (ltype != base)
        {
          error.newError("type " + etype
              + " of expression does not match type " + ltype
              + " of variable");
          return false;
        }
        return true;
      }
      if (etype instanceof ArrayType)
      {
        ArrayType ltype0 = (ArrayType)ltype;
        ArrayType etype0 = (ArrayType)etype;
        if (!match(ltype0.base, etype0.base))
        {
          error.newError("type " + etype + " of expression does not match type "
              + ltype + " of variable");
          return false;
        }
        if (isnew)
        {
          error.newError("expression does not denote a new array");
          return false;
        }
        return true;
      }
      error.newError("type " + etype + " of expression does not match type "
                             + ltype + " of variable");
      return false; 
    }
    if (ltype != etype)
    {
      error.newError("type " + etype + " of expression does not match type "
          + ltype + " of variable");
      return false;
    }
    return true;
  }
  
  /***************************************************************************
   * If exp is a null literal, annotate it with the denoted type.
   * @param exp an expression that may be a null literal.
   * @param type the type of the expression.
   **************************************************************************/
  private void annotateNullLiteral(ValueExpression exp, TypeExpression type)
  {
    if (!(exp instanceof NullLiteral)) return;
    NullLiteral exp0 = (NullLiteral)exp;
    exp0.setType(type);
  }
  
  /***************************************************************************
   * If exps contains null literals, annotate them with the denoted types.
   * @param exp expressions that may be contain null literals.
   * @param type the types of the expressions.
   **************************************************************************/
  private void annotateNullLiterals(ValueExpression[] exp, TypeExpression[] type)
  {
    int n = exp.length;
    for (int i=0; i<n; i++)
      annotateNullLiteral(exp[i], type[i]);
  }    
}
