// ----------------------------------------------------------------------------
// ClassSymbol.java
// A symbol denoting a class.
// $Id: ClassSymbol.java,v 1.44 2017/07/04 08:04:36 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Semantics;

import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Program.*;
import fmrisc.ProgramExplorer.Syntax.Logic.Formula;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Tasks.*;

import java.io.*;
import java.util.*;

/**
 * 
 * A symbol denoting a class.
 *
 */
public final class ClassSymbol extends GlobalSymbolBase 
{
  // the package holding the class
  private final PackageSymbol psymbol; 
  
  // the reader for reading the source file (may be null)
  private ClassReader reader;
  
  // the declaration of the class after reading the source file (null before)
  private ClassDeclaration decl;
  
  // indicate whether class has errors
  private boolean errors = false;
  
  // the constructors of the class
  private ParamSymbols<ConstructorSymbol> constructors = null;
  
  // the class methods of the class
  private ParamSymbolTable<ClassMethodSymbol> cmethods = null;
  
  // the object methods of the class
  private ParamSymbolTable<ObjectMethodSymbol> omethods = null;
  
  // the class variables of the class
  private SymbolTable<ClassVariableSymbol> cvars = null;
 
  // the object variables of the class
  private SymbolTable<ObjectVariableSymbol> ovars = null;
  
  // the symbols of the units (classes/theories) referenced by this class
  private Collection<Symbol> symbols = null;
  
  // the theory holding the logical type corresponding to the class type
  private TheorySymbol classTheory = null;
  
  // the theory holding the local declarations for the class
  private TheorySymbol localTheory = null;
  
  // the various task folders of this class
  private TaskFolder tasks = null;  // the tasks of this class
  private TaskFolder tccs = null;   // the type checking conditions
  
  // the "this" symbol for this class
  private ThisSymbol thisSymbol;
  
  // the type denoted by this class
  private NamedType type;
 
  // the formula representing the initial values of the object variables
  private Formula initFormula = null;
  
  // the name of the string class
  public static String stringClass = "java.lang.String";
  
  /***************************************************************************
   * Construct class symbol.
   * @param psymbol the symbol of the package holding the class.
   * @param name the name of the class
   * @param reader the reader for getting the declaration
   **************************************************************************/
  public ClassSymbol(PackageSymbol psymbol, Identifier name, ClassReader reader)
  {
    super(psymbol, name);
    this.psymbol = psymbol;
    this.reader = reader;
    
    // construct type for this class
    Name lname = getLongName();
    this.type = NamedType.construct(lname);
    
    // construct class for "String" type
    if (lname.toString().equals(stringClass))
      psymbol.getTypeTable().setStringType(this);
    
    // construct "this" symbol 
    thisSymbol = ThisSymbol.construct(this);
    
    resetTasks();
  }

  /***************************************************************************
   * Reset set of tasks related to class.
   **************************************************************************/
  private void resetTasks()
  {
    if (tasks != null)
    {
      TaskFolder parent = tasks.getParent();
      parent.remove(tasks);
      tasks.setParent(null);
      tasks.setStatus(Task.Status.REVOKED, null, null);
    }
    tasks = new TaskFolder(this, "class " + getName(), psymbol.getTasks());
    tccs = new TaskFolder(null, "type checking conditions", this.tasks);
  }
  
  /****************************************************************************
   * Get package of class.
   * @return the class's package
   ***************************************************************************/
  public PackageSymbol getPackage()
  {
    return psymbol;
  }
  
  /****************************************************************************
   * Get source of class.
   * @return the class's source.
   ***************************************************************************/
  public Source getSource()
  {
    return reader.getSource();
  }
  
  /****************************************************************************
   * Returns true if class has been read, i.e. its values are defined.
   * @return true if class has been read.
   ***************************************************************************/
  public boolean isRead()
  {
    return decl != null;
  }
  
  /****************************************************************************
   * Update declaration from reader.
   * @param error the error stream on which to print any messages.
   * @return true if everything went fine, false if an error ocurred.
   ***************************************************************************/
  public boolean read(ErrorStream error)
  {
    Name lname = getLongName();
    if (lname.toString().equals(stringClass))
      psymbol.getTypeTable().setStringType(this);
    if (decl != null && !reader.isModified() && !hasErrors()) return true;
    errors = false;
    symbols = null;
    decl = reader.read();
    if (decl == null) 
    {
      errors = true;
      if (lname.toString().equals(stringClass))
        psymbol.getTypeTable().setStringType(null);
      return false;
    }
    init(error);
    int num = error.getErrorNumber();
    Checking.process(this, error);
    // immediately compute dependencies (may be partial)
    symbols = ReferencedUnits.compute(decl);
    symbols.remove(this); // avoid circular dependency
    symbols.remove(classTheory);
    if (error.getErrorNumber() != num) 
    {
      errors = true;
      reset();
      if (lname.toString().equals(stringClass))
        psymbol.getTypeTable().setStringType(null);
      return false;
    } 
    return true;
  }
  
  /***************************************************************************
   * Show whether class has errors.
   * @return true if class has errors
   **************************************************************************/
  public boolean hasErrors()
  {
    return errors;
  }
  
  /***************************************************************************
   * Set error status.
   * @param status true iff errors, false otherwise.
   **************************************************************************/
  public void setErrors(boolean status)
  {
    errors = status;
  }
  
  /***************************************************************************
   * Initialize symbol tables.
   * @param error the error stream to print messages to.
   **************************************************************************/
  private void init(ErrorStream error)
  {
    constructors = new ParamSymbols<ConstructorSymbol>(error);
    cmethods = new ParamSymbolTable<ClassMethodSymbol>(error);
    omethods = new ParamSymbolTable<ObjectMethodSymbol>(error);
    cvars = new SymbolTable<ClassVariableSymbol>(error);
    ovars = new SymbolTable<ObjectVariableSymbol>(error);
  }
  
  /***************************************************************************
   * Reset symbol.
   **************************************************************************/
  public void reset()
  {
    decl = null;
    constructors = null;
    cmethods = null;
    omethods = null;
    cvars = null;
    ovars = null;
    classTheory.reset();
    localTheory = null;
    symbols = null;
    resetTasks();
  }
  
  /***************************************************************************
   * Get the declaration of the class.
   * @return the declaration (may be null)
   **************************************************************************/
  public ClassDeclaration getDeclaration()
  {
    return decl;
  }
  
  /****************************************************************************
   * Get set of constructors of this class.
   * @return the set of constructors.
   ***************************************************************************/
  public ParamSymbols<ConstructorSymbol> getConstructors()
  {
    return constructors;
  }
  
  /****************************************************************************
   * Get set of class methods of this class.
   * @return the set of class methods.
   ***************************************************************************/
  public ParamSymbolTable<ClassMethodSymbol> getClassMethods()
  {
    return cmethods;
  }
  
  /****************************************************************************
   * Get set of object methods of this class.
   * @return the set of object methods.
   ***************************************************************************/
  public ParamSymbolTable<ObjectMethodSymbol> getObjectMethods()
  {
    return omethods;
  }
  
  /****************************************************************************
   * Get set of class values of this class.
   * @return the set of class values.
   ***************************************************************************/
  public SymbolTable<ClassVariableSymbol> getClassVariables()
  {
    return cvars;
  }
  
  /****************************************************************************
   * Get set of object values of this class.
   * @return the set of object values.
   ***************************************************************************/
  public SymbolTable<ObjectVariableSymbol> getObjectVariables()
  {
    return ovars;
  }
  
  /****************************************************************************
   * Get set of constructors of this class.
   * May be only called if isRead() is true.
   * @return the set of object methods.
   ***************************************************************************/
  public Collection<ConstructorSymbol> getConstructorSet()
  {
    return constructors.getSymbols();
  }
  
  /****************************************************************************
   * Get set of class methods of this class.
   * May be only called if isRead() is true.
   * @return the set of class methods.
   ***************************************************************************/
  public Collection<ClassMethodSymbol> getClassMethodSet()
  {
    return cmethods.getSymbols();
  }
  
  /****************************************************************************
   * Get set of object methods of this class.
   * May be only called if isRead() is true.
   * @return the set of object methods.
   ***************************************************************************/
  public Collection<ObjectMethodSymbol> getObjectMethodSet()
  {
    return omethods.getSymbols();
  }
  
  /****************************************************************************
   * Get set of class values of this class.
   * May be only called if isRead() is true.
   * @return the set of class values.
   ***************************************************************************/
  public Collection<ClassVariableSymbol> getClassVariableSet()
  {
    return cvars.getSymbols();
  }
  
  /****************************************************************************
   * Get set of object values of this class.
   * May be only called if isRead() is true.
   * @return the set of object values.
   ***************************************************************************/
  public Collection<ObjectVariableSymbol> getObjectVariableSet()
  {
    return ovars.getSymbols();
  }
  
  /***************************************************************************
   * Get symbols of units referenced by this unit.
   * @return the collection of symbols.
   **************************************************************************/
  public Collection<Symbol> getReferencedUnits()
  {
    return symbols;
  }
  
  /***************************************************************************
   * Set theory symbol holding the logical type declaration for this class.
   * @param tsymbol the theory symbol.
   **************************************************************************/
  public void setClassTheory(TheorySymbol tsymbol)
  {
    this.classTheory = tsymbol;
  }
  
  /***************************************************************************
   * Get theory symbol holding the logical type declaration for this class.
   * @return the theory symbol (null, if an error occurred).
   **************************************************************************/
  public TheorySymbol getClassTheory()
  {
    PrintWriter writer = new PrintWriter(System.out);
    ErrorStream error = new ErrorStream("ERROR", "   ", "", writer, null);
    return getClassTheory(error);
  }
  
  /***************************************************************************
   * Get theory symbol holding the logical type declaration for this class.
   * @param error the error stream.
   * @return the theory symbol (null, if an error occurred).
   **************************************************************************/
  public TheorySymbol getClassTheory(ErrorStream error)
  {
    // populate class theory 
    if (!classTheory.isRead())
    {
      boolean okay = classTheory.read(error);
      if (!okay) return null;
    }
    return classTheory;
  }
  
  /***************************************************************************
   * Set theory symbol holding the local declaration for this class.
   * @param tsymbol the theory symbol.
   **************************************************************************/
  public void setLocalTheory(TheorySymbol tsymbol)
  {
    this.localTheory = tsymbol;
  }
  
  /***************************************************************************
   * Get theory symbol holding the local declaration for this class.
   * @return the theory symbol.
   **************************************************************************/
  public TheorySymbol getLocalTheory()
  {
    return localTheory;
  }
  
  /***************************************************************************
   * Print symbol on output stream.
   * @param out the output stream.
   **************************************************************************/
  public void print(PrintWriter out)
  {
    out.print("class ");
    getLongName().print(out);
    if (reader == null) return;
    Source source = reader.getSource();
    out.print(" (file ");
    out.print(source.getFile().getAbsolutePath());
    out.print(")");
  }
  
  /***************************************************************************
   * Get the type checking conditions.
   * @return the folder holding the type checking conditions.
   **************************************************************************/
  public TaskFolder getTCCs()
  {
    return tccs;
  }
  
  /***************************************************************************
   * Get the folder of tasks related to this class.
   * @return the task folder.
   **************************************************************************/
  public TaskFolder getTasks()
  {
    return tasks;
  }
  
  /***************************************************************************
   * Get "this" symbol for this class.
   * @return the symbol for "this".
   **************************************************************************/
  public ThisSymbol getThisSymbol()
  {
    return thisSymbol;
  }
  
  /***************************************************************************
   * Get type for for this class.
   * @return the type for this class.
   **************************************************************************/
  public NamedType getType()
  {
    return type;
  }
  
  /***************************************************************************
   * Set the formula representing the initial values of the object variables.
   * @param formula the formula.
   **************************************************************************/
  public void setInitFormula(Formula formula)
  {
    initFormula = formula;
  }
  
  /***************************************************************************
   * Get the formula representing the initial values of the object variables.
   * @return the formula (may be null).
   **************************************************************************/
  public Formula getInitFormula()
  {
    return initFormula;
  }
}

