// ----------------------------------------------------------------------------
// ConstructorSymbol.java
// A symbol denoting a class constructor.
// $Id: ConstructorSymbol.java,v 1.10 2010/03/24 17:01:16 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Semantics;

import java.io.PrintWriter;

import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Program.ConstructorDeclaration;
import fmrisc.ProgramExplorer.Syntax.Program.Parameter;
import fmrisc.ProgramExplorer.Syntax.Program.TypeExpression;
import fmrisc.ProgramExplorer.InOut.*;

/**
 * 
 * A symbol denoting a class constructor.
 *
 */
public final class ConstructorSymbol extends ParamSymbol
{
  private final ConstructorDeclaration decl;
  
  /***************************************************************************
   * Construct class method symbol from declaration (NULL in case of error).
   * @param csymbol the symbol of the class holding the declaration.
   * @param name the name of the symbol.
   * @param params the parameters.
   * @param atypes the argument types of its arguments.
   * @param decl the declaration introducing the symbol.
   * @param error the stream to be used for printing errors.
   * @return the symbol or null if a problem occurred.
   **************************************************************************/
  public static ConstructorSymbol construct(ClassSymbol csymbol, 
    Identifier name, Parameter[] params, TypeExpression[] atypes, 
    ConstructorDeclaration decl, ErrorStream error)
  {
    ConstructorSymbol symbol = 
      new ConstructorSymbol(csymbol, name, params, atypes, decl, error);
    // symbol construction may have failed
    if (symbol.getParams() == null) return null;
    return symbol;
  }
  
  /***************************************************************************
   * Construct class method symbol from declaration.
   * Attention: may fail on parameters with duplicate names; in this case, 
   * getParams() returns null.
   * @param csymbol the symbol of the class holding the declaration.
   * @param name the name of the symbol.
   * @param params the parameters.
   * @param atypes the argument types of its arguments.
   * @param decl the declaration introducing the symbol.
   * @param error the stream to be used for printing errors.
   **************************************************************************/
  private ConstructorSymbol(ClassSymbol csymbol, 
    Identifier name, Parameter[] params, TypeExpression[] atypes,
    ConstructorDeclaration decl, ErrorStream error)
  {
    super(csymbol, name, params, atypes, error);
    this.decl = decl;
  }
  
  /***************************************************************************
   * Get declaration of method.
   * @return the declaration.
   **************************************************************************/
  public ConstructorDeclaration getDeclaration()
  {
    return decl;
  }
  
  /***************************************************************************
   * Print symbol on output stream.
   * @param out the output stream.
   **************************************************************************/
  public void print(PrintWriter out)
  {
    out.print("constructor ");
    out.print(Syntax.procHeader(getLongName().toString(), getArgumentTypes()));
    Declaration decl = getDeclaration();
    if (decl == null) return;
    SourcePosition pos = decl.getPosition();
    out.print(" (");
    out.print(pos.toString());
    out.print(")");
  }
}


