// ----------------------------------------------------------------------------
// LogicChecking.java
// Type-check a logical theory.
// $Id: LogicChecking.java,v 1.57 2017/12/11 15:55:03 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Semantics;

import java.util.*;
import java.io.*;

import fmrisc.ProgramExplorer.Main;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Syntax.Program.Import;
import fmrisc.ProgramExplorer.Syntax.Spec.*;
import fmrisc.ProgramExplorer.Tasks.*;
import fmrisc.ProgramExplorer.InOut.*;

/**
 *
 * Type-check a logical theory.
 *
 */
public final class LogicChecking
{
  /****************************************************************************
   * Process the theory symbol i.e. type-check the associated declaration
   * and populate the content of the symbol with it.
   * @param tsymbol the theory symbol with a non-empty declaration.
   * @param error the stream on which to write the error messages.
   * @return true iff no problem occurred in type-checking.
   **************************************************************************/
  public static boolean process(TheorySymbol tsymbol, ErrorStream error)
  {
    PackageSymbol psymbol = tsymbol.getPackage();
    PackageSymbol top = psymbol.getTopPackage();
    Environment penv = new Environment(top, error);
    penv.enterPackage(psymbol);
    return new LogicChecking(penv, error).process(tsymbol);
  }
  
  /***************************************************************************
   * Create checker for processing logical entities in environment set up
   * by the (already processed) local theory of a class.
   * @param csymbol the class symbol
   * @param penv the program environment used by the checker.
   * @param error the stream on which to write error messages.
   * @return a checker for processing logical entities in theory environment.
   **************************************************************************/
  public static LogicChecking createChecker(ClassSymbol csymbol, 
    Environment penv, ErrorStream error)
  {
    LogicChecking checker = new LogicChecking(penv, error);
    TheorySymbol tsymbol = csymbol.getLocalTheory();
    LogicEnvironment env = new LogicEnvironment(tsymbol, error);
    boolean okay = checker.prepare(tsymbol, env);
    if (!okay) Main.internalError("error in already processed theory");
    return checker;
  }
  
  /**************************************************************************
   * Create checker for processing logical entities in empty environment.
   * @param error the stream on which to write error messages.
   * @return a checker for processing logical entities.
   ************************************************************************/
  public static LogicChecking createChecker(ErrorStream error)
  {
    // no references to program environment are allowed
    LogicChecking checker = new LogicChecking(null, error);
    LogicEnvironment env = new LogicEnvironment(error);
    checker.prepare(env);
    return checker;
  }
  
  /*** set by constructor ***/
  
  // the program environment in which references are looked up
  private Environment penv;
  
  // the stream for generating error messages
  private ErrorStream error;
    
  /*** set by prepare() ***/
  
  // the current theory
  private TheorySymbol tsymbol;
  
  // the environment to be used for looking up logic entities
  private LogicEnvironment env;

  // the type table to be used for checking
  private LogicTypeTable ttable;

  // the type translator
  private TypeTranslator translator;

  // the type checking condition generator
  private TCCGenerator tcc;
  
  /*** set by setStateContext() ***/
  
  // the precondition of the current program state (null, if there is none)
  private Formula preCondition;
  
  // the type of the current program state (null, if there is no such state)
  private StateType resultType;
  
  // the variable frame of the current program state (null, if no such state)
  private Name[] stateVars;
  
  // the exceptions that may arise in the current state (null, if no such state)
  private Name[] stateExcs;
  
  // the old tcc directory
  private File oldTccDir;
  
  // the task folder for type checking conditions, set by setStateContext()
  private TaskFolder tccFolder;     
  
  /****************************************************************************
   * Construct a checker for logic entities.
   * @param penv the program environment to be used in checking.
   * @param error the stream on which to write the error messages.
   **************************************************************************/
  private LogicChecking(Environment penv, ErrorStream error)
  {
    this.penv = penv;
    this.error = error;
  }
  
  /****************************************************************************
   * Get the environment for looking up logic symbols.
   * @return the environment.
   ***************************************************************************/
  public LogicEnvironment getLogicEnvironment()
  {
    return env;
  }
  
  /****************************************************************************
   * Get the generator for producing type checking conditions
   * @return the type checking condition generator
   ***************************************************************************/
  public TCCGenerator getTCCGenerator()
  {
    return tcc;
  }
  
  /****************************************************************************
   * Type-check declarations in theory symbol and annotate it semantically and
   * populate the theory symbol with the symbols of the declarations.
   * @return true if no error occurred, false otherwise.
   ***************************************************************************/
  private boolean process(TheorySymbol tsymbol)
  {    
    if (tsymbol.getClassSymbol() == null && 
        !tsymbol.getName().toString().equals("(local)"))
      error.newMessage("processing theory " + tsymbol.getLongName() + "...");
    
    // the theory declaration
    TheoryDeclaration tdecl = tsymbol.getDeclaration();
    PackageSymbol psymbol = tsymbol.getPackage();
    PackageSymbol top = psymbol.getTopPackage();

    // initialize error stream appropriately
    SourcePosition pos = tdecl.getPosition();
    SourcePosition pos0 = null;
    if (pos != null)
    {
      error.clearErrors(pos.getSource());
      pos0 = error.getPosition();
      error.setPosition(pos);
    }

    // check name of package
    Name pname = tdecl.getPackage();
    if (pname == null) pname = top.getLongName();
    Name pname0 = psymbol.getLongName();
    if (!pname.equals(pname0))
    {
      error.newError(pname.idents[0], "declaration of wrong package " + pname + 
        " in theory " + tsymbol.getLongName());
      return false;
    }
    
    // check name of theory
    Identifier tident = tsymbol.getName();
    String tname = tident.getString();
    Identifier dident = tdecl.getName();
    String dname = dident.getString();
    if (tname != dname)
    {
      error.newError(dident, "found declaration of theory " + dname + 
        " rather than declaration of theory " + tname);
      return false;
    }
    dident.setSymbol(tsymbol);

    // prepare checker for processing declarations
    boolean okay = prepare(tsymbol, new LogicEnvironment(top, error));
    if (!okay) return false;
    
    // process declarations
    LogicChecking old = ttable.setChecker(this);
    env.enter();
    Declaration[] decls = tdecl.getDeclarations();
    okay = process(decls);
    env.exit();
    ttable.setChecker(old);
    
    // indicate theory which had problems
    if (!okay)
    {
      if (pos != null)
      {
        error.explainError(pos + ": declaration of theory " + tsymbol.getLongName());
        error.explainError("Source file " + pos.getSource().getFile().getAbsolutePath());
      }
      return false;
    }
    
    // add symbols to theory
    for (Declaration d : decls)
    {
      Symbol symbol = d.getName().getSymbol();
      if (symbol instanceof TypeSymbol)
        tsymbol.getTypes().put((TypeSymbol)symbol);
      else if (symbol instanceof FormulaSymbol)
        tsymbol.getFormulas().put((FormulaSymbol)symbol);
      else if (symbol instanceof ValueSymbol)
        tsymbol.getValues().put((ValueSymbol)symbol);
      else
        Main.internalError("unknown symbol");
    }
    
    // register type checking tasks in theory
    TaskFolder folder = tsymbol.getTCCs();
    folder.clear();
    tcc.addTasks(folder);
    
    if (pos != null) error.setPosition(pos0);
    return okay;
  }
  
  /****************************************************************************
   * Prepare checker for processing entities.
   * @param tsymbol the symbol of the current theory.
   * @param env the environment to be used for processing
   * @return true if everything went fine, false otherwise.
   ***************************************************************************/
  private boolean prepare(TheorySymbol tsymbol, LogicEnvironment env)
  {
    // set global entities
    this.tsymbol = tsymbol;
    this.ttable = tsymbol.getLogicTypeTable();
    this.translator = tsymbol.getPackage().getTopPackage().getTypeTranslator();
    this.env = env;
    
    // set theories that may be used by environment
    // by processing the theory imports and the uses declarative
    TheoryDeclaration tdecl = tsymbol.getDeclaration();
    Import[] imports = tdecl.getImports();
    Name[] tnames = tdecl.getTheories();
    PackageSymbol psymbol = tsymbol.getPackage();
    ClassSymbol csymbol = tsymbol.getHostClass();
    TheorySymbol classTheory = null;
    if (csymbol != null) classTheory = csymbol.getClassTheory();
    LogicChecking old = ttable.setChecker(this);
    Vector<TheorySymbol> theories = 
      env.setTheories(psymbol, translator.getBaseTheory(), classTheory,
          imports, tnames);
    ttable.setChecker(old);
    if (theories == null) return false;
    
    // set type checking condition generator
    File tdir = tsymbol.getTCCs().getDirectory();
    this.tcc = new TCCGenerator(tsymbol, theories, env, tdir, error);
    return true;
  }
  
  /****************************************************************************
   * Prepare checker for processing entities.
   * @param env the environment to be used for processing
   ***************************************************************************/
  private void prepare(LogicEnvironment env)
  {
    // set global entities (no theory and no type translation)
    this.tsymbol = null;
    this.ttable = new LogicTypeTable(error);
    this.translator = null;
    this.env = env;
    
    // set dummy type checking condition generator
    this.tcc = new TCCGenerator(null, null, env, null, error);
    
    // set checker for table
    this.ttable.setChecker(this);
  }
  
  /****************************************************************************
   * Set the current program state context.
   * @param pre the precondition of the method (may be null).
   * @param type the current state type (may be null).
   * @param vars the variable frame of the current program state (may be null).
   * @param excs the exceptions of the current program state (may be null).
   * @param folder the folder for generating type checking conditions
   ***************************************************************************/
  public void setStateContext(Formula pre, StateType type,
    Name[] vars, Name[] excs, TaskFolder folder)
  {
    // the precondition
    preCondition = pre;
    
    // construct state type
    if (type == null)
      resultType = null;
    else
    {
      LogicChecking old = ttable.setChecker(this);
      resultType = (StateType)ttable.process(type);
      ttable.setChecker(old);
    }
    
    // construct variable frame
    stateVars = vars;
    
    // determine the corresponding types
    /*
    if (stateVars != null)
    {
      int n = stateVars.length;
      LogicChecking old = ttable.setChecker(this);
      for (int i=0; i<n; i++)
      {
        VariableSymbol symbol = (VariableSymbol)(vars[i].getSymbol());
        Type type0 = logicType(symbol.getType());
        symbol.setLogicType(type0);
      }
      ttable.setChecker(old);
    }
    */
    
    // propagate information to type table
    ttable.setStateContext(resultType, stateVars, stateExcs);
    
    // set up tcc folder and directory
    tccFolder = folder;
    oldTccDir = tcc.setDirectory(folder.getDirectory());
  }
  
  /****************************************************************************
   * Reset the program state context.
   ***************************************************************************/
  public void resetStateContext()
  {
    resultType = null;
    stateVars = null;
    stateExcs = null;
    ttable.resetStateContext();
    tccFolder = null;
    tcc.setDirectory(oldTccDir);
    oldTccDir = null;
  }
  
  /****************************************************************************
   * decide whether definition type is first order, i.e.,
   * - a function type is only allowed at the top level
   * - the type BOOLEAN is only allowed at the top level
   *   or as the range of the function type
   * @param type an already processed type arising from a type definition
   * @return true iff the type is first order
   ***************************************************************************/
  public static boolean validDefinitionType(Type type)
  {
    if (!Main.firstOrder()) return true;
    if (isBooleanType(type))
      return true;
    if (!(type instanceof FunctionType))
      return validTermType(type);
    FunctionType type0 = (FunctionType)type;
    for (Type t : type0.getDomain())
    {
      if (!validTermType(t)) return false;
    }
    Type t = type0.getRange();
    return isBooleanType(t) || validTermType(t);
  }

  
  /****************************************************************************
   * decide whether type does not contain a function type or BOOLEAN.
   * @param type an already processed type
   * @return true iff the type does not contain a function type or type BOOLEAN
   ***************************************************************************/
  public static boolean validTermType(Type type)
  {
    if (!Main.firstOrder()) return true;
    if (type instanceof Reference)
    {
      Reference ident = (Reference)type;
      Type type0 = ident.getType();
      return validTermType(type0);
    }
    if (type instanceof AtomicType)
    {
      return !isBooleanType(type);
    }
    if (type instanceof FunctionType)
    {
      return false;
    }
    if (type instanceof ArrayType)
    {
      ArrayType type0 = (ArrayType)type;
      return validTermType(type0.getIndex()) 
          && validTermType(type0.getBase());
    }
    if (type instanceof RecordType)
    {
      RecordType type0 = (RecordType)type;
      for(TypedIdentifier t : type0.getBase())
      {
        ValueSymbol symbol = (ValueSymbol)t.getName().getSymbol();
        Type type1 = symbol.getType();
        if (!validTermType(type1))
          return false;
      }
      return true;
    }
    if (type instanceof TupleType)
    {
      TupleType type0 = (TupleType)type;
      for(Type t : type0.getBase())
      {
        if (!validTermType(t)) return false;
      }
      return true;
    }
    if (type instanceof SubrangeType)
    {
      return true;
    }
    if (type instanceof SubType)
    {
      SubType type0 = (SubType)type;
      return validTermType(type0.getBase());
    }
    if (type instanceof StateType)
    {
      StateType type0 = (StateType)type;
      if (type0.getReturn() == null) return true;
      return validTermType(type0.getReturn());
    }
    Main.internalError("unknown kind of type: " + type);
    return false;
  }
  
  /***************************************************************************
   * Process formula and create type-checking conditions in context
   * set up by setStateContext().
   * @param formula the formula to be processed.
   * @return true, if everything went fine, false otherwise.
   **************************************************************************/
  public boolean processFormula(Formula formula)
  {
    // add precondition to type checking context
    tcc.open();
    if (preCondition != null) tcc.assume(preCondition);
    
    // process formula in context of current state
    LogicChecking old = ttable.setChecker(this);
    env.enter();
    boolean okay = process(formula);
    env.exit();
    ttable.setChecker(old);
    if (!okay) 
    {
      tcc.abort();
      return false;
    }
    tcc.commit();
    
    // add task to folder
    tcc.addTasks(tccFolder);
    return true;
  }
  
  /***************************************************************************
   * Process term denoting an integer number and create type-checking 
   * conditions in context set up by setStateContext().
   * @param term the term to be processed.
   * @return true, if everything went fine, false otherwise.
   **************************************************************************/
  public boolean processIntTerm(Term term)
  { 
    // add precondition to type checking context
    tcc.open();
    if (preCondition != null) tcc.assume(preCondition);
    
    // process term in context of current state
    LogicChecking old = ttable.setChecker(this);
    env.enter();
    Type type = process(term);
    boolean okay = type != null;
    if (okay) okay = hasIntType(term, type, true);
    env.exit();
    ttable.setChecker(old);
    if (!okay) 
    {
      tcc.abort();
      return false;
    }
    tcc.commit();
    
    // add task to folder
    tcc.addTasks(tccFolder);
    return true;
  }
  
  /****************************************************************************
   * Type-check declaration sequence in current theory.
   * @param decls the declarations to be type-checked.
   * @return true iff everything went okay.
   ***************************************************************************/
  public boolean process(Declaration[] decls)
  {
    for (Declaration d: decls)
    {
      SourcePosition p = d.getPosition();
      error.setPosition(p);
      Symbol symbol = process(d);
      if (symbol == null) 
      { 
        error.explainError(p + ": declaration of " + d.getName());   
        return false;
      }
    }
    return true;
  }
  
  /****************************************************************************
   * Type-check declaration in current theory.
   * @param decl the declaration to be type-checked.
   * @return the symbol introduced by the declaration (null, in case of error)
   ***************************************************************************/
  private Symbol process(Declaration decl)
  {
    if (decl instanceof TypeDeclaration)
      return process((TypeDeclaration)decl);
    if (decl instanceof ValueDeclaration)
      return process((ValueDeclaration)decl);
    if (decl instanceof FormulaDefinition)
      return process((FormulaDefinition)decl);
    Main.internalError("unknown declaration: " + decl);
    return null;
  }
  
  /****************************************************************************
   * Type-check type declaration in current theory.
   * @param decl the declaration to be type-checked.
   * @return the symbol introduced by the declaration (null, in case of error)
   ***************************************************************************/
  private TypeSymbol process(TypeDeclaration decl)
  {
    Identifier name = decl.getName();
    Type type;
    if (decl instanceof TypeDefinition)
      type = process(((TypeDefinition)decl).getValue());
    else
      type = process(AtomicType.construct(name.getString()));
    if (type == null) return null;
    TypeSymbol result = env.putType(name, type, decl);
    if (result == null)
    {
      error.newError("a type with name " + name + " is already declared in the current scope");
      return null;
    }
    return result;
  }
  
  /****************************************************************************
   * Type-check value declaration in current theory.
   * @param decl the declaration to be type-checked.
   * @return the symbol introduced by the declaration (null, in case of error)
   ***************************************************************************/
  private ValueSymbol process(ValueDeclaration decl)
  {
    Identifier name = decl.getName();
    Type type = process(decl.getType());
    if (type == null) return null;
    if (!validDefinitionType(type))
    {
      error.newError(decl, "type " + type + " in declaration of value " + name
          + " is not first order (you must not use type " 
          + (Main.toLogical() ? "LOGICAL" : "BOOLEAN")
          + " or a function type as an argument type)");
      return null;
    }
    if (decl instanceof ValueDeclarationInitialized)
    {
      ValueDeclarationInitialized def = (ValueDeclarationInitialized)decl;
      Term value = def.getValue();
      Type vtype = processDefinitionTerm(value);
      if (vtype == null) return null;
      tcc.open();
      boolean okay = hasType(value, vtype, type, true);
      if (!okay)
      {
        error.newError(def, "constant " + name +
            " is declared with type " + type + 
            " but defined by value of type " + vtype);
        tcc.abort();
        return null;
      }
      tcc.commit();
    }
    else if (decl instanceof ValueDeclarationFormula)
    {
      ValueDeclarationFormula def = (ValueDeclarationFormula)decl;
      Formula formula = def.getFormula();
      boolean okay = process(formula);
      if (!okay) return null;
      tcc.open();
      if (!hasType(formula, type, true))
      {
        error.newError(def, "constant " + name +
            " is declared with type " + type + 
            " but defined by formula");
        tcc.abort();
        return null;
      }
      tcc.commit();
    }
    ValueSymbol result = env.putValue(name, type, decl);
    if (result == null)
    {
      error.newError("a value with name " + name + " is already declared in the current scope");
      return null;
    }
    return result;
  }
  
  /****************************************************************************
   * Type-check formula declaration in current theory.
   * @param decl the declaration to be type-checked.
   * @return the symbol introduced by the declaration (null, in case of error)
   ***************************************************************************/
  private FormulaSymbol process(FormulaDefinition decl)
  {
    Identifier name = decl.getName();
    Formula formula = decl.getValue();
    boolean okay = process(formula);
    if (!okay) return null;
    FormulaSymbol symbol = env.putFormula(name, formula, decl, tsymbol);
    if (symbol == null)
    {
      error.newError("a formula with name " + name + " is already defined in the current scope");
      return null;
    }
    
    // generate task for proving formula that is not an axiom
    if (decl instanceof AxiomDefinition) return symbol;
    TaskFolder ftasks = tsymbol.getFormulaTasks();
    FormulaTask task = new FormulaTask(decl, tsymbol, ftasks.getDirectory(), error);
    ftasks.add(task);
    return symbol;
  }
  
  /****************************************************************************
   * Type-check formula in current theory.
   * @param formula the formula to be type-checked.
   * @return true if everything went fine, false otherwise
   ***************************************************************************/
  public boolean process(Formula formula)
  {
    int depth = tcc.getDepth();
    boolean okay;
    if (formula instanceof AtomicFormula)
      okay = process((AtomicFormula)formula);
    else if (formula instanceof BinaryAtomicFormula)
      okay = process((BinaryAtomicFormula)formula);
    else if (formula instanceof BinaryFormula)
      okay = process((BinaryFormula)formula);
    else if (formula instanceof BooleanLiteral)
      okay = processFormula((BooleanLiteral)formula);
    else if (formula instanceof WritesOnlyFormula)
      okay = process((WritesOnlyFormula)formula);
    else if (formula instanceof IfThenElseFormula)
      okay = process((IfThenElseFormula)formula);
    else if (formula instanceof LetFormula)
      okay = process((LetFormula)formula);
    else if (formula instanceof PostfixFormula)
      okay = process((PostfixFormula)formula);
    else if (formula instanceof QuantifiedFormula)
      okay = process((QuantifiedFormula)formula);
    else if (formula instanceof Reference)
      okay = processFormula((Reference)formula);
    else if (formula instanceof ProgramVariableBase)
    {
      if (Main.firstOrder())
      {
        error.newError("formula " + formula 
            + " is not first order (you must not use a program variable)");
        return false;
      }
      okay = processFormula((ProgramVariableBase)formula);
    }
    else if (formula instanceof SelectionTerm)
    {
      if (Main.firstOrder())
      {
        error.newError("formula " + formula 
            + " is not first order (you must not use a selection term)");
        return false;
      }
      okay = processFormula((SelectionTerm)formula);
    }
    else if (formula instanceof UnaryFormula)
      okay = process((UnaryFormula)formula);
    else
    {
      Main.internalError("Unknown kind of formula " + formula);
      okay = false;
    }
    if (depth != tcc.getDepth())
      Main.internalError("error in tcc stack");
    return okay;
  }
  
  private boolean process(AtomicFormula formula)
  {
    Type type = processTermApplication(formula);
    if (type == null) return false;
    if (!isBooleanType(type))
    {
      error.newError(formula, "application " + formula + " has type " + type + 
          " but should have type " + LogicTypeTable.booleanType);
      return false;
    }
    return true;
  }
  
  private Type processTermApplication(TermApplication exp)
  {
    Term fun = exp.getFunction();
    Term[] args = exp.getArguments();
    Type type = processDefinitionTerm(fun);
    if (type == null) return null;
    tcc.open();
    FunctionType ftype = getFunctionType(type);
    if (ftype == null)
    {
      error.newError(exp, "function " + fun +
          " has type " + type +
          " (function type is expected in application)");
      tcc.abort();
      return null;
    }
    Type[] domain = ftype.getDomain();
    Type range = ftype.getRange();
    int n = domain.length;
    if (args.length != n)
    {
      error.newError(exp, "illegal application " + exp +
        " of function with type " + ftype + 
        " to " + args.length + " arguments (" + n + 
        " arguments expected)");
      tcc.abort();
      return null;
    }
    for (int i=0; i<n; i++)
    {
      Type atype = process(args[i]);
      if (atype == null) 
      {
        tcc.abort();
        return null;
      }
      if (hasType(args[i], atype, domain[i], true)) continue;
      error.newError(exp, "illegal application " + exp + 
        " of function with type " + ftype +
        " to argument #" + i + " " + args[i] +
        " of type " + atype + 
        " (argument of type " + domain[i] + " expected)");
      tcc.abort();
      return null;
    }
    tcc.commit();
    return range;
  }
  
  private boolean process(BinaryAtomicFormula formula)
  {
    if (formula instanceof EqualsFormula)
      return processEquality((EqualsFormula)formula);
    if (formula instanceof GreaterEqualFormula)
      return processInequality((GreaterEqualFormula)formula);
    if (formula instanceof GreaterFormula)
      return processInequality((GreaterFormula)formula);
    if (formula instanceof LessEqualFormula)
      return processInequality((LessEqualFormula)formula);
    if (formula instanceof LessFormula)
      return processInequality((LessFormula)formula);
    if (formula instanceof NotEqualsFormula)
      return processEquality((NotEqualsFormula)formula);
    if (formula instanceof SimilarFormula)
      return processSimilarity((SimilarFormula)formula);
    Main.internalError("Unknown kind of binary atomic formula " + formula);
    return false;
  }
  
  private boolean processEquality(BinaryAtomicFormula formula)
  {
    Term first = formula.getFirst();
    Type ftype = process(first);
    if (ftype == null) return false; 
    
    Term second = formula.getSecond();
    Type stype = process(second);
    if (stype == null) return false; 
    
    if (superType(ftype, stype) == null)
    {
      error.newError(first, "term " + first +
        " has type " + ftype +
        " while expression " + second +
        " has type " + stype +
        " (matching types expected in (non-)equality)");
      return false;
    }

    return true;
  }
  
  private boolean processInequality(BinaryAtomicFormula formula)
  {
    tcc.open();
    Term first = formula.getFirst();
    Type ftype = process(first);
    if (ftype == null) 
    { 
      tcc.abort();
      return false;
    }
    if (!hasRealType(first, ftype, true))
    {
      error.newError(first, "term " + first.toString() +
        " has type " + ftype.toString() +
        " (arithmetic type expected in inequality)");  
      tcc.abort();
      return false;
    }
    
    Term second = formula.getSecond();
    Type stype = process(second);
    if (stype == null) 
    {
      tcc.abort();
      return false;
    }
    if (!hasRealType(second, stype, true))
    {
      error.newError(second, "expression " + second.toString() +
        " has type " + stype.toString() +
        " (arithmetic type expected in in equality)"); 
      tcc.abort();
      return false;
    }  
   
    tcc.commit();
    return true;
  }

  private boolean processSimilarity(SimilarFormula formula)
  {
    Term term1 = formula.getFirst();
    Type type1 = process(term1);
    if (type1 == null) return false;
    type1 = getStateType(type1);
    if (type1 == null)
    {
      error.newError(term1, "term " + term1 +
          " has type " + type1 +
      " (state type expected in state similarity)");
      return false;
    }
    Term term2 = formula.getSecond();
    Type type2 = process(term2);
    if (type2 == null) return false;
    type2 = getStateType(type2);
    if (type2 == null)
    {
      error.newError(term2, "term " + term2 +
          " has type " + type2 +
      " (state type expected in state similarity)");
      return false;
    }
    return true;
  }
  
  private boolean process(BinaryFormula formula)
  {    
    if (formula instanceof AndFormula)
      return process((AndFormula)formula);
    if (formula instanceof EquivalentFormula)
      return process((EquivalentFormula)formula);
    if (formula instanceof ImpliesFormula)
      return process((ImpliesFormula)formula);
    if (formula instanceof NotEquivalentFormula)
      return process((NotEquivalentFormula)formula);
    if (formula instanceof OrFormula)
      return process((OrFormula)formula);
    Main.internalError("Unknown kind of binary formula " + formula);
    return false;
  }
  
  private boolean process(AndFormula formula)
  {
    return process(formula, 1);
  }
  
  private boolean process(OrFormula formula)
  {
    return process(formula, -1);
  }
  
  private boolean process(ImpliesFormula formula)
  {
    return process(formula, 1);
  }
  
  private boolean process(EquivalentFormula formula)
  {
    return process(formula, 0);
  }
  
  private boolean process(NotEquivalentFormula formula)
  {
    return process(formula, 0);
  }
  
  private boolean process(BinaryFormula formula, int context)
  {
    // set up context first because tccs are added as assumptions
    tcc.open();
    Formula first = formula.getFirst();
    boolean okay = process(first);
    if (!okay) 
    {
      tcc.abort();
      return false; 
    }
    
    // second subformula may be processed in context set up by first formula
    // context =  1: first subformula represents context
    // context = -1: negation of first subformula represents context
    // context = 0:  there is no context
    if (context != 0)
    {
      Formula assumption = first;
      if (context < 0) assumption = NotFormula.construct(assumption);
      tcc.assume(assumption);
    }
    
    Formula second = formula.getSecond();
    okay = process(second);
    if (!okay) 
    {
      tcc.abort();
      return false;
    }
    
    tcc.commit();
    return true;
  }
  
  private boolean processFormula(BooleanLiteral formula)
  {
    formula.setType(LogicTypeTable.booleanType);
    return true;
  }
  
  private boolean process(WritesOnlyFormula formula)
  {
    if (formula instanceof ReadsOnlyFormula)
      return process((ReadsOnlyFormula)formula);
    Name[] vars = formula.getVariables();
    for (Name var : vars)
    {
      VariableSymbol symbol = penv.getVariable(var);
      if (symbol == null)
      {
        error.newError(var.idents[0], 
            "there is no program variable " + var + " visible");
        return false;
      }
    }
    return true;
  }
    
  private boolean process(ReadsOnlyFormula formula)
  {
    return true;
  }
  
  private boolean process(IfThenElseFormula formula)
  {
    Formula condition = formula.getCondition();
    Formula thenBranch = formula.getThenBranch();
    Formula elseBranch = formula.getElseBranch();
    boolean okay = process(condition);
    if (!okay) return false;
    tcc.open();
    tcc.open();
    tcc.assume(condition);
    okay = process(thenBranch);
    if (!okay)
    {
      tcc.abort();
      tcc.abort();
      return false;
    }
    tcc.commit();
    tcc.open();
    tcc.assume(NotFormula.construct(condition));
    okay = process(elseBranch);
    if (!okay)
    {
      tcc.abort();
      tcc.abort();
      return false;
    }
    tcc.commit();
    tcc.commit();
    return true;
  }
  
  private boolean process(LetFormula formula)
  {
    env.enter();
    tcc.open();
    ValueDefinition[] defs = formula.getDefinitions();
    int n = defs.length;
    int i = process(defs);
    if (i < n) 
    { 
      tcc.abort();
      env.exit();
      return false; 
    }
    Formula base = formula.getBody();
    boolean okay = process(base);
    if (!okay) 
    { 
      tcc.abort();
      env.exit();
      return false; 
    }
    tcc.commit();
    env.exit();
    return true;
  }
    
  private int process(ValueDefinition[] defs)
  {
    int n = defs.length;
    for (int i=0; i<n; i++)
    {
      ValueDefinition def = defs[i];
      Identifier name = def.getName();
      Term value = def.getValue();
      if (def instanceof ValueDeclarationInitialized)
      {
        // attached type may narrow value type
        ValueDeclarationInitialized def0 = (ValueDeclarationInitialized)def;
        ValueSymbol vsymbol = process(def0);
        if (vsymbol == null) continue;
      }
      else
      {
        Type type = process(value);
        if (type == null) return i;
        ValueSymbol vsymbol = env.putValue(name, type, def);
        if (vsymbol == null) return i;
      }
      Reference ref = Reference.construct(name);
      Formula cond = EqualsFormula.construct(ref, value);
      boolean okay = process(cond);
      if (!okay) Main.internalError("error in value definition");
      tcc.assume(cond);
    }
    return n;
  }
  
  private boolean process(PostfixFormula formula)
  {
    Term arg = formula.getArgument();
    Type type = process(arg);
    StateType stype = getStateType(type);
    if (stype == null)
    {
      error.newError(arg, "argument " + arg + " is of type " + type +
          " but should denote a state");
      return false;
    }
    if (formula instanceof StateThrowsException)
    {
      StateThrowsException f = (StateThrowsException)formula;
      Name exception = f.getException();
      ClassSymbol csymbol = penv.getClass(exception);
      if (csymbol == null) return false;
    }
    return true;
  }
  
  private boolean process(QuantifiedFormula formula)
  {
    TypedIdentifier[] vars = formula.getVariables();
    Formula body = formula.getBody();
    env.enter();
    boolean okay = process(vars);
    if (!okay) 
    {
      env.exit();
      return false;
    }
    okay = process(body);
    if (!okay) 
    {
      env.exit();
      return false;
    }
    env.exit();
    return true;
  }
  
  private boolean processFormula(Reference ref)
  {
    ValueSymbol symbol = env.getValueSymbol(ref);
    if (symbol == null) return false;
    Type type = symbol.getType();
    if (!isBooleanType(type))
    {
      error.newError(ref, "reference " + ref + " has type " + type + 
          " but should have type " + LogicTypeTable.booleanType);
      return false;
    }
    return true;
  }
  
  private boolean processFormula(ProgramVariableBase var)
  {
    Type type = process(var);
    if (type == null) return false;
    if (!isBooleanType(type))
    {
      error.newError(var, "value " + var + " has type " + type + 
          " but should have type " + LogicTypeTable.booleanType);
      return false;
    }
    return true;
  }
  
  private boolean processFormula(SelectionTerm term)
  {
    Type type = process(term);
    if (type == null) return false;
    if (!isBooleanType(type))
    {
      error.newError(term, "selection " + term + " has type " + type + 
          " but should have type " + LogicTypeTable.booleanType);
      return false;
    }
    return true;
  }
  
  private boolean process(UnaryFormula formula)
  {    
    if (formula instanceof NotFormula)
      return process((NotFormula)formula);
    Main.internalError("Unknown kind of unary formula " + formula);
    return false;
  }
  
  private boolean process(NotFormula formula)
  {    
    return process(formula.getBase());
  }
  
  /****************************************************************************
   * Type-check term in current theory and, as a side effect,
   * annotate the term with the type.
   * @param term the term to be type-checked.
   * @return the resulting type (null, if an error occurred)
   ***************************************************************************/
  public Type process(Term term)
  {
    Type result = processTermCore(term);
    if (result == null) return result;
    if (!validTermType(result))
    {
      error.newError(term, "type " + result + " of term " + term
          + " is not first order (you must not use type " 
          + (Main.toLogical() ? "LOGICAL" : "BOOLEAN")
          + " or a function type here)");
      return null;
    }
    term.setType(result);
    return result;
  }
  
  /****************************************************************************
   * Type-check term in current theory and, as a side effect,
   * annotate the term with the type. The term may appear
   * as the value of a (also function or predicate) definition.
   * @param term the term to be type-checked.
   * @return the resulting type (null, if an error occurred)
   ***************************************************************************/
  private Type processDefinitionTerm(Term term)
  {
    Type result = processTermCore(term);
    if (result == null) return result;
    if (!validDefinitionType(result))
    {
      error.newError(term, "type " + result + " of definition term " + term
          + " is not first order (you must not use type " 
          + (Main.toLogical() ? "LOGICAL" : "BOOLEAN")
          + " or a function type as an argument type)");
      return null;
    }
    term.setType(result);
    return result;
  }
  
  /****************************************************************************
   * Type-check term in current theory and, as a side effect,
   * annotate the term with the type.
   * @param term the term to be type-checked.
   * @return the resulting type (null, if an error occurred)
   ***************************************************************************/
  public Type processTermCore(Term term)
  {
    int depth = tcc.getDepth();
    Type result;
    if (term instanceof BinaryTerm)
      result = process((BinaryTerm)term);
    else if (term instanceof BooleanLiteral)
      result = processTerm((BooleanLiteral)term);
    else if (term instanceof BitLiteral)
      result = process((BitLiteral)term);
    else if (term instanceof FunctionApplication)
      result = process((FunctionApplication)term);
    else if (term instanceof IfThenElseTerm)
      result = process((IfThenElseTerm)term);
    else if (term instanceof IntLiteral)
      result = process((IntLiteral)term);
    else if (term instanceof StringLiteral)
      result = process((StringLiteral)term);
    else if (term instanceof LetTerm)
      result = process((LetTerm)term);
    else if (term instanceof PostfixTerm)
      result = process((PostfixTerm)term);
    else if (term instanceof QuantifiedTerm)
      result = process((QuantifiedTerm)term);
    else if (term instanceof Reference)
      result = processTerm((Reference)term);
    else if (term instanceof StateLiteral)
      result = process((StateLiteral)term);
    else if (term instanceof TermBase)
      result = process((TermBase)term);
    else if (term instanceof UnaryTerm)
      result = process((UnaryTerm)term);
    else
    {
      Main.internalError("unknown kind of term " + term);
      return null;
    }
    if (depth != tcc.getDepth())
      Main.internalError("error in tcc stack");
    return result;
  }
  
  private Type process(BinaryTerm term)
  {
    if (term instanceof DividesTerm)
      return process((DividesTerm)term);
    if (term instanceof MinusTerm)
      return process((MinusTerm)term);
    if (term instanceof PlusTerm)
      return process((PlusTerm)term);
    if (term instanceof PowerTerm)
      return process((PowerTerm)term);
    if (term instanceof TimesTerm)
      return process((TimesTerm)term);
    Main.internalError("unknown kind of binary term " + term);
    return null;
  }
  
  private final class TypeTuple
  {
    Type first;
    Type second;
    TypeTuple(Type first, Type second)
    {
      this.first = first;
      this.second = second;
    }
  }
  
  private Type process(PlusTerm term)
  {
    TypeTuple types = processBinaryArithmetic(term, false);
    if (types == null) return null;
    if (hasNatType(term, types.first, false) &&
        hasNatType(term, types.second, false))
      return LogicTypeTable.natType;
    if (hasIntType(term, types.first, false) &&
        hasIntType(term, types.second, false))
      return LogicTypeTable.intType;
    return LogicTypeTable.realType;
  }
  
  private Type process(MinusTerm term)
  {
    TypeTuple types = processBinaryArithmetic(term, false);
    if (types == null) return null;
    if (hasIntType(term, types.first, false) &&
        hasIntType(term, types.second, false))
      return LogicTypeTable.intType;
    return LogicTypeTable.realType;
  }
  
  private Type process(TimesTerm term)
  {
    TypeTuple types = processBinaryArithmetic(term, false);
    if (types == null) return null;
    if (hasNatType(term, types.first, false) &&
        hasNatType(term, types.second, false))
      return LogicTypeTable.natType;
    if (hasIntType(term, types.first, false) && 
        hasIntType(term, types.second, false))
      return LogicTypeTable.natType;
    return LogicTypeTable.realType;
  }
  
  private Type process(DividesTerm term)
  {
    TypeTuple types = processBinaryArithmetic(term, false);
    if (types == null) return null;
    // do not allow partial functions because of CVCL
    Formula formula =
      NotEqualsFormula.construct(term.getSecond(), IntLiteral.construct("0"));
    tcc.prove("divisor is not zero", 
        term.getPosition(), formula, resultType, stateVars, stateExcs);
    return LogicTypeTable.realType;
  }
  
  private Type process(PowerTerm term)
  {
    TypeTuple types = processBinaryArithmetic(term, true);
    if (types == null) return null;
    if (hasNatType(term, types.second, false))
      return types.first;
    return LogicTypeTable.realType;
  }
  
  /***************************************************************************
   * Process binary arithmetic term
   * @param term the term
   * @param integer true iff second argument must denote an integer.
   * @return a tuple of the argument types (null, if an error occurred).
   **************************************************************************/
  private TypeTuple processBinaryArithmetic(BinaryTerm term, boolean integer)
  {
    Term first = term.getFirst();
    Type ftype = process(first);
    if (ftype == null) return null; 
    
    Term second = term.getSecond();
    Type stype = process(second);
    if (stype == null) return null; 
    
    tcc.open();
    if (!hasRealType(first, ftype, true))
    {
      error.newError(first, "term " + first +
        " has type " + ftype +
        " (INT or REAL expected in arithmetic expression)");    
      tcc.abort();
      return null;
    }
    if (( integer && !hasIntType(second, stype, true)) ||
        (!integer && !hasRealType(second, stype, true)))
    {
      error.newError(second, "term " + second +
          " has type " + stype +
          " (INT or REAL expected in arithmetic expression)");  
      tcc.abort();
      return null;
    }
    tcc.commit();
    return new TypeTuple(ftype, stype);
  }
  
  private Type processTerm(BooleanLiteral term)
  {
    return LogicTypeTable.booleanType;
  }
  
  private Type process(BitLiteral term)
  {
    return LogicTypeTable.bitType;
  }
  
  private Type process(FunctionApplication term)
  {
    return processTermApplication(term);
  }
  
  private Type process(IfThenElseTerm term)
  {
    Formula condition = term.getCondition();
    boolean okay = process(condition);
    if (!okay) return null;
    tcc.open();
    
    tcc.open();
    tcc.assume(condition);
    Term thenBranch = term.getThenBranch();
    Type thenType = process(thenBranch);
    if (thenType == null)
    {
      tcc.abort();
      tcc.abort();
      return null;
    }
    tcc.commit();
    
    tcc.open();
    tcc.assume(NotFormula.construct(condition));
    Term elseBranch = term.getElseBranch();
    Type elseType = process(elseBranch);
    if (elseType == null)
    {
      tcc.abort();
      tcc.abort();
      return null;
    }
    tcc.commit();
        
    // first try to determine result type without generating TCC
    if (hasType(thenBranch, thenType, elseType, false)) 
    {
      tcc.commit();
      return elseType;
    }
    if (hasType(elseBranch, elseType, thenType, false))
    {
      tcc.commit();
      return thenType;
    }
    Type superType = superType(thenType, elseType);
    if (superType != null)
    {
      tcc.commit();
      return superType;
    }
    
    // now try to determine result type with generating TCC
    if (hasType(thenBranch, thenType, elseType, true)) 
    {
      tcc.commit();
      return elseType;
    }
    if (hasType(elseBranch, elseType, thenType, true)) 
    { 
      tcc.commit();
      return thenType;
    }
    
    // all attempts to determine common result type failed
    tcc.abort();
    error.newError(term, "term " + thenBranch +
        " has type " + thenType +
        " while expression " + elseBranch +
        " has type " + elseType +
    " (matching types expected in conditional branches)");
    return null;
  }
  
  private Type process(IntLiteral term)
  {
    // no sign included
    return LogicTypeTable.natType;
    
    // more precise approximation
    // return ttable.processSubrangeType(SubrangeType.construct(term, term));
  }
  
  private Type process(StringLiteral term)
  {
    return translator.getStringType();
  }
  
  private Type process(LetTerm term)
  {
    env.enter();
    tcc.open();
    ValueDefinition[] defs = term.getDefinitions();
    int n = defs.length;
    int i = process(defs);
    if (i < n) 
    { 
      tcc.abort();
      env.exit();
      return null; 
    }
    Term base = term.getBody();
    Type type = process(base);
    if (type == null) 
    { 
      tcc.abort();
      env.exit();
      return null; 
    }
    tcc.commit();
    env.exit();
    return type;
  }
  
  private Type process(PostfixTerm term)
  {
    if (term instanceof StateValue)
      return process((StateValue)term);
    if (term instanceof StateMessage)
      return process((StateMessage)term);
    Main.internalError("unknown postfix term " + term);
    return null;
  }
  
  private Type process(StateValue term)
  {
    Term arg = term.getArgument();
    Type type = process(arg);
    if (type == null) return null;
    StateType stype = getStateType(type);
    if (stype == null)
    {
      error.newError(arg, "term " + arg + " is of type " + type +
          " but should denote a state");
      return null;
    }
    Type rtype = stype.getReturn();
    if (rtype == null)
    {
      error.newError(arg, "term " + arg + " is of type " + stype +
          " which does not provide .VALUE");
      return null;
    }
    return rtype;
  }
  
  private Type process(StateMessage term)
  {
    Term arg = term.getArgument();
    Type type = process(arg);
    if (type == null) return null;
    StateType stype = getStateType(type);
    if (stype == null)
    {
      error.newError(arg, "term " + arg + " is of type " + type +
      " which does not provide .MESSAGE");
      return null;
    }
    return translator.getStringType();
  }
  
  private Type process(QuantifiedTerm term)
  {
    if (term instanceof ArrayTerm)
      return process((ArrayTerm)term);
    if (term instanceof LambdaFormula)
      return process((LambdaFormula)term);
    if (term instanceof LambdaTerm)
      return process((LambdaTerm)term);
    Main.internalError("unknown kind of quantified term: " + term);
    return null;
  }
  
  private Type process(ArrayTerm term)
  {
    TypedIdentifier[] vars = term.getVariables();
    Term body = term.getBody();
    env.enter();
    boolean okay = process(vars);
    if (!okay) 
    {
      env.exit();
      return null;
    }
    Type type = process(body);
    if (type == null) 
    {
      env.exit();
      return null;
    }
    int n = vars.length;
    if (n != 1)
    {
      error.newError(term, "ARRAY expression " + term +
        " has " + n + " parameters (1 expected)");
      env.exit();
      return null;
    }
    Type ptype = vars[0].getType();
    Type result = ttable.process(ArrayType.construct(ptype, type));
    env.exit();
    return result;
  }
  
  private Type process(LambdaTerm term)
  {
    TypedIdentifier[] vars = term.getVariables();
    Term body = term.getBody();
    env.enter();
    boolean okay = process(vars);
    if (!okay) 
    {
      env.exit();
      return null;
    }
    Type type = process(body);
    if (type == null) 
    {
      env.exit();
      return null;
    }
    int n = vars.length;
    Type[] ptypes = new Type[n];
    for (int i=0; i<n; i++)
      ptypes[i] = vars[i].getType();
    Type result = ttable.process(FunctionType.construct(ptypes, type));
    env.exit();
    return result;
  }
  
  private Type process(LambdaFormula term)
  {
    TypedIdentifier[] vars = term.getVariables();
    Formula body = term.getBody();
    env.enter();
    boolean okay = process(vars);
    if (!okay) 
    {
      env.exit();
      return null;
    }
    okay = process(body);
    if (!okay) 
    {
      env.exit();
      return null;
    }
    env.exit();
    int n = vars.length;
    Type[] ptypes = new Type[n];
    for (int i=0; i<n; i++)
      ptypes[i] = vars[i].getType();
    return ttable.process(FunctionType.construct(ptypes, 
        LogicTypeTable.booleanType));
  }
  
  private Type processTerm(Reference ref)
  {
    ValueSymbol symbol = env.getValueSymbol(ref);
    if (symbol == null) return null;
    return symbol.getType();
  }
  
  private Type process(StateLiteral term)
  {
    if (resultType == null)
    {
      error.newError(term, "literal " + term + 
          " may be only used in the context of a program state ");
      return null;
    }
    return resultType;
  }
  
  private Type process(TermBase term)
  {
    if (term instanceof ProgramVariableBase)
      return process((ProgramVariableBase)term);
    if (term instanceof RecordTerm)
      return process((RecordTerm)term);
    if (term instanceof SelectionTerm)
      return process((SelectionTerm)term);
    if (term instanceof TupleTerm)
      return process((TupleTerm)term);
    if (term instanceof UpdateTerm)
      return process((UpdateTerm)term);
    Main.internalError("unknown term derived from TermBase");
    return null;
  }
  
  private Type process(ProgramVariableBase term)
  {
    Name name = term.getName();
    VariableSymbol symbol = penv.getVariable(name, false);
    if (symbol == null)
    {
      error.newError(name.idents[0], 
          "there is no program variable " + name + " visible");
      return null;
    }
    Type type = ttable.process(symbol.getLogicType());
    return type;
  }
  
  private Type process(RecordTerm term)
  {
    RecordTerm v = (RecordTerm)term;
    ValuedIdentifier[] base = v.getBase();
    int n =  base.length;
    TypedIdentifier[] types = new TypedIdentifier[n];
    for (int i=0; i<n; i++)
    {
      Type t = process(base[i].getValue());
      if (t == null) return null;
      Identifier ident = 
        Identifier.construct(base[i].getIdentifier().getString());
      types[i] = TypedIdentifier.construct(ident, t);
    }
    return ttable.process(RecordType.construct(types));
  }
  
  private Type process(SelectionTerm term)
  {
    SelectionTerm v = (SelectionTerm)term; 
    Term base = v.getBase();
    Selector sel = v.getSelector();
    Type btype = process(base);
    if (btype == null) return null;
    return getSelectedType(base, btype, sel);
  }
  
  private Type process(TupleTerm term)
  {
    TupleTerm v = (TupleTerm)term;
    Term[] base = v.getBase();
    int n =  base.length;
    Type[] types = new Type[n];
    for (int i=0; i<n; i++)
    {
      Type t = process(base[i]);
      if (t == null) return null;
      types[i] = t;
    }
    return ttable.process(TupleType.construct(types));
  }
  
  private Type process(UpdateTerm term)
  {
    UpdateTerm t = (UpdateTerm)term;
    Term base = t.getBase();
    Selector[] sels = t.getSelectors();
    Term update = t.getValue();
    Type btype = process(base);
    if (btype == null) return null;
    Term exp = base;
    Type type = btype;
    int n = sels.length;
    for (int i=0; i<n; i++)
    {
      Selector sel = sels[i];
      type = getSelectedType(exp, type, sel);
      if (type == null) return null;
      exp = SelectionTerm.construct(exp, sel);
    }
    Type utype = process(update);
    if (utype == null) return null;
    if (hasType(update, utype, type, true)) return btype;
    error.newError(term, " type " + utype.toString() + 
        " of update value does not match type " + utype +
        " of " + exp);
    return null;
  }
  
 
  private Type process(UnaryTerm term)
  {
    if (term instanceof NegationTerm)
      return process((NegationTerm)term);
    Main.internalError("unknown unary term");
    return null;
  }
  
  private Type process(NegationTerm term)
  {
    Term base = term.getBase();
    Type btype = process(base);
    if (btype == null) return null; 
    tcc.open();
    if (!hasRealType(base, btype, true))
    {
      error.newError(base, "term " + base +
        " has type " + btype +
        " (INT or REAL expected in arithmetic expression)");    
      tcc.abort();
      return null;
    }
    tcc.commit();
    return btype;
  }
  
  /****************************************************************************
   * Returns type of selected expression
   * @param base the base expression
   * @param btype the type of the base expression
   * @param sel the selector
   * @return the selected type or null, if something went wrong
   ***************************************************************************/
  private Type getSelectedType(Expression base, Type btype, Selector sel)
  {
    // first strip subtypes from btype
    while (btype instanceof SubType)
    {
      SubType stype = (SubType)btype;
      btype = stype.getBase();
    }
    if (sel instanceof SelectorNumber)
    {
      if (!(btype instanceof TupleType))
      {
        error.newError(base, "value " + base.toString() +
            " has type " + btype +
        " (TUPLE type expected for component selection by number)");
        return null;
      }
      Type[] btypes = ((TupleType)btype).getBase();
      Integer number = parseInt(base.getPosition(),
          ((SelectorNumber)sel).getNumber().getName());
      if (number == null) return null;
      int n = number.intValue();
      int l = btypes.length;
      if (n < 0 || l <= n)
      {
        error.newError(base, "type " + btype.toString() +
            " of term " + base + 
            " only allows selection of components 0.." + (n-1) +
            " (component " + n + " selected)");
        return null;
      }
      return btypes[n];
    }
    if (sel instanceof SelectorIdentifier)
    {     
      if (!(btype instanceof RecordType))
      {
        error.newError(base, "value " + base.toString() +
            " has type " + btype +
        " (RECORD type expected for component selection by identifier)");
        return null;
      }
      Identifier ident = ((SelectorIdentifier)sel).getIdentifier();
      String name = ident.getString();
      TypedIdentifier[] tids = ((RecordType)btype).getBase();
      int l = tids.length;
      for (int i=0; i<l; i++)
      {
        TypedIdentifier tid = tids[i];
        Identifier id = tid.getName();
        if (name.equals(id.getString()))
        {
          ident.setSymbol(id.getSymbol());
          return tid.getType();
        }
      }
      error.newError(base, "type " + btype.toString() +
          " of term " + base.toString() + 
          " does not have a component called " + name);
      return null;
    }
    if (sel instanceof SelectorIndex)
    {
      if (!(btype instanceof ArrayType))
      {
        error.newError(base, "value " + base.toString() +
            " has type " + btype.toString() +
        " (ARRAY type expected for component selection by index)");
        return null;
      }
      Type itype = ((ArrayType)btype).getIndex();
      Type etype = ((ArrayType)btype).getBase();
      Term index = ((SelectorIndex)sel).getIndex();
      Type stype = process(index);
      if (stype == null) return null;
      if (!hasType(index, stype, itype, true))
      {
        error.newError(base, "index " + index.toString() +
            " has type " + stype.toString() +
            " which does not match index type " + itype.toString() +
            " of array " + base.toString());
        return null;
      }
      return etype;
    }
    Main.internalError("unknown kind of selector: " + sel.toString());
    return null;
  }
  
  /****************************************************************************
   * get number from digit representation
   * @param pos a source position for printing error messages
   * @param digits the representation of the number as a decimal digit sequence
   * @return the number or null if conversion not possible
   ***************************************************************************/
  private Integer parseInt(SourcePosition pos, String digits)
  {
    try
    {
      return new Integer(Integer.parseInt(digits));
    }
    catch(NumberFormatException e)
    {
      error.newError(pos, "number " + digits + "cannot be represented as an int");
      return null;
    }
  }
  
  /****************************************************************************
   * Process type expression in current theory.
   * @param type the type expression to be type-checked.
   * @return the resulting canonical type (null, if an error occurred)
   ***************************************************************************/
  public Type process(Type type)
  {
    return ttable.process(type);
  }
  
  /***************************************************************************
   * Checks if term exp of type etype matches type.
   * May generate as a side effect a type checking condition.
   * 
   * The following general rules succeed without generating TCCs:
   *    
   *  exp: T                  matches T
   *  exp: SUBTYPE(p:T->BOOL) matches T
   * 
   * The following general rule succeeds by generating a TCC:
   * 
   *   exp: T matches SUBTYPE(p:T->BOOL) if p(exp)
   *   
   * the following structural rules do not generate TCCs:
   * 
   *   exp: A->B         matches A0->B0,         if A0 matches A and B matches B0
   *   exp: ARRAY A OF B matches ARRAY A0 OF B0, if A0 matches A and B matches B0   
   *   
   * the following structural rules *may* generate TCCs:
   * 
   *   exp: [T,...]      matches [T0,...],       if exp.0: T matches T0, ...
   *   exp: [#f:T,...#]  matches [#f:T0,...],    if exp.f: T matches T0, ...
   * 
   * as for arithmetic, the relationship is as follows:
   *
   *   [a..b] < INT   < REAL
   *   NAT    < INT
   *
   * The corresponding conversions from left to right succeed without
   * generating TCCs; conversions are applied transitively.
   * 
   * As for the conversion from right to left, we have:
   * 
   *  [ // disabled because of CVCL problems
   *    exp: REAL   matches INT      if (EXISTS (x:INT): x=exp)
   *  ]
   *  
   *   exp: INT    matches [a..b]   if a <= exp && exp <= b
   *   exp: INT    matches NAT      if 0 <= exp
   *    
   * As for subranges of INT, we have
   * 
   *   exp: NAT matches [a..b]      if (a <= 0 || a <= exp) && (exp <= b)
   *   exp: [a..b] matches NAT      if 0 <= a || 0 <= exp
   *   exp: [a..b] matches [c..d]   if (c <= a || c <= exp) &&
   *                                   (b <= d || exp <= d)
   *                                                                 
   * If rules are applied transitively, the conditions are appropriately
   * conjoined, e.g. exp: REAL -> SUBTYPE(p:[a..b]->BOOL) generates the TCC
   * 
   *   (EXISTS (x:INT): x=exp) && a <= exp && exp <= b && p(exp)
   * 
   * @param exp the term (may be null if gen is false)
   * @param etype its type
   * @param type the type which expr:etype is expected to match
   * @param gen true iff a tcc may be generated
   * @return true iff the matching succeeds
   **************************************************************************/
  private boolean hasType(Term exp, Type etype, Type type, boolean gen)
  {
    // exp: T matches T
    if (etype == type) return true;
    
    // exp: SUBTYPE(p:T->BOOL) matches T (with assumption p(exp))
    if (etype instanceof SubType)
    {
      SubType stype = (SubType)etype;
      Type domain = stype.getBase();
      if (gen)
      {
        tcc.open();
        tcc.assume(AtomicFormula.construct(
                     stype.getPredicate(), new Term[] { exp }));
      }
      boolean result = hasType(exp, domain, type, gen);
      if (gen) 
      {
        if (result)
          tcc.commit();
        else
          tcc.abort();
      }
      return result;
    }
    
    // exp: T matches SUBTYPE(p:T->BOOL) if p(exp)
    // in non-generating type mode, we just check for base type compatibility
    if (type instanceof SubType)
    {
      SubType stype = (SubType)type;
      Type domain = stype.getBase();
      tcc.open();
      if (!hasType(exp, etype, domain, gen))
      {
        tcc.abort();
        return false;
      }
      if (!gen) 
      {
        tcc.abort();
        return true;
      }
      Formula tccFormula = 
        AtomicFormula.construct(stype.getPredicate(), new Term[] { exp });
      tcc.prove("value satisfies subtype property", 
          exp.getPosition(), tccFormula, resultType, stateVars, stateExcs);
      tcc.commit();
      return true;
    }
    
    // exp: A->B matches A0->B0, if A0 matches A and B matches B0
    if (etype instanceof FunctionType && type instanceof FunctionType)
    {
      FunctionType etype0 = (FunctionType)etype;
      FunctionType type0 = (FunctionType)type;
      
      // matching of domain
      Type[] domain = type0.getDomain();
      Type[] edomain = etype0.getDomain();
      int n = domain.length;
      if (n != edomain.length) return false;
      for (int i=0; i<n; i++)
      {
        // contravariant matching of domain
        if (!hasType(null, domain[i], edomain[i], false)) return false;
      }

      // covariant matching of range
      return hasType(null, etype0.getRange(), type0.getRange(), false);
    }
   
    // exp: ARRAY A OF B matches ARRAY A0 OF B0, if A0 matches A and B matches B0
    if (etype instanceof ArrayType && type instanceof ArrayType)
    {
      ArrayType etype0 = (ArrayType)etype;
      ArrayType type0 = (ArrayType)type;
      
      // array type is contravariant in index, no TCCs may be generated
      return (hasType(null, type0.getIndex(), etype0.getIndex(), false) &&
              hasType(null, etype0.getBase(), type0.getBase(), false));
    }
    
    // exp: [T,...] matches [T0,...], if exp.0: T matches T0, ...
    if (etype instanceof TupleType && type instanceof TupleType)
    {
      TupleType etype0 = (TupleType)etype;
      TupleType type0 = (TupleType)type;
      Type[] etype1 = etype0.getBase();
      Type[] type1 = type0.getBase();
      int n = etype1.length;
      if (n != type1.length) return false;
      if (gen) tcc.open();
      for (int i=0; i<n; i++)
      {
        Term exp0 = null;
        if (gen)
          exp0 = SelectionTerm.construct(exp,
            SelectorNumber.construct(
              IntLiteral.construct(Integer.toString(i))));
        if (!hasType(exp0, etype1[i], type1[i], gen)) 
        {
          if (gen) tcc.abort();
          return false;  
        }
      }
      if (gen) tcc.commit();
      return true;
    }
    
    // exp: [#f:T,...#]  matches [#f:T0,...],    if exp.f: T matches T0, ...
    if (etype instanceof RecordType && type instanceof RecordType)
    {
      RecordType etype0 = (RecordType)etype;
      RecordType type0 = (RecordType)type;
      TypedIdentifier[] etype1 = etype0.getBase();
      TypedIdentifier[] type1 = type0.getBase();
      int n = etype1.length;
      if (n != type1.length) return false;
      if (gen) tcc.open(); 
      for (int i=0; i<n; i++)
      {
        Identifier ident = etype1[i].getName();
        if (!ident.getString().equals(type1[i].getName().getString())) 
        {
          if (gen) tcc.abort();
          return false;
        }
        Term exp0 = null;
        if (gen)
        {
          // identifier shared between record type and record expression
          exp0 = SelectionTerm.construct(exp, SelectorIdentifier.construct(ident));
        }
        if (!hasType(exp0, etype1[i].getType(), type1[i].getType(), gen))
        {
          if (gen) tcc.abort();
          return false;
        }
      }
      if (gen) tcc.commit();
      return true;
    }
    
    // arithmetic conversions without tccs
    if (etype instanceof SubrangeType && 
        (type instanceof IntType || type instanceof RealType)) 
      return true;
    if (etype instanceof NatType && 
        (type instanceof IntType || type instanceof RealType))
      return true;
    if (etype instanceof IntType && type instanceof RealType)
      return true;
    
    // after that, only chance is to generate tccs
    if (!gen) return false;

    // INT->[c..d]      c <= exp && exp <= d
    // NAT->[c..d]      (c <= 0 || c <= exp) && exp <= d
    // [a..b]->[c..d]   (c <= a || c <= exp) && (b <= d || exp <= d)   
    if ((etype instanceof IntType || etype instanceof NatType || 
         etype instanceof SubrangeType) && 
           type instanceof SubrangeType)
    {
      SubrangeType stype = (SubrangeType)type;
      Term from = stype.getFrom();
      Term to = stype.getTo();
      Formula tcc1 = LessEqualFormula.construct(from, exp);
      Formula tcc2 = LessEqualFormula.construct(exp, to);
      if (etype instanceof NatType)
        tcc1 = OrFormula.construct(
         LessEqualFormula.construct(from, IntLiteral.construct("0")),
         tcc1);
      else if (etype instanceof SubrangeType)
      {
        SubrangeType etype0 = (SubrangeType)etype;
        tcc1 = OrFormula.construct(
            LessEqualFormula.construct(from, etype0.getFrom()),
            tcc1);
        tcc2 = OrFormula.construct(
            LessEqualFormula.construct(etype0.getTo(), to),
            tcc2);
      }
      Formula tccFormula = AndFormula.construct(tcc1, tcc2);
      tcc.prove("value is in interval", 
          exp.getPosition(), tccFormula, resultType, stateVars, stateExcs);
      return true;
    }
    
    // INT->NAT         0 <= exp
    if (etype instanceof IntType && type instanceof NatType)
    {
      Formula tccFormula =
          GreaterEqualFormula.construct(exp, IntLiteral.construct("0"));
      tcc.prove("value is natural number", 
          exp.getPosition(), tccFormula, resultType, stateVars, stateExcs);
      return true;
    }
    
    // [a..b]->NAT         (0 <= a) || (0 <= exp)
    if (etype instanceof SubrangeType && type instanceof NatType)
    {
      Term from = ((SubrangeType)etype).getFrom();
      Formula tccFormula =
          OrFormula.construct(
            GreaterEqualFormula.construct(from, IntLiteral.construct("0")),
            GreaterEqualFormula.construct(exp, IntLiteral.construct("0"))
          );
      tcc.prove("value is natural number", 
          exp.getPosition(), tccFormula, resultType, stateVars, stateExcs);
      return true;
    }
    
    /* disabled because of CVCL problems
     
    // REAL->INT        EXISTS(x:INT):x=exp
    if (etype instanceof RealType && type instanceof IntType)
    {
      Collection<String> free = FreeVariables.compute(exp);
      Identifier ident = Identifier.constructNew("x", free);
      TypedIdentifier[] vars = 
        { TypedIdentifier.construct(ident, IntType.construct()) } ;
      Formula tccFormula =
        ExistsFormula.construct(
          vars,
          EqualsFormula.construct(Reference.construct(ident), exp));
      boolean okay = processTCC(tccFormula); // set reference value symbol
      if (!okay) Main.internalError("error in processing tcc formula");
      tcc.prove("value is integer", 
          exp.getPosition(), tccFormula, resultType, stateVars, stateTypes, stateExcs);
      return true;
    }
    
    */
    
    // REAL->NAT        EXISTS(x:NAT):x=exp
    if (etype instanceof RealType && type instanceof NatType)
    {
      Collection<String> free = FreeVariables.compute(exp);
      Identifier ident = Identifier.constructNew("x", free);
      TypedIdentifier[] vars = 
        { TypedIdentifier.construct(ident , NatType.construct()) } ;
      Formula tccFormula =
        ExistsFormula.construct(
          vars,
          EqualsFormula.construct(Reference.construct(ident), exp));
      boolean okay = processTCC(tccFormula); // set reference value symbol
      if (!okay) Main.internalError("error in processing tcc formula");
      tcc.prove("value is natural number", 
          exp.getPosition(), tccFormula, resultType, stateVars, stateExcs);
      return true;
    }
    
    // REAL->[a..b]     (EXISTS(x:INT):x=exp && a<=x && x<=b)
    if (etype instanceof RealType && type instanceof SubrangeType)
    {
      SubrangeType stype = (SubrangeType)type;
      Term from = stype.getFrom();
      Term to = stype.getTo();
      Collection<String> free = FreeVariables.compute(exp);
      FreeVariables.add(from, free);
      FreeVariables.add(to, free);
      Identifier ident = Identifier.constructNew("x", free);
      TypedIdentifier[] vars = 
        { TypedIdentifier.construct(ident,IntType.construct()) } ;
      Formula tccFormula =
        ExistsFormula.construct(
          vars,
          AndFormula.construct(
            AndFormula.construct(
              EqualsFormula.construct(Reference.construct(ident), exp),
              LessEqualFormula.construct(from, Reference.construct(ident))),
            LessEqualFormula.construct(Reference.construct(ident), to)));
      boolean okay = processTCC(tccFormula); // set reference value symbol
      if (!okay) Main.internalError("error in processing tcc formula");
      tcc.prove("value is bounded integer", 
          exp.getPosition(), tccFormula, resultType, stateVars, stateExcs);
      return true;   
    }
    
    // bad luck
    return false;
  }
  
  /***************************************************************************
   * Process tcc formula without generating new type-checking conditions.
   * @param formula tcc formula to be processed.
   * @return true iff everything went fine.
   **************************************************************************/
  private boolean processTCC(Formula formula)
  {
    tcc.open();
    boolean okay = process(formula);
    tcc.abort();
    return okay;
  }
  
  /****************************************************************************
   * Returns true if truth value of formula has denoted type;
   * may generate a type-checking condition.
   * @param formula the formula.
   * @param type the expected type for the truth value of the formula.
   * @param gen true iff a type-checking condition may be generated.
   * @return true iff truth value of formula has denoted type.
   ***************************************************************************/
  private boolean hasType(Formula formula, Type type, boolean gen)
  {
    if (type instanceof BooleanType) return true;
    if (!gen) return false;
    if (!(type instanceof SubType)) return false;
    SubType stype = (SubType)type;
    Type base = stype.getBase();
    if (!(base instanceof BooleanType)) return false;
    Term predicate = stype.getPredicate();
    Formula tccFormula =
      IfThenElseFormula.construct(formula,
          AtomicFormula.construct(predicate, new Term[] { True.construct() }),
          AtomicFormula.construct(predicate, new Term[] { False.construct() }));
    tcc.prove("formula value satisfies subtype property", 
        formula.getPosition(), tccFormula, resultType, stateVars, stateExcs);
    return true;
  }
  
  /***************************************************************************
   * Checks if term has type NAT, may generate type checking condition.
   * @param term the term (may be null, if gen is false).
   * @param type its type determined by the type checker.
   * @param gen true if type checking condition may be generated.
   * @return true iff term has type NAT.
   **************************************************************************/
  public boolean hasNatType(Term term, Type type, boolean gen)
  {
    return hasType(term, type, LogicTypeTable.natType, gen);
  }
  
  /***************************************************************************
   * Checks if term has type INT, may generate type checking condition.
   * @param term the term (may be null, if gen is false).
   * @param type its type determined by the type checker.
   * @param gen true if type checking condition may be generated.
   * @return true iff term has type INT.
   **************************************************************************/
  public boolean hasIntType(Term term, Type type, boolean gen)
  {
    return hasType(term, type, LogicTypeTable.intType, gen);
  }
  
  /***************************************************************************
   * Checks if term has type REAL, may generate type checking condition.
   * @param term the term (may be null, if gen is false).
   * @param type its type determined by the type checker.
   * @param gen true if type checking condition may be generated.
   * @return true iff term has type REAL.
   **************************************************************************/
  private boolean hasRealType(Term term, Type type, boolean gen)
  {
    return hasType(term, type, LogicTypeTable.realType, gen);
  }
  
  /****************************************************************************
   * Returns type as a function type.
   * @param type a type that should denote a function type.
   * @return the corresponding function type or null, if type is not a function
   * type or not a subtype of a function type.
   ***************************************************************************/
  private FunctionType getFunctionType(Type type)
  {
    if (type instanceof FunctionType) return (FunctionType)type;
    if (!(type instanceof SubType)) return null;
    SubType stype = (SubType)type;
    Type base = stype.getBase();
    if (!(base instanceof FunctionType)) return null;
    return (FunctionType)base;
  }
  
  /****************************************************************************
   * Returns type as a state type.
   * @param type a type that should denote a state type.
   * @return the corresponding state type or null, if type is not a state
   * type or not a subtype of a state type.
   ***************************************************************************/
  private StateType getStateType(Type type)
  {
    if (type instanceof StateType) return (StateType)type;
    if (!(type instanceof SubType)) return null;
    SubType stype = (SubType)type;
    Type base = stype.getBase();
    if (!(base instanceof StateType)) return null;
    return (StateType)base;
  }
  
  /***************************************************************************
   * Returns least upper bound of type0 and type1 (null, if none exists);
   * does not generate type-checking conditions..
   * @param type0 a type
   * @param type1 a type
   * @return the least upper bound of type0 and type1 (or null, if none exists)
   *************************************************************************/
  private Type superType(Type type0, Type type1)
  {
   if (type0 == type1) return type0;
    
    if (type0 instanceof SubType)
    {
      SubType stype = (SubType)type0;
      return superType(stype.getBase(), type1);
    }
    if (type1 instanceof SubType)
    {
      SubType stype = (SubType)type1;
      return superType(type0, stype.getBase());
    }
    
    if ((type0 instanceof SubrangeType || type0 instanceof NatType) &&
         (type1 instanceof IntType || type1 instanceof RealType)) return type1;
    if ((type1 instanceof SubrangeType || type1 instanceof NatType) &&
        (type0 instanceof IntType || type0 instanceof RealType)) return type0;
    
    if ((type0 instanceof SubrangeType || type0 instanceof NatType) && 
        (type1 instanceof SubrangeType || type1 instanceof NatType))
      return LogicTypeTable.intType;
    
    if (type0 instanceof IntType && type1 instanceof RealType) return type1;
    if (type1 instanceof IntType && type0 instanceof RealType) return type0;
    
    
    if (type0 instanceof FunctionType && type1 instanceof FunctionType)
    {
      FunctionType ftype0 = (FunctionType)type0;
      FunctionType ftype1 = (FunctionType)type1;
      Type[] domain0 = ftype0.getDomain();
      Type range0 = ftype0.getRange();
      Type[] domain1 = ftype1.getDomain();
      Type range1 =ftype1.getRange();
      int n = domain0.length;
      if (n != domain1.length) return null;
      Type[] domain = new Type[n];
      for (int i=0; i<n; i++)
      {
        domain[i] = subType(domain0[i], domain1[i]);
        if (domain[i] == null) return null;
      }
      Type range = superType(range0, range1);
      if (range == null) return null;
      return ttable.process(FunctionType.construct(domain, range));
    }
    
    if (type0 instanceof ArrayType && type1 instanceof ArrayType)
    {
      ArrayType ftype0 = (ArrayType)type0;
      ArrayType ftype1 = (ArrayType)type1;
      Type domain0 = ftype0.getIndex();
      Type range0 = ftype0.getBase();
      Type domain1 = ftype1.getIndex();
      Type range1 =ftype1.getBase();
      Type domain = subType(domain0, domain1);
      if (domain == null) return null;
      Type range = superType(range0, range1);
      if (range == null) return null;
      return ttable.process(ArrayType.construct(domain, range));
    }
    
    if (type0 instanceof TupleType && type1 instanceof TupleType)
    {
      Type[] base0 = ((TupleType)type0).getBase();
      Type[] base1 = ((TupleType)type1).getBase();
      int n = base0.length;
      if (n != base1.length) return null;
      Type[] base = new Type[n];
      for (int i=0; i<n; i++)
      {
        base[i] = superType(base0[i], base1[i]);
        if (base[i] == null) return null;
      }
      return ttable.process(TupleType.construct(base));
    }
    
    if (type0 instanceof RecordType && type1 instanceof RecordType)
    {
      TypedIdentifier[] base0 = ((RecordType)type0).getBase();
      TypedIdentifier[] base1 = ((RecordType)type1).getBase();
      int n = base0.length;
      if (n != base1.length) return null;
      TypedIdentifier[] base = new TypedIdentifier[n];
      for (int i=0; i<n; i++)
      {
        String name = base0[i].getName().getString();
        if (!name.equals(base1[i].getName().getString()))
          return null;
        Type type = superType(base0[i].getType(), base1[i].getType());
        if (type == null) return null;
        base[i] = TypedIdentifier.construct(Identifier.construct(name), type);
      }
      return ttable.process(RecordType.construct(base));
    }
    
    return null;
  }
  
  /***************************************************************************
   * Returns greatest lower bound of type0 and type1 (null, if none exists);
   * does not generate tccs.
   * @param type0 a type
   * @param type1 a type
   * @return the greatest lower bound of type0 and type1 (null, if none exists)
   *************************************************************************/
  public Type subType(Type type0, Type type1)
  {
    if (type0 == type1) return type0;
    
    if (type0 instanceof SubType && hasType(null, type0, type1, false))
      return type0;
    if (type1 instanceof SubType && hasType(null, type1, type0, false))
      return type1;
    
    if ((type0 instanceof SubrangeType || type0 instanceof NatType) &&
         (type1 instanceof IntType || type1 instanceof RealType)) return type0;
    if ((type1 instanceof SubrangeType || type1 instanceof NatType) &&
        (type0 instanceof IntType || type0 instanceof RealType)) return type1;
    
    if (type0 instanceof IntType && type1 instanceof RealType) return type0;
    if (type1 instanceof IntType && type0 instanceof RealType) return type1;
    
    if (type0 instanceof FunctionType && type1 instanceof FunctionType)
    {
      FunctionType ftype0 = (FunctionType)type0;
      FunctionType ftype1 = (FunctionType)type1;
      Type[] domain0 = ftype0.getDomain();
      Type range0 = ftype0.getRange();
      Type[] domain1 = ftype1.getDomain();
      Type range1 =ftype1.getRange();
      int n = domain0.length;
      if (n != domain1.length) return null;
      Type[] domain = new Type[n];
      for (int i=0; i<n; i++)
      {
        domain[i] = superType(domain0[i], domain1[i]);
        if (domain[i] == null) return null;
      }
      Type range = subType(range0, range1);
      if (range == null) return null;
      return ttable.process(FunctionType.construct(domain, range));
    }
    
    if (type0 instanceof ArrayType && type1 instanceof ArrayType)
    {
      ArrayType ftype0 = (ArrayType)type0;
      ArrayType ftype1 = (ArrayType)type1;
      Type domain0 = ftype0.getIndex();
      Type range0 = ftype0.getBase();
      Type domain1 = ftype1.getIndex();
      Type range1 =ftype1.getBase();
      Type domain = superType(domain0, domain1);
      if (domain == null) return null;
      Type range = subType(range0, range1);
      if (range == null) return null;
      return ttable.process(ArrayType.construct(domain, range));
    }
    
    if (type0 instanceof TupleType && type1 instanceof TupleType)
    {
      Type[] base0 = ((TupleType)type0).getBase();
      Type[] base1 = ((TupleType)type1).getBase();
      int n = base0.length;
      if (n != base1.length) return null;
      Type[] base = new Type[n];
      for (int i=0; i<n; i++)
      {
        base[i] = subType(base0[i], base1[i]);
        if (base[i] == null) return null;
      }
      return ttable.process(TupleType.construct(base));
    }
    
    if (type0 instanceof RecordType && type1 instanceof RecordType)
    {
      TypedIdentifier[] base0 = ((RecordType)type0).getBase();
      TypedIdentifier[] base1 = ((RecordType)type1).getBase();
      int n = base0.length;
      if (n != base1.length) return null;
      TypedIdentifier[] base = new TypedIdentifier[n];
      for (int i=0; i<n; i++)
      {
        String name = base0[i].getName().getString();
        if (!name.equals(base1[i].getName().getString()))
          return null;
        Type type = subType(base0[i].getType(), base1[i].getType());
        if (type == null) return null;
        base[i] = TypedIdentifier.construct(Identifier.construct(name), type);
      }
      return ttable.process(RecordType.construct(base));
    }
    
    return null;
  }
  
  /***************************************************************************
   * Get core type of a type (stripping references and subtype declarations)
   * @param type the basic type
   * @return its core type
   **************************************************************************/
  private static Type getCoreType(Type type)
  {
    if (type instanceof Reference)
    {
      Reference type0 = (Reference)type;
      TypeSymbol symbol = (TypeSymbol)type0.getName().getSymbol();
      return getCoreType(symbol.getType());
    }
    if (type instanceof SubType)
    {
      SubType type0 = (SubType)type;
      FunctionType ftype = (FunctionType)type0.getPredicate().getType();
      Type atype = ftype.getDomain()[0];
      return getCoreType(atype);
    }
    if (type == null) Main.internalError("null type");
    return type;
  }
  
  /***************************************************************************
   * Determine whether type is (a subtype of) the integer type.
   * @param type the type.
   * @return true if the type is an integer type.
   *************************************************************************/
  public static boolean isIntegerType(Type type)
  {
    type = getCoreType(type);
    if (type instanceof IntType) return true;
    if (type instanceof NatType) return true;
    if (type instanceof SubrangeType) return true;
    return false;
  }
  
  /****************************************************************************
   * Check whether type denotes a boolean.
   * @param type the type.
   * @return true if the type denotes a boolean.
   ***************************************************************************/
  public static boolean isBooleanType(Type type)
  {
    type = LogicChecking.getCoreType(type);
    return type instanceof BooleanType;
  }
  
  /***************************************************************************
   * Checks if type is (a subtype of) BOOLEAN (no tcc is generated).
   * @param type the type to be checked.
   * @return true iff type is (a subtype of) BOOLEAN.
   ***************************************************************************
  private boolean isBooleanType(Type type)
  {
    return hasType(null, type, LogicTypeTable.booleanType, false);
  }
  */
}
