// ----------------------------------------------------------------------------
// LogicTypeTable.java
// Make logic type expressions unique such that they can be compared by reference.
// $Id: LogicTypeTable.java,v 1.29 2012/10/18 16:12:58 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2009, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Semantics;

import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;

/**
 * 
 * Make logic type expressions unique such that they can be compared by reference.
 *
 */
public class LogicTypeTable
{
  // the stream to which to write error messages
  private ErrorStream error;
  
  // the type of the current state (null, if none)
  private StateType stateType;
  
  // the assignable variables of the current state (null, if none)
  private Name[] stateVars;
  
  // the throwable exceptions of the current state (null, if none)
  private Name[] stateExcs;
  
  // tables of types encountered so far
  // as far as possible, equivalent types do not appear twice in the table
  private Collection<AtomicType>   ptable; // primitive (atomic) types
  private Collection<FunctionType> ftable; // function types
  private Collection<TupleType>    ttable; // tuple types
  private Collection<RecordType>   rtable; // record types
  private Collection<ArrayType>    atable; // array types
  private Collection<StateType>    xtable; // state types
  private Collection<SubType>      stable; // subtypes
  private Collection<SubrangeType> utable; // subrange types
  
  // derived tables
  private SetClass<Type[]> seqtable;             // type sequences
  private SetClass<TypedIdentifier> tidtable;    // typed identifiers
  private SetClass<TypedIdentifier[]> tidstable; // typed identifier seqs
  
  // types used in type checking
  public static final BooleanType booleanType = BooleanType.construct();
  public static final BitType bitType = BitType.construct();
  public static final NatType natType = NatType.construct();
  public static final IntType intType = IntType.construct();
  public static final RealType realType = RealType.construct();
  public static final StringType stringType = StringType.construct();
  
  /****************************************************************************
   * Construct a table for making types unique.
   * @param error the stream on which to print errors.
   ***************************************************************************/
  public LogicTypeTable(ErrorStream error)
  {
    this.error = error;
    this.stateType = null;
    
    // canonical type tables
    ptable = new SetClass<AtomicType>();   // primitive (atomic) types
    ftable = new SetClass<FunctionType>(); // function types
    ttable = new SetClass<TupleType>();    // tuple types
    rtable = new SetClass<RecordType>();   // record types
    atable = new SetClass<ArrayType>();    // array types
    xtable = new SetClass<StateType>();    // state types
    stable = new SetClass<SubType>();      // subtypes
    utable = new SetClass<SubrangeType>(); // subrange types
    
    // derived tables
    seqtable = new SetClass<Type[]>();             // type sequences
    tidtable = new SetClass<TypedIdentifier>();    // typed identifiers
    tidstable = new SetClass<TypedIdentifier[]>(); // typed identifiers
    
    // types used in type checking
    ptable.add(booleanType); 
    ptable.add(bitType); 
    ptable.add(natType);
    ptable.add(intType);
    ptable.add(realType);
    ptable.add(stringType);
  }
  
  // to be used for type-checking terms
  private LogicChecking checker;
  private LogicEnvironment env;
  private TCCGenerator tcc;
  
  /****************************************************************************
   * Set checker to be used for type-checking terms in types;
   * must be called before process() is called.
   * @param checker the checker for type-checking terms in types
   *        (null means that no checker is in use)
   * @return old the old checker (null, if none)
   ***************************************************************************/
  public LogicChecking setChecker(LogicChecking checker)
  {
    LogicChecking old = this.checker;
    if (checker == null)
    {
      this.checker = null;
      this.env = null;
      this.tcc = null;
    }
    else
    {
      this.checker = checker;
      this.env = checker.getLogicEnvironment();
      this.tcc = checker.getTCCGenerator();
    }
    return old;
  }
  
  /****************************************************************************
   * Set the current state context.
   * @param stateType the current state type.
   * @param stateVars the assignable variables.
   * @param stateExcs the throwable exceptions.
   ***************************************************************************/
  public void setStateContext(StateType stateType, 
    Name[] stateVars, Name[] stateExcs)
  {
    this.stateType = stateType;
    this.stateVars = stateVars;
    this.stateExcs = stateExcs;
  }
  
  /****************************************************************************
   * Reset the current state context.
   ***************************************************************************/
  public void resetStateContext()
  {
    this.stateType = null;
    this.stateVars = null;
    this.stateExcs = null;
  }
  
  // --------------------------------------------------------------------------
  //
  // Construct semi-canonical types
  // 
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Return (if possible) canonical version of the type
   * @param type the type to be checked
   * @return a type equivalent to the input type or null, if an error occurred
   ***************************************************************************/
  public Type process(Type type)
  {
    if (checker == null) Main.internalError("setChecker() not called");
    if (type instanceof Reference) return process((Reference)type);
    if (type instanceof AtomicType) return process((AtomicType)type);
    if (type instanceof FunctionType) return process((FunctionType)type);
    if (type instanceof TupleType) return process((TupleType)type);
    if (type instanceof RecordType) return process((RecordType)type);
    if (type instanceof ArrayType) return process((ArrayType)type);
    if (type instanceof StateType) return process((StateType)type);
    if (type instanceof SubType) return process((SubType)type);
    if (type instanceof SubrangeType) return process((SubrangeType)type);
    Main.internalError("unknown type: " + type);
    return null;
  }
  
  private Type[] process(Type[] type)
  {
    if (seqtable.contains(type)) return type;
    int n = type.length;
    Type[] result = new Type[n];
    for (int i=0; i<n; i++)
    {
      result[i] = process(type[i]);
      if (result[i] == null) return null;
    }
    Iterator<Type[]> iterator = seqtable.iterator();
    L: while (iterator.hasNext())
    {
      Type[] stype = iterator.next();
      if (stype.length != n) continue;
      for (int i=0; i<n; i++)
      {
        if (result[i] != stype[i]) continue L;
      }
      return stype;
    }
    seqtable.add(result);
    return result;
  }
  
  private TypedIdentifier process(TypedIdentifier tid)
  {
    if (tidtable.contains(tid)) return tid;
    // identifier strings are unique by construction
    Identifier ident = tid.getName();
    String name = ident.getString();
    Type type = process(tid.getType());
    if (type == null) return null;
    Iterator<TypedIdentifier> iterator = tidtable.iterator();
    while (iterator.hasNext())
    {
      TypedIdentifier utid = iterator.next();
      if (utid.getName().getString() == name && utid.getType() == type)
        return utid;
    }
    TypedIdentifier result = TypedIdentifier.construct(ident, type);
    tidtable.add(result);
    ValueSymbol symbol = ValueSymbol.construct(ident, type, result);
    ident.setSymbol(symbol);
    return result;
  }
  
  private TypedIdentifier[] process(TypedIdentifier[] tids)
  {
    if (tidstable.contains(tids)) return tids;
    int n = tids.length;
    TypedIdentifier[] result = new TypedIdentifier[n];
    Collection<String> names = new SetClass<String>();
    for (int i=0; i<n; i++)
    {
      // identifier names are unique by construction
      TypedIdentifier tid = tids[i];
      Identifier id = Identifier.construct(tid.getName().getString());
      String name = id.getString();
      if (names.contains(name))
      {
        error.newError(tids[i],
            "record field " + name + " is multiply declared");
        return null;
      }
      result[i] = process(tids[i]);
      if (result[i] == null) return null;
      ValueSymbol symbol = ValueSymbol.construct(id, result[i].getType(), tid);
      id.setSymbol(symbol);
    }
    Iterator<TypedIdentifier[]> iterator = tidstable.iterator();
    L: while (iterator.hasNext())
    {
      TypedIdentifier[] utids = iterator.next();
      if (utids.length != n) continue;
      for (int i=0; i<n; i++)
      {
        if (utids[i] != result[i]) continue L;
      }
      return utids;
    }
    tidstable.add(result);
    return result;
  }
  
  private Type process(Reference ref)
  {
    TypeSymbol symbol = env.getTypeSymbol(ref);
    if (symbol == null) return null;
    return symbol.getType();
  }
  
  private AtomicType process(AtomicType type)
  {
    if (ptable.contains(type)) return type;
    // atomic types are identical if their names are identical
    Iterator<AtomicType> iterator = ptable.iterator();
    while (iterator.hasNext())
    {
      AtomicType ptype = iterator.next();
      if (ptype.getName().equals(type.getName()))
        return ptype;
    }
    ptable.add(type);
    return type;
  }
  
  private FunctionType process(FunctionType type)
  {
    if (ftable.contains(type)) return type;
    Type[] domain = type.getDomain();
    Type[] udomain = process(domain);
    if (udomain == null) return null;
    Type range = type.getRange();
    Type urange = process(range);
    if (urange == null) return null;
    Iterator<FunctionType> iterator = ftable.iterator();
    while (iterator.hasNext())
    {
      FunctionType ftype = iterator.next();
      if (ftype.getDomain() == udomain && ftype.getRange() == urange)
        return ftype;
    }
    FunctionType result = FunctionType.construct(udomain, urange);
    ftable.add(result);
    return result;
  }
  
  private TupleType process(TupleType type)
  {
    if (ttable.contains(type)) return type;
    Type[] base = type.getBase();
    Type[] ubase = process(base);
    if (ubase == null) return null;
    Iterator<TupleType> iterator = ttable.iterator();
    while (iterator.hasNext())
    {
      TupleType ttype = iterator.next();
      if (ttype.getBase() == ubase) return ttype;
    }
    TupleType result = TupleType.construct(ubase);
    ttable.add(result);
    return result;
  }
  
  private RecordType process(RecordType type)
  {
    if (rtable.contains(type)) return type;
    TypedIdentifier[] base = type.getBase();
    TypedIdentifier[] ubase = process(base);
    if (ubase == null) return null;
    Iterator<RecordType> iterator = rtable.iterator();
    while (iterator.hasNext())
    {
      RecordType ttype = iterator.next();
      if (ttype.getBase() == ubase) return ttype;
    }
    RecordType result = RecordType.construct(ubase);
    rtable.add(result);
    return result;
  }
  
  private ArrayType process(ArrayType type)
  {
    if (atable.contains(type)) return type;
    Type index = type.getIndex();
    Type uindex = process(index);
    if (uindex == null) return null;
    Type base = type.getBase();
    Type ubase = process(base);
    if (ubase == null) return null;
    Iterator<ArrayType> iterator = atable.iterator();
    while (iterator.hasNext())
    {
      ArrayType atype = iterator.next();
      if (atype.getIndex() == uindex && atype.getBase() == ubase)
        return atype;
    }
    ArrayType result = ArrayType.construct(uindex, ubase);
    atable.add(result);
    return result;
  }
  
  private StateType process(StateType type)
  {
    if (xtable.contains(type)) return type;
    Type ret = type.getReturn();
    Type uret = null;
    if (ret != null)
    {
      uret = process(ret);
      if (uret == null) return null;
    }
    Iterator<StateType> iterator = xtable.iterator();
    while (iterator.hasNext())
    {
      StateType stype = iterator.next();
      if (stype.getReturn() == uret)
        return stype;
    }
    StateType result = StateType.construct(uret);
    xtable.add(result);
    return result;
  }
  
  private SubrangeType process(SubrangeType type)
  {
    if (utable.contains(type)) return type;
    Term from = type.getFrom();
    Term to = type.getTo();
    Type ftype = checker.process(from);
    if (ftype == null) return null;
    if (!checker.hasIntType(from, ftype, true))
    {
      error.newError(from, "lower bound " + from +
        " of subrange type has type " + ftype + " (INT expected)");
      return null;
    }
    Type ttype = checker.process(to);
    if (ttype == null) return null;
    if (!checker.hasIntType(to, ttype, true))
    {
      error.newError(to, "upper bound " + to +
        " of subrange type has type " + ttype + " (INT expected)");
      return null;
    }
    return processSubrangeType(type);
  }
  
  /***************************************************************************
   * Process subrange type assuming that expressions in type have been processed.
   * @param type the subrange type with processed expressions.
   * @return the processed version of the type.
   **************************************************************************/
  private SubrangeType processSubrangeType(SubrangeType type)
  {
    Term from = type.getFrom();
    Term to = type.getTo();
    
    // we share subrange types only if they refer to same constants/literals
    for (SubrangeType utype : utable)
    {
      if (!equals(from, utype.getFrom())) continue;
      if (!equals(to, utype.getTo())) continue;
      return utype;
    }
    utable.add(type);

    // subrange type must not be empty
    Formula f = LessEqualFormula.construct(from, to);
    boolean okay = checker.process(f);
    if (!okay) Main.internalError("invalid interval condition");
    tcc.prove("Interval " + type + " is not empty", 
        type.getPosition(), f, stateType, stateVars, stateExcs);
    return type;
  }   
  
  /***************************************************************************
   * Semi-decide whether two terms denote the same value.
   * @param term1 the first term.
   * @param term2 the second term.
   * @return true if it can be guaranteed that both terms denote the same value.
   **************************************************************************/
  private boolean equals(Term term1, Term term2)
  {
    if (term1 instanceof IntLiteral)
    {
      if (!(term2 instanceof IntLiteral)) return false;
      IntLiteral t1 = (IntLiteral)term1;
      IntLiteral t2 = (IntLiteral)term2;
      return (t1.getName().equals(t2.getName()));
    }
    if (term1 instanceof Reference)
    {
      if (!(term2 instanceof Reference)) return false;
      Reference t1 = (Reference)term1;
      Reference t2 = (Reference)term2;
      return (t1.getName().getSymbol() == t2.getName().getSymbol());
    }
    return false;
  }
  
  private SubType process(SubType type)
  {
    if (stable.contains(type)) return type;
    Term predicate = type.getPredicate();
    Type ptype = checker.process(predicate);
    if (ptype == null) return null;
    if (!(ptype instanceof FunctionType))
    {
      error.newError(predicate, "subtype predicate " + predicate + 
          " has type " + ptype + 
          " which is not a function type");
      return null;
    }
    FunctionType ftype = (FunctionType)ptype;
    Type range = ftype.getRange();
    if (!(range instanceof BooleanType))
    {
      error.newError(predicate, "subtype predicate " + predicate + 
          " has type " + ftype + 
          (Main.toLogical() ?
              " whose range is not LOGICAL" :
              " whose range is not BOOLEAN"));
      return null;
    }
    Type[] domain = ftype.getDomain();
    if (domain.length != 1)
    {
       error.newError(predicate, "subtype predicate " + predicate + 
         " has type " + ftype + 
         " with " + domain.length + " parameters (1 expected)");
       return null;
    }
    
    // set (non-subtype) base type for use by type checker
    Type base = process(domain[0]);
    if (base instanceof SubType) base = ((SubType)base).getBase();
    type.setBase(base);
    
    // type checking condition: type must not be empty
    Collection<String> free = FreeVariables.compute(predicate);
    Identifier ident = Identifier.constructNew("x", free);
    Formula tccFormula = (Formula)ASTCloner.clone(
      ExistsFormula.construct(
        new TypedIdentifier[] { 
            TypedIdentifier.construct(ident , domain[0]) },
        AtomicFormula.construct(
          predicate, 
          new Term[] { Reference.construct(Name.construct(ident)) })
      ));
    boolean okay = checker.process(tccFormula);
    if (!okay) Main.internalError("invalid tcc " + tccFormula);
    tcc.prove("subtype is not empty", type.getPosition(), 
        tccFormula, stateType, stateVars, stateExcs);
    
    // subrange types are not shared
    stable.add(type);
    return type;
  }
}
