// ----------------------------------------------------------------------------
// ParamSymbolTableBase.java
// A table of parameterized symbols.
// $Id: ParamSymbolTable.java,v 1.12 2017/07/04 08:04:37 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Semantics;

import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Program.TypeExpression;

/**
 * 
 * A table of parameterized symbols.
 *
 */
public final class ParamSymbolTable<S extends ParamSymbol> 
{
  // our collection of symbol sets
  private final MapClass<String,ParamSymbols<S>> table = 
    new MapClass<String,ParamSymbols<S>>();
  
  // the stream to print messages on
  private ErrorStream error;
  
  /****************************************************************************
   * Construct symbol table with denoted error stream.
   * @param error the error stream.
   ***************************************************************************/
  public ParamSymbolTable(ErrorStream error)
  {
    this.error = error;
  }
  
  /****************************************************************************
   * Return collection of symbols in table.
   * @return the collection of all symbols.
   ***************************************************************************/
  public Collection<S> getSymbols()
  {
    Collection<ParamSymbols<S>> sets = table.values();
    Collection<S> result = new SetClass<S>();
    for (ParamSymbols<S> set : sets)
      result.addAll(set.getSymbols());
    return result;
  }
  
  /****************************************************************************
   * Puts symbol into table to be looked up for the symbol's 
   * name and argument types.
   * @param  value the symbol.
   * @return true, if symbol could be added; 
   *         false, if symbol with same name and argument types was in table
   *                (in this case the symbol is not added and a message is printed)
   ***************************************************************************/
  public boolean put(S value)
  {
    Identifier ikey = value.getName();
    TypeExpression[] tkey = value.getArgumentTypes();
    ParamSymbols<S> symbols = table.get(ikey.getString());
    if (symbols == null)
    {
      symbols = new ParamSymbols<S>(error);
      table.put(ikey.getString(), symbols);
    }
    else if (symbols.get(tkey) != null) 
    {
      StringBuffer b = new StringBuffer();
      b.append("symbol " + ikey + "(");
      int n = tkey.length;
      if (n > 0)
      {
        for (int i=0; i<n-1; i++)
          b.append(tkey[i] + ", ");
        b.append(tkey[n-1]);
      }
      b.append(") is already defined in current environment");
      error.newError(b.toString());
      return false;
    }
    symbols.put(value);
    return true;
  }
  
  /****************************************************************************
   * Returns symbol associated to key consisting of an identifier and
   * a sequence of types and links key to symbol.
   * @param ikey the identifier part of the key.
   * @param tkey the type sequence part of the key.
   * @return the associated symbol or null if the symbol is not there.
   ***************************************************************************/
  public S get(Identifier ikey, TypeExpression[] tkey)
  {
    ParamSymbols<S> symbols = table.get(ikey.getString());
    if (symbols == null) return null;
    S symbol = symbols.get(tkey);
    if (symbol != null) ikey.setSymbol(symbol);
    return symbol;
  }
}