// ----------------------------------------------------------------------------
// UpdateTerm.java
// A term from which a component is updated.
// $Id: UpdateTerm.java,v 1.7 2011/09/30 15:47:33 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2009, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the Formulas of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Syntax.Logic;

import fmrisc.ProgramExplorer.Syntax.*;

/**
 * 
 * A term from which a component is updated.
 *
 */
public final class UpdateTerm extends TermBase
{
  private static final int precedence = 60; // see SpecLang.g
  public final static String withKeyword = "WITH";
  public final static String assignToken = ":=";
  
  private final Term base;
  private final Selector[] selectors;
  private final Term value;
  
  /****************************************************************************
   * Construct selection term.
   * @param base the base term.
   * @param selectors the sequence of selectors.
   * @param value the update value.
   * @return the term.
   ***************************************************************************/
  public static UpdateTerm construct(Term base, Selector[] selectors, Term value)
  {
    return new UpdateTerm(base, selectors, value);
  }
  
  /****************************************************************************
   * Construct selection term.
   * @param base the base term.
   * @param selectors the sequence of selectors.
   * @param value the update value.
   ***************************************************************************/
  private UpdateTerm(Term base, Selector[] selectors,  Term value)
  {
    this.base = base;
    this.selectors = selectors;
    this.value = value;
  }
  
  /****************************************************************************
   * Get base term.
   * @return the base term.
   ***************************************************************************/
  public Term getBase()
  {
    return base;
  }
  
  /****************************************************************************
   * Get selectors.
   * @return the selector.
   ***************************************************************************/
  public Selector[] getSelectors()
  {
    return selectors;
  }
  
  /****************************************************************************
   * Get update value.
   * @return the update value.
   ***************************************************************************/
  public Term getValue()
  {
    return value;
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitorOld visitor)
  {
    boolean done = visitor.visit(this);
    if (done) return;
    super.accept(visitor);
    base.accept(visitor);
    accept(selectors, visitor);
    value.accept(visitor);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitor visitor)
  {
    visitor.visit(this);
  }
  
  /***************************************************************************
   * Accept visitor for a visit to the children of this node.
   * @param visitor the visitor who is accepted by the children of this node.
   **************************************************************************/
  public void acceptChildren(ASTVisitor visitor)
  {
    super.acceptChildren(visitor);
    base.accept(visitor);
    accept(selectors, visitor);
    value.accept(visitor);
  }
  
  /***************************************************************************
   * Get precedence value of the expression operator 
   * (higher precedence value means higher binding power)
   **************************************************************************/
  public int getPrecedence()
  {
    return precedence;
  }
}
