// ----------------------------------------------------------------------------
// Identifier.java
// A name i.e. a structured identifier.
// $Id: Name.java,v 1.19 2017/07/04 08:04:37 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Syntax;

import java.util.*;
import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Semantics.*;

/**
 * 
 * A name i.e. a structured identifier.
 * 
 */
public final class Name extends ASTBase
{
  public final Identifier[] idents;
  
  /***************************************************************************
   * Construct name.
   * @param prefix the prefix of the name.
   * @param ident the base identifier of the name.
   * @return the name.
   **************************************************************************/
  public static Name construct(Name prefix, Identifier ident)
  {
    int n = prefix.idents.length;
    Identifier idents[] = new Identifier[n+1];
    System.arraycopy(prefix.idents, 0, idents, 0, n);
    idents[n] = ident;
    return construct(idents);
  }
  
  /***************************************************************************
   * Construct name.
   * @param idents the sequence of identifiers of which the name is composed.
   * @return the name.
   **************************************************************************/
  public static Name construct(Vector<Identifier> idents)
  {
    return construct(idents.toArray(new Identifier[idents.size()]));
  }
  
  /***************************************************************************
   * Construct name.
   * @param ident a single identifier making up the name
   * @return the name.
   **************************************************************************/
  public static Name construct(Identifier ident)
  {
    return construct(new Identifier[]{ ident });
  }
  
  /***************************************************************************
   * Construct name.
   * @param idents the sequence of identifiers of which the name is composed.
   * @return the name.
   **************************************************************************/
  public static Name construct(Identifier[] idents)
  {
    return new Name(idents);
  }
  
  /***************************************************************************
   * Get the length of the name.
   * @return the number of identifiers in the name (greater equal 1)
   **************************************************************************/
  public int getLength()
  {
    return idents.length;
  }
  
  /***************************************************************************
   * Get the base identifier of the name.
   * @return the last identifier in the identifier sequence of the name.
   **************************************************************************/
  public Identifier getBase()
  {
    int n = idents.length;
    return idents[n-1];
  }
  
  /***************************************************************************
   * Get the prefix of the name.
   * @return the name without the base identifier.
   **************************************************************************/
  public Name getPrefix()
  {
    int n=idents.length-1;
    if (n < 0) Main.internalError("empty identifier encountered");
    Identifier[] prefix = new Identifier[n];
    for (int i=0; i<n; i++) prefix[i] = idents[i];
    return construct(prefix);
  }
  
  /***************************************************************************
   * Get the postfix of the name.
   * @return the name without the first identifier.
   **************************************************************************/
  public Name getPostfix()
  {
    int n=idents.length-1;
    if (n < 0) Main.internalError("empty identifier encountered");
    Identifier[] postfix = new Identifier[n];
    for (int i=0; i<n; i++) postfix[i] = idents[i+1];
    return construct(postfix);
  }
  
  /***************************************************************************
   * Get prefix of first n identifiers of name.
   * @param n the number of identifiers.
   * @return the prefix with n identifiers.
   **************************************************************************/
  public Name getPrefix(int n)
  {
    Identifier[] idents = new Identifier[n];
    for (int i=0; i<n; i++)
      idents[i] = this.idents[i];
    return construct(idents);
  }
  
  /***************************************************************************
   * Construct name.
   * @param idents the sequence of identifiers of which the name is composed.
   **************************************************************************/
  private Name(Identifier[] idents)
  {
    this.idents = idents;
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitorOld visitor)
  {
    boolean done = visitor.visit(this);
    if (done) return;
    super.accept(visitor);
    accept(idents, visitor);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitor visitor)
  {
    visitor.visit(this);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void acceptChildren(ASTVisitor visitor)
  {
    super.acceptChildren(visitor);
    accept(idents, visitor);
  }
  
  /***************************************************************************
   * Compare names for equality.
   * @param name the name with which this name is compared.
   * @return true iff both names are equal
   **************************************************************************/
  public boolean equals(Name name)
  {
    // use string comparison because unnamed (top-level) package
    // might appear in either identifier
    return this.toString().equals(name.toString());
  }
  
  /***************************************************************************
   * Get symbol identified by (last part of) name.
   * @return the symbol.
   **************************************************************************/
  public Symbol getSymbol()
  {
    return getBase().getSymbol();
  }
}

