// ----------------------------------------------------------------------------
// ConditionalStatement.java
// A conditional statement with one or two branches.
// $Id: ConditionalStatement.java,v 1.6 2017/07/04 08:04:37 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Syntax.Program;

import fmrisc.ProgramExplorer.Syntax.*;

/**
 * A conditional statement with one or two branches.
 */
public final class ConditionalStatement extends StatementBase
{
  public final ValueExpression cond;
  public final Statement thenstat;
  public final Statement elsestat; // may be null
  
  /***************************************************************************
   * Construct a conditional statement.
   * @param cond the condition.
   * @param thenstat the "then" branch
   * @param elsestat the "else" branch (may be null)
   * @return the expression.
   **************************************************************************/
  public static ConditionalStatement construct(ValueExpression cond, 
    Statement thenstat, Statement elsestat)
  {
    return new ConditionalStatement(cond, thenstat, elsestat);
  } 
  
  /***************************************************************************
   * Construct a conditional statement.
   * @param cond the condition.
   * @param thenstat the "then" branch
   * @param elsestat the "else" branch (may be null)
   **************************************************************************/
  private ConditionalStatement(ValueExpression cond, 
    Statement thenstat, Statement elsestat)
  {
    this.cond = cond;
    this.thenstat = thenstat;
    this.elsestat = elsestat;
  }    
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitorOld visitor)
  {
    boolean done = visitor.visit(this);
    if (done) return;
    super.accept(visitor);
    cond.accept(visitor);
    thenstat.accept(visitor);
    if (elsestat != null) elsestat.accept(visitor);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitor visitor)
  {
    visitor.visit(this);
  }
  
  /***************************************************************************
   * Accept visitor for a visit to the children of this node.
   * @param visitor the visitor who is accepted by the children of this node.
   **************************************************************************/
  public void acceptChildren(ASTVisitor visitor)
  {
    super.acceptChildren(visitor);
    cond.accept(visitor);
    thenstat.accept(visitor);
    if (elsestat != null) elsestat.accept(visitor);
  }
  
  /****************************************************************************
   * Get type of statement.
   * @return a string identifying the type of statement.
   ***************************************************************************/
  public String getKind()
  {
    return "conditional";
  }
}
