// ----------------------------------------------------------------------------
// ClassSpec.java
// A class specification.
// $Id: ClassSpec.java,v 1.3 2011/09/30 15:47:37 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2011-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Syntax.Spec;

import org.antlr.runtime.*;

import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Parser.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;

/**
 * 
 * A class specification.
 *
 */
public final class ClassSpec extends SpecificationBase
{
  private Formula classInvariant;
  
  public final static String invariantKeyword ="invariant";
  
  /****************************************************************************
   * Construct class specification from denoted text.
   * @param text the text of the specification.
   * @param error the stream for printing error messages
   * @param pos the source position of the specification
   * @return the class specification (null, if an error ocurred in parsing)
   ***************************************************************************/
  public static ClassSpec construct(String text, 
    ErrorStream error, SourcePosition pos)
  {
    ClassSpec spec = new ClassSpec(text);
    try
    {
      
      CharStream input = new ANTLRStringStream(text);
      SpecLangLexer lexer = new SpecLangLexer(input, error, pos);
      CommonTokenStream tokens = new CommonTokenStream(lexer);
      SpecLangParser parser = new SpecLangParser(tokens, error, pos, text);
      parser.classspec(spec);  
    }
    catch (RecognitionException e)
    {
      error.newError("ill-formed specification (" + e.getMessage() + ")");
      return null;
    }
    return spec;
  }
  
  /***************************************************************************
   * Construct loop specification from text (other fields are not set).
   * @param text the specification text.
   **************************************************************************/
  private ClassSpec(String text)
  {
    super(text);
  }

  /***************************************************************************
   * Get class invariant.
   * @return the class invariant (null, if none).
   **************************************************************************/
  public Formula getClassInvariant() { return classInvariant; }
  
  /***************************************************************************
   * Set class invariant.
   * @param inv the class invariant (may be null).
   **************************************************************************/
  public void setClassInvariant(Formula inv)
  {
    classInvariant = inv;
  }

  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitorOld visitor)
  {
    boolean done = visitor.visit(this);
    if (done) return;
    super.accept(visitor);
    if (classInvariant != null) classInvariant.accept(visitor);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitor visitor)
  {
    visitor.visit(this);
  }
  
  /***************************************************************************
   * Accept visitor for a visit to the children of this node.
   * @param visitor the visitor who is accepted by the children of this node.
   **************************************************************************/
  public void acceptChildren(ASTVisitor visitor)
  {
    super.acceptChildren(visitor);
    if (classInvariant != null) classInvariant.accept(visitor);
  }
}
