// ----------------------------------------------------------------------------
// AutoStrategy.java
// A strategy that attempts automatic before interactive proving.
// $Id: AutoStrategy.java,v 1.12 2011/04/13 16:32:28 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.ewuni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Tasks;

import fmrisc.ProgramExplorer.Proving.*;

/**
 * 
 * A strategy that attempts automatic before interactive proving.
 *
 */
public class AutoStrategy implements Strategy
{
  /****************************************************************************
   * True if this strategy is applicable to the denoted task.
   * @param task a task
   * @return true iff apply() may be invoked on task.
   ***************************************************************************/
  public boolean isApplicable(Task task)
  {
    // must be able to apply the RISC ProofNavigator
    return task instanceof ProofNavigatorTask;
  }
  
  /****************************************************************************
   * Apply this strategy to a task to which it is applicable.
   * @param task a task to which the task is applicable.
   ***************************************************************************/
  public void apply(Task task)
  {
    // task may be closed
    if (task.isFinal()) return;
    
    // generate ProofNavigatorProblem
    ProofNavigatorTask task0 = (ProofNavigatorTask)task; 
    ProofNavigatorProblem problem = task0.proofNavigatorProblem();
    
    // if task is new, attempt automatic solution
    Task.Status status = task.getStatus();
    if (status == Task.Status.NEW)
    {
      if (problem.existsProof())
      {
        fmrisc.ProofNavigator.Proving.Proof proof = problem.getProof();
        if (proof.isClosed() && proof.isAbsolute() && proof.isTrusted())  
        {
          task.setStatus(Task.Status.ALMOST,
              "skeleton for interactive proof exists", null);
          return;
        }
      }
      problem.proveAuto(task);
      if (task.isFinal()) return;
      problem.proveStrategy(task);
      if (task.isFinal()) return;
      if (problem.existsProof())
      { 
        task.setStatus(Task.Status.OPEN,
            "skeleton for interactive proof exists", null);
        return;
      }
      task.setStatus(Task.Status.OPEN,
          "interactive proof is required", null);
      return;
    }
    
    // task is not new, attempt interactive solution
    problem.proveManual(task);
  }
}
