// ----------------------------------------------------------------------------
// FlattenDeclarations.java
// Flatten nested declarations of a task.
// $Id: FlattenDeclarations.java,v 1.10 2012/10/22 15:22:30 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2009, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Tasks;

import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Syntax.Spec.TheoryDeclaration;
import fmrisc.ProgramExplorer.Semantics.*;

/**
 * 
 * Flatten nested declarations of a task.
 *
 */
public final class FlattenDeclarations
{
  /**
   * 
   * The flattening result.
   *
   */
  public static class Result
  {
    public final Declaration[] decls;     // flat sequence of declarations
    public final Formula goal;            // goal formula
    public final Map<Symbol, Symbol> map; // symbol translation
    public Result(Declaration[] decls, Formula goal, Map<Symbol, Symbol> map)
    {
      this.decls = decls;
      this.goal = goal;
      this.map = map;
    }
  }
  
  /***************************************************************************
   * Flatten a task with nested declarations.
   * @param theories the theories that are referenced in dependence order.
   * @param decls the nested declarations.
   * @param goal the goal.
   * @param theory the theory of the goal (may be null)
   * @return a pair of translated declarations and goals.
   **************************************************************************/
  public static Result flatten(Vector<TheorySymbol> theories,
    Declaration[][] decls, Formula goal, TheorySymbol theory)
  {
    FlattenDeclarations f = new FlattenDeclarations(theories, decls, goal, theory);
    Result result = f.flatten();
    return result;
  }
  
  // the flattening arguments
  private final Vector<TheorySymbol> theories;
  private final Declaration[][] decls;
  private final Formula goal;
  private final TheorySymbol theory;
  
  // mapping of symbols to unique names and of old symbols to new symbols
  private Map<Symbol,String> symbolNames;
  private Map<Symbol,Symbol> symbolSymbols;
  
  // theories that were already processed
  private Set<TheorySymbol> theorySymbols;
  
  // assumptions that were constructed
  private Vector<Formula> assumptions;
  
  private FlattenDeclarations(Vector<TheorySymbol> theories,
    Declaration[][] decls, Formula goal, TheorySymbol theory)
  {
    // the arguments
    this.theories = theories;
    this.decls = decls;
    this.goal = goal;
    this.theory = theory;
    
    // the global variables
    symbolNames = new MapClass<Symbol,String>();
    symbolSymbols = new MapClass<Symbol,Symbol>();
    theorySymbols = new SetClass<TheorySymbol>();
    assumptions = new Vector<Formula>();
  }
  
  /***************************************************************************
   * Perform the flattening.
   * @return the pair of flattened declarations and translated goal.
   **************************************************************************/
  private Result flatten()
  {
    // initialize result vector
    Vector<Declaration> result = new Vector<Declaration>();
    
    // add declarations for theories, declarations, and goal
    for (TheorySymbol tsymbol : theories)
      toDeclarations(result, tsymbol);
    for (Declaration[] d : decls)
      toDeclarations(result, d);
        
    // build the result array
    int n = result.size();
    Declaration[] decls0 = new Declaration[n];
    result.toArray(decls0);
    
    // translate the goal
    Formula goal0 = (Formula)translate(goal);
    
    // weave assumptions into goal
    n = assumptions.size();
    if (n != 0)
    {
      Formula assumption = assumptions.elementAt(0);
      for (int i=1; i<n; i++)
        assumption = AndFormula.construct(assumption, assumptions.elementAt(i));
      goal0 = ImpliesFormula.construct(assumption, goal0);
    }
    
    // construct the translation result
    return new Result(decls0, goal0, symbolSymbols);
  }
  
  /***************************************************************************
   * Create copy of expression with all references in symbols in name table
   * are replaced by new and unique names.
   * @param exp an expression.
   * @return the translated version of the expression.
   *************************************************************************/
  private Expression translate(Expression exp)
  {
    return (Expression)ReferenceCloner.clone(exp, symbolSymbols);
  }
  
  /****************************************************************************
   * Add declarations arising from denoted theory
   * @param result the result vector to which the new declarations are added. 
   * @param tsymbol the symbol denoting the theory.
   ***************************************************************************/
  private void toDeclarations(Vector<Declaration> result, TheorySymbol tsymbol)
  {
    // prevent recursion cycle
    if (theorySymbols.contains(tsymbol)) return;
    theorySymbols.add(tsymbol);
    
    // add recursively referenced theories
    TheoryDeclaration tdecl = tsymbol.getDeclaration();
    Name[] tnames = tdecl.getTheories();
    for (Name tname : tnames)
    {
      if (!(tname.getBase().getSymbol() instanceof TheorySymbol))
        Main.internalError("not a theory symbol");
      TheorySymbol tsymbol0 = (TheorySymbol)tname.getBase().getSymbol();
      toDeclarations(result, tsymbol0);
    }
    
    // add declarations from theory
    toDeclarations(result, tdecl.getDeclarations());
  }
  
  /****************************************************************************
   * Add declarations arising from denoted declarations
   * @param result the result vector to which the new declarations are added. 
   * @param decls the given declarations.
   ***************************************************************************/
  private void toDeclarations(Vector<Declaration> result, Declaration[] decls)
  {
    for (Declaration d: decls) 
      toDeclarations(result, d);
  }
  
  /****************************************************************************
   * Add declaration arising from denoted declaration
   * @param result the result vector to which the new declarations are added. 
   * @param decl the given declaration.
   ***************************************************************************/
  private void toDeclarations(Vector<Declaration> result, Declaration decl)
  {
    Symbol symbol = decl.getName().getSymbol();
    Identifier name = uniqueName(symbol);
    if (decl instanceof FormulaDefinition)
    {
      /*
      if (decl instanceof AxiomDefinition)
      {
        AxiomDefinition decl0 = (AxiomDefinition)decl;
        Formula formula = (Formula)translate(decl0.getValue());
        AxiomDefinition decl1 = AxiomDefinition.construct(name, formula);
        AxiomSymbol symbol0 = AxiomSymbol.construct(name, formula, decl1, theory);
        symbolSymbols.put(symbol, symbol0);
        result.add(decl1);
        return;
      }
      */
      // all formulas are translated to axioms
      FormulaDefinition decl0 = (FormulaDefinition)decl;
      Formula formula = (Formula)translate(decl0.getValue());
      FormulaDefinition decl1 = AxiomDefinition.construct(name, formula);
      FormulaSymbol symbol0 = AxiomSymbol.construct(name, formula, decl1, theory);
      symbolSymbols.put(symbol, symbol0);
      result.add(decl1);
      return;
    }
    if (decl instanceof ValueDefinitionClass)
    {
      // translate definition into value declaration 
      // add initialization value to assumptions
      ValueSymbol symbol1 = (ValueSymbol)symbol;
      Type type = (Type)translate(symbol1.getType());
      ValueDeclarationClass decl1 = ValueDeclarationClass.construct(name, type);
      ValueSymbol symbol0 = ValueSymbol.construct(name, type, decl1);
      symbolSymbols.put(symbol, symbol0);
      result.add(decl1);
      ValueDefinitionClass decl0 = (ValueDefinitionClass)decl;
      Term value = (Term)translate(decl0.getValue());
      Identifier name0 = Identifier.construct(name.getString());
      name0.setSymbol(symbol0);
      Reference ref = Reference.construct(name0);
      ref.setType(symbol0.getType());
      EqualsFormula assumption = EqualsFormula.construct(ref, value);
      assumptions.add(assumption);
      return;
    }
    if (decl instanceof TypeDeclaration)
    {
      if (decl instanceof TypeDefinition)
      {
        TypeDefinition decl0 = (TypeDefinition)decl;
        Type type = (Type)translate(decl0.getValue());
        TypeDefinition decl1 = TypeDefinition.construct(name, type);
        TypeSymbol symbol0 = TypeSymbol.construct(name, type, decl1);
        symbolSymbols.put(symbol, symbol0);
        result.add(decl1);
        return;
      }
      Type type = AtomicType.construct(name.getString());
      TypeDeclaration decl1 = TypeDeclaration.construct(name);
      TypeSymbol symbol0 = TypeSymbol.construct(name, type, decl1);
      symbolSymbols.put(symbol, symbol0);
      result.add(decl1);
      return;
    }
    if (decl instanceof ValueDeclarationClass)
    {
      if (decl instanceof ValueDeclarationFormula)
      {
        ValueDeclarationFormula decl0 = (ValueDeclarationFormula)decl;
        Type type = (Type)translate(((ValueSymbol)symbol).getType());
        Formula formula = (Formula)translate(decl0.getFormula());
        ValueDeclarationFormula decl1 = 
          ValueDeclarationFormula.construct(name, type, formula);
        ValueSymbol symbol0 = ValueSymbol.construct(name, type, decl1);
        symbolSymbols.put(symbol, symbol0);
        result.add(decl1);
        return;
      }
      if (decl instanceof ValueDeclarationInitialized)
      {
        ValueDeclarationInitialized decl0 = (ValueDeclarationInitialized)decl;
        Type type = (Type)translate(((ValueSymbol)symbol).getType());
        Term term = (Term)translate(decl0.getValue());
        ValueDeclarationInitialized decl1 = 
          ValueDeclarationInitialized.construct(name, type, term);
        ValueSymbol symbol0 = ValueSymbol.construct(name, type, decl1);
        symbolSymbols.put(symbol, symbol0);
        result.add(decl1);
        return;
      }
      Type type = (Type)translate(((ValueSymbol)symbol).getType());
      ValueDeclarationClass decl1 = ValueDeclarationClass.construct(name, type);
      ValueSymbol symbol0 = ValueSymbol.construct(name, type, decl1);
      symbolSymbols.put(symbol, symbol0);
      result.add(decl1);
      return;
    }
    if (decl instanceof TypedIdentifier)
    {
      Type type = (Type)translate(((ValueSymbol)symbol).getType());
      TypedIdentifier decl1 = TypedIdentifier.construct(name, type);
      ValueSymbol symbol0 = ValueSymbol.construct(name, type, decl1);
      symbolSymbols.put(symbol, symbol0);
      result.add(decl1);
      return;
    }
    Main.internalError("unknown declaration");
  }
  
  /****************************************************************************
   * Convert name to identifier.
   * @param name a name.
   * @return the name translated to an identifier.
   **************************************************************************/
  private Identifier toIdentifier(Name name)
  {
    StringBuffer b = new StringBuffer();
    int n = name.getLength();
    for (int i=0; i<n-1; i++)
    {
      b.append(name.idents[i].getString());
      b.append('_');
    }
    b.append(name.idents[n-1].getString());
    return Identifier.construct(b.toString());
  }
  
  /***************************************************************************
   * Construct unique string that does not yet occur in names table.
   * @param idstring a suggestion for the string.
   * @return a unique version for the string.
   **************************************************************************/
  private String uniqueString(String idstring)
  {
    if (symbolNames.containsValue(idstring))
    {
      int index = 0;
      String idstring0 = idstring + "_" + index;
      while (symbolNames.containsValue(idstring0))
      {
        index = index+1;
        idstring0 = idstring + "_" + index;
      }
      idstring = idstring0;
    }
    return idstring;
  }
  
  /****************************************************************************
   * Construct unique name for symbol (keeping its original name, if possible).
   * @param symbol the symbol.
   * @return its unique name which is after the call in the name table
   ***************************************************************************/
  private Identifier uniqueName(Symbol symbol)
  {
    // symbol already occurs in table
    String idstring = symbolNames.get(symbol);
    if (idstring != null) return Identifier.construct(idstring);
    
    // symbol does not yet occur in table and its name is not yet used
    Identifier name = symbol.getName();
    idstring = name.getString();
    if (!symbolNames.containsValue(idstring))
    {
      symbolNames.put(symbol, idstring);
      return Identifier.construct(idstring);
    }

    // symbol does not yet occur in table and its short name is already used
    if (symbol instanceof GlobalSymbol)
    {
      // for global symbol, try its long name
      GlobalSymbol symbol0 = (GlobalSymbol)symbol;
      Name lname = symbol0.getLongName();
      idstring = toIdentifier(lname).getString();
    }
    
    // construct unique name
    idstring = uniqueString(idstring);
    symbolNames.put(symbol, idstring);
    return Identifier.construct(idstring);
  }
}
