// ----------------------------------------------------------------------------
// FrameTask.java
// The task of checking the frame condition.
// $Id: FrameTask.java,v 1.7 2011/10/03 06:59:02 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2009, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Tasks;

import java.io.PrintWriter;
import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Spec.*;
import fmrisc.ProgramExplorer.Syntax.Program.VoidType;
import fmrisc.ProgramExplorer.Syntax.Program.NamedType;
import fmrisc.ProgramExplorer.Syntax.Program.ArrayType;
import fmrisc.ProgramExplorer.Syntax.Program.TypeExpression;
import fmrisc.ProgramExplorer.Judgements.*;

/**
 * 
 * The task of checking the frame condition
 *
 */
public final class FrameTask extends TaskBase
{
  // the relevant information from the method itself
  private final boolean isFunction; // true if it returns a value
  
  // the relevant information from the specification
  private final Collection<VariableSymbol> specVariables; // the (base) variables
  private final Collection<ClassSymbol> specExceptions;   // the exceptions
  
  // the precomputed judgement
  private final StatJudgement judgement;
  
  // the result of the checking
  private Vector<String> errors;       // the generated error messages
  public final boolean executionOkay;  // execution status okay?
  public final boolean variablesOkay;  // variables frame okay? 
  public final boolean exceptionsOkay; // exceptions frame okay

  /***************************************************************************
   * Construct the task of checking the frame condition.
   * @param method the method whose frame-condition is to be checked.
   **************************************************************************/
  public FrameTask(ParamSymbol method)
  {
    super("[" + method.getLongName() + "] effects", false,
        method.getDeclaration().getPosition(),
        method.getObligations().getDirectory(), null);
    
    if (method instanceof ConstructorSymbol)
      isFunction = false;
    else 
    {
      MethodSymbol method0 = (MethodSymbol)method;
      isFunction = !(method0.getReturnType() instanceof VoidType);
    }
    
    // determine the relevant information from the specification
    MethodSpec spec = method.getSpecification();
    if (spec == null)
    {
      specVariables = new SetClass<VariableSymbol>();
      specExceptions = new SetClass<ClassSymbol>();
    }
    else
    {
      specVariables = NameUtils.getVariablePrefixSymbols(spec.getVariables());
      specExceptions = NameUtils.getClassSymbols(spec.getExceptions());
    }
  
    // determine the judgement
    judgement = method.getBodyJudgement();
    
    // the error messages to be generated
    errors = new Vector<String>();
    
    // determine individual status information
    executionOkay = checkExecution();
    variablesOkay = checkVariables();
    exceptionsOkay = checkExceptions();
    
    // determine overall execution status and error message
    if (executionOkay && variablesOkay && exceptionsOkay)
    {
      setStatus(Task.Status.DONE, null, null);
    }
    else
    {
      StringBuffer buffer = new StringBuffer();
      int n=errors.size();
      int i=0;
      for (String message : errors)
      {
        buffer.append(message);
        i=i+1;
        if (i<n) buffer.append(", ");
      }
      setStatus(Task.Status.FAILED, buffer.toString(), null);
    }
  }
  
  /****************************************************************************
   * Check execution status.
   * @return true iff everything is okay.
   ***************************************************************************/
  private boolean checkExecution()
  {
    boolean okay = true;
    if (isFunction)
    {
      if (!judgement.returns)
      {
        okay = false;
        errors.add("execution does not end with 'return'");
      }
      else if (judgement.executes)
      {
        okay = false;
        errors.add("execution may end without 'return'");
      }
    }
    if (judgement.continues)
    {
      okay = false;
      errors.add("execution may end with 'continue'");
    }
    if (judgement.breaks)
    {
      okay = false;
      errors.add("execution may end with 'break'");
    }
    return okay;
  }
  
  /****************************************************************************
   * Check variable status.
   * @return true iff everything is okay.
   ***************************************************************************/
  private boolean checkVariables()
  {
    boolean okay = true;
    for (VariableSymbol variable : judgement.variables)
    {
      if (variable instanceof ParameterSymbol)
      {
        TypeExpression type = variable.getType();
        if (!(type instanceof NamedType || type instanceof ArrayType)) continue;
      }
      if (specVariables.contains(variable)) continue;
      okay = false;
      errors.add("variable " + NameUtils.uniqueName(variable) + " may be modified");
    }
    return okay;
  }
  
  /****************************************************************************
   * Check exception status.
   * @return true iff everything is okay.
   ***************************************************************************/
  private boolean checkExceptions()
  {
    boolean okay = true;
    for (ClassSymbol exception : judgement.exceptions)
    {
      if (specExceptions.contains(exception)) continue;
      okay = false;
      errors.add("exception " + exception + " may be thrown");
    }
    return okay;
  }
  
  /****************************************************************************
   * Print linear representation of task.
   ***************************************************************************/
  public void print(PrintWriter out)
  {
    super.print(out);
    out.print("Modified variables: ");
    if (judgement.variables.isEmpty())
      out.print("(none)");
    else
    {
      int n = judgement.variables.size();
      int i=0;
      for (VariableSymbol variable : judgement.variables)
      {
        out.print(NameUtils.uniqueName(variable));
        i = i+1;
        if (i<n) out.print(", ");
      }
    }
    out.println();
    out.print("Thrown Exceptions: ");
    if (judgement.exceptions.isEmpty())
      out.print("(none)");
    else
    {
      int n = judgement.exceptions.size();
      int i=0;
      for (ClassSymbol exception : judgement.exceptions)
      {
        out.print(exception.getLongName());
        i = i+1;
        if (i<n) out.print(", ");
      }
    }
    out.println();
    out.print("May reach end of body: " + judgement.executes);
    out.println();
    out.print("May execute 'return': " + judgement.returns);
    out.println();
  }
}
