// ----------------------------------------------------------------------------
// SpecTask.java
// A task for validating a specification.
// $Id: SpecTask.java,v 1.19 2011/11/16 13:42:38 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Tasks;

import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.InOut.ErrorStream;
import fmrisc.ProgramExplorer.Judgements.*;
import fmrisc.ProgramExplorer.Judgements.Substitution.*;
import fmrisc.ProgramExplorer.Proving.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Spec.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;

/**
 * 
 * A task for validating a specification
 *
 */
public final class SpecTask extends StateTaskBase
{
  private ParamSymbol method;
  private boolean satisfiable;
  private StateType stateType;
  private Vector<TheorySymbol> theories;
  private Declaration[][] decls;
  private ErrorStream out;
  
  /***************************************************************************
   * Construct a task for validating a specification.
   * @param method the method whose specification is to be validated
   * @param satisfiable if true, the specification is checked for satisfiablility.
   *                    if false, it is checked for non-triviality (tautology)
   * @param stateType the type of the current state (null, if none)
   * @param theories the theories on which the following depend in dependence order.
   * @param decls a sequence of declaration lists.
   * @param out the stream to which to write error messages.
   **************************************************************************/
  public SpecTask(ParamSymbol method, boolean satisfiable,
    StateType stateType, Vector<TheorySymbol> theories, Declaration[][] decls,
    ErrorStream out)
  {
    super("[" + method.getLongName() + "] " +
        (satisfiable ? "specification is satisfiable" : 
                      "specification is not trivial"), true,
          method.getDeclaration().getPosition(),
          method.getOptional().getDirectory(), new ManualStrategy());
    
    // set status
    this.method = method;
    this.satisfiable = satisfiable;
    this.stateType = stateType;
    this.theories = theories;
    this.decls = decls;
    this.out = out;
    
    // handle special cases
    MethodSpec spec = method.getSpecification();
    if (spec == null)
    {
      this.setStatus(Task.Status.DONE, "no specification", null);
      return;
    }
    if (spec.getPostCondition() == null)
    {
      this.setStatus(Task.Status.DONE, "no postcondition", null);
      return;
    }
  }

  /***************************************************************************
   * Construct a state problem from the task.
   ***************************************************************************/
  public StateProblem stateProblem()
  {
    // specification components
    MethodSpec spec = method.getSpecification();
    Formula pre = Normalization.preCondition(method, true);
    Formula formula = Normalization.postCondition(method);
    Name[] exceptions0 = spec.getExceptions();
    Name[] variables0 = spec.getVariables();

    // determine exceptions/variables frame
    Set<ClassSymbol> exceptions = new SetClass<ClassSymbol>();
    Set<VariableSymbol> variables = new SetClass<VariableSymbol>();
    if (spec != null)
    {
      if (exceptions0 != null)
      {
        for (Name ex : exceptions0)
          exceptions.add((ClassSymbol)ex.getSymbol());
      }
      if (variables0 != null)
      {
        Normalization.modifiableVariables(variables0, variables);
      }
    }
    int varN = variables.size();
    
    // construct core goal
    if (!satisfiable) formula = NotFormula.construct(formula);
    
    // normalize formula such that all parameters and
    // non-assignable global variables occur in pre-state form
    formula = Normalization.perform(formula, variables);
    
    // normalize precondition such that all variables occur in pre-state form
    if (pre != null) 
      pre = Normalization.perform(pre, new SetClass<VariableSymbol>());
    
    // determine method parameters
    ParameterSymbol[] params = method.getParameterArray();
    int paramN = params.length;
    
    // determine the number of universally quantified variables:
    // * one (pre-state) for each parameter
    // * one (pre-state) for each modifiable variable
    // * one for the pre-state itself
    int preN = paramN+varN+1;
    TypedIdentifier[] preTids = new TypedIdentifier[preN];
    
    // determine the number of existentially quantified variables:
    // * one (post-state) for each modifiable variable
    // * one for the post-state itself
    int postN = varN+1;
    TypedIdentifier[] postTids = new TypedIdentifier[postN];
    
    // determine the symbols from which to derive the pre-variable names
    VariableSymbol[] preSymbols0 = new VariableSymbol[preN-1];
    int pos = 0;
    for (int i=0; i<paramN; i++) // parameters (pre-state)
    {
      preSymbols0[pos] = (VariableSymbol)params[i].getName().getSymbol();
      pos = pos+1;
    }
    for (VariableSymbol var : variables) // modifiable variables (pre-state)
    {
      preSymbols0[pos] = var;
      pos = pos+1;
    }
    if (pos != preN-1) Main.internalError("error in pre-variables");
    
    // determine the symbols from which to derive the post-variable names
    VariableSymbol[] postSymbols0 = new VariableSymbol[postN-1];
    pos = 0;
    for (VariableSymbol var : variables) // modifiable variables (post-state)
    {
      postSymbols0[pos] = var;
      pos = pos+1;
    }
    if (pos != postN-1) Main.internalError("error in post-variables");
    
    // unused names for logical variables substituting program variables
    Collection<String> used;
    if (pre == null)
      used = UsedVariables.compute(formula);
    else
    {
      used = UsedVariables.compute(pre);
      UsedVariables.add(formula, used);
    }
    
    // logical variables for universal quantification of pre-state
    String[] preVars = NewVariables.newVariables(preSymbols0, "", used);
    UsedVariables.add(preVars, used);
    ValueSymbol[] preSymbols = new ValueSymbol[preN];
    for (int i=0; i<preN-1; i++)
    {
      Type type = preSymbols0[i].getLogicType();
      Identifier varId = Identifier.construct(preVars[i]);
      TypedIdentifier tid = TypedIdentifier.construct(varId, type);
      preSymbols[i] = ValueSymbol.construct(varId, type, tid);
      preTids[i] = tid;
    }
    String preName = NewVariables.newVariable("os", used);
    used.add(preName);
    Identifier preIdent = Identifier.construct(preName);
    TypedIdentifier preVar = TypedIdentifier.construct(preIdent, stateType);
    preSymbols[preN-1] = ValueSymbol.construct(preIdent, stateType, preVar);
    preTids[preN-1] = preVar;
    
    // logical variables for existential quantification of post-state
    String[] postVars = NewVariables.newVariables(postSymbols0, "", used);
    UsedVariables.add(postVars, used);
    ValueSymbol[] postSymbols = new ValueSymbol[postN];
    for (int i=0; i<postN-1; i++)
    {
      Type type = postSymbols0[i].getLogicType();
      Identifier varId = Identifier.construct(postVars[i]);
      TypedIdentifier tid = TypedIdentifier.construct(varId, type);
      postSymbols[i] = ValueSymbol.construct(varId, type, tid);
      postTids[i] = tid;
    }
    String postName = NewVariables.newVariable("ns", used);
    used.add(postName);
    Identifier postIdent = Identifier.construct(postName);
    TypedIdentifier postVar = TypedIdentifier.construct(postIdent, stateType);
    postSymbols[postN-1] = ValueSymbol.construct(postIdent, stateType, postVar);
    postTids[postN-1] = postVar;
    
    // construct substitution
    Map<VariableSymbol,Value> vars = new MapClass<VariableSymbol,Value>();
    
    // pre-state substitution for parameters
    for (int i=0; i<paramN; i++)
    {
      vars.put(params[i], 
          new Substitution.Value(
              Substitution.Action.term, 
              Satisfies.reference(preSymbols[i].getName()),
              Substitution.Action.none, 
              null
          ));
    } 
    
    // pre/post-state substitution for modifiable variables
    // (may overrride above substitution)
    pos = 0;
    for (VariableSymbol variable : variables)
    {
      vars.put(variable, 
          new Substitution.Value(
              Substitution.Action.term, 
              Satisfies.reference(preSymbols[paramN+pos].getName()),
              Substitution.Action.term,
              Satisfies.reference(postSymbols[pos].getName())
          ));
      pos = pos+1;
    } 
    
    // perform substitutions in core formula and precondition
    formula = (Formula)Substitution.perform(formula, 
        Satisfies.reference(preIdent), Satisfies.reference(postIdent), vars);
    if (pre != null) 
      pre = (Formula)Substitution.perform(pre, 
        Satisfies.reference(preIdent), Satisfies.reference(preIdent), vars);
    
    // quantify the formula
    if (pre == null)
      formula = ForallFormula.construct(preTids, 
          ExistsFormula.construct(postTids, formula));
    else
      formula = ForallFormula.construct(preTids, 
          ImpliesFormula.construct(pre,
              ExistsFormula.construct(postTids, formula)));
    
    // simplify the formula
    formula = Simplification.perform(formula);
    
    // construct state problem
    ClassSymbol csymbol = method.getClassSymbol();
    FlattenDeclarations.Result result = 
      FlattenDeclarations.flatten(theories, decls, formula, csymbol.getLocalTheory());
    Type stringType = csymbol.getPackage().getTypeTranslator().getStringType();
    return new StateProblem(result.decls, result.goal, result.map,
        stateType, stringType, variables, exceptions, out);
  }
}
