// ----------------------------------------------------------------------------
// TaskFolder.java
// A folder of verification tasks.
// $Id: TaskFolder.java,v 1.16 2012/10/15 15:16:48 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2009, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Tasks;

import java.util.*;
import java.io.*;

import fmrisc.ProgramExplorer.Main;
import fmrisc.ProgramExplorer.Semantics.*;

/**
 * 
 * A folder of verification tasks.
 *
 */
public class TaskFolder extends TaskBase
{
  private Symbol symbol;                  // the symbol associated to folder 
  private Collection<TaskFolder> folders; // the subfolders
  private Collection<Task> tasks;         // the tasks
  private boolean sorted;                 // contents are sorted by name.
  
  /****************************************************************************
   * Construct root folder.
   * @param symbol the symbol associated to folder (may be null)
   * @param name the name of the folder
   * @param root a writable directory 
   *             in which a directory for this task folder is created.
   ***************************************************************************/
  public TaskFolder(Symbol symbol, String name, File root)
  {
    super(name, false, null, root, new AllNewStrategy());
    this.symbol = symbol;
    this.sorted = true;
    createFolders();
    clear();
  }
  
  /****************************************************************************
   * Construct task folder with denoted name and parent
   * @param symbol the symbol associated to folder (may be null)
   * @param name name of the folder (null, if none)
   * @param parent the parent of the task
   * @param sorted true iff the contents are to be sorted by name.
   **************************************************************************/
  public TaskFolder(Symbol symbol, String name, TaskFolder parent, boolean sorted)
  {
    super(name, false, null, parent.getDirectory(), new AllNewStrategy());
    this.symbol = symbol;
    this.sorted = sorted;
    parent.add(this);
    createFolders();
    clear();
  }
  
  /****************************************************************************
   * Construct sorted task folder with denoted name and parent
   * @param symbol the symbol associated to folder (may be null)
   * @param name name of the folder (null, if none)
   * @param parent the parent of the task
   **************************************************************************/
  public TaskFolder(Symbol symbol, String name, TaskFolder parent)
  {
    this(symbol, name, parent, true);
  }
  
  /****************************************************************************
   * Set task status.
   * @param evidence the evidence for the status (may be null).
   * @param info any additional information about the evidence (may be null).
   ***************************************************************************/
  public void setStatus(Status status, String evidence, Object info)
  {
    if (status == Status.REVOKED)
    {
      for (TaskFolder f : folders)
        f.setStatus(status, evidence, info);
      for (Task t : tasks)
        t.setStatus(status, evidence, info);
    }
    super.setStatus(status, evidence, info);
  }
  
  /***************************************************************************
   * Create folders.
   **************************************************************************/
  private void createFolders()
  {
    if (!sorted)
    {
      this.folders = new Vector<TaskFolder>();
      return;
    }
    this.folders = new TreeSet<TaskFolder>(
        new Comparator<Object>() {
          public int compare(Object o1, Object o2)
          {
            if ((o1 instanceof TaskFolder) && (o2 instanceof TaskFolder))
            {
              TaskFolder f1 = (TaskFolder)o1;
              TaskFolder f2 = (TaskFolder)o2;
              return f1.getName().compareTo(f2.getName());
            }
            Main.internalError("unknown object in task folder set");
            return 0; 
          }
        });
  }
  
  /***************************************************************************
   * Clear content of task folder
   **************************************************************************/
  public void clear()
  {        
    // tasks sorted by order of creation
    this.tasks = new Vector<Task>();

    // initial task status
    setStatus(Status.DONE, "folder is done", null);
  }
  
  /****************************************************************************
   * Compute new folder status.
   ***************************************************************************/
  public void computeStatus()
  {
    boolean hasNew = false;
    boolean hasOpen = false;
    boolean hasFailed = false;
    for (Task t : tasks)
    {
      if (t.isOptional()) continue;
      Status status = t.getStatus();
      if (status == Task.Status.NEW) hasNew = true;
      else if (status == Task.Status.OPEN) hasOpen = true;
      else if (status == Task.Status.FAILED) hasFailed = true;
    }
    for (Task t : folders)
    {
      if (t.isOptional()) continue;
      Status status = t.getStatus();
      if (status == Task.Status.NEW) hasNew = true;
      else if (status == Task.Status.OPEN) hasOpen = true;
      else if (status == Task.Status.FAILED) hasFailed = true;
    }
    if (hasFailed) 
      setStatus(Task.Status.FAILED, null, null);
    else if (hasOpen) 
      setStatus(Task.Status.OPEN, null, null);
    else if (hasNew) 
      setStatus(Task.Status.NEW, null, null);
    else
      setStatus(Task.Status.DONE, null, null);
  }

  /***************************************************************************
   * Get subfolders appropriately sorted for display.
   * @return the subfolders.
   **************************************************************************/
  public TaskFolder[] getFolders()
  {
    return folders.toArray(new TaskFolder[folders.size()]);
  }
  
  /***************************************************************************
   * Get tasks appropriately sorted for display
   * @return the tasks.
   **************************************************************************/
  public Task[] getTasks()
  {
    return tasks.toArray(new Task[tasks.size()]);
  }
  
  /***************************************************************************
   * Add subfolder to folder.
   * @param folder subfolder to be added.
   **************************************************************************/
  private void add(TaskFolder folder)
  {
    folders.add(folder);
    folder.setParent(this);
  }
  
  /***************************************************************************
   * Add task to folder.
   * @param task task to be added.
   **************************************************************************/
  public void add(Task task)
  {
    tasks.add(task);
    task.setParent(this);
  }
  
  /***************************************************************************
   * Remove subfolder from folder.
   * @param folder subfolder to be removed.
   **************************************************************************/
  public void remove(TaskFolder folder)
  {
    folders.remove(folder);
  }
  
  /****************************************************************************
   * Get the symbol associated to the folder.
   * @return the associated symbol (may be null).
   ***************************************************************************/
  public Symbol getSymbol()
  {
    return symbol;
  }
}
