// ----------------------------------------------------------------------------
// FormulaC.java
// The virtual "declaration" command.
// $Id: DeclarationC.java,v 1.30 2017/07/04 08:04:36 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Commands;

import java.io.*;
import java.util.*;

import org.w3c.dom.*;

import fmrisc.ProofNavigator.*;
import fmrisc.ProofNavigator.Communication.*;
import fmrisc.ProofNavigator.Proving.*;
import fmrisc.ProofNavigator.Semantics.*;
import fmrisc.ProofNavigator.Syntax.*;

/**
 * 
 * Virtual command "declaration": process declaration in current environment
 *
 */
public final class DeclarationC extends CommandBase
{
  public final static String Name = "declaration";
  private Declaration decl;
  
  /***************************************************************************
   * Create a "declaration" command.
   * @param decl the declaration to be processed
   **************************************************************************/
  public DeclarationC(Declaration decl)
  {
    super(Name);
    this.decl = decl;
  }
  
  /****************************************************************************
   * Process declaration. This means the following:
   * - At the global level, type checking conditions are immediately checked
   *   and the declaration is added to the current environment.
   * - Within a proof, a type checking condition lead to an additional
   *   child state; the declaration is added to a child state.
   *   E: A, ... |- B, ... implies
   *   ( E: A, ... |- TCC_D )
   *     E U {D}: A, ... |- B
   ***************************************************************************/
  public void process()
  {
    process(true);
  }
  
  /****************************************************************************
   * Process declaration. 
   * @param display true iff declaration is to be displayed.
   ***************************************************************************/
  public void process(boolean display)
  {
    PrintWriter out = State.getSystemOut();
    
    Environment senv = State.getEnvironment();
    Environment oenv = getEnvironment();
    State.setEnvironment(oenv);
    
    Environment env;
    Proof proof = State.getProof();
    if (proof == null)
    {
      env = oenv;
      Checking.resetReferencedSymbols();
    }
    else
      env = oenv.cloneGlobal();

    Identifier ident = decl.getIdentifier();
    Context context = State.getContext();
    context.open(); 
    Symbol symbol = null;
    if (decl instanceof TypeDeclaration)
    {
      if (proof != null)
      {
        State.setError("type declaration not allowed within proof");
        State.setEnvironment(senv);
        context.abort();
        return;
      }
      TypeDeclaration tdecl = (TypeDeclaration)decl;
      TypeSymbol tsymbol = env.putTypeDeclaration(tdecl);
      if (tsymbol == null)    
      {
        State.setEnvironment(senv);
        context.abort();
        return;
      }
      symbol = tsymbol;
      out.print("Type " + ident.toString());
      if (tdecl.getType() != null)
      {
        out.print(" = " + tsymbol.getType().toString());
      }
      out.println(".");
    }  
    else if (decl instanceof FormulaDeclaration)
    {
      if (proof != null)
      {
        State.setError("formula declaration not allowed within proof");
        State.setEnvironment(senv);
        context.abort();
        return;
      }
      FormulaDeclaration fdecl = (FormulaDeclaration)decl;
      FormulaSymbol fsymbol = env.putFormulaDeclaration(fdecl);
      if (fsymbol == null)    
      {
        State.setEnvironment(senv);
        context.abort();
        return;
      }
      symbol = fsymbol;
      out.println("Formula " + ident.toString() + ".");
    } 
    else if (decl instanceof ValueDeclaration)
    {
      ValueDeclaration vdecl = (ValueDeclaration)decl;
      ValueSymbol vsymbol = env.putValueDeclaration(vdecl);
      if (vsymbol == null)
      {
        State.setEnvironment(senv);
        context.abort();
        return;
      }
      symbol = vsymbol;
      // make sure that symbol is not treated as global symbol
      // when collecting symbols referenced in proof
      Type type = vsymbol.getType();
      out.print("Value " + ident.toString());
      out.print(":" + type.toString());
      out.println(".");
    }
    else
    {
      error("Unknown kind of declaration: " + decl.toString());
      State.setEnvironment(senv);
      context.abort();
      return;
    }
    Expression tcc = context.getFormula();
    context.commit();
    State.setEnvironment(senv);
    
    // at the global level:
    // type checking condition is checked;
    // proof of formula is read
    // tcc and declaration is added to prover state
    // declaration is signalled and written to store
    if (proof == null)
    {       
      // determine symbols referenced in declaration
      // (local value symbols must be filtered)
      Vector<Symbol> referenced = getReferenced(Checking.getReferencedSymbols(), env);
      Store store = State.getStore();
      boolean okay = store.write(decl, referenced);
      if (!okay) return;
      
      // check type checking condition
      State.setLastTCC(tcc);
      boolean checked = checkTCC(tcc);
      Prover prover = State.getProver();
      if (decl instanceof TypeDeclaration || decl instanceof ValueDeclaration)
        prover.addDeclaration(decl);
      if (tcc != null && !checked) env.addTCC(tcc);
      
      // experimental: make user axiom visible to type checking
      if (decl instanceof AxiomDeclaration)
      {
        Expression formula = ((AxiomDeclaration)decl).getFormula();
        prover.assertFormula(formula);
      }
      
      if (symbol instanceof FormulaSymbol)
      {
        FormulaSymbol fsymbol = (FormulaSymbol)symbol;
        if (store.existsProof(fsymbol))
        {
          Proof fproof = store.readProof(fsymbol);
          if (fproof != null) 
          {
            fsymbol.setProof(fproof);
            out.print("Proof read (");
            fproof.printStatus(out);
            out.println(").");
          }
        }
        // we have already cloned the symbol environment
        if (tcc != null && !checked) fsymbol.getEnvironment().addTCC(tcc);
      }
      
      // signal declaration (at the end because proof status may be included)
      if (display) okay = Main.newDeclaration(decl);
      return;
    }
    
    // within a proof: generate new proof states.
    ProofState state = proof.getCurrent();
    ProofState state0 = null;
    if (tcc != null) 
    {
      state0 = new ProofState(state, oenv);
      state0.setAssumptions(state0.copy(state.getAssumptions()));
      state0.setGoals(new Formula[] { new Formula(state0, tcc, true) });
    }
    ProofState state1 = new ProofState(state, env);
    // gets also instantiated by auto!!!
    if (tcc != null) 
      state1.setAxioms(state1.add(state1.getAxioms(),
        new Formula(state1, tcc, false)));
    /* 
    if (decl instanceof ValueDeclaration)
    {
      ValueDeclaration vdecl = (ValueDeclaration)decl;
      TypeExpression te = 
        vdecl.getValueDeclIdentifier().getSymbol().getClearType();
      if (te != null)
      {
        Expression formula = te.getExpression();
        state1.setAxioms(state1.add(state1.getAxioms(),
            new Formula(state1, formula, false)));        
      }
    }
    */
    state1.setAssumptions(state1.copy(state.getAssumptions()));
    state1.setGoals(state1.copy(state.getGoals()));
    if (state0 == null)
      state.extend(this, new ProofState[]{ state1 });
    else
      state.extend(this, new ProofState[]{ state0, state1 });
  }
  
  /***************************************************************************
   * Get value symbols from collection that are contained in environment.
   * @param symbols a collection of Symbol values.
   * @param env the environment
   * @return a vector of the value symbols from the symbols
   *         that are contained in the environment.
   **************************************************************************/
  public static Vector<Symbol> getReferenced(Collection<Symbol> symbols, Environment env)
  {
    Collection <Symbol> vsymbols = env.getValueSymbols();
    Vector<Symbol> referenced = new Vector<Symbol>();
    Iterator<Symbol> iterator = symbols.iterator();
    while (iterator.hasNext())
    {
      Symbol rsymbol = iterator.next();
      if (rsymbol instanceof ValueSymbol)
      {
        if (vsymbols.contains(rsymbol)) referenced.add(rsymbol);
      }
      else
        referenced.add(rsymbol);
    }
    return referenced;
  }
  
  /***************************************************************************
   * check type checking condition
   * @param tcc the type checking condition to be checked
   * @return true iff checking was successful
   **************************************************************************/
  private static boolean checkTCC(Expression tcc)
  {
    if (tcc == null) return true;
    
    PrintWriter out = State.getSystemOut();
    Prover prover = State.getProver();
    if (prover == null)
    {
      out.println("Warning: prover not running, cannot verify type checking condition.");
      return false;
    }
    Answer answer = prover.queryFormula(tcc);
    if (answer != null && answer instanceof ValidAnswer) return true;

    out.println("Warning: cannot verify type checking condition:");
    tcc.print(out);
    out.println("");
    if (answer == null)
    {
      out.println(prover.getError());
      State.resetError();
    }
    else
    {
      String justification = answer.getJustification();
      if (justification == null)
        out.println("No justification given.");
      else
        out.println(justification);
    }
    out.print("Warning: type checking condition is added as an axiom, ");
    out.println("the remainder of the proof may be unsound.");
    prover.assertFormula(tcc);
    return false;
  }
  
  /***************************************************************************
   * Set external representation of command in a DOM node.
   * @param node the node in which the representation is to be set
   *        (by default none, to be overridden by command class)
   **************************************************************************/
  protected void setRepresentation(Node node)
  {
    node.appendChild(toNode(decl));
  }
  
  /***************************************************************************
   * Convert DOM node to command.
   * @param node the DOM representation of the command.
   * @return the command (null, if an error occured).
   **************************************************************************/
  public static Command newCommand(Node node)
  {
    NodeList children = node.getChildNodes();
    if (children.getLength() != 1) 
    {
      State.setError("child expected: " + node.toString());
      return null;
    }
    Declaration decl = toDeclaration(children.item(0));
    if (decl == null) return null;
    return new DeclarationC(decl);
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print(getName());
    out.print(" ");
    decl.print(out);
  }
}
