// ----------------------------------------------------------------------------
// Store.java
// Persistent store of declarations and proofs.
// $Id: Store.java,v 1.26 2020/07/21 07:38:25 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Communication;

import java.io.*;
import java.util.*;
import java.util.zip.*;

import nl.tue.win.riaca.openmath.lang.OMObject;

import org.w3c.dom.*;
import org.xml.sax.*;
import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;

import fmrisc.External.*;
import fmrisc.ProofNavigator.*;
import fmrisc.ProofNavigator.Proving.*;
import fmrisc.ProofNavigator.Semantics.*;
import fmrisc.ProofNavigator.Syntax.*;

/**
 * Persistent store of declarations and proofs.
 * 
 * For a declaration NAME of KIND (where KIND in {type,value,formula}) 
 * the following files are created:
 * 
 * KIND_NAME_decl.xgz
 *   Gzipped XML file holding the OMDoc representation of the declaration.
 *   (Since the declaration is actually not needed, we don't generate it any more).
 *   
 * KIND_NAME_hash.txt (not for proofs)
 *   A single text line with a hash value for the representation (a 
 *   non-negative integer number in decimal representation).
 *   
 * KIND_NAME_time.txt
 *   A single text line with a time stamp (a non-negative integer number
 *   in decimal representation).
 *   
 * KIND_NAME_refs.xml
 *   XML file representing the symbols referenced in the declaration:
 *   fmrisc:refs
 *     fmrisc:KIND time="timestampvalue" NAME /fmrisc:KIND
 *     ...
 *   /fmrisc:refs
 * 
 * Writing to the Store:
 * 
 * If an entity is to be written to the store, it is assumed that all the 
 * symbols it depends on have already been updated in the store.
 * 
 * If the entity is not yet contained in the store, we write it to the 
 * store and are done.
 * 
 * Otherwise, we check whether the
 * symbols listed in the corresponding "refs" file are stored with the
 * same time stamps as listed in the file. If yes, then the hash value
 * of the internal representation is computed. If this value is the same
 * as the hash value of the "hash" file, we assume that the definition
 * has not changed and we are done without updating the entity. 
 * Otherwise (i.e., if one of the previous conditions is not satisfied), 
 * we write the entity to the store updating all files.
 *  
 * Currently, we do not read declaration contents from the store.
 * 
 * For a proof, the following files are created:
 * 
 * proof_NAME_decl.xgz
 *   Gzipped XML file representing the proof of the formula NAME.
 *   fmrisc:proof name="formulaname" closed="true|false"
 *     fmrisc:state
 *       fmrisc:command /fmrisc:command
 *       fmrisc:state.../fmrisc:state
 *       ...
 *     /fmrisc:state
 *   /fmrisc:proof
 *   
 * KIND_NAME_refs.xml
 *   XML file representing the symbols referenced in the declaration:
 * 
 * Writing a proof proceeds as explained for the declarations above.
 * 
 * When a proof is read from the store, it is assumed that all the 
 * symbols in memory have already been updated in the store.
 * 
 * The proof is returned with status "skeleton=yes" and the status
 * "closed" as listed in the file.
 * 
 * The proof's "trusted" value is set to "true" iff the symbols listed 
 * in the "refs" file are stored with the same time stamps as listed in 
 * the file. Replaying such a proof must not cause errors.
 * 
 * The proof's "absolute" value is set to "true" iff 
 * the proof is "trusted" and "closed" and all formulas
 * listed in the "refs" file have absolute proofs.
 *****************************************************************************/
public final class Store
{
  private File directory; // the directory
  private Cache cache;    // the cache
  
  // true if declarations are written to file (not needed)
  private static boolean WRITE_DECLARATIONS = false;
  
  // true if cache is to be used
  private static boolean USE_CACHE = true;
  
  // default values, if the corresponding info could not be determined
  public final static long HASH_DEFAULT = -1;
  public final static long STAMP_DEFAULT = -1;
  
  /***************************************************************************
   * Create instance of store for denoted directory.
   * @param directory for storing the declared objects
   **************************************************************************/
  public Store(File directory)
  {
    this.directory = directory;
    if (USE_CACHE)
      this.cache = new Cache();
    else
      this.cache = null;
  }
  
  /****************************************************************************
   * Construct file from name.
   * @param name of file
   * @return the file
   ***************************************************************************/
  private File getFile(String name)
  {
    return new File(directory, name);
  }
  
  /***************************************************************************
   * Write proof to store updating the dependencies file.
   * @param proof the proof
   * @return true iff no problem occured in storing
   **************************************************************************/
  public boolean write(Proof proof)
  {
    // writing the proof also constructs collection of referenced symbols 
    boolean okay = writeProof(proof);
    if (!okay) return false;
    
    // write referenced symbols
    File dfile = getFile(getProofDepName(proof.getSymbol()));
    okay = writeDep(dfile, proof.getReferenced());
    return okay;
  }
  
  /***************************************************************************
   * Write proof to store.
   * @param proof the proof
   * @return true iff no problem occured in storing
   **************************************************************************/
  private boolean writeProof(Proof proof)
  {
    String name = getProofDeclName(proof.getSymbol());
    try
    {
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      DocumentBuilder builder = factory.newDocumentBuilder();
      Document document = builder.newDocument();
      Node node = proof.toNode(document);
      document.appendChild(node);
      TransformerFactory tfactory = TransformerFactory.newInstance();
      Transformer transformer = tfactory.newTransformer();
      // causes problem with Java 11 when restoring proofs
      // transformer.setOutputProperty(OutputKeys.INDENT, "yes");
      DOMSource source = new DOMSource(document);
      File file = getFile(name);
      OutputStream out = 
        new GZIPOutputStream (new BufferedOutputStream(new FileOutputStream(file)));
      StreamResult result = new StreamResult(out);
      transformer.transform(source, result);
      out.close();
      return true;
    }
    catch(ParserConfigurationException e)
    {
      State.setError("exception " + e.getMessage());
      return false;
    }
    catch(TransformerConfigurationException e)
    {
      State.setError("exception " + e.getMessage());
      return false;
    }
    catch(TransformerException e)
    {
      State.setError("exception " + e.getMessage());
      return false;
    }
    catch(IOException e)
    {
      State.setError("exception " + e.getMessage());
      return false;
    }
  }
 
  /***************************************************************************
   * Check whether store holds proof of formula.
   * @param name the name of a formula.
   * @return true iff store holds a proof of the formula.
   **************************************************************************/
  public boolean existsProof(String name)
  {
    File file = getFile(getProofDeclName(name));
    return file.exists();
  }
  
  /***************************************************************************
   * Check whether store holds proof of formula.
   * @param symbol a formula symbol.
   * @return true iff store holds a proof of the formula.
   **************************************************************************/
  public boolean existsProof(FormulaSymbol symbol)
  {
    return existsProof(symbol.getIdentifier().getName());
  }
  
  /***************************************************************************
   * Read proof for formula from store.
   * @param symbol a formula symbol.
   * @return the proof of the formula (or null, if an error occurred).
   **************************************************************************/
  public Proof readProof(FormulaSymbol symbol)
  {
    try
    {
      File file = getFile(getProofDeclName(symbol));
      InputStream in = 
        new GZIPInputStream(new BufferedInputStream(new FileInputStream(file)));
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      factory.setNamespaceAware(true);
      DocumentBuilder builder = factory.newDocumentBuilder();
      Document document = builder.parse(in);
      if (document == null) return null;
      Node root = document.getFirstChild();
      if (root == null) return null;
      removeWhitespace(root);
      Proof proof = Proof.toProof(symbol, root);
      if (proof == null) return null;
      File dfile = getFile(getProofDepName(symbol));
      if (dfile.isFile())
      {
        Map<Symbol,Long>dep = readDep(dfile, null);
        if (dep == null) return null;
        if (proof.isTrusted())   // proof may be already untrusted from reading
          proof.setTrusted(!hasChanged(dep));
        proof.setAbsolute(true); // symbol itself occurs in dependence map
        proof.setAbsolute(absoluteOnly(dep));
      }
      else
      {
        proof.setTrusted(false);
        proof.setAbsolute(false);
      }
      return proof;
    }
    catch(ParserConfigurationException e)
    {
      State.setError("exception " + e.getMessage());
      return null;
    }
    catch(IOException e)
    {
      State.setError("exception " + e.getMessage());
      return null;
    }
    catch(SAXException e)
    {
      State.setError("exception " + e.getMessage());
      return null;
    }
  }
  
  /***************************************************************************
   * Write declaration to store.
   * @param decl the declaration
   * @param referenced the symbols referenced in the definition of the symbol
   * @return true iff no problem occured in storing
   **************************************************************************/
  public boolean write(Declaration decl, Collection<Symbol> referenced)
  {
    Symbol symbol = decl.getSymbol();
    if (hasChanged(decl))
    {
      boolean okay = write(decl);
      if (!okay) return false;
      okay = writeDep(symbol, referenced);
      if (!okay) return false;
      if (cache != null) return true; // writeDep() wrote hash and stamp
    }
    else
    {
      Map<Symbol,Long>dep = null;
      if (cache == null)
      {
        File file = getFile(getDepName(symbol));
        if (!file.isFile()) 
        {
          State.setError("file " + file + " not found");
          return false;
        }
        dep = readDep(file, null);
      }
      else
        dep = cache.getDep(symbol);
      if (dep == null) return false;
      if (hasChanged(dep, referenced))
      {
        // update time stamp in declaration, even if decl is not changed
        boolean okay = write(decl);
        if (!okay) return false;
        okay = writeDep(symbol, referenced);
        if (!okay) return false;
        if (cache != null) return true; // writeDep() wrote hash and stamp
      }
      else 
      {
        if (dep.size() != referenced.size())
        {      
          // update time stamp in declaration
          boolean okay = write(decl);
          if (!okay) return false;
          okay = writeDep(symbol, referenced);
          if (!okay) return false;
        }
        return true;
      }
    }
    // only reached if cache == null
    boolean okay = writeHash(decl);
    if (!okay) return false;
    okay = writeStamp(decl.getSymbol());
    return okay;
  }
  
  /***************************************************************************
   * Signal whether declaration has changed compared to that in store.
   * @param decl a declaration
   * @return true iff store declaration is different from current one
   **************************************************************************/
  public boolean hasChanged(Declaration decl)
  {
    long hash0 = getHash(decl.getSymbol());
    if (hash0 == HASH_DEFAULT) return true;
    long hash1 = getHash(decl);
    return hash0 != hash1;
  }
  
  /***************************************************************************
   * Get hash code associated to symbol.
   * @param a symbol.
   * @return the hash code (HASH_DEFAULT, if none could be determined).
   **************************************************************************/
  private long getHash(Symbol symbol)
  {
    if (cache == null)
      return getHashFromFile(symbol);
    else
      return cache.getHash(symbol);
  }
  
  /***************************************************************************
   * Get hash code associated to symbol from file.
   * @param a symbol.
   * @return the hash code (HASH_DEFAULT, if none could be determined).
   **************************************************************************/
  private long getHashFromFile(Symbol symbol)
  {
    File file = getFile(getHashName(symbol));
    try
    {
      BufferedReader reader = new BufferedReader(new FileReader(file));
      long hash = Long.parseLong(reader.readLine());
      reader.close();
      return hash;
    }
    catch(FileNotFoundException e)
    {
      // file not found
      return HASH_DEFAULT;
    }
    catch(IOException e)
    {
      // cannot read file
      return HASH_DEFAULT;
    }
    catch(NumberFormatException e)
    {
      // invalid hash code
      return HASH_DEFAULT;
    }
  }
  
  /***************************************************************************
   * Signal whether dependencies have changed compared to that in store.
   * @param dep a dependencies map.
   * @param referenced collection of symbols of referenced values
   * @return true iff stored dependencies different from current one
   **************************************************************************/
  private boolean hasChanged(Map<Symbol,Long>dep, Collection<Symbol> referenced)
  {
    Iterator<Symbol> iterator = referenced.iterator();
    while (iterator.hasNext())
    {
      Symbol symbol = (Symbol)(iterator.next());
      Long stamp0 = (Long)dep.get(symbol);
      if (stamp0 == null) return true;
      long stamp1 = getStamp(symbol);
      if (stamp0.longValue() != stamp1) return true;
    }
    return false;
  }
  
  /***************************************************************************
   * Signal whether dependencies have changed time stamps
   * compared to that listed in dependency map.
   * @param dep map of dependency mappings.
   * @return true iff time stamps have changed.
   **************************************************************************/
  private boolean hasChanged(Map<Symbol,Long>dep)
  {
    Iterator<Symbol> iterator = dep.keySet().iterator();
    while (iterator.hasNext())
    {
      Symbol symbol = (Symbol)(iterator.next());
      Long stamp0 = (Long)dep.get(symbol);
      if (stamp0 == null) return true;
      long stamp1 = getStamp(symbol);
      if (stamp0.longValue() != stamp1) return true;
    }
    return false;
  }
   
  /***************************************************************************
   * Signal whether all formulas in dependency map have absolute proofs.
   * @param dep map of dependencies.
   * @return true iff all formulas in map have absolute proofs.
   **************************************************************************/
  private boolean absoluteOnly(Map<Symbol,Long>dep)
  {
    Iterator<Symbol> iterator = dep.keySet().iterator();
    while (iterator.hasNext())
    {
      Symbol symbol = (Symbol)(iterator.next());
      if (!(symbol instanceof FormulaSymbol)) continue;
      FormulaSymbol fsymbol = (FormulaSymbol)symbol;
      if (fsymbol.isAxiom()) continue;
      Proof proof = fsymbol.getProof();
      if (proof == null) return false;
      if (!proof.isAbsolute()) return false;
    }
    return true;
  }
  
  /***************************************************************************
   * Read dependencies.
   * @param file a file holding the dependencies.
   * @param string the already loaded content of the file (may be null).
   * @return map of Symbol->Long pairs representing the dependencies
   *         (new MapClass<Symbol,Long>(), if some error occured)
   **************************************************************************/
  private Map<Symbol,Long>readDep(File file, String string)
  {
    try
    {
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      factory.setNamespaceAware(true);
      DocumentBuilder builder = factory.newDocumentBuilder();
      Document document = null;
      if (string == null)
        document = builder.parse(file);
      else
      { 
        Reader reader = new StringReader(string);
        document = builder.parse(new InputSource(reader));
      }
      if (document == null) return new MapClass<Symbol,Long>();
      Node root = document.getFirstChild();
      if (root == null) return new MapClass<Symbol,Long>();
      removeWhitespace(root);
      String name = root.getLocalName();
      if (!name.equals("refs"))
      {
        State.setError("root node in file " + file.getAbsolutePath() + 
            " has name " + name + " (\"refs\" expected)");
        return new MapClass<Symbol,Long>();
      }
      Map<Symbol,Long> result = new MapClass<Symbol,Long>();
      NodeList list = root.getChildNodes();
      int n = list.getLength();
      for (int i=0; i<n; i++)
      {
        Node node = list.item(i);
        name = node.getLocalName();
        NamedNodeMap attributes = node.getAttributes();
        long stamp = STAMP_DEFAULT;
        if (attributes != null)
        {
          Node attribute = attributes.getNamedItem("time");
          if (attribute != null)
            stamp = Long.parseLong(((Attr)attribute).getValue());
        }
        if (stamp == STAMP_DEFAULT)
        {
          State.setError("Invalid symbol number " + i + " in file " +
              file.getAbsolutePath() +
              " (no valid \"time\" attribute):\n" + node.toString());
          return new MapClass<Symbol,Long>();
        }
        Node child = node.getFirstChild();
        if (child == null || !(child instanceof Text))
        {
          State.setError("Invalid symbol number " + i + " in file " +
              file.getAbsolutePath() +
              " (no text):\n" + node.toString());
          return new MapClass<Symbol,Long>();
        } 
        Text text = (Text)child;
        Identifier ident = Construct.reference(text.getData());
        Environment env = State.getEnvironment();
        Symbol csymbol = null;
        if (name.equals("type"))
          csymbol = env.getTypeSymbol(ident);
        else if (name.equals("value"))
          csymbol = env.getValueSymbol(ident);
        else if (name.equals("formula"))
          csymbol = env.getFormulaSymbol(ident);
        else
          State.setError("Invalid symbol number " + i + " in file " +
              file.getAbsolutePath() +
              " (unknown element):\n" + node.toString());
        if (csymbol == null) 
          return new MapClass<Symbol,Long>();
        result.put(csymbol, new Long(stamp));
      }  
      return result;
    }
    catch(ParserConfigurationException e)
    {
      State.setError("exception " + e.getMessage());
      return new MapClass<Symbol,Long>();
    }
    catch(IOException e)
    {
      State.setError("exception " + e.getMessage());
      return new MapClass<Symbol,Long>();
    }
    catch(SAXException e)
    {
      State.setError("exception " + e.getMessage());
      return new MapClass<Symbol,Long>();
    }
  }
  
  /***************************************************************************
   * Write dependencies to file.
   * @param file the file to write to.
   * @param referenced collection of symbols of referenced values
   * @return true iff no problem occured in writing.
   **************************************************************************/
  private boolean writeDep(File file, Collection<Symbol> referenced)
  {
    try
    {
      FileWriter writer = new FileWriter(file);
      return writeDep(writer, referenced.iterator());
    }
    catch(IOException e)
    {
      return false;
    }
  }
  
  /***************************************************************************
   * Write dependencies of symbol to file.
   * @param symbol the symbol
   * @param referenced collection of symbols of referenced values
   * @return true iff no problem occured in writing.
   **************************************************************************/
  private boolean writeDep(Symbol symbol, Collection<Symbol> referenced)
  {
    if (cache == null)
    {
      File file = getFile(getDepName(symbol));
      return writeDep(file, referenced);
    }
    else
      return cache.writeDep(symbol, referenced);
  }
  
  /***************************************************************************
   * Write dependencies to file.
   * @param writer the writer to write to.
   * @param iterator an iterator of the symbols to be written.
   * @return true iff no problem occured in writing.
   **************************************************************************/
  private boolean writeDep(Writer writer, Iterator<Symbol> iterator)
  {
    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    try
    {
      DocumentBuilder builder = factory.newDocumentBuilder();
      Document document = builder.newDocument();
      Element root = document.createElementNS(
        State.NamespaceURI, State.NamespacePrefix + ":refs");
      document.appendChild(root);
      root.setAttribute("xmlns:" + State.NamespacePrefix, State.NamespaceURI);
      while (iterator.hasNext())
      {
        Symbol symbol = (Symbol)iterator.next();
        Identifier ident = symbol.getIdentifier();
        String kind = null;
        if (symbol instanceof TypeSymbol)
          kind = "type";
        else if (symbol instanceof ValueSymbol)
          kind = "value";
        else if (symbol instanceof FormulaSymbol)
          kind = "formula";
        Element ref = document.createElementNS(
          State.NamespaceURI, State.NamespacePrefix + ":" + kind);
        root.appendChild(ref);
        ref.setAttribute("time", Long.toString(getStamp(symbol)));
        Text text = document.createTextNode(ident.getName());
        ref.appendChild(text);
      }
      TransformerFactory tfactory = TransformerFactory.newInstance();
      Transformer transformer = tfactory.newTransformer();
      transformer.setOutputProperty(OutputKeys.INDENT, "yes");
      DOMSource source = new DOMSource(document);
      StreamResult result = new StreamResult(writer);
      transformer.transform(source, result);
    }
    catch(ParserConfigurationException e)
    {
      State.setError("exception " + e.getMessage());
      return false;
    }
    catch(TransformerException e)
    {
      State.setError("exception " + e.getMessage());
      return false;
    }
    catch(NumberFormatException e)
    {
      State.setError("exception " + e.getMessage());
      return false;
    }
    return true;
  }
 
  /****************************************************************************
   * write time stamp of symbol
   * @param symbol a symbol
   * @return true iff everything went fine.
   **************************************************************************/
  private boolean writeStamp(Symbol symbol)
  {
    String name = getStampName(symbol);
    File file = getFile(name);
    try
    {
      PrintWriter writer = new PrintWriter(new BufferedWriter(new FileWriter(file)));
      writer.println(new Date().getTime());
      writer.close();
    }
    catch(IOException e)
    {
      State.setError("exception " + e.getMessage());
      return false;
    }
    return true;
  }
  
  /****************************************************************************
   * get time stamp of symbol
   * @param symbol a symbol
   * @return the time stamp of the symbol (STAMP_DEFAULT, if an error occured)
   **************************************************************************/
  private long getStamp(Symbol symbol)
  {
    if (cache == null)
      return getStampFromFile(symbol);
    else
      return cache.getStamp(symbol);
  }
  
  /****************************************************************************
   * get time stamp of declaration from file 
   * @param symbol a symbol
   * @return the time stamp of the symbol (STAMP_DEFAULT, if an error occured)
   **************************************************************************/
  private long getStampFromFile(Symbol symbol)
  {
    String name = getStampName(symbol);
    File file = getFile(name);
    try
    {
      BufferedReader reader = new BufferedReader(new FileReader(file));
      long time = Long.parseLong(reader.readLine());
      reader.close();
      return time;
    }
    catch(FileNotFoundException e)
    {
      return STAMP_DEFAULT;
    }
    catch(IOException e)
    {
      return STAMP_DEFAULT;
    }
    catch(NumberFormatException e)
    {
      return STAMP_DEFAULT;
    }
  }
  
  /****************************************************************************
   * Compute hash code for declaration.
   * @param decl the declaration.
   * @return its hash code.
   ***************************************************************************/
  private long getHash(Declaration decl)
  {
    String rep = decl.toString();
    return GeneralHashFunctionLibrary.RSHash(rep);
  }
  
  /****************************************************************************
   * write hash code of declaration
   * @param a declaration
   * @return true iff everything went fine.
   **************************************************************************/
  private boolean writeHash(Declaration decl)
  {
    String name = getHashName(decl.getSymbol());
    File file = getFile(name);
    try
    {
      PrintWriter writer = new PrintWriter(new BufferedWriter(new FileWriter(file)));
      writer.println(getHash(decl));
      writer.close();
    }
    catch(IOException e)
    {
      State.setError("exception " + e.getMessage());
      return false;
    }
    return true;
  }
  
  /***************************************************************************
   * Write declaration to store.
   * @param decl the declaration
   * @return true iff no problem occured in storing
   **************************************************************************/
  private boolean write(Declaration decl)
  {
    // updates cache (without writing to file)
    if (cache != null) return cache.write(decl);
    
    // we never use the declarations on file
    if (!WRITE_DECLARATIONS) return true;
    String name = getDeclName(decl.getSymbol());
    File file = getFile(name);
    try
    {
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      DocumentBuilder builder = factory.newDocumentBuilder();
      Document document = builder.newDocument(); 
      OMDoc omdoc = new OMDoc(document);
      Element node = omdoc.toNode(decl);
      document.appendChild(node);
      TransformerFactory tfactory = TransformerFactory.newInstance();
      Transformer transformer = tfactory.newTransformer();
      transformer.setOutputProperty(OutputKeys.INDENT, "yes");
      DOMSource source = new DOMSource(document);
      OutputStream out = 
        new GZIPOutputStream(
          new BufferedOutputStream(new FileOutputStream(file)));
      StreamResult result = new StreamResult(out);
      transformer.transform(source, result);
      out.close();
    }
    catch(IOException e)
    {
      State.setError("exception" + e.toString());
      return false;
    }
    catch(ParserConfigurationException e)
    {
      State.setError("exception" + e.toString());
      return false;
    }
    catch(TransformerConfigurationException e)
    {
      State.setError("exception" + e.toString());
      return false;
    }
    catch(TransformerException e)
    {
      State.setError("exception" + e.toString());
      return false;
    }
    return true;
  }
  
  /***************************************************************************
   * Read declaration from file.
   * @param file path to a file holding the declaration.
   * @return the declaration (null, if an error occured).
   **************************************************************************/
  public static Declaration readDeclaration(File file)
  {
    try
    {
      InputStream in = 
        new GZIPInputStream(new BufferedInputStream(new FileInputStream(file)));
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      factory.setNamespaceAware(true);
      DocumentBuilder builder = factory.newDocumentBuilder();
      Document document = builder.parse(in);
      if (document == null) return null;
      Node root = document.getFirstChild();
      if (root == null) return null;
      removeWhitespace(root);
      return toDeclaration(root);
    }
    catch(ParserConfigurationException e)
    {
      State.setError("exception " + e.getMessage());
      return null;
    }
    catch(IOException e)
    {
      State.setError("exception " + e.getMessage());
      return null;
    }
    catch(SAXException e)
    {
      State.setError("exception " + e.getMessage());
      return null;
    }
  }
  
  /***************************************************************************
   * Get file base name for symbol.
   * @param symbol a symbol.
   **************************************************************************/
  private static String getBaseName(Symbol symbol)
  {
    return getPrefix(symbol) + "_" + symbol.getIdentifier().getName();
  }
  
 
  /***************************************************************************
   * Get file name for content of symbol.
   * @param symbol a symbol.
   **************************************************************************/
  private static String getDeclName(Symbol symbol)
  {
    return getBaseName(symbol) + "_decl.xgz";
  }
  
  /***************************************************************************
   * Get file name for time stamp of symbol.
   * @param symbol a symbol
   **************************************************************************/
  private static String getStampName(Symbol symbol)
  {
    return getBaseName(symbol) + "_time.txt";
  }
  
  /***************************************************************************
   * Get file name for hash code of symbol
   * @param symbol a symbol
   **************************************************************************/
  private static String getHashName(Symbol symbol)
  {
    return getBaseName(symbol) + "_hash.txt";
  }
  
  /***************************************************************************
   * Get dependencies file name for symbol.
   * @param symbol a symbol
   **************************************************************************/
  private static String getDepName(Symbol symbol)
  {
    return getBaseName(symbol) + "_refs.xml";
  }
  
  /***************************************************************************
   * Get combined info file name for symbol.
   * @param symbol a symbol
   **************************************************************************/
  private static String getInfoName(Symbol symbol)
  {
    return getBaseName(symbol) + "_info.xml";
  }
  
  /***************************************************************************
   * Return unique prefix for symbol kind.
   * @return a string denoting the symbol kind
   **************************************************************************/
  private static String getPrefix(Symbol symbol)
  {
    if (symbol instanceof TypeSymbol)
      return "type";
    else if (symbol instanceof ValueSymbol)
      return "value";
    else if (symbol instanceof FormulaSymbol)
      return "formula";
    return null;
  }
  
  /***************************************************************************
   * Get proof base name for formula with denoted name.
   * @param name the name of a formula.
   **************************************************************************/
  private static String getProofBaseName(String name)
  {
    return "proof_" + name;
  }
  
  /***************************************************************************
   * Get proof file name for formula with denoted name.
   * @param symbol a symbol.
   **************************************************************************/
  private static String getProofDeclName(String name)
  {
    return getProofBaseName(name) + "_decl.xgz";
  }
  
  /***************************************************************************
   * Get proof dependencies file name for formula with denoted name
   * @param symbol a symbol
   **************************************************************************/
  private static String getProofDepName(String name)
  {
    return getProofBaseName(name) + "_refs.xml";
  }
  
  /***************************************************************************
   * Get proof file name for content of symbol.
   * @param symbol a symbol.
   **************************************************************************/
  private static String getProofDeclName(FormulaSymbol symbol)
  {
    return getProofDeclName(symbol.getIdentifier().getName());
  }
  
  /***************************************************************************
   * Get proof dependencies file name for symbol.
   * @param symbol a symbol
   **************************************************************************/
  private static String getProofDepName(FormulaSymbol symbol)
  {
    return getProofDepName(symbol.getIdentifier().getName());
  }
  
  /***************************************************************************
   * Convert DOM node to OpenMath object.
   * @param node the DOM node holding the OpenMath representation.
   * @return the corresponding OpenMath node.
   **************************************************************************/
  private static OMObject toOpenMath(Node node)
  {
    NewOMDOMReader reader = new NewOMDOMReader(node);
    try
    {
      return reader.readObject();
    }
    catch(IOException e)
    {
      State.setError("exception " + e.getMessage());
      return null;
    }
  }
  
  /***************************************************************************
   * Convert DOM node to expression.
   * @param node the DOM representation of the expression.
   * @return the expression (null, if an error occured).
   **************************************************************************/
  public static Expression toExpression(Node node)
  {
    OMObject om = toOpenMath(node);
    if (om == null) return null;
    return OpenMath.toExpression(om);
  }
  
  /***************************************************************************
   * Convert DOM node to type.
   * @param node the DOM representation of the type.
   * @return the type (null, if an error occured).
   **************************************************************************/
  public static Type toType(Node node)
  {
    OMObject om = toOpenMath(node);
    if (om == null) return null;
    return OpenMath.toType(om);
  }
  
  /***************************************************************************
   * Convert DOM node to identifier.
   * @param node the DOM representation of the identifier.
   * @return the type (null, if an error occured).
   **************************************************************************/
  public static Identifier toIdentifier(Node node)
  {
    OMObject om = toOpenMath(node);
    if (om == null) return null;
    return OpenMath.toIdentifier(om);
  }
  
  /***************************************************************************
   * Convert DOM node to declaration.
   * @param node the DOM representation of the declaration.
   * @return the declaration (null, if an error occured).
   **************************************************************************/
  public static Declaration toDeclaration(Node node)
  {
    return OMDoc.toDeclaration(node);
  }
  
  /***************************************************************************
   * Removes all text nodes that only contain whitespace from DOM tree.
   * @param node the root of the tree that is to be cleaned from whitespace.
   **************************************************************************/
  public static void removeWhitespace(Node node)
  {
    Vector<Node> empty = new Vector<Node>();
    NodeList children = node.getChildNodes();
    int n = children.getLength();
    for (int i=0; i<n; i++)
    {
      Node child = children.item(i);
      if (!(child instanceof Text))
      {
        removeWhitespace(child);
        continue;
      }
      String data = ((Text)child).getData();
      boolean iswhite = true;
      int l = data.length();
      for (int j=0; j<l; j++)
      {
        char ch = data.charAt(j);
        if (!Character.isWhitespace(ch))
        {
          iswhite = false;
          break;
        }
      }
      if (iswhite) empty.add(child);
    }
    Iterator<Node> iterator = empty.iterator();
    while (iterator.hasNext())
      node.removeChild((Node)iterator.next());
  }      
    
  /***************************************************************************
   * 
   * A cache of symbol information entries.
   *
   **************************************************************************/
  private final class Cache
  {
    private final class Entry
    {
      public long hash;
      public long stamp;
      public Map<Symbol,Long>dep;
      public Entry(long hash, long stamp, Map<Symbol,Long>dep)
      {
        this.hash = hash;
        this.stamp = stamp;
        this.dep = dep;
      }
    }
    
    private final static int MAX = 1000;
    private final Map<Symbol, Entry> map = new MapClass<Symbol, Entry>();
    private int count = 0; 
    
    /*************************************************************************
     * Get hash code of symbol.
     * @param symbol
     * @return the hash code (Entry.HASH_DEFAULT, if it could not be determined).
     ************************************************************************/
    public long getHash(Symbol symbol)
    {
      Entry entry = getEntry(symbol);
      return entry.hash;
    }
    
    /*************************************************************************
     * Get time stamp of symbol.
     * @param symbol
     * @return the time stamp (Entry.STAMP_DEFAULT, if it could not be determined).
     ************************************************************************/
    public long getStamp(Symbol symbol)
    {
      Entry entry = getEntry(symbol);
      return entry.stamp;
    }
    
    /*************************************************************************
     * Get dependencies of symbol.
     * @param symbol
     * @return the dependencies (empty, if they could not be determined).
     ************************************************************************/
    public Map<Symbol,Long>getDep(Symbol symbol)
    {
      Entry entry = getEntry(symbol);
      return entry.dep;
    }
    
    /*************************************************************************
     * Update cache from symbol declaration.
     * @param decl the declaration.
     * @return true iff everything went fine.
     ************************************************************************/
    public boolean write(Declaration decl)
    {
      Symbol symbol = decl.getSymbol();
      Entry entry = getEntry(symbol);
      entry.hash = Store.this.getHash(decl);
      entry.stamp = new Date().getTime();
      return true;
    }
    
    /*************************************************************************
     * write dependency information of symbol.
     * @param symbol the symbol.
     * @param referenced the collection of referenced symbols.
     * @return true iff everything went fine.
     ************************************************************************/
    public boolean writeDep(Symbol symbol, Collection<Symbol> referenced)
    {
      Entry entry = getEntry(symbol);
      entry.dep.clear();
      Iterator<Symbol> iterator = referenced.iterator();
      while (iterator.hasNext())
      {
        Symbol ref = (Symbol)iterator.next();
        long stamp = getStamp(ref);
        entry.dep.put(ref, new Long(stamp));
      }
      return write(symbol, entry);
    }
    
    /*************************************************************************
     * Get entry associated to symbol.
     * @param symbol the symbol.
     * @return the entry.
     ************************************************************************/
    private Entry getEntry(Symbol symbol)
    {
      Entry entry = (Entry)map.get(symbol);
      if (entry != null) return entry;
      if (count == MAX)
      {
        map.clear();
        count = 0;
      }
      entry = readInfo(symbol);
      map.put(symbol, entry);
      count = count+1;
      return entry;
    }
    
    /*************************************************************************
     * Write symbol entry information to store.
     * @param symbol the symbol.
     * @param entry the entry.
     * @return true iff everything went fine.
     ************************************************************************/
    private boolean write(Symbol symbol, Entry entry)
    {
      if (entry.hash == HASH_DEFAULT)
      {
        State.setError("invalid hash for symbol " + 
          symbol.getIdentifier() + " in " + directory);
        return false;
      }
      if (entry.stamp == STAMP_DEFAULT)
      {
        State.setError("invalid stamp for symbol " + 
            symbol.getIdentifier() + " in " + directory);
        return false;
      }
      File file = getFile(getInfoName(symbol));
      try
      {
        PrintWriter writer = new PrintWriter(new BufferedWriter(new FileWriter(file)));
        writer.println(entry.hash);
        writer.println(entry.stamp);
        Store.this.writeDep(writer, entry.dep.keySet().iterator());
        writer.close();
      }
      catch(IOException e)
      {
        State.setError("exception " + e.getMessage());
        return false;
      }
      return true;
    }
    
    /*************************************************************************
     * Get symbol entry from new version of information file.
     * @param symbol the symbol.
     * @return the entry (newly created, if information could not be found).
     ************************************************************************/
    private Entry readInfo(Symbol symbol)
    {
      File file = getFile(getInfoName(symbol));
      String hashString = null;
      String stampString = null;
      String depString = null;
      
      // determine strings from file, any may be null
      try
      {
        BufferedReader reader = new BufferedReader(new FileReader(file));
        hashString = reader.readLine();
        if (hashString != null)
        {
          stampString = reader.readLine();
          if (stampString != null)
          {
            StringBuffer buffer = new StringBuffer();
            while (true)
            {
              String line = reader.readLine();
              if (line == null) break;
              buffer.append(line);
            }
            depString = buffer.toString();
          }
        }
        reader.close();
      }
      catch(FileNotFoundException e)
      {
        // file not found, try to read info from old files
        return readInfoOld(symbol);
      }
      catch(IOException e)
      {
        // cannot read file
        return new Entry(HASH_DEFAULT, STAMP_DEFAULT, new MapClass<Symbol,Long>());
      }
      
      // convert strings to information
      long hash = HASH_DEFAULT;
      long stamp = STAMP_DEFAULT;
      try { hash = Long.parseLong(hashString); }
      catch(NumberFormatException e) { }
      try { stamp = Long.parseLong(stampString); }
      catch(NumberFormatException e) { }
      Map<Symbol,Long>dep = readDep(file, depString);
      
      // create result
      return new Entry(hash, stamp, dep);
    }
    
    /*************************************************************************
     * Get symbol entry from old information files.
     * @param symbol the symbol.
     * @return the entry (null, if information could not be found).
     ************************************************************************/
    private Entry readInfoOld(Symbol symbol)
    {
      long hash = getHashFromFile(symbol);
      long stamp = getStampFromFile(symbol);
      File file = getFile(getDepName(symbol));
      Map<Symbol,Long>dep = null;
      if (file.exists())
        dep = readDep(file, null);
      else
        dep = new MapClass<Symbol,Long>();
      Entry entry = new Entry(hash, stamp, dep);
      if (hash != HASH_DEFAULT && stamp != STAMP_DEFAULT)
        write(symbol, entry);
      return entry;
    }
  }
}