// ----------------------------------------------------------------------------
// Main.java
// Command line interface to ProofNavigator.
// $Id: Main.java,v 1.90 2017/07/04 08:04:35 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator;

import java.io.*;
import java.util.*;

import fmrisc.ProofNavigator.SWT.*;
import fmrisc.ProofNavigator.Commands.*;
import fmrisc.ProofNavigator.Proving.*;
import fmrisc.ProofNavigator.Syntax.*;
import fmrisc.ProofNavigator.Semantics.*;
import fmrisc.ProofNavigator.Communication.*;

/**
 * Command line interface to ProofNavigator.
 */
public final class Main 
{  
  // software version and message printed on startup
  private static final String version = fmrisc.ProgramExplorer.Main.version;
  private static final String copyright = fmrisc.ProgramExplorer.Main.copyright;
  
  // true if we are executed by the ProgramExplorer
  private static boolean programExplorer = false;
  
  // command line options
  private static boolean silent = false; // true iff no copyright information
  private static boolean nogui = false;  // true iff command line to be used
  private static String fileName = null; // file name to be read on startup
  private static String contextName = null; // name of context directory
  private static String cvclPath = null; // path to  prover executable
  private static State.ProverVersion proverVersion = null; // the prover version
  
  // values depending on current input Stream
  private static BufferedReader reader = null; // the reader
  private static PNLexer lexer = null;         // the lexer based on reader
  private static PNParser parser = null;       // the parser based on lexer
  
  // is command being processed?
  private static boolean running = false;
  
  /***************************************************************************
   * Command line interface to ProofNavigator.
   * @param args command line arguments (call with "--help" to see which)
   **************************************************************************/
  public static void main(String[] args) 
  { 
    // process arguments
    processArgs(args);
    
    // initialize state
    init();
    
    // by default, GUI interface is used
    if (!nogui) 
    {
      CommandServer server = new CommandServer();
      State.setServerPort(server.getPortNumber());
      try
      {
        MainSWT.run();
      }
      catch(NoClassDefFoundError e)
      {
        help(false);
        System.out.println("ERROR: Class not found: " + e.getMessage());
        System.out.println("ERROR: Please read the INSTALL notes.");
        System.out.println("ERROR: You may try \"ProofNavigator --nogui\".");
        System.exit(-1);
      }
    }
    
    // automatic flushing enabled for output stream
    State.setSystemOut(new PrintWriter(System.out, true));
    State.setSystemIn(new BufferedReader(new InputStreamReader(System.in)));
     
    // run interpreter
    run();
    
    // terminate program
    Main.exit(0);
  }
  
  /***************************************************************************
   * ProgramExplorer interface to ProofNavigator
   * @param top the Top instance used by the ProgramExplorer
   **************************************************************************/
  public static void mainInternal(Top top)
  {
    // initialize state
    programExplorer = true;
    init();
    
    // start command server
    CommandServer server = new CommandServer();
    State.setServerPort(server.getPortNumber());
    
    // create thread executing the interpreter
    Thread thread = new Thread(new Runnable() 
    {
      public void run() 
      {
        Main.run();
      }
    });
 
    // initialize GUI
    MainSWT.runInternal(top, thread);
    
    // start interpreter thread
    thread.start();
  }
  
  /****************************************************************************
   * Return version string.
   * @return the string.
   ***************************************************************************/
  public static String getVersion()
  {
    return version;
  }
  
  /****************************************************************************
   * Print help message.
   * @param exit iff true exit program after message
   ***************************************************************************/
  private static void help(boolean exit)
  {
    System.out.println(copyright);
    System.out.println("Usage: ProofNavigator [OPTION]... [FILE]");
    System.out.println("FILE:   name of file to be read on startup.");
    System.out.println("OPTION: one of the following options:");
    System.out.println("\t-n, --nogui:  use command line interface.");
    System.out.println("\t-c, --context NAME: use subdirectory NAME to store context.");
    System.out.println("\t--cpath PATH: PATH refers to checker executable (default is \"cvcl\").");
    System.out.println("\t--ctype TYPE: TYPE denotes checker type (default is \"CVCL\", others are \"CVC3\" and \"CVC4\")");
    System.out.println("\t-s, --silent: omit startup message.");
    System.out.println("\t-h, --help:   print this message.");
    if (exit) System.exit(0);
  }
  
  /****************************************************************************
   * Process arguments and set corresponding flags.
   * @param args the command line arguments.
   ***************************************************************************/
  private static void processArgs(String[] args)
  {
    int n = args.length;
    int i=0;
    while (i<n)
    {
      String arg = args[i];
      if (arg.equals("-n") || arg.equals("--nogui"))
        nogui = true;
      else if (arg.equals("-s") || arg.equals("--silent"))
        silent = true;
      else if (arg.equals("-h") || arg.equals("--help"))
        help(true);
      else if (arg.equals("-c") || arg.equals("--context"))
      {
        i = i+1;
        if (i == n) help(true);
        String dir = args[i];
        if (dir.charAt(0) == '-') help(true);
        contextName = dir;
      }
      else if (arg.equals("--cpath"))
      {
        i = i+1;
        if (i == n) help(true);
        String path = args[i];
        if (path.charAt(0) == '-') help(true);
        cvclPath = path;
      }
      else if (arg.equals("--ctype"))
      {
        i = i+1;
        if (i == n) help(true);
        String ctype = args[i];
        if (ctype.charAt(0) == '-') help(true);
        switch (ctype)
        {
          case "CVCL": 
            proverVersion = fmrisc.ProofNavigator.State.ProverVersion.CVCL;
            break;
          case "CVC3": 
            proverVersion = fmrisc.ProofNavigator.State.ProverVersion.CVC3;
            break;
          case "CVC4": 
            proverVersion = fmrisc.ProofNavigator.State.ProverVersion.CVC4;
            break;
          default:
            System.out.println("unsupported value of environment variable PE_CHECKER");
            System.out.println();
            help(true);
        }
      }
      else if (arg.charAt(0) != '-' && fileName == null)
        fileName = args[i];
      else
        help(true);
      i = i+1;
    }
  }
  
  /***************************************************************************
   * Initialize the system state.
   **************************************************************************/
  public static void init()
  {
    boolean okay = State.init(contextName, cvclPath, proverVersion);
    if (!okay)
    {
      System.out.println("Cannot initialize state: " + State.getError());
      System.out.println("Probably invalid checker path or working directory specified.");
      System.out.println("Start the program with option -h for help.");
      System.exit(-1);
    }
  }
  
  /***************************************************************************
   * Run ProofNavigator interpreter.
   * Assumes that input/output interface has been initialized.
   **************************************************************************/
  public static void run()
  {    
    // set standard system output
    PrintWriter out = State.getSystemOut();
    
    // copyright message
    if (!silent) out.println(copyright);

    // initially input is from stdin
    reader = State.getSystemIn();
    lexer = new PNLexer(reader);
    State.setLexer(lexer);
    parser = new PNParser(lexer);;
    
    // if there is a file, we switch to reading that file
    if (fileName != null) open(fileName);
    
    ProofState ostate = null;
    while (true)
    {
      Proof proof = State.getProof();
      if (proof == null)
        ostate = null;
      else
      {
        ProofState state = proof.getCurrent();
        if (state == null)
        {
          Identifier ident = proof.getSymbol().getIdentifier();
          out.println("Formula " + ident.toString() + " is proved. QED.");
          quit();
          ostate = null;
        }
        else
        {
          if (state != ostate)
            state.print(out);
          ostate = state;
          out.print("prove> ");
          out.flush();
        }
      }
      if (!nogui) MainSWT.setInputFocus();
      Command command = parse();
      String msg = State.getError();
      if (msg != null)
      {
        out.println("Syntax error: " + msg);
        State.resetError();
        continue;
      }
      if (command == null) 
      {
        if (reader == State.getSystemIn())
        {
          out.println("End of input stream.");
          break;
        }
        else
        {
          close(); 
          continue;
        }
      }
      running = true;
      blockGUI(true);
      command.process();
      blockGUI(false);
      running = false;
      String error = State.getError();
      if (error != null)
      {
        out.println("Error: " + error + ".");
        State.resetError();
      }
    }
    
    // terminate prover
    State.getProver().terminate(true);
  }
  
  /****************************************************************************
   * Parse next command.
   * @return next command parsed (null, if a syntax error occured).
   ***************************************************************************/
  public static Command parse()
  {
    try
    {
      if (isRunning()) blockGUI(false);
      Command command = parser.main();
      if (isRunning()) blockGUI(true);
      return command;
    }
    catch (antlr.ANTLRException e) 
    {
      PrintWriter out = State.getSystemOut();
      out.println("ANTLR exception: " + e);
      Main.exit(-1);
      return null;
    }
  }
  
  /****************************************************************************
   * Parse string as command.
   * @param string the string to be passed.
   * @return the command (null, if a syntax error occured).
   ***************************************************************************/
  public static Command parseString(String string)
  {
    try
    {
      BufferedReader reader = new BufferedReader(new StringReader(string));
      PNLexer lexer = new PNLexer(reader);
      PNParser parser = new PNParser(lexer);
      Command command = parser.main();
      return command;
    }
    catch (antlr.ANTLRException e) 
    {
      PrintWriter out = State.getSystemOut();
      out.println("ANTLR exception: " + e);
      Main.exit(-1);
      return null;
    }
  }
  
  /****************************************************************************
   * Terminate program with denoted exit code.
   * @param code the exit code
   ***************************************************************************/
  public static void exit(int code)
  {
    Prover p = State.getProver();
    if (p != null) p.terminate(true);
    System.exit(code);
  }
  
  /***************************************************************************
   * Quit (session or current proof) after asking user whether to do so.
   * Saves the proof in current store, if this is what the user wants.
   **************************************************************************/
  public static void quit()
  {
    PrintWriter out = State.getSystemOut();
    Proof proof = State.getProof();
    if (proof == null)
    {
      if (programExplorer)
      {
        out.println("Command \"quit\" not allowed outside proof.");
        return;
      }
      out.println("Quit session.");
      Main.exit(0);
    }
    if (!proof.isClosed())
    {
      boolean yes = ask("Do you really want to quit");
      if (!yes) return; 
    }
    Store store = State.getStore();
    boolean overwrite = true;
    if (store.existsProof(proof.getSymbol()))
      overwrite = ask("Save this proof and overwrite the previous one");
    if (overwrite) 
    {
      // set proof in symbol new
      FormulaSymbol symbol = proof.getSymbol();
      symbol.setProof(proof);
      proof.computeAbsolute();
      
      // recalculate status of other proofs
      Iterator<Symbol> iterator = 
        State.getEnvironment().getFormulaSymbols().iterator();
      while (iterator.hasNext())
      {
        FormulaSymbol s = (FormulaSymbol)iterator.next();
        if (s == symbol) continue;
        Proof p = s.getProof();
        if (p == null) continue;
        Environment e = s.getEnvironment();
        if (!e.getFormulaSymbols().contains(symbol)) continue;
        p.computeAbsolute();
      }
      
      // write proof
      boolean okay = store.write(proof);
      if (!okay) return;
      
      // generate index file
      File file = State.getPresenter().writeProofIndex(proof);
      if (file == null) return;
      
      out.println("Proof saved (browse file " + file.getName() + ").");
    }
    
    Identifier ident = proof.getSymbol().getIdentifier();
    out.println("Quit proof of formula " + ident.toString() + 
     " (use 'proof " + ident.toString() + "' to see proof).");
    State.setProof(null);
    
    // signal that proof is done
    proofDone(proof);
  }
  
  /***************************************************************************
   * Read line of input from console.
   * @return the line (null, if an error occured)
   *************************************************************************/
  public static String readLine()
  {
    try
    {
      State.getSystemOut().flush();
      if (isRunning()) blockGUI(false);
      State.skipNL();
      BufferedReader in = State.getSystemIn();
      String line = in.readLine();
      if (isRunning()) blockGUI(true);
      return line;
    }
    catch(IOException e)
    {
      State.setError("exception " + e.getMessage());
      return null;
    }
  }
  
  /***************************************************************************
   * Prompt a question and read the answer.
   * @param question the question.
   * @return the answer.
   **************************************************************************/
  public static String askText(String question)
  {
    if (!nogui) return MainSWT.askText(question);
    PrintWriter out = State.getSystemOut();
    out.print(question);
    out.flush();
    return readLine();
  }
  
  /***************************************************************************
   * Ask user for a number in the range a..b.
   * @param prompt the message printed for asking. 
   * @param a the lower bound of the number.
   * @param b the upper bound of the number.
   * @return number in range a..b or -1 (if no number was given)
   **************************************************************************/
  public static int readNumber(String prompt, int a, int b)
  {
    while (true)
    {
      String line = askText(prompt + " (" + a + ".." + b + ", -1 to abort): ");
      if (line == null) return -1;
      try
      {
        int pos = Integer.parseInt(line);
        if (pos == -1 || (a <= pos && pos <= b)) return pos;
      }
      catch (NumberFormatException e)
      { }
    }
  } 
  
  /***************************************************************************
   * Ask question to be answered by user with 'y' or 'n'
   * @param question the question asked
   * @return true iff answer was 'y'
   **************************************************************************/
  public static boolean ask(String question)
  {
    if (!nogui) return MainSWT.ask("Question", question);
    PrintWriter out = State.getSystemOut();
    State.skipNL();
    char ch = 0;
    do
    {
      String line = askText(question + " (y/n)?");
      if (line == null)
      {
        if (nogui)
        {  
          out.println("End of input stream.");
          Main.exit(0);
        }
        // what should we do else?
        Main.exit(0);
      }
      if (line.length() == 0) continue;
      ch = Character.toUpperCase(line.charAt(0));
    }
    while (ch != 'Y' && ch != 'N');
    return ch == 'Y';
  }
  
  /****************************************************************************
   * Signal current proof.
   * @param proof the current proof
   ***************************************************************************/
  public static void setProof(Proof proof)
  {
    // always leave last proof on screen
    // if (proof == null) return;
    if (!nogui) MainSWT.setProof(proof, true);
    if (proof == null) 
      setProofState(null);
    else if (proof.getCurrent() == null)
      setProofState(proof.getRoot());
    else
      setProofState(proof.getCurrent());
  }
  
  /****************************************************************************
   * Signal there is no current proof.
   ***************************************************************************/
  public static void resetProof()
  {
    if (!nogui) MainSWT.setProof(null, true);
    setProofState(null);
  }
  
  /****************************************************************************
   * Signal current proof state.
   * @param state the current proof state.
   ***************************************************************************/
  public static void setProofState(ProofState state)
  {
    if (state != null) State.getPresenter().writeProofState(state);
    if (!nogui) MainSWT.setProofState(state);
  }
  
  /****************************************************************************
   * Signal new declaration.
   * @param decl the new declaration.
   * @return true iff everything went okay.
   **************************************************************************/
  public static boolean newDeclaration(Declaration decl)
  {
    if (nogui) return true;
    boolean okay = State.getPresenter().writeDeclaration(decl);
    if (!okay) return false;
    displayDeclarations();
    return true;
  }
  
  /****************************************************************************
   * Display goal
   * @param decls the declaration list.
   * @param goal the goal declaration.
   * @return true iff everything went okay.
   **************************************************************************/
  public static boolean displayGoal(Declaration[] decls, 
    FormulaDeclaration goal)
  {
    if (nogui) return true;
    setVisible(true);
    Presenter presenter = State.getPresenter();
    boolean okay = presenter.newDeclarations();
    if (!okay) return false;
    for (Declaration decl : decls)
    {
      okay = presenter.writeDeclaration(decl);
      if (!okay) return false;
    }
    okay = presenter.writeDeclaration(goal);
    if (!okay) return false;
    displayDeclarations();
    return true;
  }
  
  /****************************************************************************
   * Signal that declarations are to be displayed.
   **************************************************************************/
  public static void displayDeclarations()
  {
    if (nogui) return;
    MainSWT.displayFile("Declarations", 
      State.getPresenter().getDeclFile(), false);
  }
  
  /***************************************************************************
   * Read command from file with denoted path.
   * @param path the (absolute or relative) path to the file.
   * @return true iff everything went okay.
   **************************************************************************/
  public static boolean open(String path)
  {
    try
    {
      Stack<File> fileStack = State.getFileStack();
      Stack<BufferedReader> readerStack = State.getReaderStack();
      File file = new File(path);
      String path0 = file.getCanonicalPath();
      int n = fileStack.size();
      for (int i=0; i<n; i++)
      {
        File file0 = (File)fileStack.elementAt(i);
        if (path0.equals(file0.getCanonicalPath()))
        {
          State.setError("file " + path + " must not read itself recursively");
          return false;
        }
      }
      fileStack.push(file);
      reader = new BufferedReader(new FileReader(file));
      readerStack.push(reader);
      lexer = new PNLexer(reader);
      State.setLexer(lexer);
      parser = new PNParser(lexer);  
    }
    catch(FileNotFoundException e)
    {
      State.setError("error in opening file: " + e.getMessage());
      State.getFileStack().pop();
      return false;
    }
    catch(IOException e)
    {
      State.setError("error in opening file: " + e.getMessage());
      State.getFileStack().pop();
      return false;
    }
    return true;
  }
  
  /***************************************************************************
   * Close current input stream and switch to previous stream.
   **************************************************************************/
  private static void close()
  {
    Stack<File> fileStack = State.getFileStack();
    Stack<BufferedReader> readerStack = State.getReaderStack();
    if (!fileStack.isEmpty())
    {
      File file = (File)fileStack.pop();
      State.getSystemOut().println("File " + file.getName() + " read.");
    }
    if (!readerStack.isEmpty())
      readerStack.pop();
    if (readerStack.isEmpty())
      reader = State.getSystemIn();
    else
      reader = (BufferedReader)readerStack.peek();
    lexer = new PNLexer(reader);
    State.setLexer(lexer);
    parser = new PNParser(lexer); 
  }
  
  /***************************************************************************
   * Set context directory.
   * @param path the (absolute or relative) path to context directory
   *        (if null, a default path is used).
   * @return true iff everything went okay.
   **************************************************************************/
  public static boolean setContext(String path)
  {
    boolean okay = State.init(path, cvclPath, proverVersion);
    displayDeclarations();
    resetProof();
    return okay;
  }
  
  /***************************************************************************
   * Signal to GUI (blocking interaction) that command is being processed.
   * @param blocked true iff command is being processed.
   **************************************************************************/
  private static void blockGUI(boolean blocked)
  {
    if (!nogui) MainSWT.signalRunning(blocked);
  }
  
  /***************************************************************************
   * Signal whether command is being processed.
   * @return true iff command is being processed.
   *************************************************************************/
  public static boolean isRunning()
  {
    return running;
  }
  
  /***************************************************************************
   * Make ProofNavigator visible or invisible.
   * @param visible the requested visibility status.
   **************************************************************************/
  private static void setVisible(boolean visible)
  {
    if (nogui) return;
    MainSWT.setVisible(visible);
  }
  
  /***************************************************************************
   * Signal that proof is completed.
   * @param proof the completed proof.
   **************************************************************************/
  public static void proofDone(Proof proof)
  {
    displayDeclarations();
    if (programExplorer) fmrisc.ProgramExplorer.Main.proofDone(proof);
  }
  
  /***************************************************************************
   * Set various properties.
   * @param context the context directory.
   * @param cvcl the prover command path.
   * @param version the prover version.
   * @param firstOrder true if logic is first order
   * @param toLogical true if formula type is LOGICAL.
   **************************************************************************/
  public static void setProperties(String context, String cvcl, 
    State.ProverVersion version, boolean firstOrder, boolean toLogical)
  {
    contextName = context;
    cvclPath = cvcl;
    proverVersion = version;
    State.setFirstOrderCheck(firstOrder);
    State.setBooleanAsLogical(toLogical);
  }
}
