// ----------------------------------------------------------------------------
// MainSWT.java
// Main method for ProofNavigator with SWT-based GUI.
// $Id: MainSWT.java,v 1.37 2017/07/04 08:04:35 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.SWT;

import org.eclipse.swt.*;
import org.eclipse.swt.widgets.*;
import org.eclipse.swt.graphics.*;

import fmrisc.ProofNavigator.*;
import fmrisc.ProofNavigator.Commands.*;
import fmrisc.ProofNavigator.Proving.*;

import java.io.*;

/**
 * 
 * Main method for ProofNavigator with SWT-based GUI.
 *
 */
public final class MainSWT
{
  // the interpreter thread
  private static Thread thread;
  
  // flag to indicate restart after shell is closed
  private static boolean restart;

  // application shell
  private static Shell shell;
  
  // the current window content
  private static Top top;
  
  // various colors
  private static Color openColor;       // color of open tree nodes
  private static Color closedColor;     // color of closed tree nodes
  private static Color currentColor;    // color of current tree node
  private static Color backgroundColor; // background color
  
  // the application icon
  private static String riscImageName = "risc-logo-transparent.png";
  private static Image riscImage = null;
  
  /****************************************************************************
   * Run program after command line processing.
   **************************************************************************/
  public static void run()
  {
    Display display = new Display(); 
    openColor = new Color(display, new RGB(255, 0, 63));
    closedColor = new Color(display, new RGB(0, 0, 255));
    currentColor = new Color(display, new RGB(0, 255, 0));
    backgroundColor = new Color(display, new RGB(255, 255, 255));
    do
    {
      restart = false;
      Main.init();
      runShell(display);
    } 
    while (restart);
    Main.exit(0);
  }

  /****************************************************************************
   * ProgramExplorer interface to ProofNavigator
   * @param top the Top instance used by the ProgramExplorer
   * @param t the interpreter thread
   **************************************************************************/
  public static void runInternal(Top top, Thread t)
  {
    Display display = top.getDisplay();
    openColor = new Color(display, new RGB(255, 0, 63));
    closedColor = new Color(display, new RGB(0, 0, 255));
    currentColor = new Color(display, new RGB(0, 255, 0));
    backgroundColor = new Color(display, new RGB(255, 255, 255));
    restart = false;
    thread = t;
    shell = top.getShell();
    MainSWT.top = top;
  }
  
  /****************************************************************************
   * Create shell for running the program in denoted display.
   * @param Display the display used for the shell
   **************************************************************************/
  private static void runShell(Display display)
  {  
    // create shell
    shell = new Shell(display, 
      SWT.CLOSE | SWT.MIN | SWT.MAX | SWT.RESIZE | SWT.TITLE | SWT.BORDER);
    resize();
    shell.setText("RISC ProofNavigator");
    riscImage = new Image(shell.getDisplay(),
        Main.class.getResourceAsStream("/" + riscImageName));
    shell.setImage(riscImage);
    top = new Top(shell, 0);
    top.pack();
    
    // start thread executing the interpreter
    thread = new Thread(new Runnable() 
    {
      public void run() 
      {
        Main.run();
      }
    });
    thread.start();
    
    // close splash screen
    java.awt.SplashScreen screen = java.awt.SplashScreen.getSplashScreen();
    if (screen != null) screen.close();
    
    // open shell and process events
    shell.open();
    while (!shell.isDisposed())
    {
      if (!display.readAndDispatch())
        display.sleep();
    }
    
    // give interpreter thread time to terminate
    // this should leave no external process running
    // while (thread.isAlive()) Thread.yield();
  }
  
  /****************************************************************************
   * Get interpreter thread
   * @return the thread
   ***************************************************************************/
  public static Thread getThread()
  {
    return thread;
  }
  
  /****************************************************************************
   * Set signal whether program is to be restarted.
   * @param signal true iff program is to be restarted.
   ***************************************************************************/
  public static void restart(boolean signal)
  {
    MainSWT.restart = signal;
  }
  
  /****************************************************************************
   * Check whether program is to be restarted.
   * @return true if program is to be restarted.
   ***************************************************************************/
  public static boolean isRestarted()
  {
    return MainSWT.restart;
  }
  
  /****************************************************************************
   * Set current proof.
   * @param proof the current proof.
   * @param enabled true if the proof is to be enabled.
   ***************************************************************************/
  public static void setProof(Proof proof, boolean enabled)
  {
    top.setProof(proof, enabled);
    updateMenu();
  }
  
  /****************************************************************************
   * Signal current proof state.
   * @param state the current proof state.
   ***************************************************************************/
  public static void setProofState(ProofState state)
  {
    top.setProofState(state);
  }
  
  /****************************************************************************
   * Display file.
   * @param title the title of the display.
   * @param file the file to be displayed.
   * @param state true iff file displays proof state.
   ***************************************************************************/
  public static void displayFile(String title, File file, boolean state)
  {
    top.displayFile(title, file, state);
  }
  
  /****************************************************************************
   * Get color for open proof tree nodes.
   * @return the color.
   ***************************************************************************/
  public static Color getOpenColor()
  {
    return openColor;
  }
  
  /****************************************************************************
   * Get color for closed proof tree nodes.
   * @return the color.
   ***************************************************************************/
  public static Color getClosedColor()
  {
    return closedColor;
  }
  
  /****************************************************************************
   * Get color for current proof node.
   * @return the color.
   ***************************************************************************/
  public static Color getCurrentColor()
  {
    return currentColor;
  }
  
  /****************************************************************************
   * Get background color.
   * @return the color.
   ***************************************************************************/
  public static Color getBackgroundColor()
  {
    return backgroundColor;
  }
  
  /****************************************************************************
   * Signals that prover is running in the background.
   * @param running true if running starts, false if running stops.
   ***************************************************************************/
  public static void signalRunning(boolean running)
  {
    // may be invoked by interpreter after termination
    if (top.isDisposed()) return;
    top.signalRunning(running);
  }
  
  /****************************************************************************
   * Updates menu depending on current system state.
   ***************************************************************************/
  public static void updateMenu()
  {
    // may be invoked by interpreter after termination
    if (top.isDisposed()) return;
    top.updateMenu();
  }
  
  /****************************************************************************
   * Resize shell based on font setting
   ***************************************************************************/
  public static void resize()
  {
    if (State.useBigFont())
      shell.setSize(1200, 900);
    else
      shell.setSize(1200, 900);
  }
  
  /****************************************************************************
   * Resize font (from small to big and vice versa)
   ***************************************************************************/
  public static void resizeFont()
  {
    top.resizeFont();
  }
  
  /***************************************************************************
   * Process command denoted by string.
   * @param cstring the string denoting the command
   **************************************************************************/
  public static void process(String cstring)
  {
    top.process(cstring);
  }
  
  /***************************************************************************
   * Process command.
   * @param command the command to be processed.
   **************************************************************************/
  public static void process(Command command)
  {
    process(command.toString());
  }
  
  /***************************************************************************
   * Set focus to input field.
   **************************************************************************/
  public static void setInputFocus()
  {
    top.setInputFocus();
  }
  
  /****************************************************************************
   * Enter input denoted by string
   * @param string the input to be added.
   **************************************************************************/
  public static void enter(final String string)
  {
    top.getDisplay().syncExec(new Runnable() {
      public void run() {
        top.enter(string);
      }});
    setInputFocus();
  }
  
  /***************************************************************************
   * Make ProofNavigator visible or invisible.
   * @param visible the requested visibility status.
   **************************************************************************/
  public static void setVisible(boolean visible)
  {
    top.setVisible(visible);
  }
  
  // a question that is answered with yes or no
  private static class YesNoQuestion implements Runnable
  {
    private String question;
    public boolean answer;
    public YesNoQuestion(String question)
    {
      this.question = question;
    }
    public void run()
    {
      answer = top.ask("Question", question + "?");
    }
  }
  
  /***************************************************************************
   * Open dialog to ask yes/no question.
   * @param title the dialog title
   * @param question the question asked
   * @return true iff question is answered positively.
   **************************************************************************/
  public static boolean ask(String title, String question)
  {
    YesNoQuestion q = new YesNoQuestion(question);
    top.getDisplay().syncExec(q);
    return q.answer;
  }
  
  // a question that is answered by text
  private static class TextQuestion implements Runnable
  {
    private String question;
    public String answer;
    public TextQuestion(String question)
    {
      this.question = question;
    }
    public void run()
    {
      answer = top.askText("Question", question);
    }
  }
  
  /***************************************************************************
   * Open dialog to ask yes/no question.
   * @param question the question asked
   * @return the answer to the question;
   **************************************************************************/
  public static String askText(String question)
  {
    TextQuestion q = new TextQuestion(question);
    top.getDisplay().syncExec(q);
    // shell may be closed, don't return answer any more in current thread
    while (q.answer == null) Thread.yield();
    return q.answer;
  }
}
