// ----------------------------------------------------------------------------
// ProofTreeItem.java
// Graphical representation of proof state.
// $Id: ProofTreeItem.java,v 1.13 2011/03/31 16:55:06 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.SWT;

import fmrisc.ProofNavigator.Commands.*;
import fmrisc.ProofNavigator.Proving.*;

import org.eclipse.swt.*;
import org.eclipse.swt.widgets.*;
import org.eclipse.swt.graphics.*;

/**
 * 
 * Graphical representation of proof state.
 *
 */
public final class ProofTreeItem implements ProofStateListener
{
  private ProofState state; // the state
  private TreeItem item;    // its visualization
  
  /****************************************************************************
   * Set tree to visualize the state tree with the denoted root state.
   * @param tree the tree.
   * @param root the root state.
   ***************************************************************************/
  public static void setNodes(Tree tree, ProofState root)
  {
    ProofTreeItem rootItem = new ProofTreeItem(tree, root);
    rootItem.addChildren(root);
    rootItem.changeAncestors();
  }
  
  /***************************************************************************
   * Return tree item associated to item.
   * @return the associated tree item.
   **************************************************************************/
  public TreeItem getTreeItem()
  {
    return item;
  }
  
  /***************************************************************************
   * Return proof state associated to item.
   * @return the associated proof state.
   **************************************************************************/
  public ProofState getState()
  {
    return state;
  }
  
  /***************************************************************************
   * Return state label for display.
   * @param state the state to be displayed.
   * @return its label
   **************************************************************************/
  private static String getLabel(ProofState state)
  {
    String label = "[" + state.getLabel() + "]";
    Command command = state.getCommand();
    if (command != null) label = label + ": " + command.toString();
    return label;
  }
  
  /****************************************************************************
   * Construct tree item for denoted parent component.
   * @param tree the tree of which this node is the root.
   * @param state the state that is to be visualized.
   ***************************************************************************/
  private ProofTreeItem(final Tree tree, ProofState state)
  {
    this.state = state;
    state.setListener(this);
    final String label = getLabel(state);
    tree.getDisplay().syncExec(new Runnable() {
      public void run() {
        item = new TreeItem(tree, SWT.NULL);
        item.setText(label);
        tree.showItem(item);
      }
    });
  }
  
  /***************************************************************************
   * Get color of item (openColor or closedColor)
   * @return the color
   **************************************************************************/
  private Color getColor()
  {
    Color color = item.getForeground();
    if (color.equals(MainSWT.getBackgroundColor()))
      color = item.getBackground();
    return color;
  }
  
  /***************************************************************************
   * Auxiliary class to set the item to its color and remember that color.
   **************************************************************************/
  private class SetColor implements Runnable
  {
    private boolean current;
    private boolean isClosed;
    private Color color;
    
    public SetColor(boolean current, boolean isClosed)
    {
      this.current = current;
      this.isClosed = isClosed;
    }
    
    public void run() {
      Color openColor = MainSWT.getOpenColor();
      Color closedColor = MainSWT.getClosedColor();
      Color backgroundColor = MainSWT.getBackgroundColor();
      item.getParent().showItem(item);
      if (current)
      {
        color = openColor;
        item.setForeground(backgroundColor);
        item.setBackground(color);
        return;
      }
      item.setBackground(backgroundColor);
      boolean closed = isClosed;
      if (closed)
      {
        int n = item.getItemCount();
        for (int i=0; i<n; i++)
        {
          TreeItem item0 = item.getItem(i);
          if (item0.getForeground().equals(openColor) ||
              item0.getBackground().equals(openColor))
          {
            closed = false;
            break;
          }
        }
      }
      if (closed)
        color = closedColor;
      else
        color = openColor;
      item.setForeground(color);
    }
  }
  
  /****************************************************************************
   * Add to item the items visualizing the children of the denoted state.
   * @param state the state whose children are to be visualized.
   * @return the color of the item (openColor or closedColor)
   ***************************************************************************/
  private Color addChildren(ProofState state)
  {
    ProofState[] children = state.getChildren();
    if (children != null)
    {
      int n = children.length;
      for (int i=0; i<n; i++)
      {
        ProofState child = children[i];
        ProofTreeItem childItem = new ProofTreeItem(this, child);
        childItem.addChildren(child);
      }
    }
    
    final boolean current = state == state.getProof().getCurrent();
    final boolean isClosed = state.isClosed(); 
    SetColor setColor = new SetColor(current, isClosed);
    item.getDisplay().syncExec(setColor);
    // speed up proof replay, do here if not in ProofState.hasChanged()
    // State.getPresenter().writeProofState(state);
    return setColor.color;
  }
  
  /****************************************************************************
   * Construct tree item for denoted parent component.
   * @param parent the item to which the new items are to be added
   * @param state the state that is to be visualized.
   ***************************************************************************/
  private ProofTreeItem(final ProofTreeItem parent, final ProofState state)
  {
    this.state = state;
    state.setListener(this);
    final String label = getLabel(state);
    parent.item.getDisplay().syncExec(new Runnable() {
      public void run() {
        item = new TreeItem(parent.item, SWT.NULL);
        item.setText(label);
        parent.item.setExpanded(true);
        parent.item.getParent().showItem(item);
      }
    });
  }
  
  /***************************************************************************
   * Auxiliary class to reset an item and remember its previous color.
   **************************************************************************/
  private final class Reset implements Runnable
  {
    Color color;
    public void run()
    {
      color = getColor();
      String label = getLabel(state);
      item.removeAll();
      item.setText(label);
      item.getParent().setSelection(new TreeItem[] { });
      item.getParent().showItem(item);
    }
  }
  
  /***************************************************************************
   * Signal that proof state has changed.
   **************************************************************************/
  public void hasChanged()
  {
    if (item.isDisposed()) return;
    Reset reset = new Reset();
    item.getDisplay().syncExec(reset);
    Color color = addChildren(state);
    if (!color.equals(reset.color))
      changeAncestors();
    MainSWT.setInputFocus();
  }
  
  /****************************************************************************
   * Recompute state colors of ancestor nodes. 
   ***************************************************************************/
  private void changeAncestors()
  {
    if (item.isDisposed()) return;
    final boolean closed = state.isClosed();
    item.getDisplay().syncExec(new Runnable() {      
      public void run() {
        Color openColor = MainSWT.getOpenColor();
        Color closedColor = MainSWT.getClosedColor();
        TreeItem item0 = item.getParentItem();
        while (item0 != null)
        {
          if (closed)
          {
            if (item0.getForeground().equals(closedColor)) return;
            int n = item0.getItemCount();
            for (int i=0; i<n; i++)
            {
              if (!item0.getItem(i).getForeground().equals(closedColor)) return;
            }
            item0.setForeground(closedColor);
          }
          else
          {
            if (item0.getForeground().equals(openColor)) return;
            item0.setForeground(openColor);
          }
          item0 = item0.getParentItem();
        }
      }
    }); 
  }
}
