// ----------------------------------------------------------------------------
// ValueSymbol.java
// Symbols in ValueTable.
// $Id: ValueSymbol.java,v 1.4 2017/07/04 08:04:35 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Semantics;

import fmrisc.ProofNavigator.Syntax.*;

/**
 * 
 * Symbols in ValueTable.
 *
 */
public final class ValueSymbol extends SymbolBase
{
  private Type type;                 // the type
  private Expression value;          // the value
  private TypeExpression clearType;  // handling of subtypes, subranges, NAT
  private int depth = -1;            // the environment depth
  private int varNumber = -1;        // the variable number (local vars only)      
  
  /****************************************************************************
   * Construct value symbol from ident (type and value are both null).
   * @param ident the identifier of the symbol.
   ***************************************************************************/
  public ValueSymbol(ValueDeclIdentifier ident) 
  {
    super(ident);
  }
  
  /****************************************************************************
   * Return identifier of symbol.
   * @return the symbol identifier
   ***************************************************************************/
  public ValueDeclIdentifier getValueDeclIdentifier()
  {
    return (ValueDeclIdentifier)getIdentifier();
  }
  
  /****************************************************************************
   * Return type of symbol.
   * @return the symbol type
   ***************************************************************************/
  public Type getType()
  {
    return type;
  }
  
  /****************************************************************************
   * Return value of symbol
   * @return the symbol value (may be null)
   ***************************************************************************/
  public Expression getValue()
  {
    return value;
  }
  
  /****************************************************************************
   * Return environment depth of symbol.
   * @return the environment depth
   ***************************************************************************/
  public int getDepth()
  {
    return depth;
  }
  
  /****************************************************************************
   * Return variable number of symbol.
   * @return the variable number (only if depth gt 0)
   ***************************************************************************/
  public int getVarNumber()
  {
    return varNumber;
  }
  
  /****************************************************************************
   * Set type of symbol.
   * @param type the symbol type
   ***************************************************************************/
  public void setType(Type type)
  {
    this.type = type;
  }
   
  /****************************************************************************
   * Set value of symbol.
   * @param value the symbol value 
   ***************************************************************************/
  public void setValue(Expression value)
  {
    this.value = value;
  }
  
  /****************************************************************************
   * Set environment depth of symbol.
   * @param depth the symbol environment depth
   ***************************************************************************/
  public void setDepth(int depth)
  {
    this.depth = depth;
  }
  
  /****************************************************************************
   * Set variable number of symbol.
   * @param varNumber variable number of symbol
   ***************************************************************************/
  public void setVarNumber(int varNumber)
  {
    this.varNumber = varNumber;
  }
  
  /****************************************************************************
   * Set clear type to typexp.
   * @param clearType the type/expression pair representing the cleared type
   ***************************************************************************/
  public void setClearType(TypeExpression clearType)
  {
    this.clearType = clearType;
  }
  
  /****************************************************************************
   * Get clear type.
   * @return the type/expression pair representing the cleared type
   ***************************************************************************/
  public TypeExpression getClearType()
  {
    return clearType;
  }
  
  /***************************************************************************
   * Instantiate symbol with copy of denoted symbol.
   * The identifier is *not* copied from the denoted symbol
   * also the cleartype is not yet instantiated.
   * @param symbol the symbol from which the instantiation is derived
   **************************************************************************/
  public void instantiate(ValueSymbol symbol)
  {
    Type type = symbol.getType();
    if (type != null) setType(Checking.canonical(type.instantiateType()));
    Expression value = symbol.getValue();
    if (value != null) setValue(value.instantiate());
    setDepth(symbol.getDepth());
    setVarNumber(symbol.getVarNumber());
  }
}
