// ----------------------------------------------------------------------------
// ASTUtil.java
// Execution context for the proof navigator.
// $Id: State.java,v 1.52 2012/10/29 13:27:42 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator;

import fmrisc.ProofNavigator.Semantics.*;
import fmrisc.ProofNavigator.Syntax.*;
import fmrisc.ProofNavigator.Proving.*;
import fmrisc.ProofNavigator.Proving.CVCL.*;
import fmrisc.ProofNavigator.Communication.*;

import java.io.*;
import java.util.*;

/**
 * 
 * Execution context for the proof navigator.
 * 
 */
public final class State
{
  // default name of context directory
  private static final String contextDirNameDefault = ".ProofNavigator";
  private static String contextDirName;
  
  // default name of cvcl (cvc3, cvc4) path
  private static final String cvclPathDefault = "cvcl";
  private static String cvclPath;

  // the version of the prover used
  public enum ProverVersion { CVCL, CVC3, CVC4 }
  private static final ProverVersion proverVersionDefault = ProverVersion.CVCL;
  private static ProverVersion proverVersion;
  
  // the name of the base mode log file
  private static final String logFileName = "cvcl.log";
  
  // the name of the proof mode log file
  private static final String logFileName0 = "cvcl0.log";
  
  // the prefixes and namespace URIs used in generated XML documents
  public final static String NamespacePrefix = "fmrisc";
  public final static String NamespaceURI = 
    "http://www.risc.uni-linz.ac.at/FM-RISC";
  
  public final static String OpenMathPrefix = "om";
  public final static String OpenMathURI =
    "http://www.openmath.org/OpenMath";
  
  public final static String OMDocPrefix = "omdoc";
  public final static String OMDocURI =
    "http://www.mathweb.org/omdoc";
  
  public final static String XMLPrefix = "xml";
  public final static String XMLURI = 
    "http://www.w3.org/XML/1998/namespace";
  
  // port used by command server
  private static int serverPort;
  
  // stack of File/BufferedReader objects representing open files
  private static Stack<File> fileStack = new Stack<File>();
  private static Stack<BufferedReader> readerStack = new Stack<BufferedReader>();
   
  // writer as replacement for stream, automatic flushing enabled
  private static PrintWriter systemOut;
  
  // reader to read user commands from
  private static BufferedReader systemIn;
  
  // type checking conditons generated in last declaration
  private static Expression lastTCC;
  
  // lexer in use
  private static PNLexer lexer; 
  
  // whether lexer shall return labels rather than idents
  private static boolean lexerLabels;
  
  // error message (null signals no error)
  private static String error;  
  
  // the global environment used by the proof navigator
  private static Environment global;
  
  // the environment currently in use (different from global during proofs)
  private static Environment environment;
  
  // the context used to collect type checking conditions (tccs)
  private static Context context;
    
  // if alternative context is used, the old context
  private static Context oldContext;
  
  // the directory for the store
  private static File contextDir; 
  
  // the global default prover
  private static Prover prover;
  
  // if alternative prover is used, the old prover
  private static Prover oldProver;
  
  // the currently running prover
  private static Prover runningProver;
  
  // the current proof
  private static Proof proof;
  
  // the current store
  private static Store store;
  
  // the current presenter
  private static Presenter presenter;
  
  // the log file for the base prover
  private static PrintWriter log;
  
  // the log file for the proof state prover
  private static PrintWriter log0;
  
  // the formula mask currently used 
  private static CVCLFormulaMask formulaMask;
  
  // prover has been aborted
  private static boolean aborted;
  
  // children of current state during proof replay
  private static ProofState[] replayChildren;
  
  // big fonts shall be displayed (remains set across restarts)
  private static boolean bigFont = false;

  // automatic proof state simplification is on (remains set across restarts)
  private static boolean autoSimplify = true ;
  
  // check value types to be first-order
  private static boolean firstOrderCheck = false;
      
  // externally represent type BOOLEAN as LOGICAL (and type BIT as BOOLEAN)
  private static boolean booleanAsLogical = false;
  
  /***************************************************************************
   * Initialize prover state using denoted context.
   * @param name the name (path) of the context 
   *        (if null, default name is used)
   * @param cvcl the name (path) to the prover executable 
   *        (if null, default name and type is used)
   * @param version the version of the prover used 
   *        (ignored, if cvcl is null)
   * @return true iff everything went okay
   **************************************************************************/
  public static boolean init(String name, String cvcl, ProverVersion version)
  {
    return init(name, cvcl, version, firstOrderCheck, booleanAsLogical);
  }
  
  /***************************************************************************
   * Initialize prover state using denoted context.
   * @param name the name (path) of the context 
   *        (if null, default name is used)
   * @param cvcl the name (path) to the prover executable 
   *        (if null, default name and type is used)
   * @param version the version of the prover used 
   *        (ignored, if cvcl is null)
   * @param firstOrder true, if logic is first order (must be true for CVCL)
   * @param toLogical true, if formula type is LOGICAL
   * @return true iff everything went okay
   **************************************************************************/
  public static boolean init(String name, String cvcl, ProverVersion version,
    boolean firstOrder, boolean toLogical)
  {
    if (name == null) 
      contextDirName = contextDirNameDefault;
    else
      contextDirName = name;
    if (cvcl == null)
    {
      cvclPath = cvclPathDefault;
      proverVersion = proverVersionDefault;
    }
    else
    {
      cvclPath = cvcl;
      proverVersion = version;
    }
    boolean okay = setContextDir(new File(contextDirName));
    if (!okay) return false;
    if (prover != null) prover.terminate(true);
    if (proverVersion == ProverVersion.CVCL)
      firstOrderCheck = firstOrder;
    else
      firstOrderCheck = true;
    booleanAsLogical = toLogical;
    prover = newProver(log, 0, false);
    if (!prover.isOkay())
    {
      State.setError("could not start prover: " + prover.getError());
      return false;
    }
    global = new Environment();
    environment = global;
    context = new Context();
    proof = null;
    aborted = false;
    replayChildren = null;
    Checking.init();
    return true;
  }
  
  /***************************************************************************
   * Get current context directory.
   * @return the current context directory.
   **************************************************************************/
  public static File getContextDir()
  {
    return contextDir;
  }
  
  /***************************************************************************
   * Set context directory to denoted directory (creating it, if necessary).
   * @param dir directory where all generated files are stored.
   * @return true iff directory exists or can be generated and is writable
   **************************************************************************/
  public static boolean setContextDir(File dir)
  {
    if (dir.exists())
    {
      if (!dir.isDirectory())
      {
        State.setError("file " + dir.getAbsolutePath() + 
          " conflicts context directory");
        return false;
      }
    }
    else
    {
      boolean okay = dir.mkdir();
      if (!okay)
      {
        State.setError("cannot create context directory " + 
          dir.getAbsolutePath());
        return false;
      }
    }
    if (!dir.canWrite())
    {
      State.setError("cannot write to context directory " + 
        dir.getAbsolutePath());
      return false;
    }
    contextDir = dir;
    store = new Store(contextDir);
    presenter = new Presenter(contextDir);
    try
    {
      if (log != null) log.close();
      log = new PrintWriter(new BufferedWriter(new FileWriter(
        new File(contextDir, logFileName))));
      if (log0 != null) log0.close();
      log0 = new PrintWriter(new BufferedWriter(new FileWriter(
        new File(contextDir, logFileName0))));
    }
    catch(IOException e)
    {
      State.setError("Cannot create log file: " + e.getMessage());
      return false;
    }
    return true;
  }
  
  /***************************************************************************
   * get new instance of a prover
   * @param log stream for prover to write log messages to (null, if none)
   * @param timeout timeout for prover in seconds (0, if none)
   * @param mask true iff quantified formulas are to be masked
   * @return a new prover instance
   **************************************************************************/
  public static Prover newProver(PrintWriter log, int timeout, boolean mask)
  {
    String command;
    if (proverVersion == ProverVersion.CVC4)
      command = cvclPath + " --interactive --incremental --produce-models --simplification=incremental --tlimit=" + timeout;
    else
      command = cvclPath + " +interactive -indent -timeout " + timeout;
    return new CVCL(command, proverVersion, log, mask);
  }
  
  /***************************************************************************
   * get standard output stream
   * @return standard output stream used by system
   **************************************************************************/
  public static PrintWriter getSystemOut()
  {
    return systemOut;
  }
  
  /***************************************************************************
   * set standard output stream
   * @param out standard output stream used by system
   **************************************************************************/
  public static void setSystemOut(PrintWriter out)
  {
    systemOut = out;
  }
  
  /***************************************************************************
   * get standard input stream
   * @return standard input stream used by system
   **************************************************************************/
  public static BufferedReader getSystemIn()
  {
    return systemIn;
  }

  /***************************************************************************
   * set standard input stream
   * @param in standard input stream used by system
   **************************************************************************/
  public static void setSystemIn(BufferedReader in)
  {
    systemIn = in;
  }
  
  /***************************************************************************
   * Skip new line from input stream
   **************************************************************************/
  public static void skipNL()
  {
    try
    {
      if (!systemIn.ready()) return;
      systemIn.mark(1);
      int ch = systemIn.read();
      if (ch == '\n') return;
      systemIn.reset();
    }
    catch(IOException e)
    {
      System.out.println("Exception " + e.getMessage() + ".");
      System.exit(-1);
    }
  }
  
  /***************************************************************************
   * set last type checking condition
   * @param tcc the type checking condition generated last
   **************************************************************************/
  public static void setLastTCC(Expression tcc)
  {
    State.lastTCC = tcc;
  }
  
  /***************************************************************************
   * get last type checking condition
   * @return the type checking condition generated last
   **************************************************************************/
  public static Expression getLastTCC()
  {
    return lastTCC;
  }
  
  /***************************************************************************
   * set lexer used by proof navigator
   * @param lexer used by the proof navigator
   **************************************************************************/
  public static void setLexer(PNLexer lexer)
  {
    State.lexer = lexer;
  }
  
  /***************************************************************************
   * get lexer used by proof navigator
   * @return lexer used by proof navigator
   **************************************************************************/
  public static PNLexer getLexer()
  {
    return lexer;
  }
  
  /***************************************************************************
   * signal that lexer shall produce labels rather than idents
   * @param signal true iff lexer shall produce labels
   **************************************************************************/
  public static void setLabels(boolean signal)
  {
    lexerLabels = signal;
  }
  
  /***************************************************************************
   * ask whether lexer shall produce labels rather than idents
   * @return true ff lexer shall produce labels
   **************************************************************************/
  public static boolean getLabels()
  {
    return lexerLabels;
  }
  
  /***************************************************************************
   * set error message, does not override previously set message
   * @param error the error message
   **************************************************************************/
  public static void setError(String error)
  {
    // System.out.println(error);
    // new Exception().printStackTrace(System.out);
    if (State.error == null) State.error = error;
  }
  
  /***************************************************************************
   * reset error message
   **************************************************************************/
  public static void resetError()
  {
    State.error = null;
  }
  
  /***************************************************************************
   * get error message set
   * @return the error message (null if none set)
   **************************************************************************/
  public static String getError()
  {
    return error;
  }
  
  /***************************************************************************
   * set environment used by proof navigator
   * @param env environment to be used by proof navigator
   **************************************************************************/
  public static void setEnvironment(Environment env)
  {
    State.environment = env;
  }
  
  /***************************************************************************
   * get environment used by proof navigator
   * @return environment used by proof navigator
   **************************************************************************/
  public static Environment getEnvironment()
  {
    return environment;
  }
  
  /***************************************************************************
   * get global environment of proof navigator
   * @return flobal environment of proof navigator
   **************************************************************************/
  public static Environment getGlobal()
  {
    return global;
  }
  
  /***************************************************************************
   * get context used by proof navigator
   * @return context used by proof navigator
   **************************************************************************/
  public static Context getContext()
  {
    return context;
  }
  
  /***************************************************************************
   * get prover currently in use
   * @return prover currently in use
   **************************************************************************/
  public static Prover getProver()
  {
    return prover;
  }
  
  /***************************************************************************
   * get version of prover currently in use
   * @return version of the prover currently in use
   **************************************************************************/
  public static ProverVersion getProverVersion()
  {
    return proverVersion;
  }
  
  /***************************************************************************
   * Set new prover.
   **************************************************************************/
  public static void setNewProver()
  {
    if (oldProver != null) setError("attempt to set new prover twice");
    oldProver = prover;
    prover = newProver(log, 0, false);
    oldContext = context;
    context = new Context();
  }
  
  /***************************************************************************
   * Restore old prover.
   **************************************************************************/
  public static void restoreOldProver()
  {
    if (oldProver == null) setError("no old prover to restore");
    prover.terminate(true);
    prover = oldProver;
    context = oldContext;
    oldProver = null;
    oldContext = null;
  }
  
  /***************************************************************************
   * Set prover currently running.
   * @param prover currently running prover.
   **************************************************************************/
  public static void setRunningProver(Prover prover)
  {
    runningProver = prover;
  }
  
  /***************************************************************************
   * Get prover currently running.
   * @return prover currently running.
   **************************************************************************/
  public static Prover getRunningProver()
  {
    return runningProver;
  }
  
  /***************************************************************************
   * set current proof
   * @param proof proof currently under development (null, if none)
   **************************************************************************/
  public static void setProof(Proof proof)
  {
    State.proof = proof;
    Main.setProof(proof);
  }
  
  /***************************************************************************
   * get current proof
   * @return proof currently under development (null, if none)
   **************************************************************************/
  public static Proof getProof()
  {
    return proof;
  }
  
  /***************************************************************************
   * Get current store.
   * @return current store.
   **************************************************************************/
  public static Store getStore()
  {
    return store;
  }
  
  /***************************************************************************
   * Get current presenter.
   * @return current presenter.
   **************************************************************************/
  public static Presenter getPresenter()
  {
    return presenter;
  }
  
  /***************************************************************************
   * Get prover log 
   * @return the log file
   **************************************************************************/
  public static PrintWriter getProverLog()
  {
    return log0;
  }
  
  /***************************************************************************
   * Get file stack.
   * @return the stack.
   **************************************************************************/
  public static Stack<File> getFileStack()
  {
    return fileStack;
  } 
  
  /***************************************************************************
   * Get reader stack.
   * @return the stack.
   **************************************************************************/
  public static Stack<BufferedReader> getReaderStack()
  {
    return readerStack;
  } 
  
  /***************************************************************************
   * Get formula mask.
   * @return the formula mask.
   **************************************************************************/
  public static CVCLFormulaMask getFormulaMask()
  {
    return formulaMask;
  }
  
  /***************************************************************************
   * Set formula mask.
   * @param mask the formula mask.
   **************************************************************************/
  public static void setFormulaMask(CVCLFormulaMask mask)
  {
    formulaMask = mask;
  }
  
  /***************************************************************************
   * Set server port.
   * @param port the port used by the command server
   **************************************************************************/
  public static void setServerPort(int port)
  {
    serverPort = port;
  }
  
  /***************************************************************************
   * Get server port.
   * @return the port used by the command server
   **************************************************************************/
  public static int getServerPort()
  {
    return serverPort;
  }
  
  /***************************************************************************
   * Set prover abort state.
   * @param aborted true if the prover has been aborted.
   **************************************************************************/
  public static void setAborted(boolean aborted)
  {
    State.aborted = aborted;
  }
  
  /***************************************************************************
   * Get prover abort state.
   * @return true if the prover has been aborted.
   **************************************************************************/
  public static boolean isAborted()
  {
    return aborted;
  }
  
  /***************************************************************************
   * Set font size.
   * @param big true iff big fonts are to be displayed.
   **************************************************************************/
  public static void setBigFont(boolean big)
  {
    State.bigFont = big;
  }
  
  /***************************************************************************
   * Get font size.
   * @return true if big fonts are to be displayed.
   **************************************************************************/
  public static boolean useBigFont()
  {
    return bigFont;
  }
  
  /***************************************************************************
   * Set children of current proof state during proof replay.
   * @param children the children states (null, if none).
   **************************************************************************/
  public static void setReplayChildren(ProofState[] children)
  {
    State.replayChildren = children;
  }
  
  /***************************************************************************
   * Set children of current proof state during proof replay.
   * @return the children (null, if none).
   **************************************************************************/
  public static ProofState[] getReplayChildren()
  {
    return replayChildren;
  }
  
  /***************************************************************************
   * Turn on/off automatic simplification
   * @param on true iff automatic simplification is turned on.
   **************************************************************************/
  public static void setAutoSimplify(boolean on)
  {
    State.autoSimplify = on;
  }
  
  /***************************************************************************
   * Is automatic simplification used?
   * @return true iff automatic simplification is used.
   **************************************************************************/
  public static boolean useAutoSimplify()
  {
    return autoSimplify;
  }
  
  /***************************************************************************
   * Are value types to be checked to be first-order?
   * @return true iff value types are to be checked to be first-order.
   **************************************************************************/
  public static boolean firstOrderCheck()
  {
    return firstOrderCheck;
  }
  
  /***************************************************************************
   * Is type BOOLEAN externally represented as LOGICAL (and type BIT as BOOLEAN)?
   * @return true iff BOOLEAN is externally represented as LOGICAL.
   **************************************************************************/
  public static boolean booleanAsLogical()
  {
    return firstOrderCheck && booleanAsLogical;
  }

  /***************************************************************************
   * Turn on/off whether value types to be checked to be first-order?
   * @param value true iff value types are to be checked to be first-order.
   **************************************************************************/
  public static void setFirstOrderCheck(boolean value)
  {
    if (proverVersion == ProverVersion.CVCL) firstOrderCheck = value;
  }
  
  /***************************************************************************
   * Turn on/off whether BOOLEAN is externally represented as LOGICAL.
   * @param value true iff BOOLEAN is to be externally represented as LOGICAL.
   **************************************************************************/
  public static void setBooleanAsLogical(boolean value)
  {
    booleanAsLogical = value;
  }

}
