// ----------------------------------------------------------------------------
// ASTUtil.java
// Collect state related to handling of abstract syntax trees
// $Id: ASTUtil.java,v 1.28 2017/07/04 08:04:36 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Syntax;
import java.util.*;

/**
 * Collect state related to handling of abstract syntax trees.
 */
public final class ASTUtil
{
  /** every (sub)expression is printed with enclosing parentheses */
  private static boolean printParens = false;
  
  /** every identifier is printed with its unique name */
  private static boolean printUnique = false;
  
  /** every local variable is printed with its variable number */
  private static boolean printVarNumber = false;
  
  /** current variable number */
  private static int varNumber = 0;
  
  /** unique identifiers for global environment **/
  private static UniqueNameTable uniqueNameTable = new UniqueNameTable();
  
   /***************************************************************************
   * signal whether every (sub)expression is to be printed with 
   * enclosing parentheses
   * @param flag true iff enclosing parenthesis are to be printed
   **************************************************************************/
  public static void setPrintParens(boolean flag)
  {
    printParens = flag;
  }
  
  /***************************************************************************
   * show whether every (sub)expression is to be printed with 
   * enclosing parentheses
   * @return true iff enclosing parenthesis are to be printed
   **************************************************************************/
  public static boolean getPrintParens()
  {
    return printParens;
  }
  
  /***************************************************************************
   * signal whether every identifier is printed with its unique name
   * @param flag true iff unique name is to be printed
   **************************************************************************/
  public static void setPrintUnique(boolean flag)
  {
    printUnique = flag;
  }
  
  /***************************************************************************
   * show whether every identifier is printed with its unique name
   * @return true iff unique name is printed
   **************************************************************************/
  public static boolean getPrintUnique()
  {
    return printUnique;
  }
  
  /***************************************************************************
   * signal whether every identifier is printed with its variable number
   * @param flag true iff variable number is to be printed
   **************************************************************************/
  public static void setPrintVarNumber(boolean flag)
  {
    printVarNumber = flag;
    // varNumber = 0;
  }
  
  /***************************************************************************
   * show whether every local variable is printed with its variable number
   * @return true iff variable number is printed
   **************************************************************************/
  public static boolean getPrintVarNumber()
  {
    return printVarNumber;
  }
  
  /***************************************************************************
   * set the current variable number for printing
   * @param varNumber the variable number
   **************************************************************************/
  public static void setVarNumber(int varNumber)
  {
    ASTUtil.varNumber = varNumber;
  }
  
  /***************************************************************************
   * increase the current variable number for printing by one
   * @return the old variable number
   **************************************************************************/
  public static int incVarNumber()
  {
    int v = varNumber;
    varNumber = varNumber+1;
    return v;
  }
  
  /***************************************************************************
   * get current variable number for printing
   * @return the variable number
   **************************************************************************/
  public static int getVarNumber()
  {
    return varNumber;
  }
  
  /***************************************************************************
   * converts vector v of String objects to String array
   * @param v a vector holding String objects
   * @return an array holding these objects
   **************************************************************************/
  public static String[] toStringArray(Vector<String> v)
  {
    return (String[])v.toArray(new String[v.size()]);
  }
  
  /***************************************************************************
   * converts vector v of Reference objects toarray
   * @param v a vector holding Reference objects
   * @return an array holding these objects
   **************************************************************************/
  public static Reference[] toReferenceArray(Vector<Reference> v)
  {
    return (Reference[])v.toArray(new Reference[v.size()]);
  }
  
  /***************************************************************************
   * converts vector v of Declaration objects to Declaration array
   * @param v a vector holding Declaration objects
   * @return an array holding these objects
   **************************************************************************/
  public static Declaration[] toDeclarationArray(Vector<Declaration> v)
  {
    return (Declaration[])v.toArray(new Declaration[v.size()]);
  }
  
  /***************************************************************************
   * converts vector v of ValueDeclaration objects to ValueDeclaration array
   * @param v a vector holding ValueDeclaration objects
   * @return an array holding these objects
   **************************************************************************/
  public static ValueDeclaration[] toValueDeclarationArray(Vector<ValueDeclaration> v)
  {
    return (ValueDeclaration[])v.toArray(new ValueDeclaration[v.size()]);
  }
  
  /***************************************************************************
   * converts vector v of TypeDeclaration objects to TypeDeclaration array
   * @param v a vector holding TypeDeclaration objects
   * @return an array holding these objects
   **************************************************************************/
  public static TypeDeclaration[] toTypeDeclarationArray(Vector<TypeDeclaration> v)
  {
    return (TypeDeclaration[])v.toArray(new TypeDeclaration[v.size()]);
  }
  
  /***************************************************************************
   * converts vector v of Type objects to Type array
   * @param v a vector holding AST objects
   * @return an array holding these objects
   **************************************************************************/
  public static Type[] toTypeArray(Vector<Type> v)
  {
    return (Type[])v.toArray(new Type[v.size()]);
  }
  
  /***************************************************************************
   * converts vector v of Expression objects to Expression array
   * @param v a vector holding Expression objects
   * @return an array holding these objects
   **************************************************************************/
  public static Expression[] toExpressionArray(Vector<Expression> v)
  {
    return (Expression[])v.toArray(new Expression[v.size()]);
  }
  
  /***************************************************************************
   * converts vector v of Selector objects to Selector array
   * @param v a vector holding Selector objects
   * @return an array holding these objects
   **************************************************************************/
  public static Selector[] toSelectorArray(Vector<Selector> v)
  {
    return (Selector[])v.toArray(new Selector[v.size()]);
  }
  
  /***************************************************************************
   * converts vector v of TypedIdentifer objects to TypedIdentifier array
   * @param v a vector holding TypedIdentifier objects
   * @return an array holding these objects
   **************************************************************************/
  public static TypedIdentifier[] toTypedIdentifierArray(Vector<TypedIdentifier> v)
  {
    return (TypedIdentifier[])v.toArray(new TypedIdentifier[v.size()]);
  }

  /***************************************************************************
   * converts vector v of ValuedIdentifer objects to ValuedIdentifier array
   * @param v a vector holding ValuedIdentifier objects
   * @return an array holding these objects
   **************************************************************************/
  public static ValuedIdentifier[] toValuedIdentifierArray(Vector<ValuedIdentifier> v)
  {
    return (ValuedIdentifier[])v.toArray(new ValuedIdentifier[v.size()]);
  }
    
  /****************************************************************************
   * Create unique identifier name from given name.
   * @param name of an identifier
   * @return a unique version of name not in use anywhere else as a unique id.
   **************************************************************************/
  public static String getUniqueName(String name)
  {
    return uniqueNameTable.getUniqueName(name);
  }
  
  /***************************************************************************
   * Get base name from name.
   * @param name a name (possible unique as returned by getUniqueName()).
   * @return the original base name (which may be the name itself).
   **************************************************************************/
  public static String getBaseName(String name)
  {
    return UniqueNameTable.getBaseName(name);
  }
  
  /****************************************************************************
   * get canonical text of expression (variable names replaced by numbers).
   * @param exp the expression.
   * @return the canonical text of expression
   ***************************************************************************/
  public static String canonicalText(AST exp)
  {
    setPrintVarNumber(true);
    String rep = exp.toString();
    setPrintVarNumber(false);
    return rep;
  }
  
  /****************************************************************************
   * Construct conjunction of sequence of formulas.
   * @param formulas a sequence of formulas.
   * @return the conjunction of these formulas.
   ***************************************************************************/
  public static Expression toConjunction(Expression[] formulas)
  {
    int n = formulas.length;
    if (n == 0) return Construct.logical(true);
    Expression c = formulas[n-1];
    for (int i=n-2; i>=0; i--)
      c = Construct.andFormula(formulas[i], c);
    return c;
  }
  
  /****************************************************************************
   * Construct disjunction of sequence of formulas.
   * @param formulas a sequence of formulas.
   * @return the disjunction of these formulas.
   ***************************************************************************/
  public static Expression toDisjunction(Expression[] formulas)
  {
    int n = formulas.length;
    if (n == 0) return Construct.logical(false);
    Expression c = formulas[n-1];
    for (int i=n-2; i>=0; i--)
      c = Construct.orFormula(formulas[i], c);
    return c;
  }
}
