// ----------------------------------------------------------------------------
// UpdateTerm.java
// Handling of component updates by index selection.
// $Id: UpdateTerm.java,v 1.17 2015/09/10 13:07:48 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Syntax;

import java.io.*;

/**
 * Handling of component updates by index selection.
 */
public final class UpdateTerm extends ExpressionBase
{
  private Expression base1;   // the compound object
  private Selector[] base2;   // the sequence of selectors
  private Expression value;   // the update value
  
  /****************************************************************************
   * construct component update by index selection from base, base2, and base3
   * @param base1 the compound object
   * @param base2 the sequence of selectors
   * @param value the update value
   ***************************************************************************/
  public UpdateTerm(Expression base1, Selector[] base2, Expression value)
  {
    this.base1 = base1;
    this.base2 = base2;
    this.value = value;
  }
  
  /***************************************************************************
   * get binding level for printing (lower numbers mean greater binding power)
   * @return the priority level
   **************************************************************************/
  public int getPriority()
  {
    return 6;
  }
  
  /***************************************************************************
   * get the base expression
   * @return the base expression
   **************************************************************************/
  public Expression getBase()
  {
    return base1;
  }
  
  /***************************************************************************
   * get the selectors
   * @return the selectors
   **************************************************************************/
  public Selector[] getSelectors()
  {
    return base2;
  }
  
  /***************************************************************************
   * get the update value
   * @return the update value
   **************************************************************************/
  public Expression getUpdate()
  {
    return value;
  }
  
  /***************************************************************************
   * return copy of AST with value references instantiated by the substitutions
   * set in the corresponding declaration identifiers
   * @return the instantiated copy
   **************************************************************************/
  public Expression instantiate()
  {
    Expression base = getBase().instantiate();
    Selector[] sel = getSelectors();
    Expression update = getUpdate().instantiate();
    int n = sel.length;
    Selector[] sel0 = new Selector[n];
    for (int i=0; i<n; i++)
      sel0[i] = sel[i].instantiate();
    return Construct.updateTerm(base, sel0, update);
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    base1.printPriority(getPriority(), out);
    out.print(" WITH ");
    int n = base2.length;
    for (int i=0; i<n; i++)
      base2[i].print(out);
    out.print(":=");
    value.printPriority(getPriority(), out);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   * @return the result of the visit (may be null)
   **************************************************************************/
  public AST accept(ASTVisitor visitor) 
  { 
    AST result = visitor.visit(this);
    if (result != null) return result;
    AST base10 = base1.accept(visitor);
    int n = base2.length;
    Selector[] base20 = new Selector[n];
    for (int i=0; i<n; i++)
      base20[i] = (Selector)base2[i].accept(visitor);
    Object value0 = value.accept(visitor);
    if (!visitor.isCloned()) return null;
    return Construct.updateTerm((Expression)base10, base20, (Expression)value0);
  }
}